package org.zhouw.stps.monitor;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Date;
import java.util.HashMap;
import java.util.TreeMap;

import org.zhouw.stps.*;
import org.zhouw.stps.conf.SystemConfigure;

public class LogAnalyzer {

	public SystemConfigure sysConf;
	
	public LogAnalyzer(String inPath, String outPath, String confPath)
	{
		if(outPath!=null)
			{
			if(!outPath.endsWith("/"))
				outPath+="/";
			}
		this.outPath=outPath;
		if(inPath!=null)
			{
			if(!inPath.endsWith("/"))
				inPath+="/";
			}
		this.inPath=inPath;
		sysConf=new SystemConfigure(confPath);
		System.out.println("Ready to anaylze log in path:" +inPath);
		System.out.println("Configuration file path is:" +confPath);
		System.out.println("output path is:" +outPath);
		System.out.println("Experiment has lasted for " + sysConf.testDuration + " ms");
	}
	
	private String outPath;
	private String inPath;
	
	private static final long responseTimeSLA=100; //ms
	
	private static final long responseTime_interval=10; //ms
	private static final int responseTime_slotAmount=100; //ms

	private static final long checkPointResponseTime_interval=1000; //ms
	private static final int checkPointResponseTime_slotAmount=60; //ms
	
	private static final long concurrency_interval=1000; //ms
	private static final long throughput_interval=1000; //ms
	private static final long warmup_interval=300000; //ms
	private static final long endup_interval=100000; //ms
	
	public void analyzeDataServiceLog(String fileName)
	{
		analyzeClientTransactionLog(fileName);
		System.out.println("analyzeDataServiceLog has been successfully analyzed");
	}
	public void analyzeIDServiceLog(String fileName)
	{
		int total=0;
		int[] rstSlots=new int[responseTime_slotAmount];
		for(int c:rstSlots)
			c=0;
		
		long firstTimestamp=0;
		long lastTimestamp=0;
		long totalCommittedLatency=0;
		
		long curTime=0;
		int curNum=0;
		String fileNameOverview=fileName+"_Overview";
		String fileNameRs=fileName+"_Rs";
		String fileNameThroughputbyend=fileName+"_Throughputbyend";
		String fileNameRsInSlot=fileName+"_RsInSlot";
		try {
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			BufferedWriter writerRs=new BufferedWriter(new FileWriter(outPath+fileNameRs,false));
			BufferedWriter writerThroughputbyend=new BufferedWriter(new FileWriter(outPath+fileNameThroughputbyend,false));
			BufferedWriter writerRsInSlot=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot,false));
		 
			
			FileInputStream fstream = new FileInputStream(inPath+fileName);
			BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
			IDServiceLogItem log=new IDServiceLogItem();
			if(log.readFrom(in)!=-1)
				{
					firstTimestamp=log.timestamp;
					curTime=log.timestamp;
					do
					{
						
						//skip the load in warmup
						if(log.timestamp<(firstTimestamp+warmup_interval))
							continue;
						
						//skip the last time period which is doing ending
						if(log.timestamp>(firstTimestamp+ sysConf.testDuration -endup_interval))
							continue;
						
						lastTimestamp=log.timestamp;
						
						total++;
						totalCommittedLatency+=log.latency;
							
						writerRs.write((log.timestamp-firstTimestamp)
								+" "+ log.latency);
						writerRs.newLine();
						
						//
						if((log.timestamp-curTime)<throughput_interval)
							curNum++;
						else
							{
								do{
									long t=curTime-firstTimestamp;
										t=t-(t%throughput_interval);
										writerThroughputbyend.write(t+" "+ curNum);
									writerThroughputbyend.newLine();
									curTime+=throughput_interval;
									curNum=0;
								}
								while((log.timestamp-curTime)>throughput_interval);
								curNum=1;
							}
						
						
						int index=(int)(log.latency/responseTime_interval);
						if(index>=responseTime_slotAmount)
							index=responseTime_slotAmount-1;
						rstSlots[index]++;
							
					}while(log.readFrom(in)!=-1);
				}
			double totalAmount=0;
			for(int i=rstSlots.length-1;i>=0;i--)
				{
					totalAmount+=rstSlots[i];
				}
			int amount=0;
			for(int i=rstSlots.length-1;i>=0;i--)
				{
					amount+=rstSlots[i];
					writerRsInSlot.write((i+1)*responseTime_interval+" "+((double)amount/totalAmount)*100);
					writerRsInSlot.newLine();
				}
			
			long interval=lastTimestamp-firstTimestamp-warmup_interval;
			double totalThroughput=(double)total/((double)interval/1000);
			double avgResponseTime=(double)totalCommittedLatency/(double)total;
			
			writerOverview.write("ExperimentInterval="+interval);
			writerOverview.newLine();
			writerOverview.write("totalTransactionExecuted="+ total);
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput);
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgResponseTime);
			writerOverview.newLine();
			
			in.close();
			
			writerOverview.flush();
			writerRs.flush();
			writerThroughputbyend.flush();
			writerRsInSlot.flush();
			
			writerOverview.close();
			writerRs.close();
			writerThroughputbyend.close();
			writerRsInSlot.close();
			
			System.out.println("IDServiceLog has been successfully analyzed");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Generate percentage of requests response time. per 50ms
	 * Detailed points about response time by endtime timestamp  
	 * Detailed points about response time by startTime timestamp
	 * Detailed graph about throughput by timestamp per 30 seconds
	 * Percentage of committed 
	 * @param fileName
	 */
	public void analyzeClientTransactionLog(String fileName)
	{
		String fileNameOverview=fileName+"_Overview";
		String fileNameRs=fileName+"_Rs";
		String fileNameThroughputofAborted=fileName+"_ThroughputofAborted";
		String fileNameThroughputbyend=fileName+"_Throughputbyend";
		String fileNameRsInSlot=fileName+"_RsInSlot";
		String fileNameTranClasses=fileName+"_TranClasses";
		String fileNameTranRWCounts=fileName+"_RWTranCount";
		String fileNameTranROCounts=fileName+"_ROTranCount";
		
		HashMap<String,TransactionClass> tranClasses=new HashMap<String,TransactionClass>();
		
		TreeMap<Integer,Integer> roTranCount=new TreeMap<Integer,Integer>();
		TreeMap<Integer,Integer> rwTranCount=new TreeMap<Integer,Integer>();
		
		int numCommitted=0;
		int total=0;
		int[] rstSlots=new int[responseTime_slotAmount];
		for(int c:rstSlots)
			c=0;
		
		long firstTimestamp=0;
		long lastTimestamp=0;
		long totalCommittedLatency=0;
		
		long curTime=0;
		int curNum=0;
		
		
		int curNumofAborted=0;
		
		int numOfWithinSLA=0;
		int numOfOutOfSLA=0;
		
		try {
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			BufferedWriter writerRs=new BufferedWriter(new FileWriter(outPath+fileNameRs,false));
			BufferedWriter writerThroughputbyend=new BufferedWriter(new FileWriter(outPath+fileNameThroughputbyend,false));
			BufferedWriter writerThroughputofAborted=new BufferedWriter(new FileWriter(outPath+fileNameThroughputofAborted,false));
			BufferedWriter writerRsInSlot=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot,false));
			BufferedWriter writerTranClasses=new BufferedWriter(new FileWriter(this.outPath+fileNameTranClasses,false));
			
			BufferedWriter writerRWTranCount=new BufferedWriter(new FileWriter(this.outPath+fileNameTranRWCounts,false));
			BufferedWriter writerROTranCount=new BufferedWriter(new FileWriter(this.outPath+fileNameTranROCounts,false));
			
			FileInputStream fstream = new FileInputStream(inPath+fileName);
			BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
			ClientTransactionLogItem log=new ClientTransactionLogItem();
			if(log.readFrom(in)!=-1)
				{
					firstTimestamp=log.timestamp;
					curTime=log.timestamp;
					do
					{
						
						boolean isInEffectiveRange=true;
						//skip the load in warmup
						if(log.timestamp<(firstTimestamp+warmup_interval)){
							isInEffectiveRange=false;
						}
						//skip the last time period which is doing ending
						if(log.timestamp>(firstTimestamp+ sysConf.testDuration -endup_interval)) {
							isInEffectiveRange=false;
						}
						
						if(isInEffectiveRange){
								total++;
						}
						
						lastTimestamp=log.timestamp;
						
						if(log.isCommitted){
								if(isInEffectiveRange) {
										TransactionClass tranCl=tranClasses.get(log.name);
										if(tranCl==null){
											tranCl=new TransactionClass(log.name);
											tranClasses.put(log.name,tranCl);
										}
										
										TreeMap<Integer,Integer> tranCount;
										if(log.isReadonly){
											tranCount=roTranCount;
										} else {
											tranCount=rwTranCount;
										}
										
										Integer curAccessedNum=tranCount.get(log.numAccessedItems);
										if(curAccessedNum==null){
											curAccessedNum=1;
										} else {
											curAccessedNum++;
										}
										tranCount.put(log.numAccessedItems, curAccessedNum);
										
										tranCl.totalResponseTime+=log.latency;
										tranCl.totalAccessedItems+=log.numAccessedItems;
										tranCl.totalNum++;
										if(log.latency>=LogAnalyzer.responseTimeSLA) {//calculate the number of committed transactions that satisfy SLA
												numOfOutOfSLA++;
										} else {
												numOfWithinSLA++;
										}
										numCommitted++;
										totalCommittedLatency+=log.latency;
										
										int index=(int)(log.latency/responseTime_interval);
										if(index>=responseTime_slotAmount)
											index=responseTime_slotAmount-1;
										rstSlots[index]++;
								}

								writerRs.write((log.timestamp-firstTimestamp)
										+" "+ log.latency);
								writerRs.newLine();
							}
						
						//about throughput
						if((log.timestamp-curTime)<throughput_interval) {
								if(log.isCommitted)
									curNum++;
								else
									curNumofAborted++;
						}	else {
								do{
									long t=curTime-firstTimestamp;
									t=t-(t%throughput_interval);
									writerThroughputbyend.write(t+" "+ curNum);
									writerThroughputofAborted.write(t+" "+ curNumofAborted);
									writerThroughputofAborted.newLine();
									writerThroughputbyend.newLine();
									curTime+=throughput_interval;
									curNum=0;
									curNumofAborted=0;
								}
								while((log.timestamp-curTime)>throughput_interval);
								
								if(log.isCommitted)
									curNum++;
								else
									curNumofAborted++;
						}
					}while(log.readFrom(in)!=-1);
				}
			
			double totalAmount=0;
			for(int i=rstSlots.length-1;i>=0;i--)
				{
					totalAmount+=rstSlots[i];
				}
			int amount=0;
			for(int i=rstSlots.length-1;i>=0;i--)
				{
					amount+=rstSlots[i];
					writerRsInSlot.write((i+1)*responseTime_interval+" "+((double)amount/totalAmount)*100);
					writerRsInSlot.newLine();
				}
			
			for(String tranName:tranClasses.keySet()){
				TransactionClass tcl=tranClasses.get(tranName);
				double avgAccessedItem=(double)tcl.totalAccessedItems/tcl.totalNum;
				double avgRsTime=(double)tcl.totalResponseTime/tcl.totalNum;
				writerTranClasses.write("["+tcl.classname+"]");
				writerTranClasses.newLine();
				writerTranClasses.write("Average Response Time="+avgRsTime);
				writerTranClasses.newLine();
				writerTranClasses.write("Average Accessed Item="+avgAccessedItem);
				writerTranClasses.newLine();
				writerTranClasses.write("Total Number of transactions="+tcl.totalNum);
				writerTranClasses.newLine();
				writerTranClasses.write("Total Response Time="+tcl.totalResponseTime);
				writerTranClasses.newLine();
				writerTranClasses.write("Total Accessed Item="+tcl.totalAccessedItems);
				writerTranClasses.newLine();
			}
			
			
			for(int number:rwTranCount.keySet()){
				writerRWTranCount.write(number +"="+ rwTranCount.get(number));
				writerRWTranCount.newLine();
			}
			
			for(int number:roTranCount.keySet()){
				writerROTranCount.write(number +"="+ roTranCount.get(number));
				writerROTranCount.newLine();
			}
			
			
			double percentageOfCommit=((double)numCommitted/(double)total)*100;
			long interval=lastTimestamp-firstTimestamp-warmup_interval-endup_interval;
			double commitThroughput=(double)numCommitted/((double)interval/1000);
			double totalThroughput=(double)total/((double)interval/1000);
			double avgResponseTime=(double)totalCommittedLatency/(double)numCommitted;
			
			double percentageWithinSLA=(double)numOfWithinSLA/(double)(numOfWithinSLA+numOfOutOfSLA);
			
			writerOverview.write("ExperimentInterval="+interval);
			writerOverview.newLine();
			writerOverview.write("percentageWithinSLA="+percentageWithinSLA *100+ "%");
			writerOverview.newLine();
			writerOverview.write("totalTransactionExecuted="+ total);
			writerOverview.newLine();
			writerOverview.write("totalTransactionCommitted="+ numCommitted);
			writerOverview.newLine();
			writerOverview.write("percentageOfCommit="+percentageOfCommit);
			writerOverview.newLine();
			writerOverview.write("commitThroughput="+commitThroughput);
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput);
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgResponseTime);
			writerOverview.newLine();
			writerOverview.write("numOfWithinSLA="+numOfWithinSLA);
			writerOverview.newLine();
			writerOverview.write("numOfOutOfSLA="+numOfOutOfSLA);
			writerOverview.newLine();
			
			in.close();
			
			writerTranClasses.flush();
			writerTranClasses.close();
			
			writerOverview.flush();
			writerRs.flush();
			writerThroughputbyend.flush();
			writerThroughputofAborted.flush();
			writerRsInSlot.flush();
			
			writerOverview.close();
			writerRs.close();
			writerThroughputbyend.close();
			writerThroughputofAborted.close();
			writerRsInSlot.close();
			
			writerRWTranCount.flush();
			writerROTranCount.flush();
			writerRWTranCount.close();
			writerROTranCount.close();
			
			System.out.println("ClientTransactionLog has been successfully analyzed");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public void analyzeCheckPointLog(String fileName)
	{
		String fileNameOverview=fileName+"_Overview";
		String fileNameRs=fileName+"_Rs"; //include total, copy and flush latency 
		String fileNameRsInSlot=fileName+"_RsInSlot"; //include total, copy and flush latency inslot
		String fileNameRsInSlot_copy=fileName+"_RsInSlotCopyLatency"; //include total, copy and flush latency inslot
		String fileNameMem=fileName+"_memory"; //include total, copy and flush latency 
		int total=0;
		int[] totalLatencySlots=new int[checkPointResponseTime_slotAmount];
		int[] copyLatencySlots=new int[checkPointResponseTime_slotAmount];
		int[] flushLatencySlots=new int[checkPointResponseTime_slotAmount];
		int[] copyLatencySlotsInSingle=new int[responseTime_slotAmount];
		
		for(int c:totalLatencySlots)
			c=0;
		for(int c:copyLatencySlots)
			c=0;
		for(int c:flushLatencySlots)
			c=0;
		for(int c:copyLatencySlotsInSingle)
			c=0;
		
		long firstTimestamp=0;
		long lastTimestamp=0;
		long totalLatency=0;
		long totalCopyLatency=0;
		long totalFlushLatency=0;
		
		try {
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			BufferedWriter writerRs=new BufferedWriter(new FileWriter(outPath+fileNameRs,false));
			BufferedWriter writerMemory=new BufferedWriter(new FileWriter(outPath+fileNameMem,false));
			BufferedWriter writerRsInSlot=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot,false));
			BufferedWriter writerRsInSlot_copy=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot_copy,false));
			
			writerMemory.write("#Timestamp freeMemory totalMemory");
			writerMemory.newLine();
			
			FileInputStream fstream = new FileInputStream(inPath+fileName);
			BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
			CheckpointLogItem log=new CheckpointLogItem();
			if(log.readFrom(in)!=-1)
				{
					firstTimestamp=log.timestamp;
					do
					{
						
//						//skip the load in warmup
//						if(log.timestamp<(firstTimestamp+warmup_interval))
//							
//							continue;
//						//skip the last time period which is doing ending
//						if(log.timestamp>(firstTimestamp+ GlobalPara.testDuration -endup_interval))
//							continue;
						
						lastTimestamp=log.timestamp;
						
						total++;
						totalLatency+=log.totalLatency;
						totalCopyLatency+=log.copyLatency;
						totalFlushLatency+=log.totalLatency;
						
						//
						//Date curDate=new Date(log.timestamp);
						writerRs.write((log.timestamp-firstTimestamp)
								+" "+ log.totalLatency + " " + log.copyLatency + " " + log.flushLatency);
						writerRs.newLine();
						writerMemory.write((log.timestamp-firstTimestamp)
								+" "+ log.freeMemoryBefore + " " + log.totalMemoryBefore);
						writerMemory.newLine();
						writerMemory.write((log.timestamp-firstTimestamp+log.totalLatency)
								+" "+ log.freeMemoryAfter + " " + log.totalMemoryAfter);
						writerMemory.newLine();
						
						//
						int index=(int)(log.totalLatency/checkPointResponseTime_interval);
						if(index>=checkPointResponseTime_slotAmount)
							index=checkPointResponseTime_slotAmount-1;
						totalLatencySlots[index]++;
						
						index=(int)(log.copyLatency/checkPointResponseTime_interval);
						if(index>=checkPointResponseTime_slotAmount)
							index=checkPointResponseTime_slotAmount-1;
						copyLatencySlots[index]++;
						
						index=(int)(log.copyLatency/responseTime_interval);
						if(index>=responseTime_slotAmount)
							index=responseTime_slotAmount-1;
						copyLatencySlotsInSingle[index]++;
						
						index=(int)(log.flushLatency/checkPointResponseTime_interval);
						if(index>=checkPointResponseTime_slotAmount)
							index=checkPointResponseTime_slotAmount-1;
						flushLatencySlots[index]++;
							
					}while(log.readFrom(in)!=-1);
				}
			
			double totalAmount1=0;
			double totalAmount2=0;
			double totalAmount3=0;
			for(int i=totalLatencySlots.length-1;i>=0;i--)
				{
					totalAmount1+=totalLatencySlots[i];
					totalAmount2+=copyLatencySlots[i];
					totalAmount3+=flushLatencySlots[i];
				}
			int amount1=0;
			int amount2=0;
			int amount3=0;
			for(int i=totalLatencySlots.length-1;i>=0;i--)
				{
					amount1+=totalLatencySlots[i];
					amount2+=copyLatencySlots[i];
					amount3+=flushLatencySlots[i];
					writerRsInSlot.write((i+1)*checkPointResponseTime_interval+" "
							+((double)amount1/totalAmount1)*100 + " "
							+((double)amount2/totalAmount2)*100 + " "
							+((double)amount3/totalAmount3)*100 );
					writerRsInSlot.newLine();
				}
			
			double totalAmount=0;
			for(int i=copyLatencySlotsInSingle.length-1;i>=0;i--)
				{
					totalAmount+=copyLatencySlotsInSingle[i];
				}
			int amount=0;
			for(int i=copyLatencySlotsInSingle.length-1;i>=0;i--)
				{
					amount+=copyLatencySlotsInSingle[i];
					writerRsInSlot_copy.write((i+1)*responseTime_interval+" "+((double)amount/totalAmount)*100);
					writerRsInSlot_copy.newLine();
				}
			
			long interval=lastTimestamp-firstTimestamp;
			double totalThroughput=(double)total/((double)interval/(1000*60));//per minute
			double avgTotalResponseTime=(double)totalLatency/(double)total;
			double avgCopyLatency=(double)totalCopyLatency/(double)total;
			double avgFlushResponseTime=(double)totalFlushLatency/(double)total;
			
			writerOverview.write("ExperimentInterval="+interval);
			writerOverview.newLine();
			
			
//			writerOverview.write("There are totally "+total +" checkpointed executed, "+ totalThroughput +" times per second");
//			writerOverview.newLine();
			
			writerOverview.write("totalTransactionExecuted="+ total);
			writerOverview.newLine();
			writerOverview.write("#Per Minute");
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput);
			writerOverview.newLine();
		
			writerOverview.write("averageResponseTime="+avgTotalResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgCopyLatency="+avgCopyLatency);
			writerOverview.newLine();
			writerOverview.write("avgFlushResponseTime="+avgFlushResponseTime);
			writerOverview.newLine();
			
			in.close();
			
			writerOverview.flush();
			writerRs.flush();
			writerMemory.flush();
			writerRsInSlot.flush();
			writerRsInSlot_copy.flush();
			
			writerOverview.close();
			writerRs.close();
			writerMemory.close();
			writerRsInSlot.close();
			writerRsInSlot_copy.close();
			
			System.out.println("analyzeCheckPointLog has been successfully analyzed");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	public void analyzeTPNodeLog(String fileName)
	{
		String fileNameOverview=fileName+"_Overview";
		String fileNameRs=fileName+"_Rs";
		String fileNameThroughputbyend=fileName+"_Throughputbyend";
		String fileNameRsInSlot=fileName+"_RsInSlot";
		//String fileNameConcurrency=fileName+"_Concurrency";
		String fileNameConcurrencyInSlot=fileName+"_ConcurrencyInSlot";
		
		int numCommitted=0;
		int total=0;
		int[] totalRstSlots=new int[responseTime_slotAmount];
		int[] submitRstSlots=new int[responseTime_slotAmount];
		int[] waitRstSlots=new int[responseTime_slotAmount];
		int[] commitRstSlots=new int[responseTime_slotAmount];
		int[] abortRstSlots=new int[responseTime_slotAmount];
		for(int c:totalRstSlots)
			c=0;
		for(int c:submitRstSlots)
			c=0;
		for(int c:waitRstSlots)
			c=0;
		for(int c:commitRstSlots)
			c=0;
		for(int c:abortRstSlots)
			c=0;
		
		long firstTimestamp=0;
		long lastTimestamp=0;
		long totalCommittedLatency=0;
		long totalCommittedSubmitLatency=0;
		long totalCommittedWaitLatency=0;
		long totalCommittedCommitLatency=0;
		long totalCommittedAbortLatency=0;
		int totalRestartTimes=0;
		int[] totalTPNodeInvoked=new int[sysConf.nodes.getTotalNumberOfNodes()];
		int[] totalTPNodeInvokedForCommitted=new int[sysConf.nodes.getTotalNumberOfNodes()];
		int[] numOfTPNodeInslot=new int[sysConf.nodes.getServletNodes(null).length];
		
		for(int c:totalTPNodeInvoked)
			c=0;
		for(int c:totalTPNodeInvokedForCommitted)
			c=0;
		for(int c:numOfTPNodeInslot)
			c=0;
		
		long curTime=0;
		int curNum=0;
		int concuNum=0;
		try {
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			BufferedWriter writerRs=new BufferedWriter(new FileWriter(outPath+fileNameRs,false));
			BufferedWriter writerThroughputbyend=new BufferedWriter(new FileWriter(outPath+fileNameThroughputbyend,false));
			BufferedWriter writerRsInSlot=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot,false));
			
			//BufferedWriter writerConcurrency=new BufferedWriter(new FileWriter(outPath+fileNameConcurrency,false));
			BufferedWriter writerConcurrencyInSlot=new BufferedWriter(new FileWriter(outPath+fileNameConcurrencyInSlot,false));
			
			FileInputStream fstream = new FileInputStream(inPath+fileName);
			BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
			ServerTransactionLogItem log=new ServerTransactionLogItem();
			
			writerRs.write("#timestamp totalLatency submitLatency waitLatency doCommitLatency abortLatency restartTimes transaction_name");
			writerRs.newLine();
			writerThroughputbyend.write("#TimeStamp throughput ThroughputWithAllTPnodeInvoked");
			writerThroughputbyend.newLine();
			if(log.readFrom(in)!=-1)
				{
					firstTimestamp=log.timestamp;
					curTime=log.timestamp;;
					do
					{
						
						//skip the load in warmup
						if(log.timestamp<(firstTimestamp+warmup_interval))
							
							continue;
						//skip the last time period which is doing ending
						if(log.timestamp>(firstTimestamp+ sysConf.testDuration -endup_interval))
							continue;
						
						lastTimestamp=log.timestamp;
						
						total++;
						for(int i=0;i<log.STGDestTPNodeID.length;i++)
							totalTPNodeInvokedForCommitted[log.STGDestTPNodeID[i]]++;
						numOfTPNodeInslot[log.STGDestTPNodeID.length-1]++;
						
						if(log.isCommitted)
							{
							numCommitted++;
							totalCommittedLatency+=log.totalLatency;
							 totalCommittedSubmitLatency+=log.submitLatency;
							 totalCommittedWaitLatency+=log.waitLatency;
							 totalCommittedCommitLatency+=log.doCommitLatency;
							 totalCommittedAbortLatency+=log.abortLatency;
							 totalRestartTimes+=log.restartTimes;
							 for(int i=0;i<log.STGDestTPNodeID.length;i++)
									totalTPNodeInvoked[log.STGDestTPNodeID[i]]++;
							//
							//Date curDate=new Date(log.timestamp);
							writerRs.write((log.timestamp-firstTimestamp)
									+" "+ log.totalLatency
									+" "+ log.submitLatency
									+" "+ log.waitLatency
									+" "+ log.doCommitLatency
									+" "+ log.abortLatency
									+" "+ log.restartTimes
									+" "+ log.name
									);
							writerRs.newLine();
							
							
							//
							if((log.timestamp-curTime)<throughput_interval)
								{
								curNum++;
								concuNum+=log.STGDestTPNodeID.length;
								}
							else
								{
									do{
										long t=curTime-firstTimestamp;
										t=t-(t%throughput_interval);
										writerThroughputbyend.write(t+" "
												+ curNum + " "
												+ concuNum);
										writerThroughputbyend.newLine();
										curTime+=throughput_interval;
										curNum=0;
										concuNum=0;
									}
									while((log.timestamp-curTime)>throughput_interval);
									curNum=1;
								}
							
						
							int index=(int)(log.totalLatency/responseTime_interval);
							if(index>=responseTime_slotAmount)
								index=responseTime_slotAmount-1;
							totalRstSlots[index]++;
							
							index=(int)(log.submitLatency/responseTime_interval);
							if(index>=responseTime_slotAmount)
								index=responseTime_slotAmount-1;
							submitRstSlots[index]++;
							
							index=(int)(log.waitLatency/responseTime_interval);
							if(index>=responseTime_slotAmount)
								index=responseTime_slotAmount-1;
							waitRstSlots[index]++;
							
							index=(int)(log.doCommitLatency/responseTime_interval);
							if(index>=responseTime_slotAmount)
								index=responseTime_slotAmount-1;
							commitRstSlots[index]++;
							
							index=(int)(log.abortLatency/responseTime_interval);
							if(index>=responseTime_slotAmount)
								index=responseTime_slotAmount-1;
							abortRstSlots[index]++;
							}
					}while(log.readFrom(in)!=-1);
				}
			double totalAmount1=0;
			double totalAmount2=0;
			double totalAmount3=0;
			double totalAmount4=0;
			double totalAmount5=0;
			for(int i=totalRstSlots.length-1;i>=0;i--)
				{
					totalAmount1+=totalRstSlots[i];
					totalAmount2+=submitRstSlots[i];
					totalAmount3+=waitRstSlots[i];
					totalAmount4+=commitRstSlots[i];
					totalAmount5+=abortRstSlots[i];
				}
			
			writerRsInSlot.write("#timestamp totalLatency submitLatency waitLatency doCommitLatency abortLatency");
			int amount1=0;
			int amount2=0;
			int amount3=0;
			int amount4=0;
			int amount5=0;
			for(int i=totalRstSlots.length-1;i>=0;i--)
				{
					amount1+=totalRstSlots[i];
					amount2+=submitRstSlots[i];
					amount3+=waitRstSlots[i];
					amount4+=commitRstSlots[i];
					amount5+=abortRstSlots[i];
					writerRsInSlot.write((i+1)*checkPointResponseTime_interval+" "
							+((double)amount1/totalAmount1)*100 + " "
							+((double)amount2/totalAmount2)*100 + " "
							+((double)amount3/totalAmount3)*100 + " "
							+((double)amount4/totalAmount4)*100 + " "
							+((double)amount5/totalAmount5)*100);
					writerRsInSlot.newLine();
				}
			
			double percentageOfCommit=((double)numCommitted/(double)total)*100;
			long interval=lastTimestamp-firstTimestamp-warmup_interval;
			double commitThroughput=(double)numCommitted/((double)interval/1000);
			double totalThroughput=(double)total/((double)interval/1000);
			double avgResponseTime=(double)totalCommittedLatency/(double)numCommitted;
			double avgSubmitResponseTime=(double)totalCommittedSubmitLatency/(double)numCommitted;
			double avgWaitResponseTime=(double)totalCommittedWaitLatency/(double)numCommitted;
			double avgCommitResponseTime=(double)totalCommittedCommitLatency/(double)numCommitted;
			double avgAbortResponseTime=(double)totalCommittedAbortLatency/(double)numCommitted;
			double avgRestartTimes=(double)totalRestartTimes/(double)numCommitted;
			
			writerOverview.write("ExperimentInterval="+interval);
			writerOverview.newLine();
			writerOverview.write("totalTransactionExecuted="+ total);
			writerOverview.newLine();
			writerOverview.write("totalTransactionCommitted="+ numCommitted);
			writerOverview.newLine();
			writerOverview.write("percentageOfCommit="+percentageOfCommit);
			writerOverview.newLine();
			writerOverview.write("commitThroughput="+commitThroughput);
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput);
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgSubmitResponseTime="+avgSubmitResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgWaitResponseTime="+avgWaitResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgCommitResponseTime="+avgCommitResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgAbortResponseTime="+avgAbortResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgRestartTimes="+avgRestartTimes);
			writerOverview.newLine();
			int[] TPNodes=sysConf.nodes.getServletNodes(null);
			for(int i=0;i<TPNodes.length;i++)
				{
				writerOverview.write("node"+ TPNodes[i] +"InvokedForCommit="+ totalTPNodeInvokedForCommitted[TPNodes[i]]);
				writerOverview.newLine();
				}
			for(int i=0;i<TPNodes.length;i++)
				{
				writerOverview.write("node"+ TPNodes[i] +"Invoked="+ totalTPNodeInvoked[TPNodes[i]]);
				writerOverview.newLine();
				}
			
			for(int i=0;i<TPNodes.length;i++)
				{
				writerConcurrencyInSlot.write("NumOfTransactionInvoke"+(i+1)+"TPNode="+ numOfTPNodeInslot[i]);
				writerConcurrencyInSlot.newLine();
				}
			
			in.close();
			
			writerOverview.flush();
			writerRs.flush();
			writerThroughputbyend.flush();
			writerConcurrencyInSlot.flush();
			writerRsInSlot.flush();
			
			writerOverview.close();
			writerRs.close();
			writerThroughputbyend.close();
			writerConcurrencyInSlot.close();
			writerRsInSlot.close();
			
			System.out.println("TPNodeLog has been successfully analyzed");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public void analyzeLoadDataLog(String fileName)
	{
		int totalRead=0;
		int totalWrite=0;
		
		int[] rstSlotsRead=new int[responseTime_slotAmount];
		int[] rstSlotsWrite=new int[responseTime_slotAmount];
		
		for(int c:rstSlotsRead)
			c=0;
		for(int c:rstSlotsWrite)
			c=0;
		
		long firstTimestamp=0;
		long lastTimestamp=0;
		
		long totalReadLatency=0;
		long totalWriteLatency=0;
		
		
		HashMap<String, Long[]> domainRead=new HashMap<String, Long[]>();
		HashMap<String, Long[]> domainWrite=new HashMap<String, Long[]>();
		
		long curTimeRead=0;
		int curNumRead=0;
		
		long curTimeWrite=0;
		int curNumWrite=0;
		
		String fileNameOverview=fileName+"_Overview";
		
		String fileNameByDomainRead=fileName+"_DomainRead";
		String fileNameByDomainWrite=fileName+"_DomainWrite";
		
		String fileNameRs=fileName+"_Rs";
		String fileNameRsWrite=fileName+"_RsWrite";
		
		String fileNameThroughputbyendRead=fileName+"_ThroughputbyendRead";
		String fileNameThroughputbyendWrite=fileName+"_ThroughputbyendWrite";
		
		
		String fileNameRsInSlot=fileName+"_RsInSlot";
		String fileNameRsInSlotWrite=fileName+"_RsInSlotWrite";
		
		try {
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			BufferedWriter writerRs=new BufferedWriter(new FileWriter(outPath+fileNameRs,false));
			
			BufferedWriter writerThroughputbyend=new BufferedWriter(new FileWriter(outPath+fileNameThroughputbyendRead,false));
			BufferedWriter writerRsInSlot=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot,false));
		 
			
			BufferedWriter writerRsWrite=new BufferedWriter(new FileWriter(outPath+fileNameRsWrite,false));
			BufferedWriter writerByDomainRead=new BufferedWriter(new FileWriter(outPath+fileNameByDomainRead,false));
			BufferedWriter writerByDomainWrite=new BufferedWriter(new FileWriter(outPath+fileNameByDomainWrite,false));
			BufferedWriter writerRsInSlotWrite=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlotWrite,false));
			BufferedWriter writerThroughputbyendWrite=new BufferedWriter(new FileWriter(outPath+fileNameThroughputbyendWrite,false));
			
			FileInputStream fstream = new FileInputStream(inPath+fileName);
			BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
			LoadDataLogItem log=new LoadDataLogItem();
			if(log.readFrom(in)!=-1)
				{
					firstTimestamp=log.timestamp;
					curTimeRead=curTimeWrite=log.timestamp;;
					do
					{
						
						//skip the load in warmup
						if(log.timestamp<(firstTimestamp+warmup_interval))
							continue;
						
						//skip the last time period which is doing ending
						if(log.timestamp>(firstTimestamp+ sysConf.testDuration -endup_interval))
							continue;
						
						lastTimestamp=log.timestamp;
						
						int index=(int)(log.latency/responseTime_interval);
						if(index>=responseTime_slotAmount)
							index=responseTime_slotAmount-1;
						
						if(log.readWrite.compareTo("R")==0){
							Long[] domainNum=domainRead.get(log.tableName);
							if(domainNum==null){
								domainNum=new Long[2];
								domainNum[0]=1L;
								domainNum[1]=log.latency;
							} else {
								domainNum[0]++;
								domainNum[1]+=log.latency;
							}
							domainRead.put(log.tableName, domainNum);
							
							totalRead++;
							totalReadLatency+=log.latency;
							writerRs.write((log.timestamp-firstTimestamp)
									+" "+ log.latency);
							writerRs.newLine();
							rstSlotsRead[index]++;
							
							if((log.timestamp-curTimeRead)<throughput_interval)
								curNumRead++;
							else
								{
									do{
										long t=curTimeRead-firstTimestamp;
											t=t-(t%throughput_interval);
											writerThroughputbyend.write(t+" "+ curNumRead);
										writerThroughputbyend.newLine();
										curTimeRead+=throughput_interval;
										curNumRead=0;
									}
									while((log.timestamp-curTimeRead)>throughput_interval);
									curNumRead=1;
								}
							
						} else {
							Long[] domainNum=domainWrite.get(log.tableName);
							if(domainNum==null){
								domainNum=new Long[2];
								domainNum[0]=1L;
								domainNum[1]=log.latency;
							} else {
								domainNum[0]++;
								domainNum[1]+=log.latency;
							}
							domainWrite.put(log.tableName, domainNum);
							
							totalWrite++;
							totalWriteLatency+=log.latency;
							writerRsWrite.write((log.timestamp-firstTimestamp)
									+" "+ log.latency);
							writerRs.newLine();
							rstSlotsWrite[index]++;
							
							if((log.timestamp-curTimeWrite)<throughput_interval)
								curNumWrite++;
							else
								{
									do{
										long t=curTimeWrite-firstTimestamp;
											t=t-(t%throughput_interval);
											writerThroughputbyendWrite.write(t+" "+ curNumWrite);
											writerThroughputbyendWrite.newLine();
										curTimeWrite+=throughput_interval;
										curNumWrite=0;
									}
									while((log.timestamp-curTimeWrite)>throughput_interval);
									curNumWrite=1;
								}
						}
						
					}while(log.readFrom(in)!=-1);
				}
			double totalAmountRead=0;
			double totalAmountWrite=0;
			for(int i=rstSlotsRead.length-1;i>=0;i--)
				{
					totalAmountRead+=rstSlotsRead[i];
				}
			
			for(int i=rstSlotsWrite.length-1;i>=0;i--)
			{
				totalAmountWrite+=rstSlotsWrite[i];
			}
			
			int amountRead=0;
			for(int i=rstSlotsRead.length-1;i>=0;i--)
				{
					if(totalAmountRead!=0){
						amountRead+=rstSlotsRead[i];
						writerRsInSlot.write((i+1)*responseTime_interval+" "+((double)amountRead/totalAmountRead)*100);
					} else {
						writerRsInSlot.write((i+1)*responseTime_interval+" "+ 0);
					}
					writerRsInSlot.newLine();
					
				}
			
			int amountWrite=0;
			for(int i=rstSlotsWrite.length-1;i>=0;i--)
			{
				if(totalAmountWrite!=0){
					amountWrite+=rstSlotsWrite[i];
					writerRsInSlotWrite.write((i+1)*responseTime_interval+" "+((double)amountWrite/totalAmountWrite)*100);
				} else {
					writerRsInSlotWrite.write((i+1)*responseTime_interval+" "+ 0);
				}
				writerRsInSlotWrite.newLine();
			}
			
			
			long interval=lastTimestamp-firstTimestamp-warmup_interval;
			double totalReadThroughput=(double)totalRead/((double)interval/1000);
			double totalWriteThroughput=(double)totalWrite/((double)interval/1000);
			
			double avgReadResponseTime;
			
			if(totalRead!=0)
				avgReadResponseTime=(double)totalReadLatency/(double)totalRead;
			else
				avgReadResponseTime=0;
			
			double avgWriteResponseTime;
			if(totalWrite!=0)
				avgWriteResponseTime=(double)totalWriteLatency/(double)totalWrite;
			else
				avgWriteResponseTime=0;
			
			writerOverview.write("ExperimentInterval="+interval);
			writerOverview.newLine();
			writerOverview.write("totalDataItemLoaded="+ totalRead);
			writerOverview.newLine();
			writerOverview.write("totalThroughputOfLoad="+totalReadThroughput);
			writerOverview.newLine();
			writerOverview.write("averageResponseTimeOfLoad="+avgReadResponseTime);
			writerOverview.newLine();
			writerOverview.write("totalDataItemCheckpointed="+ totalWrite);
			writerOverview.newLine();
			writerOverview.write("totalThroughputOfCheckpoint="+totalWriteThroughput);
			writerOverview.newLine();
			writerOverview.write("averageResponseTimeOfCheckpoint="+avgWriteResponseTime);
			writerOverview.newLine();
			
			in.close();
			
			interval/=1000;
			
			for(String domainName:domainWrite.keySet()){
				Long[] domainCount= domainWrite.get(domainName);
				writerByDomainWrite.write(domainName+ " "+(double)domainCount[0]/(double)interval+" "+(double)domainCount[1]/(double)domainCount[0]);
				writerByDomainWrite.newLine();
			}
			
			for(String domainName:domainRead.keySet()){
				Long[] domainCount= domainRead.get(domainName);
				writerByDomainRead.write(domainName+ " "+(double)domainCount[0]/(double)interval+" "+(double)domainCount[1]/(double)domainCount[0]);
				writerByDomainRead.newLine();
			}
			
			writerOverview.flush();
			writerRs.flush();
			writerThroughputbyend.flush();
			//writerThroughputbystart.flush();
			writerRsInSlot.flush();
			
			writerOverview.close();
			writerRs.close();
			writerThroughputbyend.close();
			//writerThroughputbystart.close();
			writerRsInSlot.close();
			
			
			writerRsWrite.flush();
			writerRsWrite.close();
			
			writerByDomainRead.flush();
			writerByDomainRead.close();
			
			writerByDomainWrite.flush();
			writerByDomainWrite.close();
			
			writerRsInSlotWrite.flush();
			writerRsInSlotWrite.close();
			
			
			writerThroughputbyendWrite.flush();
			writerThroughputbyendWrite.close();
			
			System.out.println("LoadDataLog has been successfully analyzed");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
