package org.zhouw.stps.monitor;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Vector;

import org.zhouw.stps.conf.SystemConfigure;
import java.util.TreeMap;
public class LogReducer {

	
public SystemConfigure sysConf;

	private String outPath;
	private String inPath;

	public LogReducer(String inPath, String outPath, String confPath)
	{
		if(outPath!=null)
			{
			if(!outPath.endsWith("/"))
				outPath+="/";
			}
		this.outPath=outPath;
		if(inPath!=null)
			{
			if(!inPath.endsWith("/"))
				inPath+="/";
			}
		this.inPath=inPath;
		sysConf=new SystemConfigure(confPath);
		System.out.println("Ready to reduce logs in path:" +inPath);
		System.out.println("Configuration file path is:" +confPath);
		System.out.println("output path is:" +outPath);
	}
	
	private String[] getData(BufferedReader in, String delimiter) throws IOException
	{
		String[] result;
		String line;
		while((line=in.readLine())!=null)
			{
				if(line.startsWith("#"))
					continue;
				break;
			}
		if(line==null)
			{
			return null;
			}
		result=line.split(delimiter);
		return result;
	}
	
	public void reduceClientTransactionLog()
	{
		reduceClientTransaction_overview();
		reduceClientTransaction_RsInSlot();
		reduceClientTransaction_throughput();
		reduceClientTransaction_AbortedThroughput();
		reduceClientTransaction_Rs();
		reduceClientTransaction_tranClass();
		reduceClientTransaction_tranCount();
	}

	private void reduceClientTransaction_tranClass()
	{
		int[] nodes=sysConf.nodes.getClientNodes();
		String fileName="AllClientTransaction_TranClasses";
		String fileNameROPlot="AllClientTransaction_TranClassesROPlot";
		String fileNameRWPlot="AllClientTransaction_TranClassesRWPlot";
		String fileNameRWNumPlot="AllClientTransaction_TranClassesRWNumPlot";
		String fileNameRONumPlot="AllClientTransaction_TranClassesRONumPlot";
		HashMap<String,TransactionClass> tranClasses=new HashMap<String,TransactionClass>();
		try {
			BufferedWriter writerTranClasses=new BufferedWriter(new FileWriter(outPath+fileName,false));
			BufferedWriter writerTranClassesROPlot=new BufferedWriter(new FileWriter(outPath+fileNameROPlot,false));
			BufferedWriter writerTranClassesRWPlot=new BufferedWriter(new FileWriter(outPath+fileNameRWPlot,false));
			
			BufferedWriter writerTranClassesRONumPlot=new BufferedWriter(new FileWriter(outPath+fileNameRONumPlot,false));
			BufferedWriter writerTranClassesRWNumPlot=new BufferedWriter(new FileWriter(outPath+fileNameRWNumPlot,false));
			
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="clientTran"+nodeID+".log_TranClasses";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					
					
					TransactionClass tranCl=null;
					
					String paras[]=getData(in,"=");
					while(paras!=null)
						{
							if(paras.length==2&&tranCl!=null)
								{
									if(paras[0].compareTo("Total Response Time")==0)
										{
										long rs=Long.valueOf(paras[1]);
										tranCl.totalResponseTime+=rs;
										}
									else if(paras[0].compareTo("Total Accessed Item")==0){
										tranCl.totalAccessedItems+=Integer.valueOf(paras[1]);
									} else if(paras[0].compareTo("Total Number of transactions")==0)
										{
										tranCl.totalNum+=Integer.valueOf(paras[1]);
										}
								}
							else if(paras.length==1){
								String name=paras[0].substring(1, paras[0].length()-1);
								tranCl=tranClasses.get(name);
								if(tranCl==null){
									tranCl=new TransactionClass(name);
									tranClasses.put(name,tranCl);
								}
							}
							paras=getData(in,"=");
						}
				}
			TreeMap<Integer, Vector<TransactionClass>> readonlyTranClasses=new TreeMap<Integer, Vector<TransactionClass>>();
			TreeMap<Integer, Vector<TransactionClass>> readWriteTranClasses=new TreeMap<Integer, Vector<TransactionClass>>();
			
			for(String tranName:tranClasses.keySet()){
				TransactionClass tcl=tranClasses.get(tranName);
				
				Double avgAccessedItem=(double)tcl.totalAccessedItems/tcl.totalNum;
				double avgRsTime=(double)tcl.totalResponseTime/tcl.totalNum;
				int avgAccessedItemInInteger=avgAccessedItem.intValue();
				if(avgAccessedItem-avgAccessedItemInInteger>0.5)
					avgAccessedItemInInteger++;
				
				if(tranName.startsWith("get")){
					Vector<TransactionClass> collection=readonlyTranClasses.get(avgAccessedItemInInteger);
					if(collection==null){
						collection=new Vector<TransactionClass>();
						readonlyTranClasses.put(avgAccessedItemInInteger, collection);
					}
					collection.add(tcl);
				} else{
					Vector<TransactionClass> collection=readWriteTranClasses.get(avgAccessedItemInInteger);
					if(collection==null){
						collection=new Vector<TransactionClass>();
						readWriteTranClasses.put(avgAccessedItemInInteger, collection);
					}
					collection.add(tcl);
				}
				
				writerTranClasses.write("["+tcl.classname+"]");
				writerTranClasses.newLine();
				writerTranClasses.write("Average Response Time="+avgRsTime);
				writerTranClasses.newLine();
				writerTranClasses.write("Average Accessed Item="+avgAccessedItem);
				writerTranClasses.newLine();
				writerTranClasses.write("Total Number of transactions="+tcl.totalNum);
				writerTranClasses.newLine();
			}
		
			for(Integer num:readonlyTranClasses.keySet()){
				Vector<TransactionClass> collection=readonlyTranClasses.get(num);
				int totalNum=0;
				for(TransactionClass tcl:collection){
					double avgRsTime=(double)tcl.totalResponseTime/tcl.totalNum;
					totalNum+=tcl.totalNum;
					writerTranClassesROPlot.write(num+" "+avgRsTime+ " #"+tcl.classname);
					writerTranClassesROPlot.newLine();
				}
				writerTranClassesRONumPlot.write(num+" "+totalNum);
				writerTranClassesRONumPlot.newLine();
			}
			for(Integer num:readWriteTranClasses.keySet()){
				Vector<TransactionClass> collection=readWriteTranClasses.get(num);
				int totalNum=0;
				for(TransactionClass tcl:collection){
					double avgRsTime=(double)tcl.totalResponseTime/tcl.totalNum;
					totalNum+=tcl.totalNum;
					writerTranClassesRWPlot.write(num+" "+avgRsTime+ " #"+tcl.classname);
					writerTranClassesRWPlot.newLine();
				}
				writerTranClassesRWNumPlot.write(num+" "+totalNum);
				writerTranClassesRWNumPlot.newLine();
			}
			writerTranClassesROPlot.flush();
			writerTranClassesROPlot.close();
			writerTranClassesRWPlot.flush();
			writerTranClassesRWPlot.close();
			writerTranClassesRONumPlot.flush();
			writerTranClassesRONumPlot.close();
			writerTranClassesRWNumPlot.flush();
			writerTranClassesRWNumPlot.close();
			writerTranClasses.flush();
			writerTranClasses.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	
	private void reduceClientTransaction_tranCount()
	{
		//number: number of accessed data items
		//count: number of transactions
		
		int[] nodes=sysConf.nodes.getClientNodes();
		String fileNameRW="AllClientTransaction_RWTranCount";
		String fileNameRO="AllClientTransaction_ROTranCount";
		String fileNameAll="AllClientTransaction_AllTranCount";

		TreeMap<Integer, Integer> rwTranCount=new TreeMap<Integer, Integer>();
		TreeMap<Integer, Integer> roTranCount=new TreeMap<Integer, Integer>();
		
		try {
			BufferedWriter writerRWTranCount=new BufferedWriter(new FileWriter(outPath+fileNameRW,false));
			BufferedWriter writerROTranCount=new BufferedWriter(new FileWriter(outPath+fileNameRO,false));
			BufferedWriter writerAllTranCount=new BufferedWriter(new FileWriter(outPath+fileNameAll,false));
			
			for(int i=0;i<nodes.length;i++){
				int nodeID=nodes[i];
				String filenameRW="clientTran"+nodeID+".log_RWTranCount";
				FileInputStream fstream;
				fstream = new FileInputStream(inPath+filenameRW);
				BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
				
				String paras[]=getData(in,"=");
				while(paras!=null){
					if(paras.length==2){
						int number=Integer.parseInt(paras[0]);
						int count=Integer.parseInt(paras[1]);
						Integer curCount= rwTranCount.get(number);
						if(curCount==null){
							curCount=count;
						} else {
							curCount+=count;
						}
						rwTranCount.put(number, curCount);
					}
					paras=getData(in,"=");
				}
				
				String filenameRO="clientTran"+nodeID+".log_ROTranCount";
				fstream = new FileInputStream(inPath+filenameRO);
				in = new BufferedReader(new InputStreamReader(fstream));
				
				paras=getData(in,"=");
				while(paras!=null){
					if(paras.length==2){
						int number=Integer.parseInt(paras[0]);
						int count=Integer.parseInt(paras[1]);
						Integer curCount= roTranCount.get(number);
						if(curCount==null){
							curCount=count;
						} else {
							curCount+=count;
						}
						roTranCount.put(number, curCount);
					}
					paras=getData(in,"=");
				}
			}
			
			int totalRWCount=0;
			int totalROCount=0;
			int totalAllCount=0;
			int totalRWNumber=0;
			int totalRONumber=0;
			int totalAllNumber=0;
			
			for(int number:rwTranCount.keySet()){
				int count=rwTranCount.get(number);
				writerRWTranCount.write(number+" "+count );
				writerRWTranCount.newLine();
				totalRWCount+=count;
				totalRWNumber+=number*count;
			}
			
			for(int number:roTranCount.keySet()){
				int count=roTranCount.get(number);
				writerROTranCount.write(number+" "+count );
				writerROTranCount.newLine();
				totalROCount+=count;
				totalRONumber+=number*count;
			}
				
			totalAllCount=totalROCount+totalRWCount;
			totalAllNumber=totalRONumber+totalRWNumber;
		
			double avgRWNumber=(double)totalRWNumber/(double)totalRWCount; 
			double avgRONumber=(double)totalRONumber/(double)totalROCount;
			double avgAllNumber=(double)totalAllNumber/(double)totalAllCount;
			
			writerAllTranCount.write("Average accessed data items for RW Tran=" + avgRWNumber);
			writerAllTranCount.newLine();
			writerAllTranCount.write("Average accessed data items for RO Tran=" + avgRONumber);
			writerAllTranCount.newLine();
			writerAllTranCount.write("Average accessed data items for All Tran=" + avgAllNumber);
			writerAllTranCount.newLine();
			
			writerROTranCount.close();
			writerAllTranCount.close();
			writerRWTranCount.close();
			
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void reduceClientTransaction_Rs()
	{
		String outfileName="AllClientTransaction_Rs";
		int[] nodes=sysConf.nodes.getClientNodes();
		try {
			BufferedWriter writer=new BufferedWriter(new FileWriter(outPath+outfileName,false));
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="clientTran"+nodeID+".log_Rs";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					String line;
					while((line=in.readLine())!=null)
						{
						writer.write(line);
						writer.newLine();
						}
					in.close();
				}
			writer.flush();
			writer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void reduceClientTransaction_AbortedThroughput()
	{
		String outfileName="AllClientTransaction_AbortedThroughput";
		int[] nodes=sysConf.nodes.getClientNodes();
		try {
			TreeMap<Integer, Integer> table=new TreeMap<Integer, Integer>();
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="clientTran"+nodeID+".log_ThroughputofAborted";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					
					String paras[]=getData(in," ");
					while(paras!=null)
						{
							if(paras.length==2)
								{
									Integer key=Integer.valueOf(paras[0]);
									Integer value=Integer.valueOf(paras[1]);
									Integer curValue=table.get(key);
									if(curValue==null)
										table.put(key, value);
									else
										{
										curValue+=value;
										table.put(key, curValue);
										}
								}
							paras=getData(in," ");
						}
				}
			BufferedWriter writer=new BufferedWriter(new FileWriter(outPath+outfileName,false));
			Iterator<Integer> keys= table.keySet().iterator();
			while(keys.hasNext())
				{
				Integer key=keys.next();
				Integer value=table.get(key);
				writer.write(key +" "+value);
				writer.newLine();
				}
			
			writer.flush();
			writer.close();
	
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void reduceClientTransaction_throughput()
	{
		String outfileName="AllClientTransaction_Throughput";
		int[] nodes=sysConf.nodes.getClientNodes();
		try {
			TreeMap<Integer, Integer> table=new TreeMap<Integer, Integer>();
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="clientTran"+nodeID+".log_Throughputbyend";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					
					String paras[]=getData(in," ");
					while(paras!=null)
						{
							if(paras.length==2)
								{
									Integer key=Integer.valueOf(paras[0]);
									Integer value=Integer.valueOf(paras[1]);
									Integer curValue=table.get(key);
									if(curValue==null)
										table.put(key, value);
									else
										{
										curValue+=value;
										table.put(key, curValue);
										}
								}
							paras=getData(in," ");
						}
				}
			BufferedWriter writer=new BufferedWriter(new FileWriter(outPath+outfileName,false));
			Iterator<Integer> keys= table.keySet().iterator();
			while(keys.hasNext())
				{
				Integer key=keys.next();
				Integer value=table.get(key);
				writer.write(key +" "+value);
				writer.newLine();
				}
			
			writer.flush();
			writer.close();
	
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	 private void reduceClientTransaction_RsInSlot()
		{
		 String outfileName="AllClientTransaction_RsInSlot";
		 int[] nodes=sysConf.nodes.getClientNodes();
			try {
				TreeMap<Integer, Double> table=new TreeMap<Integer, Double>();
				for(int i=0;i<nodes.length;i++)
					{
						int nodeID=nodes[i];
						String filename="clientTran"+nodeID+".log_RsInSlot";
						FileInputStream fstream;
						fstream = new FileInputStream(inPath+filename);
						BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
						
						String paras[]=getData(in," ");
						while(paras!=null)
							{
								if(paras.length==2)
									{
										Integer key=Integer.valueOf(paras[0]);
										Double value=Double.valueOf(paras[1]);
										Double curValue=table.get(key);
										if(curValue==null){
											curValue=value*CT_numOfTranCommitted[i];
											table.put(key, curValue);
										}
										else
											{
											curValue+=value*CT_numOfTranCommitted[i];
											table.put(key, curValue);
											}
									}
								paras=getData(in," ");
							}
					}
				BufferedWriter writer=new BufferedWriter(new FileWriter(outPath+outfileName,false));
				Iterator<Integer> keys= table.keySet().iterator();
				while(keys.hasNext())
					{
					Integer key=keys.next();
					Double value=table.get(key)/CT_totalTran;
					writer.write(key +" "+value);
					writer.newLine();
					}
				
				writer.flush();
				writer.close();
		
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	 
	private void reduceClientTransaction_overview()
		{
			int[] nodes=sysConf.nodes.getClientNodes();
			CT_numOfTranCommitted=new int[nodes.length];
			long ExperimentInterval=0;
			int totalTransactionExecuted=0;
			int totalTransactionCommitted=0;
			int numOfCommitted=0;
			double totalResponseTime=0;
			double averageResponseTime=0;
			int numOfWithinSLA=0;
			int numOfOutOfSLA=0;
			String fileNameOverview="AllClientTransaction_Overview";
			try {
				for(int i=0;i<nodes.length;i++)
					{
						int nodeID=nodes[i];
						String filename="clientTran"+nodeID+".log_Overview";
						FileInputStream fstream;
						fstream = new FileInputStream(inPath+filename);
						BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
						String paras[]=getData(in,"=");
						while(paras!=null)
							{
								if(paras.length==2)
									{
										if(paras[0].compareTo("ExperimentInterval")==0)
											{
											long theInterval=Long.valueOf(paras[1]);
											if(ExperimentInterval<theInterval)
												ExperimentInterval=theInterval; //take the largest nuber
											}
										else if(paras[0].compareTo("totalTransactionExecuted")==0)
											totalTransactionExecuted+=Integer.valueOf(paras[1]);
										else if(paras[0].compareTo("totalTransactionCommitted")==0)
											{
											this.CT_numOfTranCommitted[i]=numOfCommitted=Integer.valueOf(paras[1]);
											totalTransactionCommitted+=Integer.valueOf(paras[1]);
											}
										else if(paras[0].compareTo("averageResponseTime")==0)
											averageResponseTime=Double.valueOf(paras[1]);
										else if(paras[0].compareTo("numOfWithinSLA")==0)
											numOfWithinSLA+=Integer.valueOf(paras[1]);
										else if(paras[0].compareTo("numOfOutOfSLA")==0)
											numOfOutOfSLA+=Integer.valueOf(paras[1]);
									}
								paras=getData(in,"=");
							}
						totalResponseTime+=averageResponseTime*numOfCommitted;
					}
				BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
				double percentageOfCommit=((double)totalTransactionCommitted/(double)totalTransactionExecuted)*100;
				double commitThroughput=(double)totalTransactionCommitted/((double)ExperimentInterval/1000);
				double totalThroughput=(double)totalTransactionExecuted/((double)ExperimentInterval/1000);
				double avgResponseTime=(double)totalResponseTime/(double)totalTransactionCommitted;
				this.CT_totalTran=totalTransactionCommitted;
				
				double percentageWithinSLA=(double)numOfWithinSLA/(double)(numOfWithinSLA+numOfOutOfSLA);
				
				writerOverview.write("ExperimentInterval="+ExperimentInterval/1000 + "Seconds");
				writerOverview.newLine();
				writerOverview.write("percentageWithinSLA="+percentageWithinSLA *100+ "%");
				writerOverview.newLine();
				writerOverview.write("totalTransactionExecuted="+ totalTransactionExecuted);
				writerOverview.newLine();
				writerOverview.write("totalTransactionCommitted="+ totalTransactionCommitted);
				writerOverview.newLine();
				writerOverview.write("percentageOfCommit="+percentageOfCommit  + "%");
				writerOverview.newLine();
				writerOverview.write("commitThroughput="+commitThroughput + "transaction per second");
				writerOverview.newLine();
				writerOverview.write("totalThroughput="+totalThroughput+ "transaction per second");
				writerOverview.newLine();
				writerOverview.write("averageResponseTime="+avgResponseTime + " ms");
				writerOverview.newLine();
				
				
				writerOverview.flush();
				writerOverview.close();
		
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	
	public void reduceDataServiceLog()
	{
		reduceDataService_overview();
		reduceDataService_RsInSlot();
		reduceDataService_throughput();
		reduceDataService_Rs();
	}
	
	private void reduceDataService_Rs()
	{
		String outfileName="AllDataService_Rs";
		int[] nodes=sysConf.nodes.getClientNodes();
		try {
			BufferedWriter writer=new BufferedWriter(new FileWriter(outPath+outfileName,false));
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="ds"+nodeID+".log_Rs";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					String line;
					while((line=in.readLine())!=null)
						{
						writer.write(line);
						writer.newLine();
						}
					in.close();
				}
			writer.flush();
			writer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}
	private void reduceDataService_throughput()
	{
		String outfileName="AllDataService_Throughput";
		int[] nodes=sysConf.nodes.getClientNodes();
		try {
			TreeMap<Integer, Integer> table=new TreeMap<Integer, Integer>();
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="ds"+nodeID+".log_Throughputbyend";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					
					String paras[]=getData(in," ");
					while(paras!=null)
						{
							if(paras.length==2)
								{
									Integer key=Integer.valueOf(paras[0]);
									Integer value=Integer.valueOf(paras[1]);
									Integer curValue=table.get(key);
									if(curValue==null)
										table.put(key, value);
									else
										{
										curValue+=value;
										table.put(key, curValue);
										}
								}
							paras=getData(in," ");
						}
				}
			BufferedWriter writer=new BufferedWriter(new FileWriter(outPath+outfileName,false));
			Iterator<Integer> keys= table.keySet().iterator();
			while(keys.hasNext())
				{
				Integer key=keys.next();
				Integer value=table.get(key);
				writer.write(key +" "+value);
				writer.newLine();
				}
			
			writer.flush();
			writer.close();
	
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private int[] CT_numOfTranCommitted;
	private int CT_totalTran;
	
	 private void reduceDataService_RsInSlot()
		{
		 String outfileName="AllDataService_RsInSlot";
		 int[] nodes=sysConf.nodes.getClientNodes();
			try {
				TreeMap<Integer, Double> table=new TreeMap<Integer, Double>();
				for(int i=0;i<nodes.length;i++)
					{
						int nodeID=nodes[i];
						String filename="ds"+nodeID+".log_RsInSlot";
						FileInputStream fstream;
						fstream = new FileInputStream(inPath+filename);
						BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
						
						String paras[]=getData(in," ");
						while(paras!=null)
							{
								if(paras.length==2)
									{
										Integer key=Integer.valueOf(paras[0]);
										Double value=Double.valueOf(paras[1]);
										Double curValue=table.get(key);
										if(curValue==null)
											table.put(key, value);
										else
											{
											curValue+=value*CT_numOfTranCommitted[i];
											table.put(key, curValue);
											}
									}
								paras=getData(in," ");
							}
					}
				BufferedWriter writer=new BufferedWriter(new FileWriter(outPath+outfileName,false));
				Iterator<Integer> keys= table.keySet().iterator();
				while(keys.hasNext())
					{
					Integer key=keys.next();
					Double value=table.get(key)/CT_totalTran;
					writer.write(key +" "+value);
					writer.newLine();
					}
				
				writer.flush();
				writer.close();
		
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	 
	private void reduceDataService_overview()
		{
			int[] nodes=sysConf.nodes.getClientNodes();
			CT_numOfTranCommitted=new int[nodes.length];
			long ExperimentInterval=0;
			int totalTransactionExecuted=0;
			int totalTransactionCommitted=0;
			int numOfCommitted=0;
			double totalResponseTime=0;
			double averageResponseTime=0;
			String fileNameOverview="AllDataService_Overview";
			try {
				for(int i=0;i<nodes.length;i++)
					{
						int nodeID=nodes[i];
						String filename="ds"+nodeID+".log_Overview";
						FileInputStream fstream;
						fstream = new FileInputStream(inPath+filename);
						BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
						String paras[]=getData(in,"=");
						while(paras!=null)
							{
								if(paras.length==2)
									{
										if(paras[0].compareTo("ExperimentInterval")==0)
											ExperimentInterval=Long.valueOf(paras[1]);
										else if(paras[0].compareTo("totalTransactionExecuted")==0)
											totalTransactionExecuted+=Integer.valueOf(paras[1]);
										else if(paras[0].compareTo("totalTransactionCommitted")==0)
											{
											this.CT_numOfTranCommitted[i]=numOfCommitted=Integer.valueOf(paras[1]);
											totalTransactionCommitted+=Integer.valueOf(paras[1]);
											}
										else if(paras[0].compareTo("averageResponseTime")==0)
											averageResponseTime=Double.valueOf(paras[1]);
									}
								paras=getData(in,"=");
							}
						totalResponseTime+=averageResponseTime*numOfCommitted;
					}
				BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
				double percentageOfCommit=((double)totalTransactionCommitted/(double)totalTransactionExecuted)*100;
				double commitThroughput=(double)totalTransactionCommitted/((double)ExperimentInterval/1000);
				double totalThroughput=(double)totalTransactionExecuted/((double)ExperimentInterval/1000);
				double avgResponseTime=(double)totalResponseTime/(double)totalTransactionCommitted;
				this.CT_totalTran=totalTransactionCommitted;
				
				writerOverview.write("ExperimentInterval="+ExperimentInterval/1000 + "Seconds");
				writerOverview.newLine();
				writerOverview.write("totalTransactionExecuted="+ totalTransactionExecuted);
				writerOverview.newLine();
				writerOverview.write("totalTransactionCommitted="+ totalTransactionCommitted);
				writerOverview.newLine();
				writerOverview.write("percentageOfCommit="+percentageOfCommit  + "%");
				writerOverview.newLine();
				writerOverview.write("commitThroughput="+commitThroughput + "transaction per second");
				writerOverview.newLine();
				writerOverview.write("totalThroughput="+totalThroughput+ "transaction per second");
				writerOverview.newLine();
				writerOverview.write("averageResponseTime="+avgResponseTime + " ms");
				writerOverview.newLine();
				
				writerOverview.flush();
				writerOverview.close();
		
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	
	
	public void reduceIDServiceLog()
	{
		reduceIDServiceLog_overview();
	}
	private void reduceIDServiceLog_overview()
	{
		int[] nodes=sysConf.nodes.getServletNodes(null);
		long ExperimentInterval=0;
		int totalTransactionExecuted=0;
		int transactionExecuted=0;
		double totalResponseTime=0;
		double averageResponseTime=0;
		String fileNameOverview="AllIDService_Overview";
		try {
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="idservice"+nodeID+".log_Overview";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					String paras[]=getData(in,"=");
					while(paras!=null)
						{
							if(paras.length==2)
								{
									if(paras[0].compareTo("ExperimentInterval")==0)
										ExperimentInterval=Long.valueOf(paras[1]);
									else if(paras[0].compareTo("totalTransactionExecuted")==0)
										totalTransactionExecuted+=(transactionExecuted=Integer.valueOf(paras[1]));
									else if(paras[0].compareTo("averageResponseTime")==0)
										averageResponseTime=Double.valueOf(paras[1]);
								}
							paras=getData(in,"=");
						}
					totalResponseTime+=averageResponseTime*transactionExecuted;
				}
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			double totalThroughput=(double)totalTransactionExecuted/((double)ExperimentInterval/1000);
			double avgResponseTime=(double)totalResponseTime/(double)totalTransactionExecuted;
			
			writerOverview.write("ExperimentInterval="+ExperimentInterval/1000 + "Seconds");
			writerOverview.newLine();
			writerOverview.write("totalTransactionExecuted="+ totalTransactionExecuted);
			writerOverview.newLine();
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput+ "transaction per second");
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgResponseTime + " ms");
			writerOverview.newLine();
			
			writerOverview.flush();
			writerOverview.close();
	
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	
	
	public void reduceTPNodeLog()
	{
		reduceTPNodeLog_overview();
		reduceTPNodeLog_ConcurrencySlot();
		reduceTPNodeLog_throughput();
		//reduceTPNodeLog_rs();
	}
	private int[] TP_numOfTranCommitted;
	private int TP_totalTran;
	

	private void reduceTPNodeLog_throughput()
	{
		int[] nodes=sysConf.nodes.getServletNodes(null);
		int[] numOfTrans=new int[nodes.length];
		String outFileName="AllTPNode_throughput";
		try {
			TreeMap<Integer, Integer[]> table=new TreeMap<Integer, Integer[]>();
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="tpnode"+nodeID+".log_Throughputbyend";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					
					String paras[]=getData(in," ");
					while(paras!=null)
						{
							if(paras.length==3)
								{
									Integer key=Integer.valueOf(paras[0]);
									Integer[] value=new Integer[2];
									value[0]=Integer.valueOf(paras[1]);
									value[1]=Integer.valueOf(paras[2]);
									Integer[] curValue=table.get(key);
									if(curValue==null)
										table.put(key, value);
									else
										{
										curValue[0]+=value[0];
										curValue[1]+=value[1];
										table.put(key, curValue);
										}
								}
							paras=getData(in," ");
						}
				}
			BufferedWriter writer=new BufferedWriter(new FileWriter(outPath+outFileName,false));
			writer.write("#TimeStamp ThroughputOfTCNodeOnly ThroughputWithAllTPnodeInvoked");
			writer.newLine();
			Iterator<Integer> keys= table.keySet().iterator();
			while(keys.hasNext())
				{
				Integer key=keys.next();
				Integer[] value=table.get(key);
				writer.write(key +" "+value[0] + " "+value[1]);
				writer.newLine();
				}
			
			writer.flush();
			writer.close();
	
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	private void reduceTPNodeLog_ConcurrencySlot()
	{
		int[] nodes=sysConf.nodes.getServletNodes(null);
		int[] numOfTrans=new int[nodes.length];
		for(int c:numOfTrans)
			c=0;
		String outFileName="AllTPNode_ConcurrencyInSlot";
		try {
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="tpnode"+nodeID+".log_ConcurrencyInSlot";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					String paras[]=getData(in,"=");
					while(paras!=null)
						{
							if(paras.length==2)
								{
									for(int j=0;j<nodes.length;j++)
										{
											if(paras[0].compareTo("NumOfTransactionInvoke"+(j+1)+"TPNode")==0)
												{
												int num;
												num=Integer.valueOf(paras[1]);	
												numOfTrans[j]+=num;
												break;
												}
										}
								}
							paras=getData(in,"=");
						}
				}
			BufferedWriter writer=new BufferedWriter(new FileWriter(outPath+outFileName,false));
			
			writer.write("#NumOfTPNodeInvolved AmountOfTransactions");
			writer.newLine();
			for(int j=0;j<nodes.length;j++)
			{
				writer.write((j+1)+" "+numOfTrans[j]);
				writer.newLine();
			}
			
			writer.flush();
			writer.close();
			
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	private void reduceTPNodeLog_overview()
	{
		int[] nodes=sysConf.nodes.getServletNodes(null);
		TP_numOfTranCommitted=new int[nodes.length];
		int[] invokedTimes=new int[sysConf.nodes.getTotalNumberOfNodes()];
		int[] invokedForCommitTimes=new int[sysConf.nodes.getTotalNumberOfNodes()];
		int[] invokedTimesasTC=new int[sysConf.nodes.getTotalNumberOfNodes()];
		int[] invokedTimesasTCWhenCommit=new int[sysConf.nodes.getTotalNumberOfNodes()];
		for(int c:invokedForCommitTimes)
			c=0;
		for(int c:invokedTimes)
			c=0;
		for(int c:invokedTimesasTC)
			c=0;
		for(int c:invokedTimesasTCWhenCommit)
			c=0;
		
		long ExperimentInterval=0;
		int totalTransactionExecuted=0;
		int totalTransactionCommitted=0;
		int numOfCommitted=0;
		double totalResponseTime=0;
		double averageResponseTime=0;
		
		double avgSubmitResponseTime=0;
		double avgWaitResponseTime=0;
		double avgCommitResponseTime=0;
		double avgAbortResponseTime=0.0;
		double avgRestartTimes=0.0;
		
		double totalSubmitResponseTime=0;
		double totalWaitResponseTime=0;
		double totalCommitResponseTime=0;
		double totalAbortResponseTime=0.0;
		double totalRestartTimes=0.0;
		
		String fileNameOverview="AllTPNode_Overview";
		String fileNameInvokeTimes="AllTPNode_InvokeTimes";
		try {
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="tpnode"+nodeID+".log_Overview";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					String paras[]=getData(in,"=");
					while(paras!=null)
						{
							if(paras.length==2)
								{
									if(paras[0].compareTo("ExperimentInterval")==0)
										ExperimentInterval=Long.valueOf(paras[1]);
									else if(paras[0].compareTo("totalTransactionExecuted")==0)
										{
										totalTransactionExecuted+=Integer.valueOf(paras[1]);
										invokedTimesasTC[nodes[i]]=Integer.valueOf(paras[1]);
										}
									else if(paras[0].compareTo("totalTransactionCommitted")==0)
										{
										this.TP_numOfTranCommitted[i]=numOfCommitted=Integer.valueOf(paras[1]);
										totalTransactionCommitted+=Integer.valueOf(paras[1]);
										invokedTimesasTCWhenCommit[nodes[i]]=Integer.valueOf(paras[1]);
										}
									else if(paras[0].compareTo("averageResponseTime")==0)
										averageResponseTime=Double.valueOf(paras[1]);
									else if(paras[0].compareTo("avgSubmitResponseTime")==0)
										avgSubmitResponseTime=Double.valueOf(paras[1]);
									else if(paras[0].compareTo("avgWaitResponseTime")==0)
										avgWaitResponseTime=Double.valueOf(paras[1]);
									else if(paras[0].compareTo("avgCommitResponseTime")==0)
										avgCommitResponseTime=Double.valueOf(paras[1]);
									else if(paras[0].compareTo("avgAbortResponseTime")==0)
										avgAbortResponseTime=Double.valueOf(paras[1]);
									else if(paras[0].compareTo("avgRestartTimes")==0)
										avgRestartTimes=Double.valueOf(paras[1]);
									else
										{
											for(int j=0;j<nodes.length;j++)
												{
													if(paras[0].compareTo("node"+nodes[j]+"InvokedForCommit")==0)
														{
														int num;
														num=Integer.valueOf(paras[1]);	
														invokedForCommitTimes[nodes[j]]+=num;
														break;
														}
													else if(paras[0].compareTo("node"+nodes[j]+"Invoked")==0)
														{
														int num;
														num=Integer.valueOf(paras[1]);
														invokedTimes[nodes[j]]+=num;
														break;
														}
												}
										}
								}
							paras=getData(in,"=");
						}
					totalResponseTime+=averageResponseTime*numOfCommitted;
					totalSubmitResponseTime+=avgSubmitResponseTime*numOfCommitted;
					totalWaitResponseTime+=avgWaitResponseTime*numOfCommitted;
					totalCommitResponseTime+=avgCommitResponseTime*numOfCommitted;
					totalAbortResponseTime+=avgAbortResponseTime*numOfCommitted;
					totalRestartTimes+=avgRestartTimes*numOfCommitted;
				}
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			double percentageOfCommit=((double)totalTransactionCommitted/(double)totalTransactionExecuted)*100;
			double commitThroughput=(double)totalTransactionCommitted/((double)ExperimentInterval/1000);
			double totalThroughput=(double)totalTransactionExecuted/((double)ExperimentInterval/1000);
			double avgResponseTime=(double)totalResponseTime/(double)totalTransactionCommitted;
			avgSubmitResponseTime=(double)totalSubmitResponseTime/(double)totalTransactionCommitted;
			avgWaitResponseTime=(double)totalWaitResponseTime/(double)totalTransactionCommitted;
			avgCommitResponseTime=(double)totalCommitResponseTime/(double)totalTransactionCommitted;
			avgAbortResponseTime=(double)totalAbortResponseTime/(double)totalTransactionCommitted;
			avgRestartTimes=(double)totalRestartTimes/(double)totalTransactionCommitted;
			this.TP_totalTran=totalTransactionCommitted;
			
			writerOverview.write("ExperimentInterval="+ExperimentInterval/1000 + "Seconds");
			writerOverview.newLine();
			writerOverview.write("totalTransactionExecuted="+ totalTransactionExecuted);
			writerOverview.newLine();
			writerOverview.write("totalTransactionCommitted="+ totalTransactionCommitted);
			writerOverview.newLine();
			writerOverview.write("percentageOfCommit="+percentageOfCommit  + "%");
			writerOverview.newLine();
			writerOverview.write("commitThroughput="+commitThroughput + "transaction per second");
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput+ "transaction per second");
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgResponseTime + " ms");
			writerOverview.newLine();
			writerOverview.write("avgSubmitResponseTime="+avgSubmitResponseTime + " ms");
			writerOverview.newLine();
			writerOverview.write("avgWaitResponseTime="+avgWaitResponseTime + " ms");
			writerOverview.newLine();
			writerOverview.write("avgCommitResponseTime="+avgCommitResponseTime + " ms");
			writerOverview.newLine();
			writerOverview.write("avgAbortResponseTime="+avgAbortResponseTime + " ms");
			writerOverview.newLine();
			writerOverview.write("avgRestartTimes="+avgRestartTimes + " ms");
			writerOverview.newLine();
			writerOverview.flush();
			writerOverview.close();
			
			BufferedWriter writerInvokeTimes=new BufferedWriter(new FileWriter(outPath+fileNameInvokeTimes,false));
			writerInvokeTimes.write("#NodeID InvokedTimesAsTC InvokedTimesAsTCwhenCommitted invokedTimesAsAnyRole invokedTimesAsAnyRoleWhenCommitted");
			writerInvokeTimes.newLine();
			
			for(int j=0;j<nodes.length;j++)
				{
				writerInvokeTimes.write(nodes[j]+" "+invokedTimesasTC[nodes[j]]+" "+invokedTimesasTCWhenCommit[nodes[j]]+" "+invokedForCommitTimes[nodes[j]] + " " +invokedTimes[nodes[j]]);
				writerInvokeTimes.newLine();
				}
			writerInvokeTimes.flush();
			writerInvokeTimes.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	
	
	
	
	public void reduceLoadDataLog()
	{}
	
	public void reduceCheckPointLog()
	{
		reduceCheckPoint_overview();
	}
	private int[] CP_numOfTranCommitted;
	private int CP_totalTran;
	private void reduceCheckPoint_overview()
	{
		int[] nodes=sysConf.nodes.getServletNodes(null);
		CP_numOfTranCommitted=new int[nodes.length];
		long ExperimentInterval=0;
		int totalTransactionExecuted=0;
		double totalThroughput=0;
		int transactionExecuted=0;
		double totalResponseTime=0;
		double averageResponseTime=0;
		double totalCopyResponseTime=0;
		double averageCopyResponseTime=0;
		double totalFlushResponseTime=0;
		double averageFlushResponseTime=0;
		String fileNameOverview="AllCheckPoint_Overview";
		try {
			for(int i=0;i<nodes.length;i++)
				{
					int nodeID=nodes[i];
					String filename="checkpoint"+nodeID+".log_Overview";
					FileInputStream fstream;
					fstream = new FileInputStream(inPath+filename);
					BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
					String paras[]=getData(in,"=");
					while(paras!=null)
						{
							if(paras.length==2)
								{
									if(paras[0].compareTo("ExperimentInterval")==0)
										ExperimentInterval=Long.valueOf(paras[1]);
									else if(paras[0].compareTo("totalTransactionExecuted")==0)
										totalTransactionExecuted+=(transactionExecuted=Integer.valueOf(paras[1]));
									else if(paras[0].compareTo("totalThroughput")==0)
										totalThroughput+=Double.valueOf(paras[1]);
									else if(paras[0].compareTo("averageResponseTime")==0)
										averageResponseTime=Double.valueOf(paras[1]);
									else if(paras[0].compareTo("avgCopyLatency")==0)
										averageCopyResponseTime=Double.valueOf(paras[1]);
									else if(paras[0].compareTo("avgFlushResponseTime")==0)
										averageFlushResponseTime=Double.valueOf(paras[1]);
								}
							paras=getData(in,"=");
						}
					totalResponseTime+=averageResponseTime*transactionExecuted;
					totalCopyResponseTime+=averageCopyResponseTime*transactionExecuted;
					totalFlushResponseTime+=averageFlushResponseTime*transactionExecuted;
				}
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			double avgResponseTime=(double)totalResponseTime/(double)totalTransactionExecuted;
			double avgCopyResponseTime=(double)totalCopyResponseTime/(double)totalTransactionExecuted;
			double avgFlushResponseTime=(double)totalFlushResponseTime/(double)totalTransactionExecuted;
			this.CP_totalTran=totalTransactionExecuted;
			
			writerOverview.write("ExperimentInterval="+ExperimentInterval/1000 + "Seconds");
			writerOverview.newLine();
			writerOverview.write("totalTransactionExecuted="+ totalTransactionExecuted);
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput+ " checkpoints per minute");
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgResponseTime + " ms");
			writerOverview.newLine();
			writerOverview.write("avgCopyResponseTime="+avgCopyResponseTime + " ms");
			writerOverview.newLine();
			writerOverview.write("avgFlushResponseTime="+avgFlushResponseTime + " ms");
			writerOverview.newLine();

			writerOverview.flush();
			writerOverview.close();
	
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
}
