import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Vector;
import java.util.concurrent.ArrayBlockingQueue;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.membership.msg.MembershipMsg.SendInvitationRequestMsg;
import org.zhouw.stps.membership.msg.MembershipMsg.ReceiveInvitationResponseMsg;
import org.zhouw.stps.membership.*;
import org.zhouw.stps.MsgBroker;

public class ReceiveInvitationService extends ManagedHttpServlet {

	@Override
	public void process(HttpServletRequest req, HttpServletResponse res) {
		
		InputStream in;
		OutputStream output;
		try {
			in=req.getInputStream();
		
			if(in!=null)
			{
				SendInvitationRequestMsg msg=SendInvitationRequestMsg.parseFrom(in);
				MsgBroker.cleanupInputStream(in);
				
				int version=msg.getVersion();
				int compareResult=MembershipManager.compareVersionNumber(version);
				if(compareResult==-1)
					{
					//inform the other side that our version is smaller
					ReceiveInvitationResponseMsg response=ReceiveInvitationResponseMsg.newBuilder().
					setVersion(MembershipManager.getVersionNumberWithoutLock()).setResult(0).build();
					output=res.getOutputStream();
					response.writeTo(output);
					output.flush();
					output.close();
					
					GlobalPara.logger.warning("Receiving membership version larger than local LTMs from LTM "+msg.getSrcID());
					MembershipManager.handleOutdatedMembership();
					}
				else if(compareResult==1)
					{
					GlobalPara.logger.warning("Receiving membership version smaller than local LTMs");
					
					//inform the other side that our version is larger
					ReceiveInvitationResponseMsg response=ReceiveInvitationResponseMsg.newBuilder().
					setVersion(MembershipManager.getVersionNumberWithoutLock()).setResult(0).build();
					output=res.getOutputStream();
					response.writeTo(output);
					output.flush();
					output.close();
					
					return;
					}
				
				int srcID=msg.getSrcID();
				Vector<Integer> curPartition=new Vector<Integer>();
				ArrayBlockingQueue<Integer> outputQueue =new ArrayBlockingQueue<Integer>(1);
				GlobalPara.memberComp.reportReceiveInvitation(srcID, curPartition, outputQueue);
				
				
				int result=0;
				try {
					result = outputQueue.take();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
					
//					MembershipManager.receiveInvitation(srcID, curPartition);
				
				
				ReceiveInvitationResponseMsg.Builder responseBuilder=ReceiveInvitationResponseMsg.newBuilder().setResult(result).
				setVersion(version);
				if(result==-1){
					responseBuilder.addAllCurPartition(curPartition);
				}
				ReceiveInvitationResponseMsg response=responseBuilder.build();
				output=res.getOutputStream();
				if(output!=null){
					response.writeTo(output);
					output.flush();
					output.close();
				}
			}
		} catch (IOException e) {
		e.printStackTrace();
		}
	}
}
