package org.zhouw.stps;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.Future;

import org.zhouw.stps.app.common.UpdateSubTransaction;
import org.zhouw.stps.msg.SubTransactionGroupProtos.STMsg;

public class DerivedTransaction {
	
	public List<STMsg> tempSubTrans;
	
	public HashMap<Integer, SubTransactionGroup> trans;
	public HashMap<Integer, Boolean> isCommitted;
	public HashMap<Integer, Boolean> isCompleted;
	public final int initID; //the lowest id of a STG
	public int curID;
	
	public SubTransactionGroup[] getSTGs()
	{
		if(trans==null)
			return null;
		SubTransactionGroup[] result=new SubTransactionGroup[trans.size()];
		int count=0;
		for(SubTransactionGroup stg:trans.values()){
			result[count++]=stg;
		}
		return result;
	}
	
	public String getDebugMsg()
	{
		String result="";
		if(this.trans!=null){
			result+="Count of Derived STGs:"+trans.size()+"\n";
			for(int index:trans.keySet()){
				SubTransactionGroup stg=trans.get(index);
				if(stg!=null){
					result+=stg.id+") derived STG key="+stg.key+ ", timestamp="+ stg.timestamp + 
					", destID="+ stg.TPnodeID+", isCommit="+this.isCommitted.get(index) + 
					", isCompleted="+ this.isCompleted.get(index)+"\n";
				} else {
					result+=index+") derived STG =null\n";
				}
			}
		} else {
			result+="Count of Derived STGs:null\n";
		}
		
		if(tempSubTrans!=null)
			result+="There are "+tempSubTrans.size() +" STMsgs in the list\n";
		
		
		return result;
	}
	
	public void reportDoAbortComplete(int id)
	{
		isCompleted.put(id, true);
	}
	
	public void reportDoCommitComplete(int id)
	{
		isCompleted.put(id, true);
//		if((isCompleted.get(id)!=null)) {
//			if(isCompleted.put(id, true))
//				GlobalPara.logger.warning("This subtransaction already report completed");
//		}
//		else
//			GlobalPara.logger.warning("This subtransaction does not exist in the derived transaction");
	}
	
	
	public Map<Integer, Future<Integer>> commitAll()
	{
		if(this.trans==null||this.trans.size()==0)
			return null;
		Map<Integer, Future<Integer>> result= new HashMap<Integer, Future<Integer>>();
		for(int i: this.trans.keySet()){
			SubTransactionGroup stg=this.trans.get(i);
			if(!this.isCompleted.get(i)){
				result.put(i, MsgBroker.DoCommit(stg.TPnodeID, stg.getTimeStamp(), stg.key));
			}
		}
		return result;
	}
	                       
	public boolean hasMoreNewDerivedTransactions()
	{
		if(tempSubTrans.size()==0)
			return false;
		else
			return true;
	}
	
	public void addDerivedSubTranMsg(STMsg st)
	{
		if(st!=null)
			this.tempSubTrans.add(st);
	}
	
	/**
	 * Create derived subtransactions and merge the ones which access the same data item
	 * The derived transactions can only be UpdateSubTransaction
	 * @param msgList
	 * @return A list of not duplicated subTrans, Empty List if none
	 */
	public static List<SubTransaction> genDerivedSubTranFromMsg(List<STMsg> msgList)
	{
		List<SubTransaction> stList=new LinkedList<SubTransaction>();
		if(msgList!=null){
			for(STMsg msg: msgList){
				SubTransaction st=SubTransaction.buildSubTransactionByMsg(msg);
				if(st!=null) {
					boolean isIncluded=false;
					for(SubTransaction curST:stList){
						if(curST.strDBName.compareTo(st.strDBName)==0&&
								curST.strTblName.compareTo(st.strTblName)==0&&
								curST.strRowKey.compareTo(st.strRowKey)==0)
						{
							UpdateSubTransaction updateCurST= (UpdateSubTransaction)curST;
							updateCurST.mergeFrom((UpdateSubTransaction)st);
							isIncluded=true;
							break;
						} 
					}
					if(!isIncluded)
						stList.add(st);
				}
			}
		}
			
		return stList;
	}
	
	public void clearTempSubTran()
	{
		this.tempSubTrans.clear();
	}
	
	public List<SubTransaction> collectNewDerivedSubTrans()
	{
		return DerivedTransaction.genDerivedSubTranFromMsg(this.tempSubTrans);
	}
	
	public Vector<Future<Integer>> abortAllBut(int id)
	{
		Vector<Future<Integer>> results=new Vector<Future<Integer>>();
		if(this.trans!=null)
		{
			for(SubTransactionGroup stg: this.trans.values())
			{
				if(stg.id!=id)
				{
					results.add(MsgBroker.DoAbort(stg.TPnodeID, stg.getTimeStamp(), stg.key));
				}
			}
		}
		return results;
	}
	
	/**
	 * Send abort message to all derived STGs and clear this data structure
	 * @return
	 */
	public Vector<Future<Integer>> abortAll()
	{
		Vector<Future<Integer>> results=new Vector<Future<Integer>>();
		if(this.trans!=null) {
			for(SubTransactionGroup stg: this.trans.values())
					results.add(MsgBroker.DoAbort(stg.TPnodeID, stg.getTimeStamp(), stg.key));
		}
		//this.clear();
		return results;
	}
	
	public Vector<Future<Integer>> abortAllIncompleted()
	{
		Vector<Future<Integer>> results=new Vector<Future<Integer>>();
		if(this.trans!=null) {
			for(SubTransactionGroup stg: this.trans.values()){
				if(!this.isCompleted.get(stg.id))
					results.add(MsgBroker.DoAbort(stg.TPnodeID, stg.getTimeStamp(), stg.key));
			}
		}
		return results;
	}
	
	public DerivedTransaction(int startID)
	{
		this.trans=new HashMap<Integer, SubTransactionGroup>();
		this.isCommitted=new HashMap<Integer, Boolean>();
		this.isCompleted=new HashMap<Integer, Boolean>();
		this.initID=startID;
		this.curID=startID;
		this.tempSubTrans=new LinkedList<STMsg>();
	}
	
	/**
	 * Assign the index to the derived STG, and generate the key of the STG
	 * @param stg
	 * @return
	 */
	public int addTransactionGroup(SubTransactionGroup stg)
	{
		int result;
		this.trans.put(curID, stg);
		stg.id=curID;
		this.isCommitted.put(curID, false);
		this.isCompleted.put(curID, false);
		result=curID;
		stg.genKey();
		curID++;
		return result;
	}
	
	public void clear()
	{
		this.trans.clear();
		this.curID=this.initID;
		this.isCommitted.clear();
		this.isCompleted.clear();
		this.tempSubTrans.clear();
	}
	
	public void reportCommit(int id)
	{
		if((isCommitted.get(id)!=null))
		{
			if(isCommitted.put(id, true))
				GlobalPara.logger.warning("This subtransaction already report committed");
		}
		else
			GlobalPara.logger.warning("This subtransaction does not exist in the derived transaction");
	}
	
	public boolean isAllCompleted()
	{
		boolean result=true;
		for(Integer id: trans.keySet()) {
			Boolean t=this.isCompleted.get(id);
			if(t==null||!t)
				return false;
		}
		return result;
	}
	
	public boolean isAllCommitted()
	{
		boolean result=true;
		for(Integer id: trans.keySet()) {
			boolean t=this.isCommitted.get(id);
			if(!t)
				return false;
		}
		return result;
	}
}
