package org.zhouw.stps;

import java.util.logging.*;
import java.util.concurrent.LinkedBlockingDeque;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.hbase.HBaseConfiguration;

import org.zhouw.stps.app.DataTable;
import org.zhouw.stps.components.*;
import org.zhouw.stps.conf.SystemConfigure;
import org.zhouw.stps.conf.NodeConfig.role;
import org.zhouw.stps.id.MonitorIDTask;
import org.zhouw.stps.master.MasterServer;
import org.zhouw.stps.membership.MembershipManager;
import org.zhouw.stps.memory.MonitorMemoryTask;
import org.zhouw.stps.monitor.PerformanceMonitor;
import org.zhouw.stps.tp.PushUpdatesToReplicasTask;

import java.util.concurrent.*;

	
public class GlobalPara {
	public static final Logger logger = getLogger();
	public static final Logger outputLog = getLoggerForPrint();

	public static final SystemConfigure sysConf=new SystemConfigure("../expConf/");
	//public static final SystemConfigure sysConf=new SystemConfigure("/home/expConf/");
	//public static final SystemConfigure sysConf=new SystemConfigure("G:/expConf/");

//	public static final String hadoopCfgPath=sysConf.hadoopCfgPath;
//    public static final String hbaseCfgPath=sysConf.hbaseCfgPath;
    
	public static HBaseConfiguration hbaseConf = initHbaseConf();
	
	public final static int CPBGThreadPoolSize=sysConf.CPBGThreadPoolSize;
	public final static int TPNodeBGThreadPoolSize=sysConf.TPNodeBGThreadPoolSize;
    
    public final static PerformanceMonitor monitor=new PerformanceMonitor(sysConf.outPerfLogPath);
    
    public static final long testDuration = sysConf.testDuration;
    public static final long sleepTime = sysConf.sleepTime;
    public static final int numOfClients = sysConf.numOfClients;
    public static final int timeoutToRetrySendMsg = sysConf.timeoutToRetrySendMsg;
    
    /**
     * the total number of nodes in the zone
     */
    public static final int numOfReplicas = sysConf.numOfReplicas;
    public static final int numOfVirtualNodes = sysConf.numOfVirtualNodes;
    public static final int rangeOfHash = sysConf.rangeOfHash;
    
    public static final long intervalOfSnapshot = sysConf.intervalOfSnapshot;
    
//    public static int[] backupNodes=getBackupNodes();
//    public static int[] primaryNodes=getPrimaryNodes();
    
    //public static int autoCheckpoint=0;//initAutoCheckPoint();
    //public static int autoCheckpoint=initAutoCheckPoint();
    
    public static MasterServer master=new MasterServer();
    
    
    //////////////////////////////////////////////////////////
    // initializing components
    public static int initComponents=initAllComponents();
    
    public static RWGTMComponent rwComp;
    public static ROGTMComponent roComp;
    public static LTMComponent ltmComp;
    public static MembershipComponent memberComp;
    public static AdapterCoordinator adapterComp;
    public static ClientComponent clientComp;
    
    public static int initAllComponents()
    {
    	ThreadGroup g = new ThreadGroup("Core");
    	//g.setMaxPriority(Thread.MAX_PRIORITY);
    	memberComp=new MembershipComponent();
    	initOneComponent(g,memberComp);
    	
    	role[] a= GlobalPara.sysConf.nodes.getNodeRole(GlobalPara.sysConf.getLocalNodeID());
    	for(int i=0;i<a.length;i++){
    		if(a[i]==role.SERVLET){
    	    	rwComp=new RWGTMComponent();
    	    	initOneComponent(g,rwComp);
    	    	roComp=new ROGTMComponent();
    	    	initOneComponent(g,roComp);
    	    	ltmComp=new LTMComponent();
    	    	initOneComponent(g,ltmComp);
    	    	adapterComp=new AdapterCoordinator(g);
    	    	//initOneComponent(g,adapterComp);
    		} else if(a[i]==role.CLIENT){
    			clientComp=new ClientComponent();
    			initOneComponent(g,clientComp);
    		}
    	}

    	return 0;
    }
    public static int initOneComponent(ThreadGroup g, Component c)
    {
    	Thread t=new Thread(g,c, c.getName());
    	t.setDaemon(true);
    	//t.setPriority(priority);
    	t.start();
    	return 0;
    }
    
    
//    private static int getTPCWTableIndexForSimpleDB(String baseTableName, int rowID)
//    {
//    	int index=0; 
//    	if(baseTableName.compareTo(TPCWDataModel.ITEMTABLENAME)==0)
//			{
//				index=rowID%TPCWDataModel.numOfItemTable;
//			}
//		else if(baseTableName.compareTo(TPCWDataModel.LATESTORDERTABLENAME)==0)
//			{
//				index=rowID%TPCWDataModel.numOfLatestOrderTable;
//			}
//		else if(baseTableName.compareTo(TPCWDataModel.SHOPPINGTABLENAME)==0)
//			{
//			index=rowID%TPCWDataModel.numOfShoppingTable;
//			}
//		else
//			index=rowID%sysConf.numOfTablePartitions;
//    	return index;
//    }
    
    public static int getTableIndex(int rowID, DataTable tableSchema)
    {
    	int index=0; 
   		index=rowID%sysConf.numOfTablePartitions;
    	return index;
    }
    
    public static int getTableIndex(int rowID, String tableName)
    {
    	int index=0; 
   		index=rowID%sysConf.numOfTablePartitions;
    	return index;
    }
    
    public static String getTableName(int index, String baseTableName)
    {
    	String name="";
    	name= baseTableName+"_"+index;
    	return name;
    }
    
    public static boolean isInServletRole=isInServletRole();
    
    public static boolean isInServletRole()
    {
    	boolean result=false;
    	role[] a= GlobalPara.sysConf.nodes.getNodeRole(GlobalPara.sysConf.getLocalNodeID());
    	for(int i=0;i<a.length;i++)
    		if(a[i]==role.SERVLET){
	        	result=true;
	        	break;
		}
    	return result;
    }
    
    public static int forSynchronizeReplicas =  initSynchronizeReplica();
    
    public static int initSynchronizeReplica()
	  {
	    	if(isInServletRole){
		    	ExecutorService se=Executors.newSingleThreadExecutor();
		    	PushUpdatesToReplicasTask task=new PushUpdatesToReplicasTask();
		        se.submit(task);
			}
	     return 0;
	  }
    
    public static int forMemoryManage=initAutoMemoryMonitor();
    public static int initAutoMemoryMonitor()
    {
    	if(!GlobalPara.sysConf.doMemoryManagement)
    		return 0;
    	if(isInServletRole){
    		ExecutorService se=Executors.newSingleThreadExecutor();
    		MonitorMemoryTask task=new MonitorMemoryTask();
        	se.submit(task);
        	se.shutdown();
		}
    	return 0;
    }
    
    public static int forTimestampManager=initTimestampManager();
    public static int initTimestampManager()
    {
    	boolean isInIDServiceRole=false;
    	role[] a= GlobalPara.sysConf.nodes.getNodeRole(GlobalPara.sysConf.getLocalNodeID());
    	for(int i=0;i<a.length;i++)
    		if(a[i]==role.IDService){
    			isInIDServiceRole=true;
	        	break;
		}
    	
    	if(isInIDServiceRole&&sysConf.doIdServiceServerSideMonitor){
    		ExecutorService se=Executors.newSingleThreadExecutor();
    		MonitorIDTask task=new MonitorIDTask();
        	se.submit(task);
        	se.shutdown();
		}
    	return 0;
    }
    
//    public static int initAutoCheckPoint()
//    {
//    	role[] a= GlobalPara.sysConf.nodes.getNodeRole(GlobalPara.sysConf.getLocalNodeID());
//    	for(int i=0;i<a.length;i++)
//    		if(a[i]==role.SERVLET)
//				{
//	    		ExecutorService se=Executors.newSingleThreadExecutor();
//	    		long thinkTime=new Long(GlobalPara.sysConf.CheckPointPeriod*1000);
//	        	CheckPointTask task=new CheckPointTask(thinkTime);
//	        	se.submit(task);
//	        	break;
//				}
//    	return 0;
//    }
    
	public static HBaseConfiguration initHbaseConf()
	{
		HBaseConfiguration conf=new HBaseConfiguration();
		if(sysConf.DB.compareTo("HBase")==0)
			{
				Path pathOfConf=new Path(sysConf.hbaseCfgPath_default);
				conf.addResource(pathOfConf);
				pathOfConf=new Path(sysConf.hbaseCfgPath_site);
				conf.addResource(pathOfConf);
			}
		return conf;
	}
	
	public static Logger getLogger()
	{
		Logger l=Logger.getLogger("GlobalPara");
		l.setLevel(Level.WARNING);
		//l.setLevel(Level.INFO);
		return l;
	}
	public static Logger getLoggerForPrint()
	{
		Logger l=Logger.getLogger("PrintScreen");
		//l.setLevel(Level.SEVERE);
		return l;
	}
	
	/**
	 * snapshot 0: [0,intervalOfSnapshot-1]
	 * @param timestamp
	 * @return
	 */
	public static long timestampToSnapshot(long timestamp)
	{
		long snapshotID;
		snapshotID=(timestamp/intervalOfSnapshot)+1;
		return snapshotID;
	}
	
	public static long beginTimestampOfSnapshot(long snapshotID)
	{
		return intervalOfSnapshot*snapshotID;
	}
	
	public static LinkedBlockingDeque<String> debugQueue=null;
	
	public static void shutdownThreads()
	{
	    rwComp.stop();
	    roComp.stop();
	    ltmComp.stop();
	    memberComp.stop();
	    //adapterComp.stop();
	    clientComp.stop();
	    adapterComp.shutdownWorkers();
	    
		MsgBroker.executor.shutdown();
		MembershipManager.executor.shutdown();
		 try {
			Thread.sleep(5000);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		MsgBroker.executor.shutdownNow();
		MembershipManager.executor.shutdownNow();
	}
}

