package org.zhouw.stps;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Vector;
import java.util.concurrent.*;

import org.zhouw.stps.recovery.DoRecoverDataToCommit;
import org.zhouw.stps.recovery.DoRemoveTCUpdateTask;
import org.zhouw.stps.recovery.DoReplicateTCTask;
import org.zhouw.stps.recovery.DoSynchronizeOneDataItemTask;
import org.zhouw.stps.recovery.DoUpdateTCTask;
import org.zhouw.stps.rotran.ROTranMsg.*;
import org.zhouw.stps.tp.*;
import org.zhouw.stps.id.*;
import org.zhouw.stps.monitor.IDServiceLogItem;
import org.zhouw.stps.msg.SubTransactionGroupProtos.*;
import org.zhouw.stps.msg.*;
import org.zhouw.stps.net.DoCommitTask;
import org.zhouw.stps.net.DoDataPopulateTask;
import org.zhouw.stps.net.DoPutWorkloadTask;
import org.zhouw.stps.app.otpcw.*;
import org.zhouw.stps.net.*;
import org.zhouw.stps.membership.*;
import org.zhouw.stps.rotran.*;
import org.zhouw.stps.client.*;
import org.zhouw.stps.components.Component;
import org.zhouw.stps.components.MembershipComponent;

public class MsgBroker {


	public static ExecutorService executor=Executors.newFixedThreadPool(GlobalPara.TPNodeBGThreadPoolSize); 
	
	
	/**
	 * Will block on network I/O.
	 * @param destID
	 * @param input
	 * @param servletName
	 * @return 0 if success, -1: failure
	 */
	public static int PostMessage(int destID, byte[] input, String servletName)
	{
		int result=0;
		String ip=GlobalPara.sysConf.getIP(destID);
		String url="http://"+ip + ":"+GlobalPara.sysConf.getPort(destID)+"/"+GlobalPara.sysConf.projectName+"/servlet/"+servletName;
		InputStream in= MsgBroker.doHttpPost(url, input);
		if(in!=null)
		{
			try{
				MsgBroker.cleanupInputStream(in);
			} catch (IOException e) {
				result=-1;
				GlobalPara.memberComp.reportFailure(url,ip);
				//GlobalPara.logger.warning("I/O exceptions for cleanupErrorConnection when getting "+strURL);
			}
		}
		else
			result=-1;
		
		return result;
	}
	
	public static int GetMessage(int destID, String parameters, String servletName)
	{
		int result=0;
		String ip=GlobalPara.sysConf.getIP(destID);
		String url="http://"+ip + ":"+GlobalPara.sysConf.getPort(destID)+"/"+GlobalPara.sysConf.projectName+"/servlet/"+servletName;
		url+=parameters;
		InputStream in= MsgBroker.doHttpGet(url);
		if(in!=null)
		{
			try{
				MsgBroker.cleanupInputStream(in);
			} catch (IOException e) {
				result=-1;
				GlobalPara.memberComp.reportFailure(url,ip);
			}
		}
		else
			result=-1;
		
		return result;
	}
	
	/**
	 * When ongoing transactions and backuped transactions finish recovery
	 * After all the data have completed
	 * TODO implement this method   
	 * @param destID
	 * @return
	 */
	public static Future<Integer> ReportToMasterDataReplication(int destID)
	{
		if(destID<0)
			return null;
		return null;
	}
	
	/////////////////////////////////////////////////////////
	/// For Read-only transactions
	public static Future<Integer> clientSubmitROTran(int destID, ROTransactionMsg msg)
	{
		if(destID==GlobalPara.sysConf.localNodeID)
		{
			GlobalPara.roComp.clientSubmitROTran(msg);
			return null;
		}
		else
		{
			MsgBroker.PostMessage(destID, msg.toByteArray(),"ClientSubmitQuery");
//			
//			PostMessageTask task=new PostMessageTask(destID, msg.toByteArray(), "ClientSubmitQuery");
//			try{
//				Future<Integer> returnTask=MsgBroker.executor.submit(task);
//			} catch(RejectedExecutionException e) {
//				e.printStackTrace();
////				MsgBroker.PostMessage(destID, msg.toByteArray(),"ClientSubmitQuery");
//			}
			//GlobalPara.logger.warning("ROTransaction  "+msg.getClientsideTID() +" Now submitted the Task");
			return null;
		}
	}
	
	public static Future<Integer> submitROTran(int destID, SubROTranGroupMsg msg)
	{
		if(destID==GlobalPara.sysConf.localNodeID){
			GlobalPara.ltmComp.submitROTran(msg);
			return null;
		}
		else
		{
			PostMessageTask task=new PostMessageTask(destID, msg.toByteArray(), "SubmitROTransaction");
			Future<Integer> returnTask=MsgBroker.executor.submit(task);
			return returnTask;
		}
	}
	
	public static Future<Integer> returnROTran(int destID, ROTranResultMsg msg)
	{
		if(destID<0)
			return null;
		if(destID==GlobalPara.sysConf.localNodeID)
		{
			if(GlobalPara.clientComp!=null)
				GlobalPara.clientComp.submitROTranResult(msg);
			else
				GlobalPara.logger.warning("[Error]On clinet " + GlobalPara.sysConf.localNodeID+", its clientComp is null");
			return null;
		}
		else
		{
			PostMessageTask task=new PostMessageTask(destID, msg.toByteArray(), "ReturnROTranResultToClient");
			Future<Integer> returnTask=MsgBroker.executor.submit(task);
			return returnTask;
		}
	}
	
//	public static Future<Integer> submitTOSubTrans(int destID)
//	{
//		if(destID<0)
//			return null;
//		return null;
//	}
	
	public static Future<Integer> voteForROTrans(int destID, VoteROTranMsg msg)
	{
		if(destID<0)
			return null;
		
		if(destID==GlobalPara.sysConf.localNodeID){
			GlobalPara.roComp.submitVoteMsg(msg);
			return null;
		} else {
			PostMessageTask task=new PostMessageTask(destID, msg.toByteArray(), "VoteROTransaction");
			Future<Integer> returnTask=MsgBroker.executor.submit(task);
			return returnTask;
		}
	}
	/// End For Read-only transactions
	/////////////////////////////////////////////////////////
	
	/**
	 * When ongoing transactions and backuped transactions finish recovery
	 * tell the master that the recovery process is completed
	 *  
	 * @param destID
	 * @return
	 */
	public static Future<Integer> ReportToMasterTranRecovered(int destID)
	{
		if(destID<0)
			return null;
		return null;
	}
	
	
	public static Future<Integer> RecoverFailedCommit(long GID, 
			HashMap<Integer, Vector<UpdatesOfST>> dataAsPrimary,
			HashMap<Integer, Vector<UpdatesOfST>> dataAsBackup)
	{
		if(dataAsPrimary==null||dataAsBackup==null){
			GlobalPara.logger.warning("Warning: incorrect input");
			return null;
		}
		DoRecoverDataToCommit task=new DoRecoverDataToCommit(GID, dataAsPrimary, dataAsBackup);
		Future<Integer> returnTask=MsgBroker.executor.submit(task);
		
		return returnTask;
	}

	public static Future<Integer> DoBackUpCommitService(Vector<VirtualTPNode> destID, UpdatesOfST data)
	{
		if(destID==null){
			GlobalPara.logger.warning("The input destID is null");
			return null;
		}
		if(destID.size()!=0){
			DoSynchronizeOneDataItemTask task=new DoSynchronizeOneDataItemTask(destID,data);
			return MsgBroker.executor.submit(task);
		} else {
			if(data.isToRemove)
				DataRowEntry.reportRemoveReplicaSucceed(data);
			else {
				GlobalPara.ltmComp.reportCommitDone(data.tableName, data.rowKey,data.tran, data.timestamp);
				DataRowEntry.reportSendUpdateSucceed(data);
			}
		}
		return null;
	}
	
	public static Future<Integer> DoReturnTransactionResult(int destID, TransactionResultSetMsg msg,long GID)
	{
		DoReturnTransactionResult task=new DoReturnTransactionResult(destID, msg, GID);
		Future<Integer> result=MsgBroker.executor.submit(task);
		return result;
	}
	
	/**
	 * When transactions reach a commit agreement, replicate the results
	 * Will not return until all messages have been returned
	 * @param destID
	 * @param tc
	 * @return
	 */
	public static int removeBackupTC(int[] destID, long ID)
	{
		if(destID==null||destID.length==0)
			return 0;
		//Future<Integer>[] tasks=new Future[destID.length];
		for(int i=0;i<destID.length;i++){
			DoRemoveTCUpdateTask task=new DoRemoveTCUpdateTask(destID[i], ID);
			MsgBroker.executor.submit(task);
		}
//		for(int i=0;i<destID.length;i++)
//			{
//				try {
//					tasks[i].get();
//				} catch (InterruptedException e) {
//					e.printStackTrace();
//				} catch (ExecutionException e) {
//					e.printStackTrace();
//				}
//			}
		return 0;
	}
	

//	public static int updateDerivedSubTranForBackupTC(int[] destIDs, NewDerivedSTGForBackupTCMsg msg, long GID)
//	{
//		if(destIDs==null||destIDs.length==0)
//			return 0;
//		
//		UpdateDerivedSTtoBackupTCTask task=new UpdateDerivedSTtoBackupTCTask(destIDs, msg, GID);
//		MsgBroker.executor.submit(task);
//		return 1;
//	}
	
	
	/**
	 * When transactions reach a commit agreement, replicate the results
	 * @param destID
	 * @param tc
	 * @return
	 */
	public static int updateTC(int[] destID, TransCoodinator tc)
	{
		if(destID==null)
			return 0;
		
		Future<Integer>[] tasks=new Future[destID.length];
		BackupTransactionMsg msg=tc.constructUpdateBackupTCMsg();
		DoUpdateTCTask task=new DoUpdateTCTask(destID, msg, tc.ID);
		MsgBroker.executor.submit(task);
		
//		for(int i=0;i<destID.length;i++)
//			{
//				try {
//					tasks[i].get();
//				} catch (InterruptedException e) {
//					e.printStackTrace();
//				} catch (ExecutionException e) {
//					e.printStackTrace();
//				}
//			}
		return 0;
	}
	
//	/**
//	 * Add the current TCs to the backup LTMs
//	 * @param destID
//	 * @param tc
//	 * @return
//	 */
//	public static void replicateTC(int[] destID, TransCoodinator tc)
//	{
//		if(destID==null||destID.length==0)
//			return;
//		DoReplicateTCTask task=new DoReplicateTCTask(destID,tc);
//		MsgBroker.executor.submit(task);
//	}
	
	/**
	 * Get from Master
	 * @return
	 */
	public static long reportCompletedSnapShotID(long newID)
	{
		int destID=GlobalPara.sysConf.getMasterID();
		if(GlobalPara.sysConf.getLocalNodeID()==destID)
		{
			GlobalPara.master.reportSnapshotComplete(newID, destID);
		}
		else
		{
			String ip=GlobalPara.sysConf.getIP(destID);
			String url="http://"+ip + ":"+GlobalPara.sysConf.getPort(destID)+"/"+GlobalPara.sysConf.projectName+"/servlet/ReportCompletedSnapShotID";
			url+="?TPNodeID="+GlobalPara.sysConf.getLocalNodeID()+"&SID="+newID;
			InputStream in= MsgBroker.doHttpGet(url);
			if(in!=null)
			{
				try {
					BufferedReader reader = new BufferedReader(new InputStreamReader(in));
					String line=reader.readLine();
					if(line==null)
						{
						GlobalPara.logger.warning("Failed to report to master.");
						return -1;
						}
					while(reader.readLine()!=null)
					{;}
					in.close();
				} catch (IOException e) {
					//e.printStackTrace();
				}
			}
		}
		return 0;
	}
	
	/**
	 * Get from Master
	 * @return
	 */
	public static long getCurrentSnapshotID()
	{
		return 0;
	}
	
	//public static ExecutorService executor=Executors.newFixedThreadPool(GlobalPara.msgThreadCount);
	//public static ExecutorService executor=Executors.newCachedThreadPool();
	
	public static int getNewShoppingID()
	{
		long start=System.currentTimeMillis();
		int result=0;
		int destID=GlobalPara.sysConf.getShoppingIDManagerID();
		if(GlobalPara.sysConf.getLocalNodeID()==destID)
		{
			result=IDService.getShoppingID(true);
		}
		else
		{
			String ip=GlobalPara.sysConf.getIP(destID);
			String url="http://"+ip + ":"+GlobalPara.sysConf.getPort(destID)+"/"+GlobalPara.sysConf.projectName+"/servlet/ShoppingIDService";
			InputStream in= MsgBroker.doHttpGet(url);
			if(in!=null)
			{
				try {
					BufferedReader reader = new BufferedReader(new InputStreamReader(in));
					String line=reader.readLine();
					if(line==null)
						{
						GlobalPara.logger.warning("Failed to get new shopping id.");
						return -1;
						}
					result=Integer.valueOf(line);
					while(reader.readLine()!=null)
					{;}
					in.close();
				} catch (IOException e) {
					//e.printStackTrace();
				}
			}
		}
		long end=System.currentTimeMillis();
		IDServiceLogItem l=new IDServiceLogItem(end-start, end);
		GlobalPara.monitor.writelnIDLog(l.writeToString());
		return result;
	}
	
	
	public static void getNewTimestamp(Component comp, long compID, boolean isRestart)
	{
		GetNewTimestampTask task=new GetNewTimestampTask(comp, compID, isRestart);
		executor.submit(task);
	}
	
	/**
	 * All invocation to this method is remote
	 * @param destID
	 * @return
	 */
	public static Future<String> DoPutWorkload(int destID)
	{
		DoPutWorkloadTask task=new DoPutWorkloadTask(destID);
		Future<String> result=MsgBroker.executor.submit(task);
		return result;
	}
	
	/**
	 * All invocation to this method is remote
	 * @param destID
	 * @return
	 */
	public static Future<String> DoDataPopulate(int destID)
	{
		DoDataPopulateTask task=new DoDataPopulateTask(destID);
		Future<String> result=MsgBroker.executor.submit(task);
		return result;
	}
	
	/**
	 * All invocation to this method is remote
	 * @param destID
	 * @return
	 */
	public static Future<String> DoDataPopulateForSimpleDB(int destID)
	{
		DoDataPopulateForSimpleDBTask task=new DoDataPopulateForSimpleDBTask(destID);
		Future<String> result=MsgBroker.executor.submit(task);
		return result;
	}
//	
//	//////////////////////////////////////////////////
//	// from Client to TPnode
//	public static TransactionResult ExecuteGlobalTransaction(int destID, Transaction t)
//	{
//		TransactionResult result=new TransactionResult();
////		if(destID==GlobalPara.sysConf.getLocalNodeID())
////			{
////				result.state=GlobalTransManager.runTransaction(t);
////				result.result=t.output;
////			}
////		else
//			{
//				TransactionMsg.Builder msgBuilder=TransactionMsg.newBuilder();
//				msgBuilder.setIsWeakConsistent(t.isWeakConsistent);
//				for(int i=0;i<t.trans.length;i++)
//					msgBuilder.addTrans(t.trans[i].getMsgBody());
//				msgBuilder.setClientsideTID(t.clientsideTID);
//				
//				TransactionMsg msg=msgBuilder.build();
//				
//				String ip=GlobalPara.sysConf.getIP(destID);
//				String url="http://"+ip + ":"+GlobalPara.sysConf.getPort(destID)+"/"+GlobalPara.sysConf.projectName+"/servlet/ExecuteGlobalTransactionService";
//				//url+="?tid="+tranactionId;
//				InputStream in= MsgBroker.doHttpPost(url, msg.toByteArray());
//				if(in!=null)
//				{
//					try {
//						TransactionResultSetMsg r=TransactionResultSetMsg.parseFrom(in);
//						in.close();
//						String state=r.getTranState();
//						result.state=TranState.valueOf(state);
//						int count=r.getResultCount();
//						if(count==0)
//							result.result=null;
//						else
//							{
//							result.result=new String[count][2];
//							for(int i=0;i<count;i++)
//								{
//								result.result[i][0]=r.getResult(i).getKey();
//								result.result[i][1]=r.getResult(i).getValue();
//								}
//							}
//					} catch (IOException e) {
//						//e.printStackTrace();
//					}
//				}
//			}
//		return result;
//	}
//	
	
	
	
	//////////////////////////////////////////////////
	// from TC to TPnode
	public static Future<String[][]> SubmitSTG(SubTransactionGroup[] stg)
	{
		SubmitSTGTask task=new SubmitSTGTask(stg);
		Future<String[][]> result=MsgBroker.executor.submit(task);
		return result;
	}
	
	public static Future<Integer> DoCommit(int destID, long timestamp, String key)
	{
		DoCommitTask task=new DoCommitTask(destID,timestamp,key);
		Future<Integer> result=MsgBroker.executor.submit(task);
		return result;
	}
	public static Future<Integer> DoRestart(int destID, long timestamp,String key, long newTimestamp)
	{
		DoRestartTask task=new DoRestartTask(destID,timestamp,key,newTimestamp);
		Future<Integer> result=MsgBroker.executor.submit(task);
		return result;
	}
	public static Future<Integer>  DoAbort(int destID, long timestamp,String key)
	{
		DoAbortTask task=new DoAbortTask(destID,timestamp,key);
		Future<Integer> result=MsgBroker.executor.submit(task);
		return result;
	}
	
	
	public static String DoDebug(int destID,String key)
	{
		String result="";
		if(destID==GlobalPara.sysConf.getLocalNodeID())
		{
			result=TPNode.LTM.doDebug(key);
		}
		else
		{
			String ip=GlobalPara.sysConf.getIP(destID);
			String url="http://"+ip + ":"+GlobalPara.sysConf.getPort(destID)+"/"+GlobalPara.sysConf.projectName+"/servlet/PrintDebugInfo";
			url+="?key="+key;
			InputStream in= MsgBroker.doHttpGet(url);
			
			if(in!=null)
			{
				try{
					BufferedReader reader = new BufferedReader(new InputStreamReader(in));
					String line;
					while((line=reader.readLine())!=null)
					{
						result+=line+"\n";
					}
					in.close();
				} catch (IOException e) {
					//e.printStackTrace();
				}
			}
			
		}
		return result;
	}
	
	public static void DoShutDown(int destID)
	{
//		if(destID==GlobalPara.sysConf.getLocalNodeID())
//		{
//			TPNode.LTM.checkpoint();
//		}
//		else
//		{
//			String ip=GlobalPara.sysConf.getIP(destID);
//			String url="http://"+ip + ":"+GlobalPara.sysConf.getPort(destID)+"/"+GlobalPara.sysConf.projectName+"/servlet/DoShutDownService";
//			InputStream in= MsgBroker.doHttpGet(url);
//			if(in!=null)
//			{
//				try{
//					in.close();
//				} catch (IOException e) {
//					e.printStackTrace();
//				}
//			}
//		}
//		GlobalPara.logger.info("Succesfully checkpoint TPNode "+ destID);
	}
	////////////////////////////////////////////////////
	// from TPNode to TC
	
	/**
	 * the updates and result is also returned
	 */
	public static void reportCommit(SubTransactionGroup stg)
	{
		if(stg==null){
			GlobalPara.logger.warning("[error] Input STG is null");
			return;
		}
		
		Vector<SubTransaction> derived = stg.getDerived();
		if(stg.GTPNodeID==GlobalPara.sysConf.getLocalNodeID()) {
			long timestamp=stg.getTimeStamp();

			DTMsg msg;
			if(derived==null||derived.size()==0)
				msg=new DTMsg(DTMsg.MSGTYPE.REPORTCOMMIT, stg.id, timestamp, stg.GID);
			else
				msg=new DTMsg(DTMsg.MSGTYPE.REPORTCOMMIT, stg.id, timestamp, stg.GID, derived);
			
			int count=0;
			for(int i=0;i<stg.trans.length;i++) {
				if(stg.trans[i].DataToReturn!=null)
					count+=stg.trans[i].DataToReturn.length;
			}
			
			if(count!=0) {
				msg.result=new String[count][2];
				count=0;
				for(int i=0;i<stg.trans.length;i++){
					if(stg.trans[i].DataToReturn!=null)
						for(int j=0;j<stg.trans[i].DataToReturn.length;j++)
							msg.result[count++]=stg.trans[i].DataToReturn[j];
				}
			}
			
			msg.updates=new UpdatesOfST[stg.trans.length];
			for(int i=0;i<stg.trans.length;i++) {
				msg.updates[i]=new UpdatesOfST(stg.trans[i]);
				if(stg.trans[i].rootEntry.latestWriteTransaction==-1) {//the data item is loaded the first time, whole data row is loaded
					msg.updates[i].isContainsDataRow=true;
					if(!stg.trans[i].isUpdateSameDataItem){
						msg.updates[i].dataRow=stg.trans[i].rootEntry.committedData.getRowData();
					}
				}
			}
			GlobalPara.rwComp.reportCommit(msg);
		} else {
			ResultSetMsg returnMsg=stg.constructCommitResult();
			
			ReportCommitMsg.Builder commitMsgBuilder=ReportCommitMsg.newBuilder();
			commitMsgBuilder.setGID(stg.GID).setId(stg.id).setTimestamp(stg.timestamp)
			.setData(returnMsg);
			if(derived!=null&&derived.size()!=0) {
				for(SubTransaction st: derived) {
					commitMsgBuilder.addDeTrans(st.getMsgBody());
				}
			}
			
			ReportCommitMsg commitMsg=commitMsgBuilder.build();
			PostMessageTask task=new PostMessageTask(stg.GTPNodeID, commitMsg.toByteArray(), "ReportCommitService");
			MsgBroker.executor.submit(task);
		}
	}
	
	public static void reportAbort(SubTransactionGroup stg)
	{
		if(stg.GTPNodeID==GlobalPara.sysConf.getLocalNodeID()) {
			long timestamp=stg.getTimeStamp();
			DTMsg msg=new DTMsg(DTMsg.MSGTYPE.REPORTABORT, stg.id, timestamp, stg.GID);
			GlobalPara.rwComp.reportAbort(msg);	
		} else	{
			String parameters="?timestamp="+stg.timestamp+"&id="+stg.id +"&GID=" +stg.GID;
			GetMessageTask task=new GetMessageTask(stg.GTPNodeID, parameters, "ReportAbortService");
			MsgBroker.executor.submit(task);
			
		}
	}
	
	public static void reportAbortDone(SubTransactionGroup stg)
	{
		if(stg.GTPNodeID==GlobalPara.sysConf.getLocalNodeID()){
			GlobalPara.rwComp.reportAbortDone(stg.GID, stg.id, stg.timestamp);
		} else {
			String parameters="?timestamp="+stg.timestamp+"&id="+stg.id +"&GID=" +stg.GID;
			GetMessageTask task=new GetMessageTask(stg.GTPNodeID, parameters, "ReportAbortDoneService");
			MsgBroker.executor.submit(task);
		}
	}
	
	
	public static void reportCommitDone(SubTransactionGroup stg)
	{
		//GlobalPara.logger.info("*****STG "+stg.printInfo()+" \nis reporting commit done");
		if(stg.GTPNodeID==GlobalPara.sysConf.getLocalNodeID()){
			GlobalPara.rwComp.reportCommitDone(stg.GID, stg.id);
		} else {
			String parameters="?timestamp="+stg.timestamp+"&id="+stg.id +"&GID=" +stg.GID;
			GetMessageTask task=new GetMessageTask(stg.GTPNodeID, parameters, "ReportCommitDoneService");
			MsgBroker.executor.submit(task);
		}
	}
	
	
	public static void reportRestart(SubTransactionGroup stg)
	{
		//GlobalPara.logger.info("*****STG "+stg.printInfo()+" \nis reporting restart");
		if(stg.GTPNodeID==GlobalPara.sysConf.getLocalNodeID()){
			long timestamp=stg.getTimeStamp();
			DTMsg msg=new DTMsg(DTMsg.MSGTYPE.REPORTRESTART, stg.id, timestamp, stg.GID);
			GlobalPara.rwComp.reportRestart(msg);
		} else {
			String parameters ="?timestamp="+stg.timestamp+"&id="+stg.id +"&GID=" +stg.GID;
			GetMessageTask task=new GetMessageTask(stg.GTPNodeID, parameters, "ReportRestartService");
			MsgBroker.executor.submit(task);
		}
	}
	
	/**
	 * The returned InputStream object should be close after complete read
	 * @param urlString
	 * @return
	 */
	public static InputStream doHttpGet(String urlString)
	{
		int code;
		HttpURLConnection connection=null;
		
		URL url;
		
		try {
			url = new URL(urlString);
		} catch (MalformedURLException e1) {
			e1.printStackTrace();
			return null;
		}
		
		
		
		try {
			connection = (HttpURLConnection)url.openConnection();
			code = connection.getResponseCode();
			if(code!=200){
				//GlobalPara.logger.warning("Invoking servlet failed, get response code: "+code + ", when try to get "+urlString);
				String IP=url.getHost();
				GlobalPara.memberComp.reportFailure(urlString,IP);
				cleanupConnection(connection);
				return null;
			} else {
				InputStream returnIn=connection.getInputStream();
				return returnIn;
			}
		} catch (IOException e) {
			//GlobalPara.logger.warning("I/O exceptions when getting"+urlString);
			String IP=url.getHost();
			GlobalPara.memberComp.reportFailure(urlString,IP);
//			try {
//				if(connection!=null)
//					cleanupErrorConnection(connection);
//			} catch (IOException e1) {
//				GlobalPara.logger.warning("I/O exceptions for cleanupErrorConnection when getting "+urlString);
//			}
		}
		return null;
	}
	
	
	public static void cleanupConnection(HttpURLConnection connection) throws IOException
	{
		long start=System.currentTimeMillis();
		InputStream es = connection.getInputStream();
		cleanupInputStream(es);
		long duration=System.currentTimeMillis()-start;
		if(duration>1000){
			GlobalPara.logger.info("The time for cleanup connection takes a long time: "+duration);
		}
	}
	
	public static void cleanupInputStream(InputStream es) throws IOException
	{
		int ret = 0;
		// read the response body
		while ((ret = es.read()) > 0) {
		}
		// close the errorstream
		es.close();
	}
	
	public static void cleanupErrorConnection(HttpURLConnection connection) throws IOException
	{
		InputStream es = connection.getErrorStream();
		int ret = 0;
		if(es!=null){
			// read the response body
			while ((ret = es.read()) > 0) {
			}
			// close the errorstream
			es.close();
		}
	}
	
	public static InputStream sendMsgToOneLTM(int destID, String servletName, byte[] input)
	{
		String ip=GlobalPara.sysConf.getIP(destID);
		String url="http://"+ip + ":"+GlobalPara.sysConf.getPort(destID)+"/"+GlobalPara.sysConf.projectName+"/servlet/" + servletName;
		InputStream in= MsgBroker.doHttpPost(url, input);
		return in;
	}
	
	public static InputStream doHttpPost(String strURL, byte[] input)
	{
		HttpURLConnection connection=null;
		int code;
		URL url;
		
		try {
			url = new URL(strURL);
		} catch (MalformedURLException e1) {
			e1.printStackTrace();
			return null;
		}
		
		
		
		try {
			connection = (HttpURLConnection)url.openConnection();
			connection.setDoOutput(true);
			connection.setRequestMethod("POST");
			OutputStream os = connection.getOutputStream();
			if(input!=null){
				os.write(input);
				os.flush();
			}
			os.close();
	    
			code = connection.getResponseCode();
			if(code!=HttpURLConnection.HTTP_OK) {
				//GlobalPara.logger.warning("Invoking servlet failed, get response code: "+code + ", when try to get "+strURL);
				cleanupConnection(connection);
				String IP=url.getHost();
				GlobalPara.memberComp.reportFailure(strURL,IP);
				return null;
			} else {
				//GlobalPara.logger.warning("GetInputStream " + strURL);
				InputStream returnIn=connection.getInputStream();
				return returnIn;
			}
		}
		catch( IOException e ){
			//GlobalPara.logger.warning("I/O exceptions when getting"+strURL);
			String IP=url.getHost();
			GlobalPara.memberComp.reportFailure(strURL,IP);
//			try {
//				if(connection!=null)
//					cleanupErrorConnection(connection);
//			} catch (IOException e1) {
//				GlobalPara.logger.warning("I/O exceptions for cleanupErrorConnection when getting "+strURL);
//			}
		}
		return null;
	}


//////////////////////////////////////////////////
// from Client to TPnode
	/**
	 * return after the TC has been replicated to multiple LTMs
	 */
public static TransactionResult SubmitGlobalTransaction(int destID, Transaction t)
{
	TransactionResult result=new TransactionResult();
	if(destID==GlobalPara.sysConf.getLocalNodeID()){
			GlobalPara.rwComp.submitClientTransaction(t);
	} else {
			TransactionMsg msg=t.getMsg();
			
			String ip=GlobalPara.sysConf.getIP(destID);
			String url="http://"+ip + ":"+GlobalPara.sysConf.getPort(destID)+"/"+GlobalPara.sysConf.projectName+"/servlet/ExecuteGlobalTransactionService";
			InputStream in= MsgBroker.doHttpPost(url, msg.toByteArray());
			if(in!=null)
			{
				try {
					ReceiveTransactionMsg returnMsg=ReceiveTransactionMsg.parseFrom(in);
					int version =returnMsg.getVersion();
					int compareResult=MembershipManager.compareVersionNumber(version);
					if(compareResult==-1){
						MembershipManager.handleOutdatedMembershipForClient(destID, version);
						return null;
					} else if(compareResult==1) {
						return null;
					}
					
					if(returnMsg.getIsAbort())
						return null;

					int ret = 0;
					while ((ret = in.read()) > 0) {
					}
					in.close();
					
					//MsgBroker.cleanupInputStream(in);
				} catch (IOException e) {
					result=null;
				}
			}
			else
				result=null;
		}
	return result;
}

}
