package org.zhouw.stps.adapter;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.zhouw.stps.app.*;
import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.tp.BufferForCheckpoint;
import org.zhouw.stps.tp.TableDataBuffer;
import org.zhouw.stps.tp.UpdatedDataRow;
import org.zhouw.stps.*;

/**
 * The adapter will transparently transform the logical data model and store the data
 * into the adapter-specific physical data model.
 * There is no consistency/transactional properties guaranteed by the adapter's operations,
 * which are dependent on the underlying data service.
 * The update operation of adapter will not update the affected index table defined in the
 * logical data model. 
 * All operations are returned after the operation is successful.
 * However, the updates made by the operations are possibly buffered, 
 * depending on the library implementation for specific cloud data service
 * Using commitAll can flush the buffered updates.
 * 
 * @author zwlalala
 *
 */
public interface Adapter {
	
//	/**
//	 * It should return immediately when it copies the committed data to buffer
//	 * It would do the actual checkpoint in another thread
//	 * such that the writelock for LTM datapool would not be held for too long.
//	 */
//	public void checkpoint(BufferForCheckpoint datapool);
	
	/**
	 * The input data columns are in logical names
	 *  
	 * @param snapshotID
	 * @param table
	 * @param rowKey
	 * @param value
	 * @return
	 */
	public void checkpointOneItem(long snapshotID, DataTable tableSchema, String rowKey,UpdatedDataRow value);
	
	/**
	 * 
	 * @param strTblName
	 * @param strRowKey
	 * @param row Output: The column names in it must be in logical names
	 * @return 0 if successful. -1 if not found the row
	 */
	public int loadData(DataTable tableSchema, String strRowKey, HashMap<String, String> row);
	
	
	public void createTable(Iterable <? extends DataTable> tableSchemas);
	public void putRows(DataTable tableSchema,  Iterable<Map.Entry<String,LinkedHashMap<String,String>>> rows);
	public void putOneRow(DataTable tableSchema,  String rowkey, Iterable<Map.Entry<String,String>> row);
	public void delRows(DataTable tableSchema,  Iterable<String> rows);
	public void commitAll();
}
