package org.zhouw.stps.adapter;

import org.zhouw.stps.*;
import org.zhouw.stps.app.join.JoinAppDataModel;
import org.zhouw.stps.monitor.CheckpointLogItem;
import org.zhouw.stps.monitor.LoadDataLogItem;
import org.zhouw.stps.tp.*;
import org.zhouw.stps.app.*;
import org.zhouw.stps.app.tpcw.*;

import java.io.IOException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import org.apache.hadoop.hbase.HColumnDescriptor;
import org.apache.hadoop.hbase.HTableDescriptor;
import org.apache.hadoop.hbase.KeyValue;
import org.apache.hadoop.hbase.MasterNotRunningException;
import org.apache.hadoop.hbase.client.HBaseAdmin;
import org.apache.hadoop.hbase.util.Bytes;
import org.zhouw.stps.tp.DataRowEntry;
import org.zhouw.stps.tp.TableDataBuffer;
import org.zhouw.stps.tp.UpdatedDataRow;

import org.apache.hadoop.hbase.HBaseConfiguration;
import org.apache.hadoop.hbase.client.Delete;
import org.apache.hadoop.hbase.client.Get;
import org.apache.hadoop.hbase.client.HTable;
import org.apache.hadoop.hbase.client.Put;
import org.apache.hadoop.hbase.client.Result;
import org.apache.hadoop.hbase.client.ResultScanner;
import org.apache.hadoop.hbase.client.Scan;
import org.apache.hadoop.hbase.util.Bytes;


import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.locks.*;



public class HBaseAdapter implements Adapter {
	
//	//used only for performance logging
//	private static ReentrantLock lock=new ReentrantLock();
	private static int count=0;
	
	//public ExecutorService cpService= Executors.newFixedThreadPool(GlobalPara.CPBGThreadPoolSize);
	
	public static String getDataColumnFamilyName()
	{
		return columnFamilyName;
	}
	
	protected static final String columnFamilyName="Data:";
	
	public static String getColumnFamilyName(String logicalColumnName){
		if(!DataModel.isRefCacheColumn(logicalColumnName)) {
			return columnFamilyName;
		}
		else
			return DataModel.joinCacheColumnFamily;
	}
	
	public static String getPureColumnFamilyName(String columnName){
		String column=getColumnFamilyName(columnName);
		return column.substring(0, column.length()-1);
	}
	
	public static String getQualifierName(String logicalColumnName){
		if(!DataModel.isRefCacheColumn(logicalColumnName)) {
			return logicalColumnName;
		}
		else
			return logicalColumnName.substring(DataModel.joinCacheColumnFamily.length());
	}
	
	public static String getPhysicalDataColumnName(String logicalColumnName){
		if(!DataModel.isRefCacheColumn(logicalColumnName)) {
			return columnFamilyName+logicalColumnName;
		}
		else
			return logicalColumnName;
	}
	
	HashMap<String, HTable> hTables;
	
	protected HTable getHTable(String hbaseTableName) throws IOException
	{
		HTable table=hTables.get(hbaseTableName);
		if(table==null){
			table = new HTable(GlobalPara.hbaseConf, hbaseTableName);
			table.setAutoFlush(true);
			hTables.put(hbaseTableName, table);
		}
		return table;
	}

	public void commitAll(){
		if(hTables!=null){
			for(HTable table:hTables.values()){
				try {
					table.flushCommits();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	protected static String getlogicalColumnName(String physicalDataColumnName){
		if(!DataModel.isRefCacheColumn(physicalDataColumnName)) {
			if(physicalDataColumnName.startsWith(columnFamilyName))
				return physicalDataColumnName.substring(columnFamilyName.length());
			else
				GlobalPara.logger.warning("Invalid column obtained from HBase:"+physicalDataColumnName);
		}
		return physicalDataColumnName;
	}
	
	protected class TableCreator implements Callable<Long>{
		private String tableName;
		
		public TableCreator(String tableName)
		{
			this.tableName=tableName;
		}
		
		public Long call()
		{
			HBaseAdmin admin;
			try {
				admin = new HBaseAdmin(GlobalPara.hbaseConf);
				HTableDescriptor desc = new HTableDescriptor(tableName);
				desc.addFamily(new HColumnDescriptor(HBaseAdapter.columnFamilyName));
				desc.addFamily(new HColumnDescriptor(DataModel.joinCacheColumnFamily));
				System.out.println("Creating table "+ desc.getNameAsString() +"...........");
				
				if(!admin.tableExists(desc.getNameAsString())){
					admin.createTable(desc);
					System.out.print("Done "+desc.getNameAsString()+"\n");
				}
				else
					System.out.print("Skip "+desc.getNameAsString()+"\n");
				
			}	catch (MasterNotRunningException e) {
				e.printStackTrace();
				System.err.println("\n[ERROR]Unable to create table");
		    	System.exit(1);
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
				System.err.println("\n[ERROR]Unable to create table");
		    	System.exit(1);
			} catch (IOException e) {
				e.printStackTrace();
				System.err.println("\n[ERROR]Unable to create table");
		    	System.exit(1);
			} 
			return 0L;
		}
	}
	
	public static void compactTables(Iterable <? extends DataTable> tableSchemas)
	{
		if(tableSchemas==null)
			return;
		HBaseAdmin admin;
		try {
			admin = new HBaseAdmin(GlobalPara.hbaseConf);
			for(DataTable table:tableSchemas){
				String phyTblName=table.getPhysicalTableName();
				for(int i=0;i<GlobalPara.sysConf.numOfTablePartitions;i++){
					String hbaseTblName = GlobalPara.getTableName(i,phyTblName);
					admin.majorCompact(hbaseTblName);
				}
			}
		}	catch (MasterNotRunningException e) {
			e.printStackTrace();
			System.err.println("\n[ERROR]Unable to create table");
	    	System.exit(1);
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
			System.err.println("\n[ERROR]Unable to create table");
	    	System.exit(1);
		} catch (IOException e) {
			e.printStackTrace();
			System.err.println("\n[ERROR]Unable to create table");
	    	System.exit(1);
		} 
		
		
	}
	
	public void createTable(Iterable <? extends DataTable> tableSchemas)
	{
		createTable(tableSchemas, false);
	}
	public List<Future<Long>> createTable(Iterable <? extends DataTable> tableSchemas, boolean isParallel)
	{
		if(tableSchemas==null)
			return null;
		
		Vector<HTableDescriptor> ts=new Vector<HTableDescriptor>();
		Vector<Future<Long>> results=new Vector<Future<Long>>(); 
		int count=0;
		for(DataTable table:tableSchemas){
			String phyTblName=table.getPhysicalTableName();
			for(int i=0;i<GlobalPara.sysConf.numOfTablePartitions;i++){
				String hbaseTblName = GlobalPara.getTableName(i,phyTblName);
				if(isParallel){
					TableCreator task=new TableCreator(hbaseTblName);
					results.add(MsgBroker.executor.submit(task));
					count++;
					if(count%10==0){
						try {
							Thread.sleep(7000);
						} catch (InterruptedException e) {
							e.printStackTrace();
						}
					}
				} else {
					HTableDescriptor desc = new HTableDescriptor(hbaseTblName);
					desc.addFamily(new HColumnDescriptor(HBaseAdapter.columnFamilyName));
					desc.addFamily(new HColumnDescriptor(DataModel.joinCacheColumnFamily));
					ts.add(desc);
				}
			}
		}
		
		if(isParallel){
			return results;
		}else{
			HBaseAdmin admin;
			try {
				admin = new HBaseAdmin(GlobalPara.hbaseConf);
				for(HTableDescriptor desc: ts){
					System.out.print("Creating table "+ desc.getNameAsString() +"...........");
					if(!admin.tableExists(desc.getNameAsString())){
						admin.createTable(desc);
						System.out.print("Done \n");
					}
					else
						System.out.print("Skip \n");
				}
			}	catch (MasterNotRunningException e) {
				e.printStackTrace();
				System.err.println("\n[ERROR]Unable to create table");
		    	System.exit(1);
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
				System.err.println("\n[ERROR]Unable to create table");
		    	System.exit(1);
			} catch (IOException e) {
				e.printStackTrace();
				System.err.println("\n[ERROR]Unable to create table");
		    	System.exit(1);
			} 
		}
		return null;
	}
	
	/**
	 * 
	 * @param tableSchema
	 * @param rowid The integer form of row key
	 * @return
	 */
	public String getHBaseTableName(DataTable tableSchema,int rowid)
	{
		int index=GlobalPara.getTableIndex(rowid, tableSchema);
		String hbaseTableName=GlobalPara.getTableName(index,tableSchema.getPhysicalTableName());
		return hbaseTableName;
	}
	
	public void delRows(DataTable tableSchema,  Iterable<String> rows)
	{
		HTable table;
		try {
			for(String rowKey: rows){
				try{
					int rowid=CheckpointThreadExecutor.getRowID(rowKey);
					String hbaseTableName=getHBaseTableName(tableSchema,rowid);
					table=this.getHTable(hbaseTableName);
					Delete d=new Delete(Bytes.toBytes(rowKey));
					table.delete(d);
				}catch(NumberFormatException e){
					GlobalPara.logger.warning("Input Row Key is not in type Integer: "+rowKey);
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Commit the rows
	 * @param tableSchema
	 * @param rows
	 * @return
	 */
	public void putRows(DataTable tableSchema, Iterable<Map.Entry<String,LinkedHashMap<String,String>>> rows)
	{
		for(Map.Entry<String,LinkedHashMap<String,String>> row:rows){
			String rowKey=row.getKey();
			this.putOneRow(tableSchema, rowKey, row.getValue().entrySet());
		}
	}
	
	public void putOneRow(DataTable tableSchema,  String rowKey, Iterable<Map.Entry<String,String>> row){
		try{
			int rowid=CheckpointThreadExecutor.getRowID(rowKey);
			String hbaseTableName=this.getHBaseTableName(tableSchema, rowid);
			try {
				HTable table= this.getHTable(hbaseTableName);
				Put p = new Put(Bytes.toBytes(rowKey));
				for(Map.Entry<String, String> cell:row){
					p.add(Bytes.toBytes(HBaseAdapter.getPureColumnFamilyName(cell.getKey())),
							Bytes.toBytes(HBaseAdapter.getQualifierName(cell.getKey()))
							, Bytes.toBytes(cell.getValue()));
//					p.add(Bytes.toBytes(HBaseAdapter.getPhysicalDataColumnName(cell.getKey())),
//							0, Bytes.toBytes(cell.getValue()));
				}
				table.put(p);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}catch(NumberFormatException e){
			GlobalPara.logger.warning("Input Row Key is not in type Integer: "+rowKey);
		}
	}
	
	public class RowBuffer{
		public String rowKey;
		public String[][] dataToPut;
		public String[] dataToDelete;
		public boolean isDeleteAll;
	}
	
	
	public HBaseAdapter()
	{
		this.hTables=new HashMap<String, HTable>();
	}
	
	// Flush the updates immediately to guarantee durability
	public void checkpointOneItem(long timestamp, DataTable tableSchema, String rowKey,UpdatedDataRow value)
	{
		int rowid;
		
		if(tableSchema.database.dbname==TPCWDataModel.name)
			rowid=Integer.valueOf(rowKey);
		else
			rowid=CheckpointThreadExecutor.getRowID(rowKey);
		
		String hbaseTableName=this.getHBaseTableName(tableSchema,rowid);
		HTable table;
		try {
			table = this.getHTable(hbaseTableName);
			
			if(value.updatedColumns!=null&&value.updatedColumns.size()!=0){
				Put p = new Put(Bytes.toBytes(rowKey));
				for(String key:value.updatedColumns.keySet()){
//					if(key.compareTo(DataModel.joinCacheColumnFamily)==0){
//					}
					if(tableSchema.database.dbname!=TPCWDataModel.name){//no need to use compareString()
						String qualifier=HBaseAdapter.getQualifierName(key);
						p.add(Bytes.toBytes(HBaseAdapter.getPureColumnFamilyName(key)),
								Bytes.toBytes(qualifier)
								, Bytes.toBytes(value.updatedColumns.get(key)));
					} else {
						p.add(Bytes.toBytes(key), timestamp, Bytes.toBytes(value.updatedColumns.get(key)));
					}
					
//					p.add(Bytes.toBytes(HBaseAdapter.getPhysicalDataColumnName(key)),
//							snapshotID, Bytes.toBytes(value.updatedColumns.get(key)));
				}
				table.put(p);
			}
			
			if(value.deletedColumns!=null&&value.deletedColumns.size()!=0){
				Delete d=new Delete(Bytes.toBytes(rowKey));
				for(String delCol:value.deletedColumns){
					if(tableSchema.database.dbname!=TPCWDataModel.name){
						d.deleteColumn(Bytes.toBytes(HBaseAdapter.getPureColumnFamilyName(delCol)),
								Bytes.toBytes(HBaseAdapter.getQualifierName(delCol)));
					} else {
						d.deleteColumn(Bytes.toBytes(delCol));
					}
				}
				table.delete(d);
			}
			table.flushCommits();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}

	/**
	 * The output should be logical column names
	 */
	public int loadData(DataTable tableSchema, String strRowKey, HashMap<String, String> row)
	{
		long start=System.currentTimeMillis();
		int result=0;
		if(tableSchema==null){
			GlobalPara.logger.warning("Empty schema inputted");
			return -1;
		}
		
		if(row==null){
			GlobalPara.logger.warning("The output contained is null");
			return -1;
		}
		
		try {
			int rowID;
			if(tableSchema.database.dbname==TPCWDataModel.name)//no need to use compareString()
				rowID=Integer.valueOf(strRowKey);
			else
				rowID=CheckpointThreadExecutor.getRowID(strRowKey);
			
			String hbaseTableName=this.getHBaseTableName(tableSchema,rowID );
			HTable table = this.getHTable(hbaseTableName);
			
			Get g = new Get(Bytes.toBytes(strRowKey));
		    Result r = table.get(g);
		    List<KeyValue> data=r.list();
		    if(data!=null){
			    for(KeyValue cell:data){
			    	String key=new String(cell.getColumn());
			    	String familyName=new String(cell.getFamily());
			    	String qualifier=new String(cell.getQualifier());
			    	//if(!DataModel.isRefCacheColumn(key)){
		    		if(key.startsWith(HBaseAdapter.columnFamilyName))
		    			key=qualifier;
			    	//}
			    	String value=new String(cell.getValue());
			    	row.put(key, value);
			    }
		    }
		} catch (IOException e) {
			e.printStackTrace();
			result=-2;
		} catch(NumberFormatException e){
			GlobalPara.logger.warning("Input Row Key is not in type Integer: "+strRowKey);
			result=-1;
		}
		long end=System.currentTimeMillis();
		LoadDataLogItem ld=new LoadDataLogItem(end-start, end);
		GlobalPara.monitor.writelnLDLog(ld.writeToString());
		return result;
	}

}
