package org.zhouw.stps.adapter;


import com.amazonaws.*;
//import com.amazonaws.AmazonServiceException;
import com.amazonaws.auth.*;
import com.amazonaws.services.simpledb.*;
import com.amazonaws.services.simpledb.model.*;

import org.zhouw.stps.app.otpcw.TPCWAPPDataModel; 

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.locks.ReentrantLock;

import java.util.ArrayList;
import java.util.List;


import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.MsgBroker;
import org.zhouw.stps.monitor.CheckpointLogItem;
import org.zhouw.stps.monitor.LoadDataLogItem;
import org.zhouw.stps.tp.BufferForCheckpoint;
import org.zhouw.stps.tp.UpdatedDataRow;
import org.zhouw.stps.app.*;



public class SimpleDBAdapter implements Adapter {
	
	public static long timeoutDelayTime=1000;
	
	public static AmazonSimpleDBAsync sdb = initiateSimpleDBAsyncAccess();
	
	private static AmazonSimpleDBAsync initiateSimpleDBAsyncAccess(){
		try {
			return new AmazonSimpleDBAsyncClient(new PropertiesCredentials(new File(GlobalPara.sysConf.inPath+"AwsCredentials.properties")));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public void commitAll(){
	}
	
	
	public SimpleDBAdapter() {
	}
	
	
	/**
	 * 
	 * @param tableSchema
	 * @param rowid The integer form of row key
	 * @return
	 */
	public String getSimpleDBTableName(DataTable tableSchema,int rowid)
	{
		int totalNum=getNumOfTablePartitions(tableSchema);
		int index=rowid%totalNum;
		String domainName=GlobalPara.getTableName(index,tableSchema.getPhysicalTableName());
		return domainName;
	}

	
	public void createTable(Iterable <? extends DataTable> tableSchemas){
		if(sdb==null){
			GlobalPara.logger.warning("The simpledb access object is not initiated.");
			return;
		}
		if(tableSchemas==null){
			GlobalPara.logger.warning("Input table schemas are empty");
			return;
		}
		
		ListDomainsRequest sdbRequest = new ListDomainsRequest().withMaxNumberOfDomains(100);
        ListDomainsResult sdbResult = sdb.listDomains(sdbRequest);
        List<String> domainNames=sdbResult.getDomainNames(); 
        
		List<Future<Void>> list=new LinkedList<Future<Void>>();
		for(DataTable table:tableSchemas){
			for(int i=0;i<getNumOfTablePartitions(table);i++){
				String domainName = getSimpleDBTableName(table, i);
				if(!domainNames.contains(domainName)){
					boolean success=false;
					do{
						try{
							Future<Void> task=sdb.createDomainAsync(new CreateDomainRequest(domainName));
							list.add(task);
							success=true;
						} catch(AmazonServiceException ase){
//							System.out.println("Error Message:    " + ase.getMessage());
//				            System.out.println("HTTP Status Code: " + ase.getStatusCode());
//				            System.out.println("AWS Error Code:   " + ase.getErrorCode());
//				            System.out.println("Error Type:       " + ase.getErrorType());
//				            System.out.println("Request ID:       " + ase.getRequestId());
						}catch(AmazonClientException ace){
							//System.out.println("Error Message: " + ace.getMessage());
						}
						if(!success){
							try {
								Thread.sleep(timeoutDelayTime);
							} catch (InterruptedException e) {
								e.printStackTrace();
							}
						}
					}while(!success);
				}
			}
		}
		
		for(Future<Void> task:list){
			try {
				task.get();
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
		}
	}
	
	public void delRows(DataTable tableSchema,  Iterable<String> rows)
	{
		if(sdb==null){
			GlobalPara.logger.warning("The simpledb access object is not initiated.");
			return;
		}
		boolean success=false;
		do{
			try{
				for(String rowKey: rows){
					int rowid=CheckpointThreadExecutor.getRowID(rowKey);
					String domainName=getSimpleDBTableName(tableSchema, rowid);
					sdb.deleteAttributes(new DeleteAttributesRequest(domainName, rowKey));
				}
				success=true;
			}catch(NumberFormatException e){
				GlobalPara.logger.warning("Input Row Key is not in type Integer");
			}
			catch(AmazonServiceException ase){
//				System.out.println("Error Message:    " + ase.getMessage());
//	            System.out.println("HTTP Status Code: " + ase.getStatusCode());
//	            System.out.println("AWS Error Code:   " + ase.getErrorCode());
//	            System.out.println("Error Type:       " + ase.getErrorType());
//	            System.out.println("Request ID:       " + ase.getRequestId());
			}catch(AmazonClientException ace){
				//System.out.println("Error Message: " + ace.getMessage());
			}
			if(!success){
				try {
					Thread.sleep(timeoutDelayTime);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		} while(!success);
	}
	
	public void putRows(DataTable tableSchema,  Iterable<Map.Entry<String,LinkedHashMap<String,String>>> rows)
	{
		for(Map.Entry<String,LinkedHashMap<String,String>> row:rows){
			String rowKey=row.getKey();
			this.putOneRow(tableSchema, rowKey, row.getValue().entrySet());
		}
	}
	
	public void putOneRow(DataTable tableSchema,  String rowKey, Iterable<Map.Entry<String,String>> row)
	{
		if(sdb==null){
			GlobalPara.logger.warning("The simpledb access object is not initiated.");
			return;
		}
		
		boolean success=false;
		do{
			try{
				int rowid=CheckpointThreadExecutor.getRowID(rowKey);
				String domainName=getSimpleDBTableName(tableSchema, rowid);
				
				List<ReplaceableAttribute> update=new LinkedList<ReplaceableAttribute>();
				
				for(Map.Entry<String, String> e: row){
					update.add(new ReplaceableAttribute(e.getKey(),e.getValue(), true));
				}
				sdb.putAttributes(new PutAttributesRequest(domainName, rowKey, update));
				
				success=true;
			}catch(NumberFormatException e){
				GlobalPara.logger.warning("Input Row Key is not in type Integer: "+rowKey);
			}
			catch(AmazonServiceException ase){
//				System.out.println("Error Message:    " + ase.getMessage());
//	            System.out.println("HTTP Status Code: " + ase.getStatusCode());
//	            System.out.println("AWS Error Code:   " + ase.getErrorCode());
//	            System.out.println("Error Type:       " + ase.getErrorType());
//	            System.out.println("Request ID:       " + ase.getRequestId());
			}catch(AmazonClientException ace){
				//System.out.println("Error Message: " + ace.getMessage());
			}
			if(!success){
				try {
					Thread.sleep(timeoutDelayTime);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		} while(!success);
	}
	

	
	public int getNumOfTablePartitions(DataTable table)
	{
		if(table.database.getDBName().compareTo(TPCWAPPDataModel.name)!=0){
			return GlobalPara.sysConf.numOfTablePartitions;
		}
		String name=table.tableName;
		int numOfTablePartitions;
		if(name.compareTo("shopping_cart")==0 || name.compareTo("shopping_cart_line")==0){
				numOfTablePartitions=20;
		}else if(name.compareTo("order_line")==0||name.compareTo("item")==0) {
				numOfTablePartitions=10;
		} else  if(name.compareTo("orders")==0||name.compareTo("cc_xacts")==0){
			numOfTablePartitions=5;
		} else if(name.compareTo("address")==0||name.compareTo("author")==0||name.compareTo("country")==0||name.compareTo("customer")==0
				||name.compareTo("index__customer__C_UNAME")==0){
			numOfTablePartitions=5;
		}
		else
			numOfTablePartitions=GlobalPara.sysConf.numOfTablePartitions;
		
		return numOfTablePartitions;
	}
	
	protected long totalCPTIme=0;
	protected long totalCPNumber=0;
	protected long totalLoadTIme=0;
	protected long totalLoadNumber=0;
	
	public void checkpointOneItem(long snapshotID, DataTable tableSchema, String rowKey,UpdatedDataRow value) {
		long start=System.currentTimeMillis();
		if(sdb==null){
			GlobalPara.logger.warning("The simpledb access object is not initiated.");
			return;
		}
		
		if(tableSchema==null){
			GlobalPara.logger.warning("Empty schema inputted");
		}
		
		
		UpdatedDataRow data=value;
	
		int rowid=CheckpointThreadExecutor.getRowID(rowKey);
		String domainName=getSimpleDBTableName(tableSchema, rowid);
		
		List<ReplaceableAttribute> update=new LinkedList<ReplaceableAttribute>();
		if(data.updatedColumns!=null) {
			for(String column:data.updatedColumns.keySet()){
				update.add(new ReplaceableAttribute(column, data.updatedColumns.get(column),true));
			}
		}
		long deleteTime=-1;
		long putTime=-1;
		long putStart=-1;
		long deleteStart=-1;
		boolean success=false;
		do{
			try{
				Future<Void> putTask=null;
				if(update.size()>0){
					putStart=System.currentTimeMillis();
					//putTask=
					sdb.putAttributes(new PutAttributesRequest(domainName, rowKey, update));
					putTime= System.currentTimeMillis()-putStart;
				}
				
				if(data.deletedColumns!=null&&data.deletedColumns.size()>0){
					deleteStart=System.currentTimeMillis();
					DeleteAttributesRequest delete= new DeleteAttributesRequest(domainName, rowKey);
					for(String column:data.deletedColumns){
						delete.withAttributes(new Attribute().withName(column));
					}
					sdb.deleteAttributes(delete);
					deleteTime= System.currentTimeMillis()-deleteStart;
				}
				
//				if(putTask!=null)
//					putTask.get();
				success=true;
			} catch(AmazonServiceException ase){
				System.out.println("Load data from "+tableSchema.getPhysicalTableName()
						+ "-"+ rowKey +" Error Message:    " + ase.getMessage());
//	            System.out.println("HTTP Status Code: " + ase.getStatusCode());
//	            System.out.println("AWS Error Code:   " + ase.getErrorCode());
//	            System.out.println("Error Type:       " + ase.getErrorType());
//	            System.out.println("Request ID:       " + ase.getRequestId());
			} catch(AmazonClientException ace){
				System.out.println("Error Message: " + ace.getMessage());
			}
//			catch (InterruptedException e) {
//				e.printStackTrace();
//			} catch (ExecutionException e) {
//				e.printStackTrace();
//			}
			
			if(!success){
				try {
					Thread.sleep(timeoutDelayTime);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}while(!success);
		long end=System.currentTimeMillis();
		this.totalCPTIme+=end-start;
		this.totalCPNumber++;
		if(this.totalCPNumber%1000==0){
			System.out.println("Average time for checkpoint one data item is "+(double)totalCPTIme/(double)totalCPNumber+" ms");
		}
		
		if(putTime!=-1){
			LoadDataLogItem ld=new LoadDataLogItem(putTime, putStart, domainName, false);
			GlobalPara.monitor.writelnLDLog(ld.writeToString());
		}
		
		if(deleteTime!=-1){
			LoadDataLogItem ld=new LoadDataLogItem(deleteTime, deleteStart, domainName, false);
			GlobalPara.monitor.writelnLDLog(ld.writeToString());
		}
		
	}
	
	public int checkpoint(BufferForCheckpoint datapool) {
		return 1;
	}
	
	public int loadData(DataTable schema, String strRowKey, HashMap<String, String> row)
	{
		if(sdb==null){
			GlobalPara.logger.warning("The simpledb access object is not initiated.");
			return -2;
		}
		
		long start=System.currentTimeMillis();
		int result=0;
		if(schema==null){
			GlobalPara.logger.warning("Empty schema inputted");
			return -1;
		}

		int rowid=CheckpointThreadExecutor.getRowID(strRowKey);
		String domainName=getSimpleDBTableName(schema, rowid);

		GetAttributesRequest getRequest=new GetAttributesRequest(domainName, strRowKey);
		
		boolean success=false;
		do{
			try{
				GetAttributesResult item=sdb.getAttributes(getRequest);
				List<Attribute> list=item.getAttributes();
				if(list.size()==0)
					return -1;
				for(Attribute a:item.getAttributes()){
					row.put(a.getName(), a.getValue());
				}
				 
				success=true;
			} catch(AmazonServiceException ase){
				System.out.println("Load data from "+schema.getPhysicalTableName()
						+ "-"+ strRowKey +" Error Message:    " + ase.getMessage());
//	            System.out.println("HTTP Status Code: " + ase.getStatusCode());
//	            System.out.println("AWS Error Code:   " + ase.getErrorCode());
//	            System.out.println("Error Type:       " + ase.getErrorType());
//	            System.out.println("Request ID:       " + ase.getRequestId());
			}catch(AmazonClientException ace){
				System.out.println("Error Message: " + ace.getMessage());
			}
				
			if(!success){
				try {
					Thread.sleep(timeoutDelayTime*5);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		} while(!success);
		
		long end=System.currentTimeMillis();
		this.totalLoadTIme+=end-start;
		this.totalLoadNumber++;
		if(this.totalLoadNumber%1000==0){
			System.out.println("Average time for load one data item is "+(double)totalLoadTIme/(double)totalLoadNumber+" ms");
		}
		
		LoadDataLogItem ld=new LoadDataLogItem(end-start, end, domainName, true);
		GlobalPara.monitor.writelnLDLog(ld.writeToString());
		return result;
	}
}
