package org.zhouw.stps.app;

import java.util.Collection;
import java.util.Vector;
import java.util.Hashtable;
import org.zhouw.stps.app.join.*;
import org.zhouw.stps.app.otpcw.TPCWAPPDataModel;
import org.zhouw.stps.app.tpcw.*;
import org.zhouw.stps.GlobalPara;

public abstract class DataModel {

	public static DataModel getDataModel(String dbName)
	{
		return DataModel.dataModels.get(dbName);
	}
	
	public static Collection<DataModel> getAllDataModel()
	{
		return DataModel.dataModels.values();
	}
	
	/**
	 * The key is the database Name
	 */
	public static Hashtable<String, DataModel> dataModels= initDataModels();
	public static Hashtable<String, DataModel> initDataModels()
	{
		Hashtable<String, DataModel> result=new Hashtable<String, DataModel>();
		JoinAppDataModel joinDM=new JoinAppDataModel();
		result.put(joinDM.getDBName(), joinDM);
		
		TPCWAPPDataModel tpcwDM=new TPCWAPPDataModel();
		result.put(tpcwDM.getDBName(), tpcwDM);
		
		
		TPCWDataModel shoppingDM=new TPCWDataModel();
		result.put(shoppingDM.getDBName(), shoppingDM);
		
		return result;
	}
	
	public static final String joinCacheColumnFamily="Ref:";
	public static final String refRowsCacheColDelimiter="::";
	public static final String physicalTablePrefixDelimiter="__"; 
	
	public static String getPKNameFromMultipleKeys(String[] keys)
	{
		if(keys==null||keys.length==0)
			return null;
		String result="";
		for(int i=0;i<keys.length;i++){
			if(keys[i]==null)
				return null;
			if(i==0)
				result=keys[i];
			else
				result+="_"+keys[i];
		}
		return result;
	}
	
	public static boolean isRefCacheColumn(String physicalColumnName) {
		if(physicalColumnName!=null)
			return physicalColumnName.startsWith(joinCacheColumnFamily);
		else
			return false;
	}
	
	
	public static String getLogicalTableName(String dbName, String physicalTableName)
	{
		int count=dbName.length()+physicalTablePrefixDelimiter.length();
		return physicalTableName.substring(count);
	}
	
	public static String getPhysicalTableName(String dbName, String logicalTableName)
	{
		return dbName+physicalTablePrefixDelimiter+logicalTableName;
	}
	
	
	/**
	 * 
	 * @param baseTableName
	 * @param keyName
	 * @return Logical table
	 */
	public static String getIndexTableLogicalName(String baseLogicalTableName, String keyName)
	{
		return "index"+"__"+baseLogicalTableName+"__"+keyName;
	}
	
	/**
	 * 
	 * @param keyName
	 * @return Logical table
	 */
	public static String getIndexTablePKName(String baseIndexedKeyName)
	{
		return "in"+"_"+baseIndexedKeyName;
	}
	
//	public static String getIndexRowsColumnPrefix(
//			String baseKeyName)
//	{
//		return DataModel.indexTableColumnFamily+"pk";
//	}
	
	/**
	 * The input table name and column name, must be logical
	 */
	public static String getReferenceRowsCacheColumnName(
			String joinTableLogicalName, String joinFKName, String joinRowPKValue)
	{
		return DataModel.joinCacheColumnFamily+joinTableLogicalName+DataModel.refRowsCacheColDelimiter+joinFKName+DataModel.refRowsCacheColDelimiter+joinRowPKValue;
	}
	
	/**
	 * The input table name and column name, must be logical
	 */
	public static String getReferenceRowsCacheColumnPrefix(
			String joinTableName, String joinFKName)
	{
		return DataModel.joinCacheColumnFamily+joinTableName+DataModel.refRowsCacheColDelimiter+joinFKName;
	}
	
	
	public String dbname;
	
	/**
	 * The key is the logical table name
	 */
	public Hashtable<String, DataTable> tables;
	
	/**
	 * The key is the physical table name
	 */
	protected Hashtable<String, DataTable> tablesByPhysicalName;
	
	public Vector<ForeighKey> fks;
	
	
	public DataModel(String name)
	{
		tables=new Hashtable<String, DataTable>();
		tablesByPhysicalName=new Hashtable<String, DataTable>();
		fks=new Vector<ForeighKey>();
		this.dbname=name;
	}
	
	/**
	 * FK to PK 
	 * @param baseTableName
	 * @param fkTableName
	 * @param fkAttrName
	 */
	protected void inputOneFKtoPK(String baseTableName, String fkTableName, String fkAttrName, boolean isPKtoFK) 
	{
		if(GlobalPara.sysConf.isDisableIndex)
			isPKtoFK=false;
		
		ForeighKey fk=new ForeighKey();
		fk.BKTable=this.tables.get(baseTableName);
		fk.BKName=fk.BKTable.primaryKeyName;//.toLowerCase();
		fk.FKTable=this.tables.get(fkTableName);
		fk.FKName=fkAttrName;//.toLowerCase();
		fk.isAccessByPktoFK=isPKtoFK;
		this.fks.add(fk);
		this.inputOneFK(fk);
	}
	
	protected void fillTablesByPhysical()
	{
		this.tablesByPhysicalName.clear();
		for(DataTable t:tables.values()) {
			this.tablesByPhysicalName.put(t.physicalTableName,t);
		}
	}
	
	public String getDBName()
	{
		return this.dbname;
	}
	
	
//	/**
//	 * Expected to be overidden
//	 * @param logicalColName
//	 * @return
//	 */
//	public abstract String getPhysicalColumnName(String logicalColName);
	
	
	/**
	 * the input is logical table name
	 */
	public boolean hasTable(String tblName)
	{
		if(this.tables!=null)
		{
			return this.tables.containsKey(tblName);
		}
		return false;
	}
	
	protected void inputOneTable(DataModel dm, String tablename, String pkname, String[] cols)
	{
		DataTable table=new DataTable(dm,tablename,pkname,cols);
		this.tables.put(table.tableName,table);
	}
	
	protected void inputOneIndexTable(String baseTableName, String baseKeyName)
	{
//		baseTableName=baseTableName.toLowerCase();
//		baseKeyName=baseKeyName.toLowerCase();
		DataTable table=new DataTable(this,DataModel.getIndexTableLogicalName(baseTableName, baseKeyName),
				DataModel.getIndexTablePKName(baseKeyName), 
				this.tables.get(baseTableName), baseKeyName);
		this.tables.put(table.tableName,table);
		this.tables.get(baseTableName).indexes.put(baseKeyName, table);
		inputOneFKtoPK(table.tableName, baseTableName, baseKeyName, true); //always enable PKtoFK for index table
	}
	
	protected void inputOneFK(ForeighKey fk)
	{
		fk.BKTable.linkedFKs.add(fk);
		fk.FKTable.selfFKs.add(fk);
	}
	
	public DataTable getTable(String tblName)
	{
		return this.tables.get(tblName);
	}
	
	public DataTable getTableByPhysicalName(String physicalTblName)
	{
		return this.tablesByPhysicalName.get(physicalTblName);
	}
	
	public Collection<DataTable> getAllTables()
	{
		return this.tables.values();
	}
	
	/**
	 * An empty Vector if not match
	 * @return
	 */
	public Vector<DataTable> getTableByColumn(String columnName)
	{
		Vector<DataTable> result= new  Vector<DataTable>();
		for(DataTable t:this.tables.values())
		{
			if(t.columns.contains(columnName))
				result.add(t);	
		}
		return result;
	}
}
