package org.zhouw.stps.app.common;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

import org.zhouw.stps.SubTransaction;
import org.zhouw.stps.TranState;

import org.zhouw.stps.app.*;

public class InsertSubTransaction extends SubTransaction {

	/**
	 * All input column names must be logical.
	 * The input table name must be logical, because it will be explictly transformed with other type of subtransaction.
	 * @param dbName
	 * @param strTblName
	 * @param strRowKey
	 * @param colsToBeInserted
	 */
	public InsertSubTransaction(String dbName,
			String strTblName, String strRowKey,
			Iterable <? extends Map.Entry<String, String>> colsToBeInserted) {
		super("InsertSubTransaction", dbName, strTblName, strRowKey, null,
				false, true, false);
		Hashtable<String, String> parameters=new Hashtable<String, String>();
		if(colsToBeInserted!=null){
			for(Map.Entry<String, String> entry: colsToBeInserted) {
				parameters.put(entry.getKey(),entry.getValue());
			}
		}
		super.parameters=parameters;
	}

	public InsertSubTransaction(String dbName,
			String strTblName, String strRowKey,
			 Hashtable<String, String> parameters) {
		super("InsertSubTransaction", dbName, strTblName, strRowKey, parameters,
				false, true, false);
	}
	
	public TranState run()
	{
		TranState result;
		try{
			super.run();
			HashMap<String, String> data=this.rootEntry.committedData.rowData;
			boolean isAbort=false;
			if(data.size()!=0) {
				for(String key:data.keySet()){
					if(!key.startsWith(DataModel.joinCacheColumnFamily)) {
						isAbort=true;
						break;
					}
				}
			}
			if(isAbort)
				return TranState.ABORT; // abort if the row is empty or only contains the cache of referenced columns
			
			this.DataToPut=new String[super.parameters.size()][2];
			int count=0;
			for(String key: super.parameters.keySet()) {
				String value=super.parameters.get(key);
				this.DataToPut[count][0]=key;
				this.DataToPut[count][1]=value;
				count++;
			}
			this.DataToDelete=null;
			this.DataToReturn=null;
			result=TranState.COMMIT;
		}catch(java.lang.Exception e) {
			e.printStackTrace();
			result=TranState.ABORT;
		}
		return result;
	}

}
