package org.zhouw.stps.app.common;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.lang.NumberFormatException;

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.SubTransaction;
import org.zhouw.stps.TranState;

public class UpdateSubTransaction extends SubTransaction {

	public static final String toDelete=":Delete";
	public static final String toReturn=":Return";
	public static final String formulaPrefix="::";
	protected static String getToDeleteParaName(String value)
	{
		return UpdateSubTransaction.toDelete+"_"+value;
	}
	
	protected static String getToReturnParaName(int index)
	{
		return  UpdateSubTransaction.toReturn+"_"+index;
	}
	/**
	 * All input column names must be logical.
	 * The input table name must be logical, because it will be explictly transformed with other type of subtransaction.
	 * @param dbName
	 * @param strTblName  Must be logical
	 * @param strRowKey
	 * @param colsToBeDeleted
	 * @param colsToBeUpdated
	 * @param colsToBeReturn
	 */
	public UpdateSubTransaction(String dbName,
			String strTblName, String strRowKey,
			Iterable <String> colsToBeDeleted, Iterable <? extends Map.Entry<String, String>> colsToBeUpdated, 
			Iterable <String> colsToBeReturn) {
		super("UpdateSubTransaction", dbName, strTblName, strRowKey, null,
				false, false, false);
		Hashtable<String, String> parameters=new Hashtable<String, String>();
		int count=0;
		if(colsToBeDeleted!=null){
			for(String col:colsToBeDeleted)
				parameters.put(UpdateSubTransaction.getToDeleteParaName(col), col);
		}
		
		if(colsToBeUpdated!=null){
			for(Map.Entry<String, String> entry: colsToBeUpdated) {
				parameters.put(entry.getKey(),entry.getValue());
			}
		}
		
		count=0;
		if(colsToBeReturn!=null){
			for(String col:colsToBeReturn)
				parameters.put(UpdateSubTransaction.getToReturnParaName(count++), col);
		}
		
		super.parameters=parameters;
	}

	public void mergeFrom(UpdateSubTransaction t)
	{
		if(t!=null&&t.parameters!=null)
			for(String key:t.parameters.keySet()){ 
				super.parameters.put(key, t.parameters.get(key));
			}
	}
	
	public UpdateSubTransaction(String dbName,
			String strTblName, String strRowKey, Hashtable<String, String> parameters	) {
		super("UpdateSubTransaction", dbName, strTblName, strRowKey, parameters,
				false, false, false);
	}
	
	/**
	 * It will return commit, even if the row does not exist or is empty.
	 * It will return abort, only when there is a java exception in the execution.
	 */
	public TranState run()
	{
		TranState result;
		try{
			super.run();
			HashMap<String, String> data= this.rootEntry.getCommittedData();
			List<String> colsToBeDeleted=new LinkedList<String>();
			List<String> colsToReturn=new LinkedList<String>(); 
			HashMap<String, String> colsToBeUpdated=new HashMap<String, String>(); 
	
			for(String key: super.parameters.keySet()) {
				String value=super.parameters.get(key);
				if(key.startsWith(UpdateSubTransaction.toDelete))
					colsToBeDeleted.add(value);
				else if(key.startsWith(UpdateSubTransaction.toReturn))
					colsToReturn.add(value);
				else if(key.startsWith(UpdateSubTransaction.formulaPrefix)){
					String colname=key.substring(UpdateSubTransaction.formulaPrefix.length());
					String dataToPut=computeFormula(value);
					if(dataToPut!=null)
						colsToBeUpdated.put(colname, dataToPut);
					else
						GlobalPara.logger.info("Incorrect formula input for updates: "+ value);
				} else
					colsToBeUpdated.put(key, value);
			}
			if(colsToBeDeleted.size()!=0) 
				this.DataToDelete=colsToBeDeleted.toArray(new String[colsToBeDeleted.size()]); 
			else
				this.DataToDelete=null;
				
			if(colsToBeUpdated.size()!=0){
				this.DataToPut=new String[colsToBeUpdated.size()][2];
				int count=0;
				for(String key: colsToBeUpdated.keySet()) {
					this.DataToPut[count][0]=key;
					this.DataToPut[count][1]=colsToBeUpdated.get(key);
					count++;
				}
			} else {
				this.DataToPut=null;
			}
			
			if(colsToReturn.size()!=0){
				this.DataToReturn=new String[colsToReturn.size()][2];
				int count=0;
				for(String key: colsToReturn) {
					this.DataToReturn[count][0]=key;
					String value=data.get(key);
					if(value!=null)
						this.DataToReturn[count][1]=value;
					else
						this.DataToReturn[count][1]=""; //means the column not exist
					count++;
				}
			} else {
				this.DataToReturn=null;
			}
			result=TranState.COMMIT;
		}catch(java.lang.Exception e) {
			e.printStackTrace();
			result=TranState.ABORT;
		}
		return result;
	}
	
	//Formula format: columnName+/-Integer
	protected String computeFormula(String formula)
	{
		if(formula==null)
			return null;
		try{
			int index=formula.indexOf("+");
			String operator;
			if(index!=-1)
				operator="+";
			else {
				index=formula.indexOf("-");
				if(index!=-1)
					operator="-";
				else
					return null;
			}
			
			String[] ops=formula.split(operator);
			if(ops.length!=2)
				return null;
			ops[0]=ops[0].trim();
			ops[1]=ops[1].trim();
			HashMap<String, String> row= this.rootEntry.getCommittedData();
			String value=row.get(ops[0]);
			if(value!=null){
				int oldvalue=Integer.parseInt(value);
				int operand=Integer.parseInt(ops[1]);
				int result;
				if(operator.compareTo("+")==0)
					result=oldvalue+operand;
				else
					result=oldvalue-operand;
				return String.valueOf(result);	
			} else {
				if(GlobalPara.sysConf.isDebug){
					System.out.println("[DEBUG]"+this.strTblName+":"+this.strRowKey+ " does not contain the attribute "+ops[0]);
					System.out.println("[DEBUG]operator="+operator);
					System.out.println("[DEBUG]formula="+formula+"; ops[0]="+ops[0]+";ops[1]="+ops[1]);
					System.out.println("[DEBUG]data:");
					for(String attribute:row.keySet()){
						System.out.println("[DEBUG]"+attribute+" = "+row.get(attribute));
					}
					System.out.println("[DEBUG]End.");
				}
				return null;
			}
		} catch(	NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}
}
