package org.zhouw.stps.app.common;

import java.util.Random;

public class ZipfGenerator {
 private Random rnd = new Random(System.currentTimeMillis());
 private int size;
 private double skew;
 private double bottom = 0;
 
 public ZipfGenerator(int size, double skew) {
  this.size = size;
  this.skew = skew;
 
  for(int i=1;i<size; i++) {
  this.bottom += (1/Math.pow(i, this.skew));
  }
 }
 private static ZipfGenerator zipstatic;
 // the next() method returns an rank id. The frequency of returned rank ids are follows Zipf distribution.
 public int next() {
   int rank;
   double friquency = 0;
   double dice;
 
   do {
     rank = rnd.nextInt(size)+1;
     friquency = (1.0d / Math.pow(rank, this.skew)) / this.bottom;
     dice = rnd.nextDouble();
   }while(!(dice < friquency));
 
   return rank;
 }
 
 // This method returns a probability that the given rank occurs.
 public double getProbability(int rank) {
   return (1.0d / Math.pow(rank, this.skew)) / this.bottom;
 }
 
 public static void main(String[] args) {

	 
	 
	 int number=100000;
	 int size=1000;
	 int skew=1;
	 System.out.println("With number="+number+"; size="+size+"; skew="+skew);
	 
	 ZipfGenerator.zipstatic=new ZipfGenerator(size,skew);
	 
	 
	 int[] countMath=new int[size+1];
	 int[] countGen=new int[size+1];
	 double[] proMath=new double[size+1];
	 double[] proGen=new double[size+1];
	 for(int i=0;i<size+1;i++){
		 countMath[i]=0;
		 countGen[i]=0;
	 }
   
long start=System.currentTimeMillis();

   for(int i=0;i< number;i++){
	   ZipfGenerator zipfc= new ZipfGenerator(size, skew);
		int n=zipfc.next();
   }
	long end=System.currentTimeMillis();
	
	System.out.println("Zipf without cache: "+(end-start)+"ms");
	
	start=System.currentTimeMillis();
   ZipfGenerator zipf = new ZipfGenerator(size, skew);
    for(int i=0;i< number;i++){
			int n=zipf.next();
	}
   end=System.currentTimeMillis();
   System.out.println("Zipf with cached object: "+(end-start)+"ms");
   
   start=System.currentTimeMillis();
   Random ran=new Random();  
    for(int i=0;i< number;i++){
    	zipstatic.next();
	}
   end=System.currentTimeMillis();
   System.out.println("Zipf with static object: "+(end-start)+"ms");
   
   
//   for(int i=0;i<size+1;i++){
//	   proMath[i]=(double)countMath[i]/(double)number;
//	   proGen[i]=(double)countGen[i]/(double)number;
//   }
   
//   System.out.println("distribution of Math");
//   for(int i=0;i<size+1;i++){
//	   System.out.println(i+" "+proMath[i]);
//   }
//   System.out.println("distribution of Gen");
//   for(int i=0;i<size+1;i++){
//	   System.out.println(i+" "+proGen[i]);
//   }
	
 }
}
