package org.zhouw.stps.app.otpcw;

import java.io.IOException;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.concurrent.Future;

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.app.*;
import org.zhouw.stps.app.common.DataGenCommon;
import org.zhouw.stps.rotran.QueryPlan;
import org.zhouw.stps.adapter.*;

public class TPCWDataGen {
	    
	//ATTENTION: The NUM_EBS and NUM_ITEMS variables are the only variables
	//that should be modified in order to rescale the DB.
	private static final int NUM_EBS = GlobalPara.sysConf.NUM_EBS;
	private static final int NUM_ITEMS = GlobalPara.sysConf.NUM_ITEMS;
	
	private static final int NUM_CUSTOMERS = NUM_EBS * 2880;
	private static final int NUM_ADDRESSES = 2 * NUM_CUSTOMERS;
	private static final int NUM_AUTHORS = (int) (.25 * NUM_ITEMS);
	private static final int NUM_ORDERS = (int) (.9 * NUM_CUSTOMERS);

	private static Random rand=new Random();
	
	private static DataModel dm;
	
	public static void doDataGen()
	{
		dm=DataModel.getDataModel(TPCWAPPDataModel.name);
		DataGenCommon.createAllTables(dm);
		//DataGenCommon.createATable(dm.getTable("item"));
		populateAddressTable();
		populateAuthorTable();
		populateCountryTable();
		populateCustomerTable();
		populateItemTable();
		populateOrdersAndCC_XACTSTable();
		GlobalPara.adapterComp.commitAllWorkers();
		System.out.print("Waiting for data commit to complete");
		while(true){
			if(GlobalPara.adapterComp.isAllWorkersCommitFinish()){
				break;
			} else {
				System.out.print(".");
				try {
					Thread.sleep(5000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
		System.out.println("done");
		System.out.println("Data Generation successfully completed");
		System.out.println("Compacting all tables");
		if(GlobalPara.sysConf.DB.compareTo("HBase")==0){
			HBaseAdapter.compactTables(dm.getAllTables());
		}
		System.out.println("Done");
	}
	
	public static void main(String args[])
	{
		doDataGen();
	}
	
	   private static void populateOrdersAndCC_XACTSTable(){
			GregorianCalendar cal;
			String[] credit_cards = {"VISA", "MASTERCARD", "DISCOVER", 
			                          "AMEX", "DINERS"};
			int num_card_types = 5;
			String[] ship_types = {"AIR", "UPS", "FEDEX", "SHIP", "COURIER",
					       "MAIL"};
			int num_ship_types = 6;
			
			String[] status_types = {"PROCESSING", "SHIPPED", "PENDING", 
						 "DENIED"};
			int num_status_types = 4;

			//Order variables
			int O_C_ID;
			java.sql.Timestamp O_DATE;
			double O_SUB_TOTAL;
			double O_TAX;
			double O_TOTAL;
			String O_SHIP_TYPE;
			java.sql.Timestamp O_SHIP_DATE;
			int O_BILL_ADDR_ID, O_SHIP_ADDR_ID;
			String O_STATUS;

			String CX_TYPE;
			int CX_NUM;
			String CX_NAME;
		        java.sql.Date CX_EXPIRY;
			String CX_AUTH_ID;
			double CX_XACT_AMT;
			int CX_CO_ID;
			
			System.out.println("Populating ORDERS, ORDER_LINES, CC_XACTS with "
					   + NUM_ORDERS + " orders");
			
			System.out.print("Complete (in 10,000's): ");
			
			DataTable tableOrder=dm.getTable("orders");
			DataTable tableOrderLine=dm.getTable("order_line");
			DataTable tableCC=dm.getTable("cc_xacts");
			
		    int countOrderLine=0;
		  //  List<Future<Long>> result=new LinkedList<Future<Long>>();
		    for(int i = 0; i < NUM_ORDERS; i++){
		    	if(i % 2000 == 0){
				    System.out.print(i/2000+" ");
				    try {
						Thread.sleep(5000);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				String pkValueOrder=String.valueOf(i);
				LinkedHashMap<String,String> rowDataOrder=new LinkedHashMap<String,String>();
				
				int num_items = getRandomInt(1,5);
				O_C_ID = getRandomInt(1, NUM_CUSTOMERS);
				cal = new GregorianCalendar();
				cal.add(Calendar.DAY_OF_YEAR, -1*getRandomInt(1,60));
				O_DATE = new java.sql.Timestamp(cal.getTime().getTime());
				O_SUB_TOTAL = (double) getRandomInt(1000, 999999)/100;
				O_TAX = O_SUB_TOTAL * 0.0825;
				O_TOTAL = O_SUB_TOTAL + O_TAX + 3.00 + num_items;
				O_SHIP_TYPE = ship_types[getRandomInt(0,num_ship_types-1)];
				cal.add(Calendar.DAY_OF_YEAR, getRandomInt(0,7));
				O_SHIP_DATE = new java.sql.Timestamp(cal.getTime().getTime());

				O_BILL_ADDR_ID = getRandomInt(1, 2*NUM_CUSTOMERS);
				O_SHIP_ADDR_ID = getRandomInt(1, 2*NUM_CUSTOMERS);
				O_STATUS = status_types[getRandomInt(0,num_status_types-1)];
				
				// Set parameter
				rowDataOrder.put("O_C_ID",  String.valueOf(O_C_ID));
				rowDataOrder.put("O_DATE", O_DATE.toString());
				rowDataOrder.put("O_SUB_TOTAL",  String.valueOf(O_SUB_TOTAL));
				rowDataOrder.put("O_TAX",  String.valueOf(O_TAX));
				rowDataOrder.put("O_TOTAL",  String.valueOf(O_TOTAL));
				rowDataOrder.put("O_SHIP_TYPE", O_SHIP_TYPE);
				rowDataOrder.put("O_SHIP_DATE", O_SHIP_DATE.toString());
				rowDataOrder.put("O_BILL_ADDR_ID",  String.valueOf(O_BILL_ADDR_ID));
				rowDataOrder.put("O_SHIP_ADDR_ID",  String.valueOf(O_SHIP_ADDR_ID));
				rowDataOrder.put("O_STATUS", O_STATUS);

				DataGenCommon.putOneRow(tableOrder, pkValueOrder, rowDataOrder);
				
				for(int j = 0; j < num_items; j++){
				    String pkValueOrderLine=DataModel.getPKNameFromMultipleKeys(
				    		new String[]{String.valueOf(i),String.valueOf(j)});
				    LinkedHashMap<String,String> rowDataOrderLine=new LinkedHashMap<String,String>();
				    int OL_O_ID = i;
				    int OL_I_ID = getRandomInt(1, NUM_ITEMS);
				    int OL_QTY = getRandomInt(1, 300);
				    double OL_DISCOUNT = (double) getRandomInt(0,30)/100;
				    String OL_COMMENTS = getRandomAString(20,100);
				    rowDataOrderLine.put("OL_O_ID",  String.valueOf(OL_O_ID));
				    rowDataOrderLine.put("OL_I_ID",  String.valueOf(OL_I_ID));
				    rowDataOrderLine.put("OL_QTY",  String.valueOf(OL_QTY));
				    rowDataOrderLine.put("OL_DISCOUNT",  String.valueOf(OL_DISCOUNT));
				    rowDataOrderLine.put("OL_COMMENTS", OL_COMMENTS);
				    
				    DataGenCommon.putOneRow(tableOrderLine, pkValueOrderLine, rowDataOrderLine);
				}
				
				LinkedHashMap<String,String> rowDataCC=new LinkedHashMap<String,String>();
				CX_TYPE = credit_cards[getRandomInt(0, num_card_types-1)];
				CX_NUM = getRandomNString(16);
				CX_NAME = getRandomAString(14,30);
				cal = new GregorianCalendar();
				cal.add(Calendar.DAY_OF_YEAR, getRandomInt(10, 730));
				CX_EXPIRY = new java.sql.Date(cal.getTime().getTime());
				CX_AUTH_ID = getRandomAString(15);	      		
				CX_CO_ID = getRandomInt(0,91);

				rowDataCC.put("CX_TYPE", CX_TYPE);
				rowDataCC.put("CX_NUM", String.valueOf(CX_NUM));
				rowDataCC.put("CX_NAME", CX_NAME);
				rowDataCC.put("CX_EXPIRY", String.valueOf(CX_EXPIRY));
				rowDataCC.put("CX_AUTH_ID", CX_AUTH_ID);
				rowDataCC.put("O_TOTAL", String.valueOf(O_TOTAL));
				rowDataCC.put("O_SHIP_DATE", O_SHIP_DATE.toString());
				rowDataCC.put("CX_CO_ID", String.valueOf(CX_CO_ID));

				DataGenCommon.putOneRow(tableCC, pkValueOrder, rowDataCC);//share the same PK
		    }
				
	}
	  private static void populateItemTable() {
			String I_TITLE;
			GregorianCalendar cal;
			int I_A_ID;
			java.sql.Date I_PUB_DATE;
			String I_PUBLISHER, I_SUBJECT, I_DESC;
			int I_RELATED1, I_RELATED2, I_RELATED3, I_RELATED4, I_RELATED5;
			String I_THUMBNAIL, I_IMAGE;
			double I_SRP, I_COST;
			java.sql.Date I_AVAIL;
		        int I_STOCK;
			String I_ISBN;
			int I_PAGE;
			String I_BACKING;
			String I_DIMENSIONS;

			String[] SUBJECTS = { "ARTS", "BIOGRAPHIES", "BUSINESS", "CHILDREN",
					      "COMPUTERS", "COOKING", "HEALTH", "HISTORY",
					      "HOME", "HUMOR", "LITERATURE", "MYSTERY",
					      "NON-FICTION", "PARENTING", "POLITICS",
					      "REFERENCE", "RELIGION", "ROMANCE", 
					      "SELF-HELP", "SCIENCE-NATURE", "SCIENCE_FICTION",
					      "SPORTS", "YOUTH", "TRAVEL"};
			int NUM_SUBJECTS = 24;

			String[] BACKINGS = { "HARDBACK", "PAPERBACK", "USED", "AUDIO",
					      "LIMITED-EDITION"};
			int NUM_BACKINGS = 5;
			
			System.out.println("Populating ITEM table with " + NUM_ITEMS + 
					   " items");
			
			DataTable table=dm.getTable("item");
			

			//	("INSERT INTO ITEM ( I_ID, I_TITLE , I_A_ID, I_PUB_DATE, I_PUBLISHER, I_SUBJECT, I_DESC, I_RELATED1, I_RELATED2, I_RELATED3, I_RELATED4, I_RELATED5, I_THUMBNAIL, I_IMAGE, I_SRP, I_COST, I_AVAIL, I_STOCK, I_ISBN, I_PAGE, I_BACKING, I_DIMENSIONS) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");
			    for (int i = 0; i < NUM_ITEMS; i++) {
			    	
			    	if(i % 10000 == 0){
					    System.out.print(i/10000+" ");
//					    try {
//							Thread.sleep(5000);
//						} catch (InterruptedException e) {
//							e.printStackTrace();
//						}
					}
			    	LinkedHashMap<String,String> rowData=new LinkedHashMap<String,String>();
		    		String pkvalue=String.valueOf(i);
		    		
				int month,day,year,maxday;
				I_TITLE= getRandomAString(14,60);
				if(i < (NUM_AUTHORS))
				    I_A_ID = i;
				else 
				    I_A_ID = getRandomInt(0, NUM_AUTHORS-1);

				year = getRandomInt(1930, 2000);
				month = getRandomInt(0,11);
				maxday = 31;
				if (month == 3 | month ==5 | month == 8 | month == 10)
				    maxday = 30;
				else if (month == 1)
				    maxday = 28;
				day = getRandomInt(1, maxday);
				cal = new GregorianCalendar(year,month,day);
				I_PUB_DATE = new java.sql.Date(cal.getTime().getTime());

				I_PUBLISHER = getRandomAString(14,60);
				I_SUBJECT = SUBJECTS[getRandomInt(0,NUM_SUBJECTS-1)];
				I_DESC = getRandomAString(100,500);
				
				I_RELATED1 = getRandomInt(1, NUM_ITEMS);
				do {
				    I_RELATED2 = getRandomInt(1, NUM_ITEMS);
				} while(I_RELATED2 == I_RELATED1);
				do {
				    I_RELATED3 = getRandomInt(1, NUM_ITEMS);
				} while(I_RELATED3 == I_RELATED1 || I_RELATED3 == I_RELATED2);
				do {
				    I_RELATED4 = getRandomInt(1, NUM_ITEMS);
				} while(I_RELATED4 == I_RELATED1 || I_RELATED4 == I_RELATED2
					|| I_RELATED4 == I_RELATED3);
				do {
				    I_RELATED5 = getRandomInt(1, NUM_ITEMS);
				} while(I_RELATED5 == I_RELATED1 || I_RELATED5 == I_RELATED2
					|| I_RELATED5 == I_RELATED3 ||
					I_RELATED5 == I_RELATED4);
			       
				I_THUMBNAIL = new String("img"+i%100+"/thumb_"+i+".gif");
				I_IMAGE = new String("img"+i%100+"/image_"+i+".gif");
				I_SRP = (double) getRandomInt(100, 99999);
				I_SRP/=100.0;
				
				I_COST = I_SRP-((((double)getRandomInt(0, 50)/100.0))*I_SRP);

				cal.add(Calendar.DAY_OF_YEAR, getRandomInt(1,30));
				I_AVAIL = new java.sql.Date(cal.getTime().getTime());
				I_STOCK = getRandomInt(10,30);
				I_ISBN = getRandomAString(13);
				I_PAGE = getRandomInt(20,9999);
				I_BACKING = BACKINGS[getRandomInt(0, NUM_BACKINGS-1)];
				I_DIMENSIONS= ((double) getRandomInt(1,9999)/100.0) +"x"+
				    ((double) getRandomInt(1,9999)/100.0) + "x" +
				    ((double) getRandomInt(1,9999)/100.0);

				// Set parameter
				rowData.put("I_TITLE", I_TITLE);
				rowData.put("I_A_ID", String.valueOf(I_A_ID));
				rowData.put("I_PUB_DATE", I_PUB_DATE.toString());
				rowData.put("I_PUBLISHER", I_PUBLISHER);
				rowData.put("I_SUBJECT", I_SUBJECT);
				rowData.put("I_DESC", I_DESC);
				rowData.put("I_RELATED1", String.valueOf(I_RELATED1));
				rowData.put("I_RELATED2", String.valueOf(I_RELATED2));
				rowData.put("I_RELATED3", String.valueOf(I_RELATED3));
				rowData.put("I_RELATED4", String.valueOf(I_RELATED4));
				rowData.put("I_RELATED5", String.valueOf(I_RELATED5));
				rowData.put("I_THUMBNAIL", I_THUMBNAIL);
				rowData.put("I_IMAGE", I_IMAGE);
				rowData.put("I_SRP", String.valueOf(I_SRP));
				rowData.put("I_COST", String.valueOf(I_COST));
				rowData.put("I_AVAIL", I_AVAIL.toString());
				rowData.put("I_STOCK", String.valueOf(I_STOCK));
				rowData.put("I_ISBN", I_ISBN);
				rowData.put("I_PAGE", String.valueOf(I_PAGE));
				rowData.put("I_BACKING", I_BACKING);
				rowData.put("I_DIMENSIONS", I_DIMENSIONS);
				
				DataGenCommon.putOneRow(table, pkvalue, rowData);
			 }
	  }
	  
	  private static void populateCountryTable(){
			String[] countries = {"United States","United Kingdom","Canada",
					      "Germany","France","Japan","Netherlands",
					      "Italy","Switzerland", "Australia","Algeria",
					      "Argentina","Armenia","Austria","Azerbaijan",
					      "Bahamas","Bahrain","Bangla Desh","Barbados",
					      "Belarus","Belgium","Bermuda", "Bolivia",
					      "Botswana","Brazil","Bulgaria","Cayman Islands",
					      "Chad","Chile", "China","Christmas Island",
					      "Colombia","Croatia","Cuba","Cyprus",
					      "Czech Republic","Denmark","Dominican Republic",
					      "Eastern Caribbean","Ecuador", "Egypt",
					      "El Salvador","Estonia","Ethiopia",
					      "Falkland Island","Faroe Island", "Fiji", 
					      "Finland","Gabon","Gibraltar","Greece","Guam",
					      "Hong Kong","Hungary", "Iceland","India",
					      "Indonesia","Iran","Iraq","Ireland","Israel",
					      "Jamaica", "Jordan","Kazakhstan","Kuwait",
					      "Lebanon","Luxembourg","Malaysia","Mexico", 
					      "Mauritius", "New Zealand","Norway","Pakistan",
					      "Philippines","Poland","Portugal","Romania", 
					      "Russia","Saudi Arabia","Singapore","Slovakia",
					      "South Africa","South Korea", "Spain","Sudan",
					      "Sweden","Taiwan","Thailand","Trinidad",
					      "Turkey","Venezuela", "Zambia"};

			double[] exchanges = { 1, .625461, 1.46712, 1.86125, 6.24238, 121.907,
					       2.09715, 1842.64, 1.51645, 1.54208, 65.3851,
					       0.998, 540.92, 13.0949, 3977, 1, .3757, 
					       48.65, 2, 248000, 38.3892, 1, 5.74, 4.7304,
					       1.71, 1846, .8282, 627.1999, 494.2, 8.278,
					       1.5391, 1677, 7.3044, 23, .543, 36.0127, 
					       7.0707, 15.8, 2.7, 9600, 3.33771, 8.7,
					       14.9912, 7.7, .6255, 7.124, 1.9724, 5.65822,
					       627.1999, .6255, 309.214, 1, 7.75473, 237.23, 
					       74.147, 42.75, 8100, 3000, .3083, .749481,
					       4.12, 37.4, 0.708, 150, .3062, 1502, 38.3892,
					       3.8, 9.6287, 25.245, 1.87539, 7.83101,
					       52, 37.8501, 3.9525, 190.788, 15180.2, 
					       24.43, 3.7501, 1.72929, 43.9642, 6.25845, 
					       1190.15, 158.34, 5.282, 8.54477, 32.77, 37.1414,
					       6.1764, 401500, 596, 2447.7 };

			String[] currencies = { "Dollars","Pounds","Dollars","Deutsche Marks",
						"Francs","Yen","Guilders","Lira","Francs",
						"Dollars","Dinars","Pesos", "Dram",
						"Schillings","Manat","Dollars","Dinar","Taka",
						"Dollars","Rouble","Francs","Dollars", 
						"Boliviano", "Pula", "Real", "Lev","Dollars",
						"Franc","Pesos","Yuan Renmimbi","Dollars",
						"Pesos","Kuna","Pesos","Pounds","Koruna",
						"Kroner","Pesos","Dollars","Sucre","Pounds",
						"Colon","Kroon","Birr","Pound","Krone",
						"Dollars","Markka","Franc","Pound","Drachmas",
						"Dollars","Dollars","Forint","Krona","Rupees",
						"Rupiah","Rial","Dinar","Punt","Shekels",
						"Dollars","Dinar","Tenge","Dinar","Pounds",
						"Francs","Ringgit","Pesos","Rupees","Dollars",
						"Kroner","Rupees","Pesos","Zloty","Escudo",
						"Leu","Rubles","Riyal","Dollars","Koruna",
						"Rand","Won","Pesetas","Dinar","Krona",
						"Dollars","Baht","Dollars","Lira","Bolivar", 
						"Kwacha"};
			
			int NUM_COUNTRIES = 92;
			
			System.out.println("Populating COUNTRY with " + NUM_COUNTRIES+ 
					   " countries");
			DataTable table=dm.getTable("country");

				//("INSERT INTO COUNTRY(CO_ID,CO_NAME,CO_EXCHANGE,CO_CURRENCY) VALUES (?,?,?,?)");
			    for(int i = 0; i < NUM_COUNTRIES; i++){
			    	LinkedHashMap<String,String> rowData=new LinkedHashMap<String,String>();
		    		String pkvalue=String.valueOf(i);
				// Set parameter
		    		rowData.put("CO_NAME", countries[i]);
		    		rowData.put("CO_EXCHANGE", String.valueOf(exchanges[i]));
		    		rowData.put("CO_CURRENCY", currencies[i]);
		    		DataGenCommon.putOneRow(table, pkvalue, rowData);
			  }
    }
	  
	 private static void populateCustomerTable(){
			String C_UNAME, C_PASSWD, C_LNAME, C_FNAME;
			int C_ADDR_ID, C_PHONE;
			String C_EMAIL;
			java.sql.Date C_SINCE, C_LAST_LOGIN;
			java.sql.Timestamp C_LOGIN, C_EXPIRATION;
			double C_DISCOUNT, C_BALANCE, C_YTD_PMT;
			java.sql.Date C_BIRTHDATE;
			String C_DATA;
			int i;
			System.out.println("Populating CUSTOMER Table with " + 
					   NUM_CUSTOMERS + " customers");
			System.out.print("Complete (in 10,000's): ");
			
			DataTable table=dm.getTable("customer");
			
			    for(i = 0; i < NUM_CUSTOMERS; i++) {
			    	if(i % 10000 == 0){
					    System.out.print(i/10000+" ");
					    try {
							Thread.sleep(5000);
						} catch (InterruptedException e) {
							e.printStackTrace();
						}
					}
				
				LinkedHashMap<String,String> rowData=new LinkedHashMap<String,String>();
		    		String pkvalue=String.valueOf(i);
		    		
				C_UNAME = DigSyl(i, 0);
				C_PASSWD = C_UNAME.toLowerCase();
				C_LNAME = getRandomAString(8,15);
				C_FNAME = getRandomAString(8,15);
				C_ADDR_ID = getRandomInt(1, 2*NUM_CUSTOMERS);
				C_PHONE = getRandomNString(9,16);
				C_EMAIL = C_UNAME+"@"+getRandomAString(2,9)+".com";
	
				GregorianCalendar cal = new GregorianCalendar();
				cal.add(Calendar.DAY_OF_YEAR,-1*getRandomInt(1,730));
				C_SINCE = new java.sql.Date(cal.getTime().getTime());
				cal.add(Calendar.DAY_OF_YEAR,getRandomInt(0,60)); 
				if(cal.after(new GregorianCalendar()))
				    cal = new GregorianCalendar();
				
				C_LAST_LOGIN = new java.sql.Date(cal.getTime().getTime());
				C_LOGIN = new java.sql.Timestamp(System.currentTimeMillis());
				cal = new GregorianCalendar();
				cal.add(Calendar.HOUR, 2);
				C_EXPIRATION = new java.sql.Timestamp(cal.getTime().getTime());
				
				C_DISCOUNT = (double) getRandomInt(0, 50)/100.0;
				C_BALANCE=0.00;
				C_YTD_PMT = (double) getRandomInt(0, 99999)/100.0;
				int year = getRandomInt(1880, 2000);
				int month = getRandomInt(0,11);
				int maxday = 31;
				int day;
				if (month == 3 | month ==5 | month == 8 | month == 10)
				    maxday = 30;
				else if (month == 1)
				    maxday = 28;
				day = getRandomInt(1, maxday);
				cal = new GregorianCalendar(year,month,day);
				C_BIRTHDATE = new java.sql.Date(cal.getTime().getTime());
	
				C_DATA = getRandomAString(100,500);

			    rowData.put("C_UNAME", C_UNAME);
			    rowData.put("C_PASSWD", C_PASSWD);
			    rowData.put("C_FNAME", C_FNAME);
			    rowData.put("C_LNAME", C_LNAME);
			    rowData.put("C_ADDR_ID", String.valueOf(C_ADDR_ID));
			    rowData.put("C_PHONE", String.valueOf(C_PHONE));
			    rowData.put("C_EMAIL", C_EMAIL);
			    rowData.put("C_SINCE", C_SINCE.toString());
			    rowData.put("C_LAST_LOGIN", C_LAST_LOGIN.toString());
			    rowData.put("C_LOGIN", C_LOGIN.toString());
			    rowData.put("C_EXPIRATION", C_EXPIRATION.toString());
			    rowData.put("C_DISCOUNT", String.valueOf(C_DISCOUNT));
			    rowData.put("C_BALANCE", String.valueOf(C_BALANCE));
			    rowData.put("C_YTD_PMT", String.valueOf(C_YTD_PMT));
			    rowData.put("C_BIRTHDATE", C_BIRTHDATE.toString());
			    rowData.put("C_DATA", C_DATA);

			    DataGenCommon.putOneRow(table, pkvalue, rowData);
	    }
    }
	 
	  private static void populateAddressTable(){
			System.out.println("Populating ADDRESS Table with " + 
					   NUM_ADDRESSES + " addresses");
			
			DataTable table=dm.getTable("address");
			
			String ADDR_STREET1, ADDR_STREET2, ADDR_CITY, ADDR_STATE;
			String ADDR_ZIP;
			int ADDR_CO_ID;
		    for (int i = 0; i < NUM_ADDRESSES; i++) {
		    	LinkedHashMap<String,String> rowData=new LinkedHashMap<String,String>();
		    	String pkvalue=String.valueOf(i);
		    	
			if(i % 10000 == 0){
			    System.out.print(i/10000+" ");
			    try {
					Thread.sleep(5000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			ADDR_STREET1 = getRandomAString(15,40);
			ADDR_STREET2 = getRandomAString(15,40);
			ADDR_CITY    = getRandomAString(4,30);
			ADDR_STATE   = getRandomAString(2,20);
			ADDR_ZIP     = getRandomAString(5,10);
			ADDR_CO_ID   = getRandomInt(0,91);
			
			// Set parameter
			rowData.put("ADDR_STREET1", ADDR_STREET1);
			rowData.put("ADDR_STREET2", ADDR_STREET2);
			rowData.put("ADDR_CITY", ADDR_CITY);
			rowData.put("ADDR_STATE", ADDR_STATE);
			rowData.put("ADDR_ZIP", ADDR_ZIP);
			rowData.put("ADDR_CO_ID", String.valueOf(ADDR_CO_ID));

			DataGenCommon.putOneRow(table, pkvalue, rowData);
		}
		 System.out.print("Complete (in 10,000's): ");
	}

	private static void populateAuthorTable(){
	    	String A_FNAME, A_MNAME, A_LNAME, A_BIO;
	    	java.sql.Date A_DOB;
	    	GregorianCalendar cal;
	    	
	    	System.out.println("Populating AUTHOR Table with " + NUM_AUTHORS +
	    			   " authors");

	    	DataTable table=dm.getTable("author");

	    	    for(int i = 0; i < NUM_AUTHORS; i++){
	    	    	if(i % 10000 == 0){
	    			    System.out.print(i/10000+" ");
	    			    try {
	    					Thread.sleep(5000);
	    				} catch (InterruptedException e) {
	    					e.printStackTrace();
	    				}
	    			}
	    	    	
	    	    	LinkedHashMap<String,String> rowData=new LinkedHashMap<String,String>();
	    	    	String rowkey=String.valueOf(i);
	    		int month, day, year, maxday;
	    		A_FNAME = getRandomAString(3,20);
	    		A_MNAME = getRandomAString(1,20);
	    		A_LNAME = getRandomAString(1,20);
	    		year = getRandomInt(1800, 1990);
	    		month = getRandomInt(0,11);
	    		maxday = 31;
	    		if (month == 3 | month ==5 | month == 8 | month == 10)
	    		    maxday = 30;
	    		else if (month == 1)
	    		    maxday = 28;
	    		day = getRandomInt(1, maxday);
	    		cal = new GregorianCalendar(year,month,day);
	    		A_DOB = new java.sql.Date(cal.getTime().getTime());
	    		A_BIO = getRandomAString(125, 500);
	    		// Set parameter
	    		rowData.put("A_FNAME", A_FNAME);
	    		rowData.put("A_LNAME", A_LNAME);
	    		rowData.put("A_MNAME", A_MNAME);
	    		rowData.put("A_DOB", A_DOB.toString());
	    		rowData.put("A_BIO", A_BIO);
	    		DataGenCommon.putOneRow(table, rowkey, rowData);
	    }
	}
	  
	  //UTILITY FUNCTIONS BEGIN HERE
	    public static String getRandomAString(int min, int max){
		String newstring = new String();
		int i;
		final char[] chars = {'a','b','c','d','e','f','g','h','i','j','k',
				      'l','m','n','o','p','q','r','s','t','u','v',
				      'w','x','y','z','A','B','C','D','E','F','G',
				      'H','I','J','K','L','M','N','O','P','Q','R',
				      'S','T','U','V','W','X','Y','Z','!','@','#',
				      '$','%','^','&','*','(',')','_','-','=','+',
				      '{','}','[',']','|',':',';',',','.','?','/',
				      '~',' '}; //79 characters
		int strlen = (int) Math.floor(rand.nextDouble()*((max-min)+1));
		strlen += min;
		for(i = 0; i < strlen; i++){
		    char c = chars[(int) Math.floor(rand.nextDouble()*79)];
		    newstring = newstring.concat(String.valueOf(c));
		}
		return newstring;
	    }
	    
	    public static String getRandomAString(int length){
		String newstring = new String();
		int i;
		final char[] chars = {'a','b','c','d','e','f','g','h','i','j','k',
				      'l','m','n','o','p','q','r','s','t','u','v',
				      'w','x','y','z','A','B','C','D','E','F','G',
				      'H','I','J','K','L','M','N','O','P','Q','R',
				      'S','T','U','V','W','X','Y','Z','!','@','#',
				      '$','%','^','&','*','(',')','_','-','=','+',
				      '{','}','[',']','|',':',';',',','.','?','/',
				      '~',' '}; //79 characters
		for(i = 0; i < length; i++){
		    char c = chars[(int) Math.floor(rand.nextDouble()*79)];
		    newstring = newstring.concat(String.valueOf(c));
		}
		return newstring;
	    }

	    public static int getRandomNString(int num_digits){
		int return_num = 0;
		for(int i = 0; i < num_digits; i++){
		    return_num += getRandomInt(0, 9) * 
			(int) java.lang.Math.pow(10.0, (double) i);
		}
		return return_num;
	    }

	    public static int getRandomNString(int min, int max){
		int strlen = (int) Math.floor(rand.nextDouble()*((max-min)+1)) + min;
		return getRandomNString(strlen);
	    }
	    
	    public static int getRandomInt(int lower, int upper){
		
		int num = (int) Math.floor(rand.nextDouble()*((upper+1)-lower));
		if(num+lower > upper || num+lower < lower){
		    System.out.println("ERROR: Random returned value of of range!");
		    System.exit(1);
		}
		return num + lower;
	    }
	  
	    public static String DigSyl(int D, int N){
		int i;
		String resultString = new String();
		String Dstr = Integer.toString(D);
		
		if(N > Dstr.length()){
		    int padding = N - Dstr.length();
		    for(i = 0; i < padding; i++)
			resultString = resultString.concat("BA");
		}
		
		for(i = 0; i < Dstr.length(); i++){
		    if(Dstr.charAt(i) == '0')
			resultString = resultString.concat("BA");
		    else if(Dstr.charAt(i) == '1')
			resultString = resultString.concat("OG");
		    else if(Dstr.charAt(i) == '2')
			resultString = resultString.concat("AL");
		    else if(Dstr.charAt(i) == '3')
			resultString = resultString.concat("RI");
		    else if(Dstr.charAt(i) == '4')
			resultString = resultString.concat("RE");
		    else if(Dstr.charAt(i) == '5')
			resultString = resultString.concat("SE");
		    else if(Dstr.charAt(i) == '6')
			resultString = resultString.concat("AT");
		    else if(Dstr.charAt(i) == '7')
			resultString = resultString.concat("UL");
		    else if(Dstr.charAt(i) == '8')
			resultString = resultString.concat("IN");
		    else if(Dstr.charAt(i) == '9')
			resultString = resultString.concat("NG");
		}
		
		return resultString;
	    }
	
}

