package org.zhouw.stps.app.otpcw;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.app.DataModel;
import org.zhouw.stps.app.DataTable;
import org.zhouw.stps.rotran.*;

public class TPCWQueries {

	public static DataModel dm=DataModel.getDataModel(TPCWAPPDataModel.name);
	
	private static Random ran=new Random(); 
	
	private static void makePlanReady(QueryPlan plan)
	{
		if(plan!=null){
			plan.transformIntoPhysicalTableName();
			plan.isValid=true;
		}
	}
	
	/**
	 * For depth = 2
	 * Select * from Item as I1, Item as I2, Item as I3, ... Where (I1.id = I2.i_related1 AND I3.i_related1 = I2.id)
	 *  AND I3.id= ?
	 * @param num Total number of accessed data items
	 * @param depth Depth of the query tree, 0: no join, 1: two-table join, >1: multi-join
	 * 
	 * to achieve depth 2, at least 3 data items 
	 * 
	 * So there are depth + 1 JoiningTables in the tree 
	 * Each node has only one child in the tree
	 *   
	 *  
	 */
	public static QueryPlan getMicroROTransaction(int numOfRootRecords, int depth)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable rootTable=new JoiningTable(0, "item");
		
		List<Integer> itemIds=new ArrayList<Integer>(numOfRootRecords); 
		for(int i=0;i<numOfRootRecords;i++){
			int itemid;
			do{
				itemid=ran.nextInt(GlobalPara.sysConf.NUM_ITEMS);
			}while(itemIds.contains(itemid));
			itemIds.add(itemid);
			rootTable.allowedRows.add(String.valueOf(itemid));
		}
		
		rootTable.projections.add("*");
		plan.tables.put(0, rootTable);
		plan.rootTable=rootTable;
		
		JoiningTable parentTable, childTable;
		parentTable=rootTable;
		for(int i=1;i<=depth;i++){
			childTable=new JoiningTable(i, "item");
			childTable.projections.add("*");
			JoinEdge edge=new JoinEdge(i-1, "I_RELATED1".toUpperCase(), JoinEdge.FK, i, "i_id".toUpperCase(), JoinEdge.PK);
			parentTable.edges.add(edge);
			childTable.edges.add(edge);
			plan.edges.add(edge);
			plan.tables.put(i, childTable);
			parentTable=childTable;
		}
		
		makePlanReady(plan);
		return plan;	
	}
	
	//SELECT c_fname,c_lname FROM customer WHERE c_id = ?
	public static QueryPlan getCustomerNameS(int c_id)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable table=new JoiningTable(0, "customer");
		table.allowedRows.add(String.valueOf(c_id));
		table.projections.add("C_FNAME");
		table.projections.add("C_LNAME");
		plan.tables.put(0, table);
		plan.rootTable=table;
		makePlanReady(plan);
		return plan;
	}
	
//	//SELECT J.i_id,J.i_thumbnail from item I, item J where 
//	//(I.i_related1 = J.i_id or I.i_related2 = J.i_id or I.i_related3 = J.i_id or I.i_related4 = J.i_id or I.i_related5 = J.i_id)
//	//and I.i_id = ?
//	public static QueryPlan getRelatedItemsS(int c_id)
//	{
//		QueryPlan plan=new QueryPlan(dm.getDBName());
//		JoiningTable table=new JoiningTable(0, "customer");
//		table.allowedRows.add(String.valueOf(c_id));
//		table.projections.add("C_FNAME");
//		table.projections.add("C_LNAME");
//		plan.tables.put(0, table);
//		plan.rootTable=table;
//		makePlanReady(plan);
//		return plan;
//	}
	
	//SELECT c_uname FROM customer WHERE c_id = ?
	public static QueryPlan getCustomerUserNameS(int c_id)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable table=new JoiningTable(0, "customer");
		table.allowedRows.add(String.valueOf(c_id));
		table.projections.add("C_UNAME");
		plan.tables.put(0, table);
		plan.rootTable=table;
		makePlanReady(plan);
		return plan;
	}
	
	//SELECT c_passwd FROM customer WHERE c_uname = ?
	public static QueryPlan getCustomerPasswordS(int c_id)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable table=new JoiningTable(0, "customer");
		table.allowedRows.add(String.valueOf(c_id));
		table.projections.add("C_PASSWD");
		plan.tables.put(0, table);
		plan.rootTable=table;
		makePlanReady(plan);
		return plan;
	}
	
	//SELECT i_related1 FROM item where i_id = ?
	public static QueryPlan getItemRelatedS(int itemID)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable itemTable=new JoiningTable(0, "item");
		itemTable.allowedRows.add(String.valueOf(itemID));
		itemTable.projections.add("I_RELATED1");
		plan.tables.put(0, itemTable);
		plan.rootTable=itemTable;
		makePlanReady(plan);
		return plan;
	}
	
	//SELECT i_stock FROM item WHERE i_id = ?"
	public static QueryPlan getItemStockS(int itemID)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable itemTable=new JoiningTable(0, "item");
		itemTable.allowedRows.add(String.valueOf(itemID));
		itemTable.projections.add("I_STOCK");
		plan.tables.put(0, itemTable);
		plan.rootTable=itemTable;
		makePlanReady(plan);
		return plan;
	}
	
	public static QueryPlan getItemS(int itemID)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable itemTable=new JoiningTable(0, "item");
		itemTable.allowedRows.add(String.valueOf(itemID));
		itemTable.projections.add("*");
		plan.tables.put(0, itemTable);
		plan.rootTable=itemTable;
		makePlanReady(plan);
		return plan;
	}
	
	//SELECT * FROM item,author WHERE item.i_a_id = author.a_id AND i_id = ?
	public static QueryPlan getItemAndAuthor(int itemID)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable itemTable=new JoiningTable(0, "item");
		JoiningTable authorTable=new JoiningTable(1, "author");
		itemTable.allowedRows.add(String.valueOf(itemID));
		itemTable.projections.add("*");
		authorTable.projections.add("*");
		JoinEdge edge=new JoinEdge(0, "i_a_id".toUpperCase(), JoinEdge.FK, 1, "a_id".toUpperCase(), JoinEdge.PK);
		itemTable.edges.add(edge);
		authorTable.edges.add(edge);
		plan.edges.add(edge);
		plan.tables.put(0, itemTable);
		plan.tables.put(1, authorTable);
		plan.rootTable=itemTable;
		makePlanReady(plan);
		return plan;
	}
	
//	2. SELECT * FROM customer, address, country WHERE customer.c_addr_id = address.addr_id 
//	AND address.addr_co_id = country.co_id AND customer.c_uname = ?
	public static QueryPlan getCustomer(String c_uname)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		
		JoiningTable custTable=new JoiningTable(0, "customer");
		JoiningTable addrTable=new JoiningTable(1, "address");
		JoiningTable ctryTable=new JoiningTable(2, "country");
		JoiningTable indexTable=new JoiningTable(3, DataModel.getIndexTableLogicalName("customer", "c_uname".toUpperCase()));
		
		custTable.projections.add("*");
		addrTable.projections.add("*");
		ctryTable.projections.add("*");
		plan.tables.put(0, custTable);
		plan.tables.put(1, addrTable);
		plan.tables.put(2, ctryTable);
		plan.tables.put(3, indexTable);
		
		JoinEdge edge=new JoinEdge(0, "c_addr_id".toUpperCase(), JoinEdge.FK, 1, "addr_id".toUpperCase(), JoinEdge.PK);
		JoinEdge edge2=new JoinEdge(1, "addr_co_id".toUpperCase(), JoinEdge.FK, 2, "co_id".toUpperCase(), JoinEdge.PK);
		JoinEdge edge3=new JoinEdge(0, "c_uname".toUpperCase(), JoinEdge.FK, 3, 
				DataModel.getIndexTablePKName("customer").toUpperCase(), JoinEdge.PK);
		plan.inputEdge(edge);
		plan.inputEdge(edge2);
		plan.inputEdge(edge3);
		
		plan.rootTable=indexTable;
		indexTable.allowedRows.add(c_uname);
		makePlanReady(plan);
		return plan;
	}

//	3. SELECT J.i_id,J.i_thumbnail from item I, item J 
//	where (I.i_related1 = J.i_id or I.i_related2 = J.i_id or I.i_related3 = J.i_id or I.i_related4 = J.i_id or I.i_related5 = J.i_id) 
//	and I.i_id = ?
	public static QueryPlan getRelatedItem(int itemID) //FIXME create an index table for c_uname
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable itemTableI=new JoiningTable(0, "item");
		JoiningTable itemTableJ=new JoiningTable(1, "item");
		
		itemTableJ.projections.add("i_id".toUpperCase());
		itemTableJ.projections.add("i_thumbnail".toUpperCase());
		
		plan.tables.put(0, itemTableI);
		plan.tables.put(1, itemTableJ);
		
		JoinEdge edge=new JoinEdge(0, ("i_related1").toUpperCase(), JoinEdge.FK, 1, "i_id".toUpperCase(), JoinEdge.PK);
		plan.inputEdge(edge);
		
		plan.rootTable=itemTableI;
		itemTableI.allowedRows.add(String.valueOf(itemID));
		makePlanReady(plan);
		return plan;
	}
	
//	4. SELECT orders.*, customer.*,  
//    cc_xacts.cx_type,  
//    ship.addr_street1 AS ship_addr_street1,  
//    ship.addr_street2 AS ship_addr_street2,  
//    ship.addr_state AS ship_addr_state,  
//    ship.addr_zip AS ship_addr_zip,  
//    ship_co.co_name AS ship_co_name,  
//    bill.addr_street1 AS bill_addr_street1,  
//    bill.addr_street2 AS bill_addr_street2,  
//    bill.addr_state AS bill_addr_state,  
//    bill.addr_zip AS bill_addr_zip,  
//    bill_co.co_name AS bill_co_name  
//  FROM customer, orders, cc_xacts, 
//    address AS ship,  
//    country AS ship_co,  
//    address AS bill,    
//    country AS bill_co  
//  WHERE orders.o_id = ?  
//    AND cx_o_id = orders.o_id  
//    AND customer.c_id = orders.o_c_id  
//    AND orders.o_bill_addr_id = bill.addr_id  
//    AND bill.addr_co_id = bill_co.co_id  
//    AND orders.o_ship_addr_id = ship.addr_id  
//    AND ship.addr_co_id = ship_co.co_id  
//    AND orders.o_c_id = customer.c_id
	public static QueryPlan getOrder(int orderID)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable custTable=new JoiningTable(0, "customer");
		JoiningTable orderTable=new JoiningTable(1, "orders");
		JoiningTable ccTable=new JoiningTable(2, "cc_xacts");
		JoiningTable addrTableShip=new JoiningTable(3, "address");
		JoiningTable ctryTableShip=new JoiningTable(4, "country");
		JoiningTable addrTableBill=new JoiningTable(5, "address");
		JoiningTable ctryTableBill=new JoiningTable(6, "country");
		
		plan.tables.put(0, custTable);
		plan.tables.put(1, orderTable);
		plan.tables.put(2, ccTable);
		plan.tables.put(3, addrTableShip);
		plan.tables.put(4, ctryTableShip);
		plan.tables.put(5, addrTableBill);
		plan.tables.put(6, ctryTableBill);
		
		orderTable.allowedRows.add(String.valueOf(orderID));
		orderTable.projections.add("*");
		custTable.projections.add("*");
		ccTable.projections.add("cx_type".toUpperCase());
		
		addrTableShip.projections.add("addr_street1".toUpperCase());
		addrTableShip.projections.add("addr_street2".toUpperCase());
		addrTableShip.projections.add("addr_state".toUpperCase());
		addrTableShip.projections.add("addr_zip".toUpperCase());
		
		ctryTableShip.projections.add("co_name".toUpperCase());
		
		addrTableBill.projections.add("addr_street1".toUpperCase());
		addrTableBill.projections.add("addr_street2".toUpperCase());
		addrTableBill.projections.add("addr_state".toUpperCase());
		addrTableBill.projections.add("addr_zip".toUpperCase());
		
		ctryTableBill.projections.add("co_name".toUpperCase());
		
//		plan.tables.put(0, custTable);
//		plan.tables.put(1, orderTable);
//		plan.tables.put(2, ccTable);
//		plan.tables.put(3, addrTableShip);
//		plan.tables.put(4, ctryTableShip);
//		plan.tables.put(5, addrTableBill);
//		plan.tables.put(6, ctryTableBill);
		
		
//	    AND cx_o_id = orders.o_id  
//	    AND customer.c_id = orders.o_c_id  
//	    AND orders.o_bill_addr_id = bill.addr_id  
//	    AND bill.addr_co_id = bill_co.co_id  
//	    AND orders.o_ship_addr_id = ship.addr_id  
//	    AND ship.addr_co_id = ship_co.co_id  

		
		JoinEdge edge=new JoinEdge(2, "cx_o_id".toUpperCase(), JoinEdge.PK, 1, "o_id".toUpperCase(), JoinEdge.PK);
		JoinEdge edge2=new JoinEdge(0, "c_id".toUpperCase(), JoinEdge.PK, 1, "o_c_id".toUpperCase(), JoinEdge.FK);
		JoinEdge edge3=new JoinEdge(1, "o_bill_addr_id".toUpperCase(), JoinEdge.FK, 5, "addr_id".toUpperCase(), JoinEdge.PK);
		JoinEdge edge4=new JoinEdge(5, "addr_co_id".toUpperCase(), JoinEdge.FK, 6, "co_id".toUpperCase(), JoinEdge.PK);
		
		JoinEdge edge5=new JoinEdge(1, "o_ship_addr_id".toUpperCase(), JoinEdge.FK, 3, "addr_id".toUpperCase(), JoinEdge.PK);
		JoinEdge edge6=new JoinEdge(3, "addr_co_id".toUpperCase(), JoinEdge.FK, 4, "co_id".toUpperCase(), JoinEdge.PK);
//		JoinEdge edge7=new JoinEdge(1, "o_c_id".toUpperCase(), JoinEdge.FK, 0, "c_id".toUpperCase(), JoinEdge.PK);
		
		plan.inputEdge(edge);
		plan.inputEdge(edge2);
		plan.inputEdge(edge3);
		plan.inputEdge(edge4);
		plan.inputEdge(edge5);
		plan.inputEdge(edge6);
//		plan.inputEdge(edge7);
		
		plan.rootTable=orderTable;
		makePlanReady(plan);
		return plan;
	}
	
	//SELECT * FROM shopping_cart_line, item WHERE scl_i_id = item.i_id AND scl_sc_id = ?
	public static  QueryPlan getShopppingCart(int cartID)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable shopTable=new JoiningTable(0, "shopping_cart");
		JoiningTable shopLineTable=new JoiningTable(1, "shopping_cart_line");
		JoiningTable itemTable=new JoiningTable(2, "item");
		
		shopTable.projections.add("*".toUpperCase());
		shopLineTable.projections.add("*".toUpperCase());
		itemTable.projections.add("*".toUpperCase());
		
		plan.tables.put(0, shopTable);
		plan.tables.put(1, shopLineTable);
		plan.tables.put(2, itemTable);
		
		JoinEdge edge=new JoinEdge(0, ("SC_ID").toUpperCase(), JoinEdge.PK, 1, "SCL_SC_ID".toUpperCase(), JoinEdge.FK);
		JoinEdge edge2=new JoinEdge(1, ("SCL_I_ID").toUpperCase(), JoinEdge.FK, 2, "I_ID".toUpperCase(), JoinEdge.PK);
		plan.inputEdge(edge);
		plan.inputEdge(edge2);
		
		plan.rootTable=shopTable;
		shopTable.allowedRows.add(String.valueOf(cartID));
		makePlanReady(plan);
		return plan;
	}
	
//	6. SELECT * FROM item, author WHERE item.i_a_id = author.a_id AND item.i_subject = ? 
//			ORDER BY item.i_title LIMIT 50
	
	public static QueryPlan getItemAndAuthorBySubject(String subject)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable itemTable=new JoiningTable(0, "item");
		JoiningTable authorTable=new JoiningTable(1, "author");
		JoiningTable indexTable=new JoiningTable(2, DataModel.getIndexTableLogicalName("item", "I_SUBJECT"));
		indexTable.allowedRows.add(subject);
		plan.tables.put(0, itemTable);
		plan.tables.put(1, authorTable);
		plan.tables.put(2, indexTable);
		
		itemTable.projections.add("*");
		authorTable.projections.add("*");
		JoinEdge edge=new JoinEdge(0, "i_a_id".toUpperCase(), JoinEdge.FK, 1, "a_id".toUpperCase(), JoinEdge.PK);
		JoinEdge edge2=new JoinEdge(0, "I_SUBJECT".toUpperCase(), JoinEdge.FK, 2, DataModel.getIndexTablePKName("I_ID".toUpperCase()), JoinEdge.PK);
		plan.inputEdge(edge);
		plan.inputEdge(edge2);
		
		plan.rootTable=indexTable;
		makePlanReady(plan);
		return plan;
	}
	
//	7. SELECT o_id  
//    FROM customer, orders  
//    WHERE customer.c_id = orders.o_c_id  
//    AND c_uname = ?  
//    ORDER BY o_date, orders.o_id DESC  
//    LIMIT 1
	public static QueryPlan getAllOrders(String username)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable custTable=new JoiningTable(0, "customer");
		JoiningTable orderTable=new JoiningTable(1, "orders");
		JoiningTable indexTable=new JoiningTable(2, DataModel.getIndexTableLogicalName("customer", "C_UNAME"));
		indexTable.allowedRows.add(username);
		plan.tables.put(0, custTable);
		plan.tables.put(1, orderTable);
		plan.tables.put(2, indexTable);
		
		orderTable.projections.add("*");
		custTable.projections.add("*");
		JoinEdge edge=new JoinEdge(0, "C_ID".toUpperCase(), JoinEdge.PK, 1, "O_C_ID".toUpperCase(), JoinEdge.FK);
		JoinEdge edge2=new JoinEdge(0, "C_UNAME".toUpperCase(), JoinEdge.FK, 2, DataModel.getIndexTablePKName("C_ID".toUpperCase()), JoinEdge.PK);
		plan.inputEdge(edge);
		plan.inputEdge(edge2);
		
		plan.rootTable=indexTable;
		makePlanReady(plan);
		return plan;
	}

//8. SELECT * FROM order_line, item   
//    WHERE ol_o_id = ?  
//    AND ol_i_id = i_id
	public static QueryPlan getShortOrder(int orderID)
	{
		QueryPlan plan=new QueryPlan(dm.getDBName());
		JoiningTable orderTable=new JoiningTable(0, "orders");
		JoiningTable orderlineTable=new JoiningTable(1, "order_line");
		JoiningTable itemTable=new JoiningTable(2, "item");
		
		orderlineTable.projections.add("*".toUpperCase());
		itemTable.projections.add("*".toUpperCase());
		
		plan.tables.put(0, orderTable);
		plan.tables.put(1, orderlineTable);
		plan.tables.put(2, itemTable);
		
		JoinEdge edge=new JoinEdge(0, ("O_ID").toUpperCase(), JoinEdge.PK, 1, "OL_O_ID".toUpperCase(), JoinEdge.FK);
		JoinEdge edge2=new JoinEdge(1, ("ol_i_id").toUpperCase(), JoinEdge.FK, 2, "I_ID".toUpperCase(), JoinEdge.PK);
		plan.inputEdge(edge);
		plan.inputEdge(edge2);
		
		plan.rootTable=orderTable;
		orderTable.allowedRows.add(String.valueOf(orderID));
		makePlanReady(plan);
		return plan;
	}
}
