
package org.zhouw.stps.app.otpcw.sql;
/* 
 * TPCW_Database.java - Contains all of the code involved with database
 *                      accesses, including all of the JDBC calls. These
 *                      functions are called by many of the servlets.
 *
 ************************************************************************
 *
 * This is part of the the Java TPC-W distribution,
 * written by Harold Cain, Tim Heil, Milo Martin, Eric Weglarz, and Todd
 * Bezenek.  University of Wisconsin - Madison, Computer Sciences
 * Dept. and Dept. of Electrical and Computer Engineering, as a part of
 * Prof. Mikko Lipasti's Fall 1999 ECE 902 course.
 *
 * Copyright (C) 1999, 2000 by Harold Cain, Timothy Heil, Milo Martin, 
 *                             Eric Weglarz, Todd Bezenek.
 *
 * This source code is distributed "as is" in the hope that it will be
 * useful.  It comes with no warranty, and no author or distributor
 * accepts any responsibility for the consequences of its use.
 *
 * Everyone is granted permission to copy, modify and redistribute
 * this code under the following conditions:
 *
 * This code is distributed for non-commercial use only.
 * Please contact the maintainer for restrictions applying to 
 * commercial use of these tools.
 *
 * Permission is granted to anyone to make or distribute copies
 * of this code, either as received or modified, in any
 * medium, provided that all copyright notices, permission and
 * nonwarranty notices are preserved, and that the distributor
 * grants the recipient permission for further redistribution as
 * permitted by this document.
 *
 * Permission is granted to distribute this code in compiled
 * or executable form under the same conditions that apply for
 * source code, provided that either:
 *
 * A. it is accompanied by the corresponding machine-readable
 *    source code,
 * B. it is accompanied by a written offer, with no time limit,
 *    to give anyone a machine-readable copy of the corresponding
 *    source code in return for reimbursement of the cost of
 *    distribution.  This written offer must permit verbatim
 *    duplication by anyone, or
 * C. it is distributed by someone who received only the
 *    executable form, and is accompanied by a copy of the
 *    written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this codes.
 * You are forbidden to forbid anyone else to use, share and improve what
 * you give them.
 *
 ************************************************************************/

import java.io.*;
import java.net.URL;
import java.sql.*;
import java.lang.Math.*;
import java.util.*;
import java.util.concurrent.Semaphore;
import java.sql.Connection;
import java.sql.Date;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;

import org.zhouw.stps.*;

public class TPCW_Database {
			//public static int inited=TPCW_Database.readConfig(TPCW_Util.configFile);
    //static String driverName = "COM.ibm.db2.jdbc.app.DB2Driver";
    //static String jdbcPath = "jdbc:db2:tpcw2";
    static String driverName[] = null;
    static String jdbcPath[] = null;
    static String userName[] = null;
    static String userPassword[] = null;
    
    // Pool of *available* connections.
    static Vector<Connection> availConn[];
    static int checkedOut[] = null;
    static int totalConnections[] = null;
    static int createdConnections[] = null;
    static int closedConnections[] = null;
    
    public static int num=0;
    public static int nextDB=0;
    //    private static final boolean use_connection_pool = false;
    private static final boolean use_connection_pool = true;
    public static int maxConn;
    private static Semaphore masterAvailable;
    private static Semaphore slaveAvailable;
    
    // Here's what the db line looks like for postgres
    //public static final String url = "jdbc:postgresql://eli.ece.wisc.edu/tpcwb";
    public static int init()
	{
    	int[] nodes=GlobalPara.sysConf.nodes.getDataComponentNodes(null);
    	num=nodes.length;
		driverName = new String[num];
		jdbcPath = new String[num];
		userName = new String[num];
		userPassword = new String[num];
	    
	    // Pool of *available* connections.
		availConn=new Vector[num];
	 
		checkedOut = new int[num];
		totalConnections = new int[num];
		createdConnections = new int[num];
		closedConnections = new int[num];
	 
		for(int i=0;i<num;i++) {//the first one with index 0 is the master DB
			availConn[i]=new Vector(20);
			checkedOut[i] = 0;
			totalConnections[i] = 0;
			createdConnections[i] = 0;
			closedConnections[i] = 0;
			TPCW_Database.driverName[i]="org.postgresql.Driver";	
			TPCW_Database.jdbcPath[i]= "jdbc:postgresql://"+ GlobalPara.sysConf.ips.getIP(nodes[i])  +"/tpcw";
			//TPCW_Database.userName[i]="postgres";
			TPCW_Database.userName[i]="zhouw";
			TPCW_Database.userPassword[i]="123";
			maxConn=GlobalPara.sysConf.maxConn;
			masterAvailable=new Semaphore(maxConn, true);
			slaveAvailable=new Semaphore(maxConn*(num-1), true);
		}
		
		return 1;
	}

    public static String getIDServiceAddr(String configFile)
	{
		String result="";
		FileInputStream fstream;
		BufferedReader in;
		try {
		fstream = new	FileInputStream(configFile);
		in = new BufferedReader(new InputStreamReader(fstream));
		String line;
		while((line=in.readLine())!=null)
		{
			String[] inputs=line.split("=");
			if(inputs[0].compareTo("IDService")==0)
				result=inputs[1];
		}
		in.close();
		fstream.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return result;
	}
    
    public static Connection getConnection(boolean ismaster, Vector<Integer> dbID) {
    	Connection con=null;
  		try {
  			if(ismaster||num<=1)
  				masterAvailable.acquire();
  			else
  				slaveAvailable.acquire();
  			
			con=getConnectionSyn(ismaster,dbID);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
  		return con;
    }
    // Get a connection from the pool.
    //ismaster==true return first one, false: return the random
    public  static synchronized Connection getConnectionSyn(boolean ismaster, Vector<Integer> dbID) {
    	int indexOfDB=0;
    	if(ismaster||num<=1) {//ask for the master DB
    		indexOfDB=0;
    	} else {
    		nextDB=(nextDB+1)%num;
    		if(nextDB==0)
    			nextDB=(nextDB+1)%num;
    		indexOfDB=nextDB;
   		}
    	
    	dbID.clear();
    	dbID.addElement(indexOfDB);
    	
    		
	if (!use_connection_pool) {
	    return getNewConnection(indexOfDB);
	} else {
	    Connection con = null;
	    while (availConn[indexOfDB].size() > 0) {
				// Pick the first Connection in the Vector
				// to get round-robin usage
				con = availConn[indexOfDB].firstElement();
				availConn[indexOfDB].removeElementAt(0);
				try {
				    if (con.isClosed()) {
				    	con=null;
				    	continue;
				    }
				}
				catch (SQLException e) {
				    e.printStackTrace();
				    con=null;
				    continue;
				}
		
			// Got a connection.
			checkedOut[indexOfDB]++;
			return con;
	    }
	    
		 if(checkedOut[indexOfDB]<maxConn) {
		    con = getNewConnection(indexOfDB);	
		    totalConnections[indexOfDB]++;
		    checkedOut[indexOfDB]++;
		 } else { //search other slave machines for free connections
			 boolean findFreeConnection=false;
			 for(int i=indexOfDB+1;i<num;i++){
				 if(availConn[i].size()>0||checkedOut[i]<maxConn){
					 indexOfDB=i;
					 findFreeConnection=true;
					 break;
				 }
			 }
			 if(!findFreeConnection){
				 for(int i=1;i<indexOfDB;i++){
					 if(availConn[i].size()>0||checkedOut[i]<maxConn){
						 indexOfDB=i;
						 findFreeConnection=true;
						 break;
					 }
				 }
			 }
			 
			 if(findFreeConnection){
				 dbID.clear();
			     dbID.addElement(indexOfDB);
				 nextDB=indexOfDB;
				 while (availConn[indexOfDB].size() > 0) {
						con = availConn[indexOfDB].firstElement();
						availConn[indexOfDB].removeElementAt(0);
						try {
						    if (con.isClosed()) {
						    	con=null;
						    	continue;
						    }
						}
						catch (SQLException e) {
						    e.printStackTrace();
						    con=null;
						    continue;
						}
				
					// Got a connection.
					checkedOut[indexOfDB]++;
					return con;
			    }
				 
				 if(checkedOut[indexOfDB]<maxConn) {
					con = getNewConnection(indexOfDB);	
					totalConnections[indexOfDB]++;
					checkedOut[indexOfDB]++;
				 }
			 } else {
				 System.err.println("[Error] unable to obatain a JDBC connection from the connection pool");
			 }
		 }
		 return con;
	}
    }
    
    public static void returnConnection(Connection con, int indexofDB)
    {
    	if(con==null)return;
    	try {
			returnConnectionSyn(con,indexofDB);
		} catch (SQLException e) {
			e.printStackTrace();
		}
		if(indexofDB==0)
			masterAvailable.release();
		else
			slaveAvailable.release();
    }
    
    // Return a connection to the pool.
    public static synchronized void returnConnectionSyn(Connection con, int indexofDB)
    throws java.sql.SQLException
    {	
    	if(con==null)
    		return;
    	if(indexofDB<0)return;
		if (!use_connection_pool) {
		    con.close();
		} else {
		    checkedOut[indexofDB]--;
		    availConn[indexofDB].addElement(con);
		}
    }

    // Get a new connection to DB2
    public  static Connection getNewConnection(int indexofDB) {
    	try {
    	    Class.forName(driverName[indexofDB]);
    	    // Create URL for specifying a DBMS
    	    Connection con=null;
    	  //  while(true) {
    		try {
    		    //   con = DriverManager.getConnection("jdbc:postgresql://eli.ece.wisc.edu/tpcw", "milo", "");
    		    con = DriverManager.getConnection(jdbcPath[indexofDB],userName[indexofDB],userPassword[indexofDB]);
    		   // break;  
    		} catch (java.sql.SQLException ex) {
    		    System.err.println("Error getting connection: " + 
    				       ex.getMessage() + " : " +
    				       ex.getErrorCode() + 
    				       ": trying to get connection again.");
    		    ex.printStackTrace();
    		  //  java.lang.Thread.sleep(1000);
    		}
    	    //}
    		if(con!=null)
    	    {con.setAutoCommit(false);
    	    createdConnections[indexofDB]++;}
    	    return con;
    	} catch (java.lang.Exception ex) {
    	    ex.printStackTrace();
    	}
    	return null;
    }

   
    //time .05s
    private static Cart getCart(Connection con, int SHOPPING_ID, double c_discount){
	Cart mycart = null;
	try {
	    PreparedStatement get_cart = con.prepareStatement
		("SELECT * " + 
		 "FROM shopping_cart_line, item " + 
		 "WHERE scl_i_id = item.i_id AND scl_sc_id = ?");
	    get_cart.setInt(1, SHOPPING_ID);
	    ResultSet rs = get_cart.executeQuery();
	    mycart = new Cart(rs, c_discount);
	    rs.close();
	    get_cart.close();
	}catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}
	return mycart;
    }

    

    //BUY CONFIRM 

    public static BuyConfirmResult doBuyConfirm(Connection con, int orderID, int shopping_id,
						int customer_id,
						String cc_type,
						long cc_number,
						String cc_name,
						Date cc_expiry,
						String shipping) {
	
	BuyConfirmResult result = new BuyConfirmResult();
	try {
	    double c_discount = getCDiscount(con, customer_id);
	    result.cart = getCart(con, shopping_id, c_discount);
	    int ship_addr_id = getCAddr(con, customer_id);
	    result.order_id = orderID;
	    enterOrder(con, orderID, customer_id, result.cart, ship_addr_id, shipping, c_discount);
	    enterCCXact(con, result.order_id, cc_type, cc_number, cc_name, cc_expiry, result.cart.SC_TOTAL, ship_addr_id);
	    clearCart(con, shopping_id);
	} catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}
	return result;
    }
    
    //DB query time: .05s
    public static double getCDiscount(Connection con, int c_id) {
	double c_discount = 0.0;
	try {
	    // Prepare SQL
	    PreparedStatement statement = con.prepareStatement
		("SELECT c_discount FROM customer WHERE customer.c_id = ?");
	    
	    // Set parameter
	    statement.setInt(1, c_id);
	    ResultSet rs = statement.executeQuery();
	    
	    // Results
	    rs.next();
	    c_discount = rs.getDouble(1);
	    rs.close();
	    statement.close();
	} catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}
	return c_discount;
    }

    //DB time: .05s
    public static int getCAddrID(Connection con, int c_id) {
	int c_addr_id = 0;
	try {
	    // Prepare SQL
	    PreparedStatement statement = con.prepareStatement
		("SELECT c_addr_id FROM customer WHERE customer.c_id = ?");
	    
	    // Set parameter
	    statement.setInt(1, c_id);
	    ResultSet rs = statement.executeQuery();
	    
	    // Results
	    rs.next();
	    c_addr_id = rs.getInt(1);
	    rs.close();
	    statement.close();
	} catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}
	return c_addr_id;
    }

    public static int getCAddr(Connection con, int c_id) {
	int c_addr_id = 0;
	try {
	    // Prepare SQL
	    PreparedStatement statement = con.prepareStatement
		("SELECT c_addr_id FROM customer WHERE customer.c_id = ?");
	    
	    // Set parameter
	    statement.setInt(1, c_id);
	    ResultSet rs = statement.executeQuery();
	    
	    // Results
	    rs.next();
	    c_addr_id = rs.getInt(1);
	    rs.close();
	    statement.close();
	} catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}
	return c_addr_id;
    }

    public static void enterCCXact(Connection con,
				   int o_id,        // Order id
				   String cc_type,
				   long cc_number,
				   String cc_name,
				   Date cc_expiry,
				   double total,   // Total from shopping cart
				   int ship_addr_id) {

	// Updates the CC_XACTS table
	if(cc_type.length() > 10)
	    cc_type = cc_type.substring(0,10);
	if(cc_name.length() > 30)
	    cc_name = cc_name.substring(0,30);
	
	try {
	    // Prepare SQL
	    PreparedStatement statement = con.prepareStatement
		("INSERT into cc_xacts (cx_o_id, cx_type, cx_num, cx_name, cx_expire, cx_xact_amt, cx_xact_date, cx_co_id) " + 
		 "VALUES (?, ?, ?, ?, ?, ?, date('now'), (SELECT co_id FROM address, country WHERE addr_id = ? AND addr_co_id = co_id))");
	    
	    // Set parameter
	    statement.setInt(1, o_id);           // cx_o_id
	    statement.setString(2, cc_type);     // cx_type
	    statement.setLong(3, cc_number);     // cx_num
	    statement.setString(4, cc_name);     // cx_name
	    statement.setDate(5, cc_expiry);     // cx_expiry
	    statement.setDouble(6, total);       // cx_xact_amount
	    statement.setInt(7, ship_addr_id);   // ship_addr_id
	    statement.executeUpdate();
	    statement.close();
	} catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}
    }
    
    public static void clearCart(Connection con, int shopping_id) {
	// Empties all the lines from the shopping_cart_line for the
	// shopping id.  Does not remove the actually shopping cart
	try {
	    // Prepare SQL
	    PreparedStatement statement = con.prepareStatement
		("DELETE FROM shopping_cart_line WHERE scl_sc_id = ?");
	    
	    // Set parameter
	    statement.setInt(1, shopping_id);
	    statement.executeUpdate();
	    statement.close();
	} catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}
    }

    
    public static int enterOrder(Connection con,int o_id, int customer_id, Cart cart, int ship_addr_id, String shipping, double c_discount) {

	// - Creates an entry in the 'orders' table 
	try {
	    PreparedStatement insert_row = con.prepareStatement
		("INSERT into orders (o_id, o_c_id, o_date, o_sub_total, " + 
		 "o_tax, o_total, o_ship_type, o_ship_date, " + 
		 "o_bill_addr_id, o_ship_addr_id, o_status) " + 
		 "VALUES (?, ?, date('now'), ?, 8.25, ?, ?, date('now') + ?, ?, ?, 'Pending')");
	    insert_row.setInt(2, customer_id);
	    insert_row.setDouble(3, cart.SC_SUB_TOTAL);
	    insert_row.setDouble(4, cart.SC_TOTAL);
	    insert_row.setString(5, shipping);
	    insert_row.setInt(6, TPCW_Util.getRandom(7));
	    insert_row.setInt(7, getCAddrID(con, customer_id));
	    insert_row.setInt(8, ship_addr_id);
	    insert_row.setInt(1, o_id);
		insert_row.executeUpdate();
		insert_row.close();
	} catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}

	Enumeration e = cart.lines.elements();
	int counter = 0;
	while(e.hasMoreElements()) {
	    // - Creates one or more 'order_line' rows.
	    CartLine cart_line = (CartLine) e.nextElement();
	    addOrderLine(con, counter, o_id, cart_line.scl_i_id, 
			 cart_line.scl_qty, c_discount, 
			 TPCW_Util.getRandomString(20, 100));
	    counter++;

	    // - Adjusts the stock for each item ordered
	    int stock = getStock(con, cart_line.scl_i_id);
	    if ((stock - cart_line.scl_qty) < 10) {
		setStock(con, cart_line.scl_i_id, 
			 stock - cart_line.scl_qty + 21);
	    } else {
		setStock(con, cart_line.scl_i_id, stock - cart_line.scl_qty);
	    }
	}
	return o_id;
    }
    
    public static void addOrderLine(Connection con, 
				    int ol_id, int ol_o_id, int ol_i_id, 
				    int ol_qty, double ol_discount, String ol_comment) {
	int success = 0;
	try {
	    PreparedStatement insert_row = con.prepareStatement
		("INSERT into order_line (ol_id, ol_o_id, ol_i_id, ol_qty, ol_discount, ol_comments) " + 
		 "VALUES (?, ?, ?, ?, ?, ?)");
	    
	    insert_row.setInt(1, ol_id);
	    insert_row.setInt(2, ol_o_id);
	    insert_row.setInt(3, ol_i_id);
	    insert_row.setInt(4, ol_qty);
	    insert_row.setDouble(5, ol_discount);
	    insert_row.setString(6, ol_comment);
	    insert_row.executeUpdate();
	    insert_row.close();
	} catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}
    }

    public static int getStock(Connection con, int i_id) {
	int stock = 0;
	try {
	    PreparedStatement get_stock = con.prepareStatement
		("SELECT i_stock FROM item WHERE i_id = ?");
	    
	    // Set parameter
	    get_stock.setInt(1, i_id);
	    ResultSet rs = get_stock.executeQuery();
	    
	    // Results
	    rs.next();
	    stock = rs.getInt("i_stock");
	    rs.close();
	    get_stock.close();
	} catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}
	return stock;
    }

    public static void setStock(Connection con, int i_id, int new_stock) {
	try {
	    PreparedStatement update_row = con.prepareStatement
		("UPDATE item SET i_stock = ? WHERE i_id = ?");
	    update_row.setInt(1, new_stock);
	    update_row.setInt(2, i_id);
	    update_row.executeUpdate();
	    update_row.close();
	} catch (java.lang.Exception ex) {
	    ex.printStackTrace();
	}
    }

   
}

