package org.zhouw.stps.app.tpcw;
/*
 * Cart.java - Class stores the necessary components of a shopping cart
 * 
 ************************************************************************
 *
 * This is part of the the Java TPC-W distribution,
 * written by Harold Cain, Tim Heil, Milo Martin, Eric Weglarz, and Todd
 * Bezenek.  University of Wisconsin - Madison, Computer Sciences
 * Dept. and Dept. of Electrical and Computer Engineering, as a part of
 * Prof. Mikko Lipasti's Fall 1999 ECE 902 course.
 *
 * Copyright (C) 1999, 2000 by Harold Cain, Timothy Heil, Milo Martin, 
 *                             Eric Weglarz, Todd Bezenek.
 *
 * This source code is distributed "as is" in the hope that it will be
 * useful.  It comes with no warranty, and no author or distributor
 * accepts any responsibility for the consequences of its use.
 *
 * Everyone is granted permission to copy, modify and redistribute
 * this code under the following conditions:
 *
 * This code is distributed for non-commercial use only.
 * Please contact the maintainer for restrictions applying to 
 * commercial use of these tools.
 *
 * Permission is granted to anyone to make or distribute copies
 * of this code, either as received or modified, in any
 * medium, provided that all copyright notices, permission and
 * nonwarranty notices are preserved, and that the distributor
 * grants the recipient permission for further redistribution as
 * permitted by this document.
 *
 * Permission is granted to distribute this code in compiled
 * or executable form under the same conditions that apply for
 * source code, provided that either:
 *
 * A. it is accompanied by the corresponding machine-readable
 *    source code,
 * B. it is accompanied by a written offer, with no time limit,
 *    to give anyone a machine-readable copy of the corresponding
 *    source code in return for reimbursement of the cost of
 *    distribution.  This written offer must permit verbatim
 *    duplication by anyone, or
 * C. it is distributed by someone who received only the
 *    executable form, and is accompanied by a copy of the
 *    written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this codes.
 * You are forbidden to forbid anyone else to use, share and improve what
 * you give them.
 *
 ************************************************************************/

import java.util.*;

public class Cart {
    
    public double SC_SUB_TOTAL;
    public double SC_TAX;
    public double SC_SHIP_COST;
    public double SC_TOTAL;

    public Vector<CartLine> lines=new Vector<CartLine>();
    
    public Cart(){}
    
    public Cart(int[] ids, int[] qtys)
	{
    	lines = new Vector<CartLine>();
    	for(int k=0;k<ids.length;k++){
    	    CartLine line = new CartLine("i_title",
    					 500,
    					 500,
    				"i_backing",
    					qtys[k],
    					ids[k]);
    	    lines.addElement(line);
    	    }
	}
    
    public Cart(int times)
	{
    	int i;
    	int total_items;
    	lines = new Vector<CartLine>();
    	for(int k=0;k<times;k++){
    	    CartLine line = new CartLine("i_title",
    					 500,
    					 500,
    				"i_backing",
    					10*(k+1),
    					100*(k+1));
    	    lines.addElement(line);
    	    }
    	

    	SC_SUB_TOTAL = 0;
    	total_items = 0;
    	for(i = 0; i < lines.size(); i++){
    	    CartLine thisline = (CartLine) lines.elementAt(i);
    	    SC_SUB_TOTAL += thisline.scl_cost * thisline.scl_qty;
    	    total_items += thisline.scl_qty;
    	}
    	
    	//Need to multiply the sub_total by the discount.
    	SC_SUB_TOTAL = SC_SUB_TOTAL * ((100 - 1)*.01);
    	SC_TAX = SC_SUB_TOTAL * .0825;
    	SC_SHIP_COST = 3.00 + (1.00 * total_items);
    	SC_TOTAL = SC_SUB_TOTAL + SC_SHIP_COST + SC_TAX;
        
	}
    public Cart (ShortCartLine[] scart,
    									String[] title,
    									double[] cost,
    									BookDetailAttr[] detail, 
    									double C_DISCOUNT)
    {
			int i;
			int total_items;
			lines = new Vector<CartLine>();
			int j=0;
			while(j< cost.length){//While there are lines remaining
			    CartLine line = new CartLine(
			    			title[j],
			    			cost[j],
			    			detail[j].srp,
			    			detail[j].backing,
							 scart[j].scl_qty,
							 scart[j].scl_i_id);
			    lines.addElement(line);
			    	j++;
			}
		
			SC_SUB_TOTAL = 0;
			total_items = 0;
			for(i = 0; i < lines.size(); i++){
			    CartLine thisline = (CartLine) lines.elementAt(i);
			    SC_SUB_TOTAL += thisline.scl_cost * thisline.scl_qty;
			    total_items += thisline.scl_qty;
			}
			
			//Need to multiply the sub_total by the discount.
			SC_SUB_TOTAL = SC_SUB_TOTAL * ((100 - C_DISCOUNT)*.01);
			SC_TAX = SC_SUB_TOTAL * .0825;
			SC_SHIP_COST = 3.00 + (1.00 * total_items);
			SC_TOTAL = SC_SUB_TOTAL + SC_SHIP_COST + SC_TAX;
		    }

    
    //FIXME the lines are not serialized
	  public String[] serialize()
	    {
		  	String d=TPCW_globleParam.delimiter;
		  	String strList[]=new String[1+lines.size()];
		  	strList[0]= String.format("%f%s%f%s%f%s%f",
	    			SC_SUB_TOTAL,d,
	    			SC_TAX,d,
	    			SC_SHIP_COST,d,
	    			SC_TOTAL
	    			);
		  	for(int i=1;i<=lines.size();i++)
		  			strList[i]=((CartLine)lines.get(i-1)).serialize();
	    	return strList;
	    }
	  
	  //return -1 when the input is not valid,0 means ok
	  //	FIXME the lines are not serialized
	  public int deSerialize(String[] cart)
		  {
			  	String result[]=	cart[0].split(TPCW_globleParam.delimiter);
			  	//if(result.length<5)
			  	//		return -1;
			  	try{
			   this.SC_SUB_TOTAL = Double.parseDouble(result[0]);
			   this.SC_TAX = Double.parseDouble(result[1]);
			   this.SC_SHIP_COST = Double.parseDouble(result[2]);
			   this.SC_TOTAL = Double.parseDouble(result[3]);
			   lines=new Vector<CartLine>();
			   CartLine[] l=new CartLine[cart.length-1];
			   for(int i=1;i<cart.length;i++)
			   		{
				   l[i-1]=new CartLine();
				   l[i-1].deSerialize(cart[i]);
				   lines.add(l[i-1]);
			   		}
			  	}
			 	catch (NumberFormatException e)
			  	{e.printStackTrace();}
			 	catch (NullPointerException e)
			  	{e.printStackTrace();}
			   return 0;
		  }
}
