package org.zhouw.stps.app.tpcw;

import org.apache.hadoop.hbase.HBaseConfiguration;
import org.zhouw.stps.*;
import org.zhouw.stps.app.DataModel;
import org.zhouw.stps.app.DataTable;

public class TPCWDataModel extends DataModel {
	
	public TPCWDataModel() {
		super(name);
		
		String[] cols;
		String tablename;
		String pkname;
		
		tablename=BASEITEMTABLENAME;
		pkname="I_ID";
		cols=new String[]{
			"I_ID"
		};
		inputOneTable(this, tablename, pkname, cols);
		
		
		tablename=BASESHOPPINGTABLENAME;
		pkname="S_ID";
		cols=new String[]{
			"S_ID"
		};
		inputOneTable(this, tablename, pkname, cols);
		
		tablename=BASELATESTORDERTABLENAME;
		pkname="L_ID";
		cols=new String[]{
			"L_ID"
		};
		inputOneTable(this, tablename, pkname, cols);
		
		//due to legacy problem, the tables of this database do not have prefixes
		this.tablesByPhysicalName.clear();
		for(DataTable t:super.tables.values()) {
			this.tablesByPhysicalName.put(t.tableName,t);
			t.physicalTableName=t.tableName;
		}
	}

	public final static String name="shopping";
	
	public final static String BASEITEMTABLENAME="item_stock";
	public final static String BASESHOPPINGTABLENAME="shopping";
	public final static String BASELATESTORDERTABLENAME="latestOrder";
	
	public static String ITEMTABLENAME= GlobalPara.sysConf.EXTable + BASEITEMTABLENAME;
	public static String SHOPPINGTABLENAME=GlobalPara.sysConf.EXTable + BASESHOPPINGTABLENAME;
	public static String LATESTORDERTABLENAME=GlobalPara.sysConf.EXTable + BASELATESTORDERTABLENAME;
	
    //ATTENTION: The NUM_EBS and NUM_ITEMS variables are the only variables
    //that should be modified in order to rescale the DB.
    public static final int NUM_EBS = GlobalPara.sysConf.NUM_EBS;
    public static final int NUM_ITEMS = GlobalPara.sysConf.NUM_ITEMS;
    public static final String hadoopCfgPath="/usr/local/hadoop-0.17.2/conf/hadoop-site.xml";
    public static final String hbaseCfgPath="/usr/local/hbase/conf/hbase-site.xml";
	
    //public static final int NUM_CUSTOMERS = NUM_EBS * 288;
    public static final int NUM_CUSTOMERS = NUM_EBS * 2880;
    public static final int NUM_ADDRESSES = 2 * NUM_CUSTOMERS;
    public static final int NUM_AUTHORS = (int) (.25 * NUM_ITEMS);
    public static final int NUM_ORDERS = (int) (.9 * NUM_CUSTOMERS);
	
	public final static String[] tables= new String[]{ITEMTABLENAME,SHOPPINGTABLENAME,LATESTORDERTABLENAME} ;
	public final static int[] rowCount= new int[]{NUM_ITEMS,NUM_ORDERS,NUM_CUSTOMERS} ;
	
    public static final HBaseConfiguration conf=GlobalPara.hbaseConf;
    
    public static int numOfTablePartitions=GlobalPara.sysConf.numOfTablePartitions;
    
    public static final int numOfItemTable=5;
    public static final int numOfLatestOrderTable=5;
    public static final int numOfShoppingTable=80;
    
    public static String getStrIDFromInt(int i)
	{
		return String.format("%020d",i);
	}
    

    public static final String CFLO=new String("CUSTOMER:"); 
	public static final String CLATEST_ORDER=new String("CUSTOMER:LATEST_ORDER"); 
    
	public static final String CFI=new String("ITEM:"); 
	public static final String CITEM_I_STOCK=new String("ITEM:I_STOCK"); 
	
	public static final String CFO=new String("ORDERS:"); 
	public static final String CO_C_ID=new String("ORDERS:O_C_ID"); 
	public static final String CO_DATE=new String("ORDERS:O_DATE"); 
	public static final String CO_SUB_TOTAL=new String("ORDERS:O_SUB_TOTAL");
	public static final String CO_TAX=new String("ORDERS:O_TAX"); 
	public static final String CO_TOTAL=new String("ORDERS:O_TOTAL");
	public static final String CO_SHIP_TYPE=new String("ORDERS:O_SHIP_TYPE"); 
	public static final String CO_SHIP_DATE=new String("ORDERS:O_SHIP_DATE");
	public static final String CO_BILL_ADDR_ID=new String("ORDERS:O_BILL_ADDR_ID");
	public static final String CO_SHIP_ADDR_ID=new String("ORDERS:O_SHIP_ADDR_ID");
	public static final String CO_STATUS=new String("ORDERS:O_STATUS");
	public static final String CO_NUM_LINES=new String("ORDERS:NUM_LINES"); // new added attribute for HBase
	
	public static final String CFCX=new String("CC_XACTS:");
	public static final String CCX_TYPE=new String("CC_XACTS:CX_TYPE");
	public static final String CCX_NUM=new String("CC_XACTS:CX_NUM");
	public static final String CCX_NAME=new String("CC_XACTS:CX_NAME");
	public static final String CCX_EXPIRY=new String("CC_XACTS:CX_EXPIRY");
	public static final String CCX_AUTH_ID=new String("CC_XACTS:CX_AUTH_ID");
	public static final String CCX_CO_ID=new String("CC_XACTS:CX_CO_ID");
	
	public static final String CFOL=new String("ORDER_LINE:");
	public static final String getOL_I_ID(int j){return new String("ORDER_LINE:"+j+":OL_I_ID");}
	public static final String getOL_I_SUBJECT(int j){return new String("ORDER_LINE:"+j+":OL_I_SUBJECT");}
	public static final String getOL_QTY(int j){return new String("ORDER_LINE:"+j+":OL_QTY");}
	public static final String getOL_DISCOUNT(int j){return new String("ORDER_LINE:"+j+":OL_DISCOUNT");}
	public static final String getOL_COMMENTS(int j){return new String("ORDER_LINE:"+j+":OL_COMMENTS");}

	public static final String CFSC=new String("SHOPPING_CART:");
	public static final String CSC_TIME=new String("SHOPPING_CART:SC_TIME");
	public static final String CSC_NUM_LINES=new String("SHOPPING_CART:SC_NUM_LINES"); // new added attribute for HBase
	
	public static final String CFSCL=new String("SHOPPING_CART_LINE:");
	public static final String getSCL_QTY(int j){return new String("SHOPPING_CART_LINE:"+j+":SCL_QTY");}
	public static final String getSCL_I_ID(int j){return new String("SHOPPING_CART_LINE:"+j+":SCL_I_ID");}
	
}
