package org.zhouw.stps.app.tpcw.datagen;

import java.util.Calendar;
import java.util.GregorianCalendar;

import org.apache.hadoop.hbase.HBaseConfiguration;
import org.apache.hadoop.hbase.HConstants;
import org.apache.hadoop.hbase.client.HTable;
import org.apache.hadoop.hbase.io.BatchUpdate;

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.app.tpcw.*;

public class FinancialServiceDataGen extends DataServiceDataGen {

	private int startRowOfShopping=0;
	private int endRowOfShopping=0;
	private int startRowOfItemStock=0;
	private int endRowOfItemStock=0;
	private int taskID=0;
    private String ITEM_STOCK_TABLENAME=TPCWDataModel.ITEMTABLENAME;
    private String FINA_TABLENAME=TPCWDataModel.SHOPPINGTABLENAME;
	
	public FinancialServiceDataGen() {
	}

	public FinancialServiceDataGen(HBaseConfiguration c) {
		super(c);
	}
	
	public FinancialServiceDataGen(HBaseConfiguration conf, 
			int sr_sh, 	int er_sh, 
			int sr_it, 	int er_it,
			String tableName_sh, String tableName_it,
			int tid) {
		super(conf);
		this.startRowOfShopping=sr_sh;
		this.endRowOfShopping=er_sh;
		this.startRowOfItemStock=sr_it;
		this.endRowOfItemStock=er_it;
		this.taskID=tid;
		this.ITEM_STOCK_TABLENAME=tableName_it;
		this.FINA_TABLENAME=tableName_sh;
	}
	
	public void populateData()
	{
		long start;
		long end;
		if(startRowOfShopping>0){
		start=System.currentTimeMillis();
	  	populateHbaseShoppingTable();
	  	end=System.currentTimeMillis();
	  	System.out.println("in "+ (end-start)+ " milliseconds");
		}
		if(startRowOfItemStock>0){
	  	start=System.currentTimeMillis();
	  	populateHbaseItemStockTable();
	  	end=System.currentTimeMillis();
	  	System.out.println("in "+ (end-start)+ " milliseconds");
		}
	}
	
	private void	populateHbaseShoppingTable()
	{
		//Order variables
		int O_C_ID;
		java.sql.Timestamp O_DATE;
		double O_SUB_TOTAL;
		double O_TAX;
		double O_TOTAL;
		String O_SHIP_TYPE;
		java.sql.Timestamp O_SHIP_DATE;
		int O_BILL_ADDR_ID, O_SHIP_ADDR_ID;
		String O_STATUS;
		
		String CX_TYPE;
		int CX_NUM;
		String CX_NAME;
		java.sql.Date CX_EXPIRY;
		String CX_AUTH_ID;
		double CX_XACT_AMT;
		int CX_CO_ID;
		GregorianCalendar cal;
		
    	System.out.println(taskID+"th task: Start Populating "+ this.FINA_TABLENAME+" table with rows from " 
    			+ this.startRowOfShopping + "to " + this.endRowOfShopping);
    	HTable[] table=new HTable[TPCWDataModel.numOfTablePartitions];
    	HTable[] latestOrderTable=new HTable[TPCWDataModel.numOfTablePartitions];
    	try {
    		for(int i=0;i<TPCWDataModel.numOfTablePartitions;i++)
				{
	    			table[i] = new HTable(conf, this.FINA_TABLENAME+"_"+i);
	    			latestOrderTable[i] = new HTable(conf, TPCWDataModel.LATESTORDERTABLENAME+"_"+i);
				}
			
			
    	    for (int i = startRowOfShopping; i <= endRowOfShopping; i++) {
        		
	    	String row = getStrIDFromInt(i);
	    	BatchUpdate batchUpdate = new BatchUpdate(row,0);
		
	    	int num_items = getRandomInt(1,5);
	    	O_C_ID = getRandomInt(1, TPCWPopulateHbase.NUM_CUSTOMERS);
	    	cal = new GregorianCalendar();
	    	cal.add(Calendar.DAY_OF_YEAR, -1*getRandomInt(1,60));
	    	O_DATE = new java.sql.Timestamp(cal.getTime().getTime());
	    	O_SUB_TOTAL = (double) getRandomInt(1000, 999999)/100;
	    	O_TAX = O_SUB_TOTAL * 0.0825;
	    	O_TOTAL = O_SUB_TOTAL + O_TAX + 3.00 + num_items;
	    	O_SHIP_TYPE = ship_types[getRandomInt(0,num_ship_types-1)];
	    	cal.add(Calendar.DAY_OF_YEAR, getRandomInt(0,7));
	    	O_SHIP_DATE = new java.sql.Timestamp(cal.getTime().getTime());
	    	O_BILL_ADDR_ID = getRandomInt(1, 2*TPCWPopulateHbase.NUM_CUSTOMERS);
	    	O_SHIP_ADDR_ID = getRandomInt(1, 2*TPCWPopulateHbase.NUM_CUSTOMERS);
	    	O_STATUS = status_types[getRandomInt(0,num_status_types-1)];
	    	
	    	batchUpdate.put(TPCWDataModel.CO_C_ID, getStrIDFromInt(O_C_ID).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CO_DATE, String.valueOf(O_DATE.getTime()).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CO_SUB_TOTAL, String.valueOf(O_SUB_TOTAL).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CO_TAX, String.valueOf(O_TAX).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CO_TOTAL, String.valueOf(O_TOTAL).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CO_SHIP_TYPE, O_SHIP_TYPE.getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CO_SHIP_DATE, String.valueOf(O_SHIP_DATE.getTime()).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CO_BILL_ADDR_ID, getStrIDFromInt(O_BILL_ADDR_ID).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CO_SHIP_ADDR_ID, getStrIDFromInt(O_BILL_ADDR_ID).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CO_STATUS, O_STATUS.getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CO_NUM_LINES,  String.valueOf(num_items).getBytes(HConstants.UTF8_ENCODING));

	    	
	    	BatchUpdate updateForLO = new BatchUpdate(getStrIDFromInt(O_C_ID),0);
	    	updateForLO.put(TPCWDataModel.CLATEST_ORDER, row.getBytes());
	    	
	    	int index=GlobalPara.getTableIndex(i, this.FINA_TABLENAME);
	    	
	    	int indexOfLatestOrder=GlobalPara.getTableIndex(O_C_ID, TPCWDataModel.LATESTORDERTABLENAME);
	    	latestOrderTable[indexOfLatestOrder].commit(updateForLO);
	    	
	    	for(int j = 0; j < num_items; j++){

	    	    int OL_I_ID = getRandomInt(1, TPCWPopulateHbase.NUM_ITEMS);
	    	    int OL_QTY = getRandomInt(1, 300);
	    	    double OL_DISCOUNT = (double) getRandomInt(0,30)/100;
	    	    String OL_COMMENTS = getRandomAString(20,100);
	    	    
	    	    batchUpdate.put(TPCWDataModel.getOL_I_ID(j), getStrIDFromInt(OL_I_ID).getBytes(HConstants.UTF8_ENCODING));
	    	   //FIXME the subject of item may not be the same as the other item tables
	    	    batchUpdate.put(TPCWDataModel.getOL_I_SUBJECT(j), SUBJECTS[getRandomInt(0,NUM_SUBJECTS-1)].getBytes(HConstants.UTF8_ENCODING));
	    	    batchUpdate.put(TPCWDataModel.getOL_QTY(j), String.valueOf(OL_QTY).getBytes(HConstants.UTF8_ENCODING));
	    	    batchUpdate.put(TPCWDataModel.getOL_DISCOUNT(j), String.valueOf(OL_DISCOUNT).getBytes(HConstants.UTF8_ENCODING));
	    	    batchUpdate.put(TPCWDataModel.getOL_COMMENTS(j), String.valueOf(OL_COMMENTS).getBytes(HConstants.UTF8_ENCODING));
	    	}
	    	
	    	CX_TYPE = credit_cards[getRandomInt(0, num_card_types-1)];
	    	CX_NUM = getRandomNString(16);
	    	CX_NAME = getRandomAString(14,30);
	    	cal = new GregorianCalendar();
	    	cal.add(Calendar.DAY_OF_YEAR, getRandomInt(10, 730));
	    	CX_EXPIRY = new java.sql.Date(cal.getTime().getTime());
	    	CX_AUTH_ID = getRandomAString(15);	      		
	    	CX_CO_ID = getRandomInt(1,92);
	    	batchUpdate.put(TPCWDataModel.CCX_TYPE, String.valueOf(CX_TYPE).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CCX_NUM, String.valueOf(CX_NUM).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CCX_NAME, String.valueOf(CX_NAME).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CCX_EXPIRY, String.valueOf(CX_EXPIRY.getTime()).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CCX_AUTH_ID, String.valueOf(CX_AUTH_ID).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel .CCX_CO_ID, getStrIDFromInt(CX_CO_ID).getBytes(HConstants.UTF8_ENCODING));
	    	
	    	batchUpdate.put(TPCWDataModel.CSC_NUM_LINES, String.valueOf(0).getBytes(HConstants.UTF8_ENCODING));
	    	batchUpdate.put(TPCWDataModel.CSC_TIME, String.valueOf(System.currentTimeMillis()).getBytes(HConstants.UTF8_ENCODING));

	    	table[index].commit(batchUpdate);
    		
    		if(i % 10000 == 0)
    			System.out.println(taskID+"th task: Populating "+this.FINA_TABLENAME+" in progress at row " +i);
    	    }
    	}catch (java.lang.Exception ex) {
	    	System.err.println("Unable to populate "+this.FINA_TABLENAME+" table");
	    	ex.printStackTrace();
	    	System.exit(1);
	    	}
    	System.out.println(taskID+"th task Done!!!:  Populating  "+this.FINA_TABLENAME+"  table with rows from " + 
    			this.startRowOfShopping+ " to " + this.endRowOfShopping);
    	}
	
	private void	populateHbaseItemStockTable()
	{
    	int I_STOCK;
    	System.out.println(taskID+"th task: Start Populating "+ this.ITEM_STOCK_TABLENAME+" table with items from " 
    			+ this.startRowOfItemStock + "to " + this.endRowOfItemStock);
    	HTable[] table=new HTable[TPCWDataModel.numOfTablePartitions];;

    	try {
    		for(int i=0;i<TPCWDataModel.numOfTablePartitions;i++)
				{
	    			table[i] =new HTable(conf, ITEM_STOCK_TABLENAME+"_"+i);
				}

    	    for (int i = startRowOfItemStock; i <= endRowOfItemStock; i++) {
    		I_STOCK = getRandomInt(10,30);
    		String row = getStrIDFromInt(i);
    		BatchUpdate batchUpdate = new BatchUpdate(row,0);
			///TODO Decide storage format of Date
    		batchUpdate.put(TPCWDataModel.CITEM_I_STOCK, String.valueOf(I_STOCK).getBytes(HConstants.UTF8_ENCODING));
    		
    		int index=GlobalPara.getTableIndex(i, this.ITEM_STOCK_TABLENAME);
			table[index].commit(batchUpdate);
    		
    		if(i % 1000 == 0)
    			System.out.println(taskID+"th task: Populating "+this.ITEM_STOCK_TABLENAME+" in progress at item " +i);
    	    } 	
    	}catch (java.lang.Exception ex) {
	    	System.err.println("Unable to populate "+this.ITEM_STOCK_TABLENAME+" table");
	    	ex.printStackTrace();
	    	System.exit(1);
	    	}
    	System.out.println(taskID+"th task Done!!!:  Populating  "+this.ITEM_STOCK_TABLENAME+"  table with items from " + 
    			this.startRowOfItemStock+ " to " + this.endRowOfItemStock);
    	}
}
