package org.zhouw.stps.client;

import org.zhouw.stps.*;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Random;
import java.util.Vector;

import org.zhouw.stps.monitor.*;
import org.zhouw.stps.msg.SubTransactionGroupProtos.*;
import org.zhouw.stps.rotran.QueryPlan;
import org.zhouw.stps.rotran.ROTranResult;
import org.zhouw.stps.rotran.ROTranMsg.*;
import org.zhouw.stps.tp.TPNode;
import org.zhouw.stps.tp.VirtualTPNode;

import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

/**
 * Accept transactions from programs. 
 * @author zhouw
 *
 */
public class TransExecutor {

	public static HashMap<Long, LinkedBlockingQueue<TransactionResultSetMsg>> tranTable=new HashMap<Long, LinkedBlockingQueue<TransactionResultSetMsg>>(300);
	public static HashMap<Long, LinkedBlockingQueue<ROTranResultMsg>> roTranTable=new HashMap<Long, LinkedBlockingQueue<ROTranResultMsg>>(300);

	private static long counter=0;
	private static ReentrantLock lock=new ReentrantLock();;
	private static ReentrantLock lockForDebug=new ReentrantLock();;
	
	private static ReentrantLock lockForTranTable=new ReentrantLock();
	private static ReentrantLock lockForROTranTable=new ReentrantLock();
	
	private static volatile boolean isGlobalTimeout=false;
	
	public static void startDebug(){
		if(isGlobalTimeout)
			return;
		lockForDebug.lock();
		try{
			if(!isGlobalTimeout){
				isGlobalTimeout=true;
			} else 
				return;
		}finally{
			lockForDebug.unlock();
		}
		
		GlobalPara.logger.warning("[DEBUG]Waiting for nonblocked transactions to finish.");
		try {
			Thread.sleep(GlobalPara.sysConf.clientTimeoutInterval*1000);//waiting for the not blocked transactions to finish
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		GlobalPara.logger.warning("[DEBUG]Start debug process.");
		
		printClientResult();
		debugTPNodes();
		
		do{
			try {
				Thread.sleep(GlobalPara.sysConf.clientTimeoutInterval*1000);//waiting for debug log to be written
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			GlobalPara.logger.warning("[DEBUG]Waiting for debug logs to be written.");
			boolean isContinue=false;
			
			lockForTranTable.lock();
			if(tranTable.size()!=0)
				isContinue=true;
			lockForTranTable.unlock();
			
			lockForROTranTable.lock();
			if(roTranTable.size()!=0)
				isContinue=true;
			lockForROTranTable.unlock();
			
			if(!isContinue)
				break;
		}while(true);
		
		System.exit(0);
	}

	private static void printClientResult()
	{
		lockForTranTable.lock();
		try{
			TransactionResultSetMsg msg=TransactionResultSetMsg.newBuilder().setTranState("DEBUG").build();
			
			for(LinkedBlockingQueue<TransactionResultSetMsg> queue: tranTable.values()){
				queue.offer(msg);
			}
		} finally {
			lockForTranTable.unlock();
		}
		
		lockForROTranTable.lock();
		try{
			ROTranResultMsg romsg=ROTranResultMsg.newBuilder().setTranState(5).build();
			for(LinkedBlockingQueue<ROTranResultMsg> queue:roTranTable.values()){
				queue.offer(romsg);
			}
		}finally{
			lockForROTranTable.unlock();
		}
	}
	
	private static void debugTPNodes()
	{
		Vector<Future<Integer>> results=new Vector<Future<Integer>>();
		for(int destID: GlobalPara.sysConf.TPNodeSet){
			GetMessageTask task=new GetMessageTask(destID, "", "DoDebugService");
			results.add(MsgBroker.executor.submit(task));
			GlobalPara.logger.warning("[DEBUG]Ask TPNode "+ destID + " to start debug pocess");
		}
		for(Future<Integer> result:results) {
			try {
				result.get();
				GlobalPara.logger.warning("[DEBUG]One TPNode finished debugging");
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * For identifying the results belongs to which transaction invocation
	 * @return
	 */
	private static long getCounter()
	{
		lock.lock();
		try{
			return counter++;
		}
		finally{
			lock.unlock();
		}
	}
	public static volatile boolean isMembershipUptodate=true; 
	
	public static void returnRWTranResult(TransactionResultSetMsg msg){
		if(GlobalPara.sysConf.isDebug)
			latestResponseTime=System.currentTimeMillis();
		LinkedBlockingQueue<TransactionResultSetMsg> queue=null;
		long tranid=msg.getClientsideID();
		lockForTranTable.lock();
		try{
			queue=TransExecutor.tranTable.get(tranid);
		} finally{
			lockForTranTable.unlock();
		}
		if(queue==null){
			GlobalPara.logger.warning("Unable to find the queue with tranID "+tranid);
			return;
		}
		try{
			while(!(queue.offer(msg, GlobalPara.timeoutToRetrySendMsg, TimeUnit.MILLISECONDS))){
				GlobalPara.logger.warning("Delivery of transaction result timeout");
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
	
	public static void returnROTranResult(ROTranResultMsg msg)
	{
		if(GlobalPara.sysConf.isDebug)
			latestResponseTime=System.currentTimeMillis();
		
		lockForROTranTable.lock();
		try{
			if(msg!=null){
				LinkedBlockingQueue<ROTranResultMsg> queue=TransExecutor.roTranTable.get(msg.getClientsideID());
				if(queue==null)
					GlobalPara.logger.warning("Unable to find the queue for RO tran: " + msg.getClientsideID());
				else {
					if(!queue.offer(msg))
						GlobalPara.logger.warning("Unable to offer the RO transaction result to the queue");
				}
			}
		}finally{
			lockForROTranTable.unlock();
		}
	}
	
	public static volatile long latestResponseTime=0;
	
	public static ROTranResult runROTran(String query, String dbName)
	{
		try {
			long start=System.currentTimeMillis();
			ROTranResult result;
			
			if(GlobalPara.sysConf.isDebug&&isGlobalTimeout){
				result=new ROTranResult(TranState.ABORT);
				return result;
			}
			
			if(!isMembershipUptodate)
			{
				result=new ROTranResult(TranState.ABORT);
				//GlobalPara.logger.warning("[Membership]Abort the transaction because of the membership is not uptodate");
			}
			else
			{
				long tranactionId=getCounter();
				ROTransactionMsg.Builder msgBuilder=ROTransactionMsg.newBuilder();
				msgBuilder.setClientsideTID(tranactionId);
				msgBuilder.setClientNodeID(GlobalPara.sysConf.localNodeID).setQuery(query).setDbName(dbName);
				
				LinkedBlockingQueue<ROTranResultMsg> queue=new LinkedBlockingQueue<ROTranResultMsg>();
				
				lockForROTranTable.lock();
				try{
					roTranTable.put(tranactionId, queue);
				}finally{
					lockForROTranTable.unlock();
				}
				
				int destID=getRandomTPNode();
				
				//GlobalPara.logger.warning("ROTransaction  "+tranactionId +" will be Sent to LTM " + destID);
				
				if(destID==-1){
					GlobalPara.logger.warning("[Membership]Abort the transaction because of no TP nodes are available");
					result=new ROTranResult(TranState.ABORT);
				} else {
					//GlobalPara.logger.warning("ROTransaction  "+tranactionId +" will be Sent to LTM " + destID +" Right Now");
					MsgBroker.clientSubmitROTran(destID, msgBuilder.build());
					//GlobalPara.logger.warning("RO transaction "+tranactionId+ " is submitted");
					ROTranResultMsg r;
					boolean isTimeout=false;
					while(true){
						try {
							r=queue.poll(GlobalPara.sysConf.clientTimeoutInterval,TimeUnit.SECONDS);
							if(r==null) {
								isTimeout=true;
								if(GlobalPara.sysConf.isDebug&&!isGlobalTimeout){
									StartDebugTask debugTask=new StartDebugTask();
									MsgBroker.executor.submit(debugTask);
								}
								GlobalPara.logger.warning("Timeout to get RO transaction result for id: "+tranactionId );
								continue;
							} else	{
								break;
							}
						} catch (InterruptedException e) {
							e.printStackTrace();
						}
					}
					
					result=new ROTranResult(r);
					if(isTimeout) {
						;//GlobalPara.logger.warning("ROTransaction  "+tranactionId +" completed after being timeout with " +  result.state);
					}
					if(result.state!=TranState.ERROR){
						lockForROTranTable.lock();
						try{
							roTranTable.remove(tranactionId);
						} finally{
							lockForROTranTable.unlock();
						}
					}
					
					if(GlobalPara.sysConf.isDebug&&result.state==TranState.DEBUG){
						long end=System.currentTimeMillis();
						System.out.println("[DEBUG] ROTransaction ("+tranactionId+") sended to LTM " + destID + ", blocked for "+(end-start)+ " milli-seconds.");
						System.out.flush();
					}
				}
			}
			
			long end=System.currentTimeMillis();
			
			//GlobalPara.logger.warning("ROTransaction performance result Writing to logs");
			
			if(result.state==TranState.COMMIT)
				{
				ClientTransactionLogItem msg=new ClientTransactionLogItem(query,end-start,true,end, result.numberOfAccessedItems,true);
				GlobalPara.monitor.writelnCTLog(msg.writeToString());
				}
			else
				{
				ClientTransactionLogItem msg=new ClientTransactionLogItem(query,end-start,false,end, result.numberOfAccessedItems,true);
				GlobalPara.monitor.writelnCTLog(msg.writeToString());
				}
			//GlobalPara.logger.warning("ROTransaction performance result Writing to logs..done");
		
			
			
			return result;
		
		} catch (java.lang.NullPointerException e) {
			e.printStackTrace();
		}
		
		return null;
		
	}
	
	public static ROTranResult runROTran(QueryPlan plan)
	{
		return runROTran(plan, "");
	}
	
	public static ROTranResult runROTran(QueryPlan plan, String name)
	{
		long start=System.currentTimeMillis();
		ROTranResult result;
		
		if(GlobalPara.sysConf.isDebug&&isGlobalTimeout){
			result=new ROTranResult(TranState.ABORT);
			return result;
		}
		
		if(!isMembershipUptodate)
		{
			result=new ROTranResult(TranState.ABORT);
			//GlobalPara.logger.warning("[Membership]Abort the transaction because of the membership is not uptodate");
		}
		else
		{
			if(plan.rootTable==null||plan.rootTable.allowedRows.size()==0){
				result=new ROTranResult(TranState.ABORT);
				result.errMsg="RootTable not assigned or empty set of root records";
				return result;
			}
			
			long tranactionId=getCounter();
			ROTransactionMsg.Builder msgBuilder=ROTransactionMsg.newBuilder();
			msgBuilder.setClientsideTID(tranactionId);
			msgBuilder.setClientNodeID(GlobalPara.sysConf.localNodeID).setPlan(plan.constructQueryPlanMsg());
			
			LinkedBlockingQueue<ROTranResultMsg> queue=new LinkedBlockingQueue<ROTranResultMsg>();
			
			lockForROTranTable.lock();
			try{
				roTranTable.put(tranactionId, queue);
			}finally{
				lockForROTranTable.unlock();
			}
						
			int destID;
			
			if(GlobalPara.sysConf.doRandomTCForJoins){
				destID=getRandomTPNode();
			} else {
				String rowkey=plan.rootTable.allowedRows.get(0);
				VirtualTPNode id=TPNode.getVirtualNode(plan.rootTable.tableName, rowkey);
				if(!GlobalPara.sysConf.TPNodeSet.contains(id.physicalID))//the ring has not been updated, this transaction should not be executed
					return null;
				destID=id.physicalID;
			}
			
			if(destID==-1)
			{
				GlobalPara.logger.warning("[Membership]Abort the transaction because of no TP nodes are available");
				result=new ROTranResult(TranState.ABORT);
			}
			else
			{
				MsgBroker.clientSubmitROTran(destID, msgBuilder.build());
				
				ROTranResultMsg r;
				boolean isTimeout=false;
				while(true){
					try {
						r=queue.poll(GlobalPara.sysConf.clientTimeoutInterval,TimeUnit.SECONDS);
						
						if(r==null) {
							isTimeout=true;
							GlobalPara.logger.warning("Timeout to get RO transaction result for id: "+tranactionId+" with name: "+
									name+ " root row:"+ plan.rootTable.allowedRows.get(0));
							
							if(GlobalPara.sysConf.isDebug&&!isGlobalTimeout){
								StartDebugTask debugTask=new StartDebugTask();
								MsgBroker.executor.submit(debugTask);
							}
							
							continue;
						} else {
							break;
						}
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				result=new ROTranResult(r);
				if(isTimeout){
					;//GlobalPara.logger.warning("ROTransaction  "+tranactionId +" completed after being timeout with " +  result.state);
				}
				if(result.state!=TranState.ERROR){
					lockForROTranTable.lock();
					try{
						roTranTable.remove(tranactionId);
					}finally{
						lockForROTranTable.unlock();
					}
				}
				
				if(GlobalPara.sysConf.isDebug&&result.state==TranState.DEBUG){
					long end=System.currentTimeMillis();
					System.out.println("[DEBUG] ROTransaction ("+tranactionId+") sended to LTM " + destID + ", blocked for "+(end-start)+ " milli-seconds.");
					System.out.flush();
				}
			}
		}
		
		long end=System.currentTimeMillis();
		
		if(result.state==TranState.COMMIT)
			{
			ClientTransactionLogItem msg=new ClientTransactionLogItem(name,end-start,true,end, result.numberOfAccessedItems,true);
			GlobalPara.monitor.writelnCTLog(msg.writeToString());
			}
		else
			{
			ClientTransactionLogItem msg=new ClientTransactionLogItem(name,end-start,false,end, result.numberOfAccessedItems,true);
			GlobalPara.monitor.writelnCTLog(msg.writeToString());
			}
		return result;
	}
	
	private static int getRandomTPNode()
	{
		int destID=-1;
		
		int num=ran.nextInt(GlobalPara.sysConf.TPNodeSet.size());
		int count=0;
		for(int nodeID:GlobalPara.sysConf.TPNodeSet)
		{
			destID=nodeID;
			if(count==num)
				break;
			count++;
		}
		return destID;
	}
	
	public static TransactionResult runTransaction(Transaction t, boolean isPhysicalTableName)
	{
		TransactionResult result;
		if(t==null||t.trans==null||t.trans.length==0) {
			result=new TransactionResult();
			result.state=TranState.ABORT;
			GlobalPara.logger.warning("Invalid transaction inputted.");
			return result;
		}
			
		long start=System.currentTimeMillis();
		
		if(!isPhysicalTableName){
			for(SubTransaction st: t.trans){
				st.transformIntoPhysicalTableName();
			}
		}
		
		if(GlobalPara.sysConf.isDebug&&isGlobalTimeout){
			result=new TransactionResult();
			result.state=TranState.ABORT;
			return result;
		}
		
		if(!isMembershipUptodate){
			result=new TransactionResult();
			result.state=TranState.ABORT;
			//GlobalPara.logger.warning("[Membership]Abort the transaction because of the membership is not uptodate");
		} else {
			long tranactionId=getCounter();
			t.clientsideTID=tranactionId;
			t.clientNodeID=GlobalPara.sysConf.localNodeID;
			
			long id=0;// id is not used
			TransCoodinator tc=new TransCoodinator(id,t);
			SubTransactionGroup[] stgs= tc.decomposeTrans();
			
			if(stgs==null){
				result =new TransactionResult();
				result.state=TranState.ABORT;
			} else {
				int nodeID=findNodeToTransactionCoordination(stgs);
				//GlobalPara.logger.warning("Transaction id " + tranactionId  +" select TP Node "+ nodeID +" to execute the transaction");
				
				LinkedBlockingQueue<TransactionResultSetMsg> queue=new LinkedBlockingQueue<TransactionResultSetMsg>();
				lockForTranTable.lock();
				try{
					tranTable.put(tranactionId, queue);
				} finally {
					lockForTranTable.unlock();
				}
				
				result=MsgBroker.SubmitGlobalTransaction(nodeID, t);
				
				if(result==null){//failure during submitting
					result =new TransactionResult();
					result.state=TranState.ABORT;
				} else {
					result.state=TranState.ABORT;
					TransactionResultSetMsg r;
					boolean isTimeout=false;
					while(true){//the message handling is in sequential, so no concurrent access to the data structure
						try {
								r=queue.poll(GlobalPara.sysConf.clientTimeoutInterval,TimeUnit.SECONDS);// 5 seconds is a SLA
								if(r==null) {
									isTimeout=true;
									if(GlobalPara.sysConf.isDebug&&!isGlobalTimeout){
										StartDebugTask debugTask=new StartDebugTask();
										MsgBroker.executor.submit(debugTask);
									}
									GlobalPara.logger.warning("Timeout to get RWtransaction result for id "+tranactionId+", tran name:"+t.name+", number of rows:"+ t.trans.length+", first row:" + t.trans[0].strTblName+"::"+t.trans[0].strRowKey);
									continue;//break;
								} else {
									String state=r.getTranState();
									if(r.hasNumberOfAccessedItems())
										result.numberOfAccessedItems=r.getNumberOfAccessedItems();
									result.state=TranState.valueOf(state);
									int count=r.getResultCount();
									if(count==0)
										result.result=null;
									else
									{
										result.result=new String[count][2];
										for(int i=0;i<count;i++)
										{
											result.result[i][0]=r.getResult(i).getKey();
											result.result[i][1]=r.getResult(i).getValue();
										}
									}
									if(isTimeout){
										;//GlobalPara.logger.warning("Transaction  "+tranactionId +" completed after being timeout with " +  result.state);
									}
									break;
								}
							} catch (InterruptedException e) {
								e.printStackTrace();
							}
						}
				}
				if(result.state!=TranState.ERROR) {//if the system is not in recovery
					lockForTranTable.lock();
					try{
						tranTable.remove(tranactionId);
					} finally{
						lockForTranTable.unlock();
					}
				}
				
				if(GlobalPara.sysConf.isDebug&&result.state==TranState.DEBUG){
					long end=System.currentTimeMillis();
					System.out.println("[DEBUG] RWTransaction ("+tranactionId+") sended to LTM " + nodeID + ", blocked for "+(end-start)+ " milli-seconds.");
					System.out.flush();
				}
			}
		}
		
		long end=System.currentTimeMillis();
		if(result.state==TranState.COMMIT)
		{
			ClientTransactionLogItem msg=new ClientTransactionLogItem(t.name,end-start,true,end, result.numberOfAccessedItems, false);
			GlobalPara.monitor.writelnCTLog(msg.writeToString());
		}
		else
		{
			System.out.println("Transaction "+t.name+ " is aborted");
			ClientTransactionLogItem msg=new ClientTransactionLogItem(t.name,end-start,false,end, result.numberOfAccessedItems, false);
			GlobalPara.monitor.writelnCTLog(msg.writeToString());
		}
		
		return result;
	}
	
	/**
	 * 
	 * @param t: Each subTransaction contains database Name, logical table name and logical column name
	 * @return
	 */
	public static TransactionResult runTransaction(Transaction t)
	{
		return runTransaction(t, false);
	}
	
	public static void clearTransactionWaitingTable(){
		GlobalPara.logger.warning("Clearing all RW transactions pool");
		
		lockForTranTable.lock();
		try{
			GlobalPara.logger.warning("RW transactions pool lock obtained");
			TransactionResultSetMsg msg=TransactionResultSetMsg.newBuilder().setTranState("ERROR").build();
			for(LinkedBlockingQueue<TransactionResultSetMsg> queue: tranTable.values()){
				queue.offer(msg);
			}
			tranTable.clear();
		} finally{
			lockForTranTable.unlock();
			GlobalPara.logger.warning("RW transactions pool lock released");
		}
		
		GlobalPara.logger.warning("Clearing all RO transaction pool");
		
		lockForROTranTable.lock();
		try {
			GlobalPara.logger.warning("RO transactions pool lock obtained");
			ROTranResultMsg romsg=ROTranResultMsg.newBuilder().setTranState(4).build();
			for(LinkedBlockingQueue<ROTranResultMsg> queue:roTranTable.values()){
				queue.offer(romsg);
			}
			roTranTable.clear();
		} finally {
			lockForROTranTable.unlock();
			GlobalPara.logger.warning("RO transactions pool lock released");
		}
		
		GlobalPara.logger.warning("All transaction pool clear completed");
		
	}
	
	
	/**
	 * // locate the coordinator to the node which handles the most sub-transactions
	 * random select a node which holds any required data items 
	 * @param stgs
	 * @return the id of selected TP node
	 */
	public static int findNodeToTransactionCoordination(SubTransactionGroup[] stgs)
	{
		if(stgs==null||stgs.length==0)
			{
			GlobalPara.logger.warning("input stg group is empty");
			return -1;
			}
		int index= ran.nextInt(stgs.length);
		return stgs[index].TPnodeID;
		
//		int maxNum=stgs[0].trans.length;
//		int result=stgs[0].TPnodeID;
//		for(SubTransactionGroup s: stgs)
//		{
//			if(s.trans.length>maxNum)
//			{
//				maxNum=s.trans.length;
//				result=s.TPnodeID;
//			}
//		}
//		return result;
	}
	
	private static Random ran=new Random();
}
