package org.zhouw.stps.components;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.zhouw.stps.adapter.Adapter;
import org.zhouw.stps.app.DataTable;
import org.zhouw.stps.tp.DataRowContainer;
import org.zhouw.stps.tp.DataRowEntry;
import org.zhouw.stps.tp.GlobalTransManager;
import org.zhouw.stps.tp.UpdatedDataRow;
import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.SubTransaction;

import java.util.concurrent.locks.ReentrantLock;
import org.zhouw.stps.adapter.CheckpointThreadExecutor;

public class AdapterCoordinator {// extends Component

	//This workers are all single-threaded, so as to ensure the sequence of updates checkpoint
	protected AdapterWorkerComponent[] readWorkers;
	protected AdapterWorkerComponent[] writeWorkers;
	
	protected boolean isCommitted[];
	protected ReentrantLock lock=new ReentrantLock();
	
	public void shutdownWorkers()
	{
		for(AdapterWorkerComponent w:writeWorkers)
			w.stop();
		
		if(GlobalPara.sysConf.isSeparateReadWriteWorkers)
			for(AdapterWorkerComponent r:readWorkers)
				r.stop();
	}
	
//	@Override
//	public void handler(CompMessage msg) {
//		switch(msg.type){
//		case ToAdapter_ReadARow.typeid:
//			ToAdapter_ReadARow readmsg=(ToAdapter_ReadARow)msg;
//			AdapterWorkerComponent worker=this.getAReadWorker();
//			worker.sendMsg(readmsg);
//			break;
//		}
//	}
	
	public AdapterCoordinator(ThreadGroup g) {
		//super("AdapterCoordinator");
		this.readWorkerCount=GlobalPara.sysConf.CPBGThreadPoolSize;
		curIndex=0;
		isCommitted=new boolean[GlobalPara.sysConf.CPBGThreadPoolSize];
		writeWorkers=new AdapterWorkerComponent[GlobalPara.sysConf.CPBGThreadPoolSize];
		for(int i=0;i<writeWorkers.length;i++){
			writeWorkers[i]=new AdapterWorkerComponent(i);
			isCommitted[i]=false;
			GlobalPara.initOneComponent(g,writeWorkers[i]);
		}
		
		if(GlobalPara.sysConf.isSeparateReadWriteWorkers){
			readWorkers=new AdapterWorkerComponent[readWorkerCount];
			for(int i=0;i<readWorkers.length;i++){
				readWorkers[i]=new AdapterWorkerComponent(i);
				GlobalPara.initOneComponent(g,readWorkers[i]);
			}
		}
		else
			readWorkers=writeWorkers;
		
	}
	
	/**
	 * Because load data requests only send by the single thread of data manager,
	 * so there is no need to apply an extra service component for it.
	 * @param table
	 * @param rowkey
	 */
	public void loadData(DataTable table, String rowkey)
	{
		ToAdapter_ReadARow readmsg=new ToAdapter_ReadARow(table,rowkey);
		AdapterWorkerComponent worker=this.getAReadWorker();
		worker.sendMsg(readmsg);
	}
	
	/** Now load data requests not distributed by another thread 
	public void loadData(DataTable table, String rowkey)
	{
		ToAdapter_ReadARow readmsg=new ToAdapter_ReadARow(table,rowkey);
		this.sendMsg(readmsg);
	}
	**/
	
//	public void loadDataForBackupWhenRecovery(DataTable table, String rowKey, DataRowContainer row)
//	{
//		ToAdapter_ReadARowForWhenRecovery readmsg=new ToAdapter_ReadARowForWhenRecovery(
//				table, rowKey, row);
//		AdapterWorkerComponent worker=this.getAWorker(rowKey);
//		worker.sendMsg(readmsg);
//	}
//	
//	public void loadDataForPrimaryWhenRecovery(DataRowEntry row)
//	{
//		ToAdapter_ReadARowForWhenRecovery readmsg=new ToAdapter_ReadARowForWhenRecovery(row);
//		AdapterWorkerComponent worker=this.getAWorker(row.rowKey);
//		worker.sendMsg(readmsg);
//	}
	
	/**
	 * Checkpoint the committed updates
	 * @param snapshotID
	 * @param tableSchema
	 * @param rowKey
	 * @param value
	 */
	public void checkpointOneRow(long timestamp, DataTable tableSchema, String rowKey,UpdatedDataRow value, SubTransaction t)
	{
		if(GlobalPara.sysConf.isDoCheckPoint){
			ToAdapter_CheckpointARow cpmsg=new ToAdapter_CheckpointARow(timestamp, tableSchema, rowKey,value, t);
			AdapterWorkerComponent worker=this.getAWriteWorker(rowKey);
			worker.sendMsg(cpmsg);
		}
	}
	
	/**
	 * Insert one row
	 * @param tableSchema
	 * @param rowkey
	 * @param row
	 */
	public void putOneRow(DataTable tableSchema,  String rowkey, Iterable<Map.Entry<String,String>> row)
	{
		AdapterWorkerComponent worker=this.getARadnomWorker();
		ToAdapter_PutARow msg=new ToAdapter_PutARow(tableSchema,rowkey, row );
		worker.sendMsg(msg);
	}
	
	public void delRows(DataTable dt, List<String> rowsToBeDeleted)
	{
		if(rowsToBeDeleted==null){
			GlobalPara.logger.warning("The input rows are null");
			return;
		}
		for(String rowkey:rowsToBeDeleted){
			AdapterWorkerComponent worker=this.getAWriteWorker(rowkey);
			ToAdapter_DelARow msg=new ToAdapter_DelARow(dt,rowkey);
			worker.sendMsg(msg);
		}
	}
	
	public void commitAllWorkers(){
		lock.lock();
		try{
			for(int i=0;i<isCommitted.length;i++){
				isCommitted[i]=false;
			}
		} finally{
			lock.unlock();
		}
		ToAdapter_CommitAllMsg msg=new ToAdapter_CommitAllMsg();
		for(int i=0;i<writeWorkers.length;i++){
			writeWorkers[i].sendMsg(msg);
		}
	}
	
	public void createTable(Iterable <? extends DataTable> tableSchemas)
	{
		Adapter adapter=AdapterWorkerComponent.getAdapter();
		adapter.createTable(tableSchemas);
	}
	
	public boolean isAllWorkersCommitFinish(){
		lock.lock();
		try{
			boolean result=true;
			for(int i=0;i<isCommitted.length;i++){
				if(!isCommitted[i]){
					result=false;
					break;
				}
			}
			return result;
		} finally{
			lock.unlock();
		}
		
	}
	
	public void reportCommitFinish(int index){
		lock.lock();
		try{
			isCommitted[index]=true;
		} finally{
			lock.unlock();
		}
	}
	
	private int curIndex;
	private int readWorkerCount;
	
	//distribute requests by round-robin
	protected AdapterWorkerComponent getAReadWorker(){
		int index=ran.nextInt(readWorkerCount);
		return this.readWorkers[index];
		
//		
//		if(GlobalPara.sysConf.isSeparateReadWriteWorkers){
////			curIndex++;
////			curIndex%=readWorkerCount;
////			return this.readWorkers[curIndex];
//			
//		} else {
//			int index=ran.nextInt(GlobalPara.sysConf.CPBGThreadPoolSize);
//			return this.writeWorkers[index];
//		}
	}
	
	protected AdapterWorkerComponent getAWriteWorker(String rowkey){
		int rowid=CheckpointThreadExecutor.getRowID(rowkey);
		int index=rowid% GlobalPara.sysConf.CPBGThreadPoolSize;
		return this.writeWorkers[index];
	}
	protected Random ran=new Random(System.currentTimeMillis());
	protected AdapterWorkerComponent getARadnomWorker(){
		int index=ran.nextInt(GlobalPara.sysConf.CPBGThreadPoolSize);
		return this.writeWorkers[index];
	}
}
