package org.zhouw.stps.components;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.adapter.Adapter;
import org.zhouw.stps.adapter.HBaseAdapter;
import org.zhouw.stps.adapter.SimpleDBAdapter;
import org.zhouw.stps.tp.DataRowContainer;
import org.zhouw.stps.tp.DataRowEntry;
import org.zhouw.stps.app.*;

public class AdapterWorkerComponent extends Component {
	
	public int invokedTimes; 
	protected Adapter adapter;
	protected int index;
	public int queueLength; 
	
    public static Adapter getAdapter()
    {
    	if(GlobalPara.sysConf.DB.compareTo("HBase")==0)
    		return new HBaseAdapter();
    	else
    		return new SimpleDBAdapter();
    }
    
	public AdapterWorkerComponent(int index) {
		super("AdapterWorkerComponent"+index);
		this.invokedTimes=0;
		this.adapter=getAdapter();
		this.index=index;
	}
	
	@Override
	public void handler(CompMessage msg) {
		invokedTimes++;
		if(invokedTimes%1000==0){
			int curQueueLength=super.queue.size();
			if(queueLength<curQueueLength&&curQueueLength>500){
				System.out.println("[Warining]AdapterWorker "+index +" has "+curQueueLength+" delayed messages in queue");
			}
		}
		switch(msg.type){
		case ToAdapter_ReadARow.typeid:
			ToAdapter_ReadARow readmsg=(ToAdapter_ReadARow)msg;
			handleReadARow(readmsg);
			break;
		case ToAdapter_CheckpointARow.typeid:
			ToAdapter_CheckpointARow checkpointmsg=(ToAdapter_CheckpointARow)msg;
			handleCheckpointARow(checkpointmsg);
			break;
		case ToAdapter_PutARow.typeid:
			ToAdapter_PutARow putmsg=(ToAdapter_PutARow)msg;
			handlePutARow(putmsg);
			break;
		case ToAdapter_CommitAllMsg.typeid:
			handleCommitAll();
			break;
		case ToAdapter_DelARow.typeid:
			ToAdapter_DelARow delmsg=(ToAdapter_DelARow)msg;
			handleDelARow(delmsg);
			break;
//		case ToAdapter_ReadARowForWhenRecovery.READ_BACKUP_RECOVERY:
//			ToAdapter_ReadARowForWhenRecovery rbmsg=(ToAdapter_ReadARowForWhenRecovery)msg;
//			handleReadARowWhenRecovery(rbmsg.table,rbmsg.rowKey,rbmsg.backupRow);
//			break;
//		case ToAdapter_ReadARowForWhenRecovery.READ_PRIMARY_RECOVERY:
//			ToAdapter_ReadARowForWhenRecovery rpmsg=(ToAdapter_ReadARowForWhenRecovery)msg;
//			handleReadARowWhenRecovery(rpmsg.table,rpmsg.rowKey,rpmsg.row);
//			break;
		}
	}
	
//	protected void handleReadARowWhenRecovery(DataTable table, String rowkey, DataRowContainer row)
//	{
//		HashMap<String, String> data=new HashMap<String, String>();
//		this.adapter.loadData(table, rowkey, data);
//		row.rowData.putAll(data);
//		GlobalPara.ltmComp.reportDelayedLoad_BackupDataForCommitRecoveredComplete(row);
//	}
//	
//	protected void handleReadARowWhenRecovery(DataTable table, String rowkey, DataRowEntry row)
//	{
//		HashMap<String, String> data=new HashMap<String, String>();
//		this.adapter.loadData(table, rowkey, data);
//		row.committedData.rowData.putAll(data);
//		GlobalPara.ltmComp.reportDelayedLoad_PrimaryDataForCommitRecoveredComplete(row);
//	}
	
	protected void handleReadARow(ToAdapter_ReadARow msg)
	{
		HashMap<String, String> row=new HashMap<String, String>();
		int result=this.adapter.loadData(msg.table, msg.rowkey, row);
		if(result!=-2) //if cannot contact underlying cloud data store, stop the system
			GlobalPara.ltmComp.receiveLoadedData(msg.table, msg.rowkey, row);
	}
	
	protected void handleCheckpointARow(ToAdapter_CheckpointARow msg)
	{
		this.adapter.checkpointOneItem(msg.timestamp, msg.tableSchema, msg.rowKey, msg.value);
		//FIXME it should report checkpoint done after the data item has been successfully written to the cloud data store
		//however, the adapter may cache the updates and checkpoint them in batch
		GlobalPara.ltmComp.reportCheckpointDone(msg.tableSchema.getPhysicalTableName(), msg.rowKey, msg.t, msg.timestamp);
	}
	
	protected void handlePutARow(ToAdapter_PutARow msg)
	{
		this.adapter.putOneRow(msg.tableSchema, msg.rowkey, msg.row);
	}
	
	protected void handleDelARow(ToAdapter_DelARow msg)
	{
		List<String> rows=new LinkedList<String>();
		rows.add(msg.rowkey);
		this.adapter.delRows(msg.dt,rows);
	}
	
	protected void handleCommitAll()
	{
		this.adapter.commitAll();
		GlobalPara.adapterComp.reportCommitFinish(this.index);
	}
}
