package org.zhouw.stps.components;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;


import org.zhouw.stps.GlobalPara;


public abstract class Component implements Runnable {

	protected String name;
	
	protected LinkedBlockingQueue<CompMessage> queue;
	
	public void run(){
		try {
			while(true){
				CompMessage msg=queue.poll(1000000, TimeUnit.MILLISECONDS);
				if(msg!=null){
					if(msg.type!=CompMessage.STOP)
						handler(msg);
					else
						break;
				} else
					GlobalPara.logger.warning("Component "+this.name+" timeout in waiting for incoming messages");
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
	
	
	//To be override
	public abstract void handler(CompMessage msg);

	public void stop()
	{
		CompMessage msg=new CompMessage(0,CompMessage.STOP);
		if(!this.queue.offer(msg))
			GlobalPara.logger.warning("Component "+this.name+" failed to insert message to the queue");
	}
	
	public void sendMsg(CompMessage msg)
	{
		if(!this.queue.offer(msg))
			GlobalPara.logger.warning("Component "+this.name+" failed to insert message to the queue");
	}
	
	public Component(String name) {
		super();
		this.name = name;
		this.queue=new LinkedBlockingQueue<CompMessage>();
	}


	public String getName() {
		return name;
	}
	
	public int getNumOfTasksInTheQueue()
	{
		return queue.size();
	}
}
