package org.zhouw.stps.components;

import java.util.HashMap;
import java.util.Vector;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.CountDownLatch;

import org.zhouw.stps.app.DataTable;
import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.SubTransaction;
import org.zhouw.stps.SubTransactionGroup;
import org.zhouw.stps.rotran.LocalROTM;
import org.zhouw.stps.rotran.ROTranMsg.SubROTranGroupMsg;
import org.zhouw.stps.tp.DataRowContainer;
import org.zhouw.stps.tp.DataRowEntry;
import org.zhouw.stps.tp.RingMembership;
import org.zhouw.stps.tp.TPNode;
import org.zhouw.stps.tp.LTMActions;
import org.zhouw.stps.tp.UpdatesOfST;

public class LTMComponent extends Component {

	public LTMComponent() {
		super("LTM");
	}

	@Override
	public void handler(CompMessage msg) {
		switch(msg.type){
		case CompMessage.START_DEBUG:
			TPNode.LTM.printDebugMsg();
			GlobalPara.debugQueue.offer("LTM");
			break;
		case AdapterTo_ReadARowResult.typeid:
			AdapterTo_ReadARowResult rowmsg=(AdapterTo_ReadARowResult)msg;
			LTMActions.oneRowDataLoaded(rowmsg.table, rowmsg.rowkey, rowmsg.rowData);
			break;
		case RWGTMtoLTM_NewSTG.typeid:
			RWGTMtoLTM_NewSTG stgmsg=(RWGTMtoLTM_NewSTG)msg;
			TPNode.LTM.addOneSTG(stgmsg.stg);
			break;
		case RWGTPtoLTM_DoRestart.typeid:
			RWGTPtoLTM_DoRestart restartMsg=(RWGTPtoLTM_DoRestart)msg;
			TPNode.LTM.doRestartSTG(restartMsg.timestamp, restartMsg.key, restartMsg.newTimestamp);
			break;
		case RWGTPtoLTM_AggreementMsg.DOABORT:
			RWGTPtoLTM_AggreementMsg abortmsg=(RWGTPtoLTM_AggreementMsg)msg;
			TPNode.LTM.doAbortSTG(abortmsg.stgKey,abortmsg.timestamp);
			break;
		case RWGTPtoLTM_AggreementMsg.DOCOMMIT:
			RWGTPtoLTM_AggreementMsg commitmsg=(RWGTPtoLTM_AggreementMsg)msg;
			TPNode.LTM.doCommitOneSTG(commitmsg.stgKey);
			break;
		case LTMtoLTM_BackupCommitData.UPDATE:
			LTMtoLTM_BackupCommitData backupmsg=(LTMtoLTM_BackupCommitData)msg;
			if(backupmsg.data.size()==1)
				TPNode.LTM.doBackUp(backupmsg.data.get(0));
			else
				TPNode.LTM.bactchBackUp(backupmsg.data);
			
//			GlobalPara.logger.warning("[BackupRecover]Doing batch backup data items");
			if(backupmsg.queue!=null){
				//GlobalPara.logger.warning("[BackupRecover]Offer queue result");
				backupmsg.queue.offer(0);
			}
			
			break;
		case LTMtoLTM_BackupCommitData.REMOVE:
			LTMtoLTM_BackupCommitData rmbackupmsg=(LTMtoLTM_BackupCommitData)msg;
			TPNode.LTM.removeBackUp(rmbackupmsg.data.get(0));
			break;
		case ROGTMtoLTM_NewROSTGMsg.NEWROSTG:
			ROGTMtoLTM_NewROSTGMsg romsg=(ROGTMtoLTM_NewROSTGMsg)msg;
			LocalROTM.handlingSubmitROTran(romsg.msg);
			break;
		case MemberTo_StartRecoverMsg.BEGIN_RECOVERY:
			MemberTo_StartRecoverMsg recomsg=(MemberTo_StartRecoverMsg)msg;
			TPNode.LTM.doDataReorganization(recomsg.newMembership, recomsg.failedNodes);
			break;
		case MemberTo_StartRecoverMsg.BEGIN_COMMIT:
			TPNode.LTM.commitRecoveredUpdates();
			break;
		case MemberTo_StartRecoverMsg.LOCAL_WORKINGDATA_COMPLETE:
			TPNode.LTM.reportWorkingDataPoolReorganizationComplete();
			break;
		case MemberTo_StartRecoverMsg.LOCAL_BACKUPDATA_COMPLETE:
			TPNode.LTM.reportBackupDataPoolReorganizationComplete();
			break;
//		case AdapterToLTM_LoadARow_InRecovery.PRIMARY:
//			AdapterToLTM_LoadARow_InRecovery rmsg=(AdapterToLTM_LoadARow_InRecovery)msg;
//			TPNode.LTM.delayedLoadDataForCommitRecoveredComplete(rmsg.row);
//			break;
//		case AdapterToLTM_LoadARow_InRecovery.BACKUP:
//			AdapterToLTM_LoadARow_InRecovery backmsg=(AdapterToLTM_LoadARow_InRecovery)msg;
//			TPNode.LTM.delayedLoadDataForCommitRecoveredComplete(backmsg.backupRow);
//			break;
		case ToLTM_CommitDoneMsg.COMMIT_DONE:
			ToLTM_CommitDoneMsg comsg=(ToLTM_CommitDoneMsg)msg;
			TPNode.LTM.reportCommitDone(comsg.tableName, comsg.rowKey, comsg.t, comsg.timestamp);
			break;
		case ToLTM_CommitDoneMsg.CHECKPOINT_DONE:
			ToLTM_CommitDoneMsg cpmsg=(ToLTM_CommitDoneMsg)msg;
			TPNode.LTM.reportCheckpointDone(cpmsg.tableName, cpmsg.rowKey, cpmsg.t, cpmsg.timestamp);
			break;
		case CompMessage.START_MONITOR_MM:
			TPNode.MM.startMonitor();
			break;
		case CompMessage.DO_MONITOR_MM:
			TPNode.MM.printMonitor();
			break;
		}
	}
	
	public void startMonitorMM(){
		CompMessage msg=new CompMessage(0, CompMessage.START_MONITOR_MM);
		this.sendMsg(msg);
	}
	
	public void doMonitorMM(){
		CompMessage msg=new CompMessage(0, CompMessage.DO_MONITOR_MM);
		this.sendMsg(msg);
	}
	
	public void reportCommitDone(String tableName, String rowKey,SubTransaction t, long timestamp){
		ToLTM_CommitDoneMsg msg=new ToLTM_CommitDoneMsg(ToLTM_CommitDoneMsg.COMMIT_DONE,tableName,rowKey, t, timestamp);
		this.sendMsg(msg);
	}
	
	public void reportCheckpointDone(String tableName, String rowKey,SubTransaction t, long timestamp){
		ToLTM_CommitDoneMsg msg=new ToLTM_CommitDoneMsg(ToLTM_CommitDoneMsg.CHECKPOINT_DONE,tableName,rowKey, t, timestamp);
		this.sendMsg(msg);
	}
	
//	public void reportDelayedLoad_PrimaryDataForCommitRecoveredComplete(DataRowEntry row)
//	{
//		AdapterToLTM_LoadARow_InRecovery msg=new AdapterToLTM_LoadARow_InRecovery(row);
//		this.sendMsg(msg);
//	}
//	
//	public void reportDelayedLoad_BackupDataForCommitRecoveredComplete(DataRowContainer row)
//	{
//		AdapterToLTM_LoadARow_InRecovery msg=new AdapterToLTM_LoadARow_InRecovery(row);
//		this.sendMsg(msg);
//	}
	
	public void reportWorkingDataComplete()
	{
		MemberTo_StartRecoverMsg msg=new MemberTo_StartRecoverMsg(MemberTo_StartRecoverMsg.LOCAL_WORKINGDATA_COMPLETE);
		this.sendMsg(msg);
	}
	
	public void reportDataBackupComplete()
	{
		MemberTo_StartRecoverMsg msg=new MemberTo_StartRecoverMsg(MemberTo_StartRecoverMsg.LOCAL_BACKUPDATA_COMPLETE);
		this.sendMsg(msg);
	}
	
	
	public void startCommitRecoveredUpdates()
	{
		MemberTo_StartRecoverMsg msg=new MemberTo_StartRecoverMsg(MemberTo_StartRecoverMsg.BEGIN_COMMIT);
		this.sendMsg(msg);
	}
	
	public void startRecoverDataItems(RingMembership newMembership, Vector<Integer> failedNodes)
	{
		MemberTo_StartRecoverMsg msg=new MemberTo_StartRecoverMsg(MemberTo_StartRecoverMsg.BEGIN_RECOVERY
				,newMembership, failedNodes);
		this.sendMsg(msg);
	}
	
	public void submitROTran(SubROTranGroupMsg msg) {
		ROGTMtoLTM_NewROSTGMsg romsg=new ROGTMtoLTM_NewROSTGMsg(msg);
		GlobalPara.ltmComp.sendMsg(romsg);
	}
	
	
	public void doBatchBackupData(Vector<UpdatesOfST> data, ArrayBlockingQueue<Integer> queue){
		LTMtoLTM_BackupCommitData msg=new LTMtoLTM_BackupCommitData(data, queue);
		GlobalPara.ltmComp.sendMsg(msg);
	}
	
	public void doBackupCommittedDataUpdates(UpdatesOfST data){
		LTMtoLTM_BackupCommitData msg=new LTMtoLTM_BackupCommitData(LTMtoLTM_BackupCommitData.UPDATE, data);
		GlobalPara.ltmComp.sendMsg(msg);
	}
	
	public void doRemoveBackup(UpdatesOfST data){
		LTMtoLTM_BackupCommitData msg=new LTMtoLTM_BackupCommitData(LTMtoLTM_BackupCommitData.REMOVE, data);
		GlobalPara.ltmComp.sendMsg(msg);
	}
	
	public void submitSTG(SubTransactionGroup stg){
		if(stg!=null){
			RWGTMtoLTM_NewSTG stgmsg=new RWGTMtoLTM_NewSTG(stg);
			GlobalPara.ltmComp.sendMsg(stgmsg);
		}
	}
	
	public void receiveLoadedData(DataTable table,String rowkey, HashMap<String, String> rowData)
	{
		AdapterTo_ReadARowResult resultmsg=new AdapterTo_ReadARowResult(table,rowkey, rowData);
		GlobalPara.ltmComp.sendMsg(resultmsg);
	}
	
	
	public void restartSTG(long timestamp,String key,long newTimestamp)
	{
		RWGTPtoLTM_DoRestart restartmsg=new RWGTPtoLTM_DoRestart(timestamp,key,newTimestamp);
		GlobalPara.ltmComp.sendMsg(restartmsg);
	}
	
	public void doAbortSTG(String key, long timestamp)
	{
		RWGTPtoLTM_AggreementMsg abortmsg=new RWGTPtoLTM_AggreementMsg(RWGTPtoLTM_AggreementMsg.DOABORT,key, timestamp);
		GlobalPara.ltmComp.sendMsg(abortmsg);
	}
	
	public void doCommitSTG(String key)
	{
		RWGTPtoLTM_AggreementMsg commitmsg=new RWGTPtoLTM_AggreementMsg(RWGTPtoLTM_AggreementMsg.DOCOMMIT,key);
		GlobalPara.ltmComp.sendMsg(commitmsg);
	}
}
