package org.zhouw.stps.components;

import java.util.Vector;
import java.util.concurrent.ArrayBlockingQueue;

import org.zhouw.stps.membership.MembershipManager;
import org.zhouw.stps.GlobalPara;


public class MembershipComponent extends Component {

	public MembershipComponent() {
		super("Membership");
	}

	@Override
	public void handler(CompMessage msg) {
		switch(msg.type){
		case CompMessage.START_DEBUG:
			GlobalPara.debugQueue.offer("Member");
			break;
		case ToMembership_ReportFailureMsg.typeid:
			ToMembership_ReportFailureMsg reportmsg=(ToMembership_ReportFailureMsg)msg;
			//GlobalPara.logger.warning("recoverFailure started: "+reportmsg.url+" "+ reportmsg.IP + " "+ reportmsg.failureTimes);
			MembershipManager.recoverFailure(reportmsg.url,reportmsg.IP, reportmsg.failureTimes);
			break;
		case ToMembership_PartitionMsg.RECEIVE_INVI_REQUEST:
			ToMembership_PartitionMsg invitmsg=(ToMembership_PartitionMsg)msg;
			int result=MembershipManager.receiveInvitation(invitmsg.srcID, invitmsg.curPartition);
			if(!invitmsg.outputQueue.offer(result))
				GlobalPara.logger.warning("[error]Failed to return result for invitation request");
			break;
		case ToMembership_PartitionMsg.RECEIVE_INVI_RESPONSE:
			ToMembership_PartitionMsg partitionMsg=(ToMembership_PartitionMsg)msg;
			MembershipManager.handleInvitationResponses(partitionMsg.agreed, partitionMsg.rejected,partitionMsg.aborted);
			break;
		case ToMembership_PartitionMsg.RECEIVE_MERGE_REQUEST:
			ToMembership_PartitionMsg mergemsg=(ToMembership_PartitionMsg)msg;
			int mergeResult=MembershipManager.receiveMergeRequest(mergemsg.srcID, mergemsg.curPartition);
			if(!mergemsg.outputQueue.offer(mergeResult))
				GlobalPara.logger.warning("[error]Failed to return result for invitation request");
			break;
		case ToMembership_PartitionMsg.RECEIVE_MERGE_RESPONSE:
			ToMembership_PartitionMsg mergeResponseMsg=(ToMembership_PartitionMsg)msg;
			MembershipManager.commitPartitionMembershipAfterMerge(mergeResponseMsg.agreed);
			break;
		case ToMembership_PartitionMsg.RECEIVE_COMMIT:
			ToMembership_PartitionMsg commitMsg=(ToMembership_PartitionMsg)msg;
			MembershipManager.receivePartitionMembershipCommit(commitMsg.srcID, commitMsg.curPartition, commitMsg.isKeepLock);
			break;
		case ToMembership_PartitionMsg.START_INVI_TASK:
			ToMembership_PartitionMsg ppmsg=(ToMembership_PartitionMsg)msg;
			if(ppmsg.failureTimes==MembershipManager.failureTimes)
				MembershipManager.beginInvitationTask();
			else
				GlobalPara.logger.warning("[Info]Aborting one outdated start invitation task");
			break;
		case ToMembership_PartitionMsg.RECEIVE_ABORT_MT:
			MembershipManager.receiveAbortMT(0);
			break;
		case ToMember_RecoveryMsg.LOCAL_COMMIT_COMPLETE:
			MembershipManager.reportCommitUpdatesComplete();
			break;
		case ToMember_RecoveryMsg.LOCAL_DATA_COMPLETE:
			MembershipManager.reportDataComplete();
			break;
		case ToMember_RecoveryMsg.LOCAL_ROONLY_COMPLETE:
			MembershipManager.reportROComplete();
			break;
		case ToMember_RecoveryMsg.LOCAL_RWTRAN_COMPLETE:
			MembershipManager.reportRWComplete();
			break;
		case ToMember_RecoveryMsg.REMOTE_START_UPDATECOMMIT:
			MembershipManager.receiveStartCommitRecoveredUpdates();
			break;
		case ToMember_RecoveryMsg.REMOTE_START_ACCEPT_TRANS:
			MembershipManager.receiveStartAcceptingTransactions();
			break;
		case ToMember_RecoveryMsg.REMOTE_TRANRECOVER_COMPLETE:
			ToMember_RecoveryMsg trmsg=(ToMember_RecoveryMsg)msg;
			MembershipManager.receiveTranRecoveryFinishToCoordinator(trmsg.srcID);
			break;
		case ToMember_RecoveryMsg.REMOTE_UPDATECOMMIT_COMPLETE:
			ToMember_RecoveryMsg tucmsg=(ToMember_RecoveryMsg)msg;
			MembershipManager.receiveCommitRecoveredUpdatesCompleteToCoordinator(tucmsg.srcID);
			break;
		}
	}
	
	public void reportROGTMRecoverComplete(){
		ToMember_RecoveryMsg msg=new ToMember_RecoveryMsg(ToMember_RecoveryMsg.LOCAL_ROONLY_COMPLETE);
		this.sendMsg(msg);
	}
	
	public void reportRWGTMRecoverComplete(){
		ToMember_RecoveryMsg msg=new ToMember_RecoveryMsg(ToMember_RecoveryMsg.LOCAL_RWTRAN_COMPLETE);
		this.sendMsg(msg);
	}
	
	public void reportDataReorganizationComplete(){
		ToMember_RecoveryMsg msg=new ToMember_RecoveryMsg(ToMember_RecoveryMsg.LOCAL_DATA_COMPLETE);
		this.sendMsg(msg);
	}
	
	public void reportRecoveredUpdatesCommitComplete(){
		ToMember_RecoveryMsg msg=new ToMember_RecoveryMsg(ToMember_RecoveryMsg.LOCAL_COMMIT_COMPLETE);
		this.sendMsg(msg);
	}
	
	
	public void reportRemoteTranRecoveryComplete(int srcID){
		ToMember_RecoveryMsg msg=new ToMember_RecoveryMsg(ToMember_RecoveryMsg.REMOTE_TRANRECOVER_COMPLETE,srcID);
		this.sendMsg(msg);
	}
	
	public void reportRemoteUpdateCommitComplete(int srcID){
		ToMember_RecoveryMsg msg=new ToMember_RecoveryMsg(ToMember_RecoveryMsg.REMOTE_UPDATECOMMIT_COMPLETE, srcID);
		this.sendMsg(msg);
	}
	
	public void reportRemoteStartAcceptTrans(){
		ToMember_RecoveryMsg msg=new ToMember_RecoveryMsg(ToMember_RecoveryMsg.REMOTE_START_ACCEPT_TRANS);
		this.sendMsg(msg);
	}
	
	public void reportRemoteStartCommitUpdates(){
		ToMember_RecoveryMsg msg=new ToMember_RecoveryMsg(ToMember_RecoveryMsg.REMOTE_START_UPDATECOMMIT);
		this.sendMsg(msg);
	}
	
	
	public void reportReceiveMergeResponse(Vector<Integer> agreed)
	{
		ToMembership_PartitionMsg msg=new ToMembership_PartitionMsg(agreed);
		this.sendMsg(msg);
	}
	
	public void reportReceiveAbortMT()
	{
		ToMembership_PartitionMsg msg=new ToMembership_PartitionMsg(ToMembership_PartitionMsg.RECEIVE_ABORT_MT);
		this.sendMsg(msg);
	}
	
	public void reportFailure(String url, String IP)
	{
		//GlobalPara.logger.warning("Report failure reported: "+url);
		ToMembership_ReportFailureMsg msg=new ToMembership_ReportFailureMsg(url, IP, MembershipManager.failureTimes);
		this.sendMsg(msg);
	}
	
	public void startInvitation(int failureTimes)
	{
		ToMembership_PartitionMsg msg=new ToMembership_PartitionMsg(ToMembership_PartitionMsg.START_INVI_TASK,failureTimes);
		this.sendMsg(msg);
	}
	
	public void receiveAllInvitationResponses(Vector<Integer> agreed, Vector<Integer> rejected, Vector<Integer> aborted)
	{
		ToMembership_PartitionMsg msg=new ToMembership_PartitionMsg(agreed,rejected,aborted);
		this.sendMsg(msg);
	}
	
	/**
	 * 
	 * @param srcID : coordinator TPNode ID.
	 * @param curPartition : return the list of nodes of current partition
	 * @param outputQueue : return the result of handling invitation
	 */
	public void reportReceiveInvitation(int srcID, Vector<Integer> curPartition, 
			ArrayBlockingQueue<Integer> outputQueue)
	{
		ToMembership_PartitionMsg msg=new ToMembership_PartitionMsg(
				ToMembership_PartitionMsg.RECEIVE_INVI_REQUEST,
				srcID, curPartition, outputQueue);
		this.sendMsg(msg);
	}
	
	/**
	 * 
	 * @param srcID : coordinator TPNode ID.
	 * @param curPartition : return the list of nodes of current partition
	 * @param outputQueue : return the result of handling invitation
	 */
	public void reportReceiveMergeRequest(int srcID, Vector<Integer> curPartition, 
			ArrayBlockingQueue<Integer> outputQueue)
	{
		ToMembership_PartitionMsg msg=new ToMembership_PartitionMsg(
				ToMembership_PartitionMsg.RECEIVE_MERGE_REQUEST,
				srcID, curPartition, outputQueue);
		this.sendMsg(msg);
	}
	
	public void reportReceivePartitionCommit(int srcID, Vector<Integer> curPartition, boolean isKeepLock){
		ToMembership_PartitionMsg msg=new ToMembership_PartitionMsg(		
				srcID, curPartition, isKeepLock);
		this.sendMsg(msg);
	}
}
