package org.zhouw.stps.components;

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.MsgBroker;
import org.zhouw.stps.rotran.JoiningTable;
import org.zhouw.stps.rotran.QueryPlan;
import org.zhouw.stps.rotran.ROGlobalTM;
import org.zhouw.stps.rotran.ResultOfTable;
import org.zhouw.stps.rotran.ROTranMsg.ROTranResultMsg;
import org.zhouw.stps.rotran.ROTranMsg.ROTransactionMsg;
import org.zhouw.stps.rotran.ROTranMsg.VoteROTranMsg;
import org.zhouw.stps.tp.GlobalTransManager;
import org.zhouw.stps.tp.TPNode;

public class ROGTMComponent extends Component {

	protected long compid;
	
	public ROGTMComponent()
	{
		super("RO GTM");
		compid=0;
	}
	
	@Override
	public void handler(CompMessage msg) {
		switch(msg.type){
		case CompMessage.START_DEBUG:
			ROGlobalTM.printDebugMessage();
			GlobalPara.debugQueue.offer("RO");
			break;
		case ClientToROGTM_TranMsg.typeid:
			ClientToROGTM_TranMsg tranmsg=(ClientToROGTM_TranMsg)msg;
			QueryPlan qp=tranmsg.qp;
			
			if(qp!=null&&qp.isValid){
				ROGlobalTM.processROTran(compid++, qp, tranmsg.clientNodeID, tranmsg.clientsideTID);
			} else {
				ROTranResultMsg.Builder returnMsg=ROTranResultMsg.newBuilder()
				.setClientsideID(tranmsg.clientsideTID);
				returnMsg.setTranState(3);
				if(qp!=null&&qp.errorMsg!=null)
					returnMsg.setErrMsg(qp.errorMsg);
				MsgBroker.returnROTran(tranmsg.clientNodeID, returnMsg.build());
			}
			break;
		case Timestamp_idMsg.typeid:
			Timestamp_idMsg idmsg=(Timestamp_idMsg)msg;
			if(!idmsg.isRestart)
				ROGlobalTM.beginExecutingROTran(idmsg.msgid, idmsg.timestamp);
			else
				ROGlobalTM.handleRestart(idmsg.msgid, idmsg.timestamp);
			break;
		case LTMtoROGTM_VoteMsg.typeid:
			LTMtoROGTM_VoteMsg votemsg=(LTMtoROGTM_VoteMsg)msg;
			ROGlobalTM.handleVoteMsg(votemsg.msg);
			break;
		case MemberToROGTM_FailureRecoveryMsg.typeid:
			ROGlobalTM.abortAllOnGoingROTrans();
			GlobalPara.logger.warning("[ROTRANRecover] RO tran recover compelete");
			GlobalPara.memberComp.reportROGTMRecoverComplete();
			break;
		}
	}
	
	public void reportFailureRecovery()
	{
		MemberToROGTM_FailureRecoveryMsg msg=new MemberToROGTM_FailureRecoveryMsg();
		GlobalPara.roComp.sendMsg(msg);
	}
	
	public void submitVoteMsg(VoteROTranMsg msg)
	{
		LTMtoROGTM_VoteMsg votemsg=new LTMtoROGTM_VoteMsg(msg);
		GlobalPara.roComp.sendMsg(votemsg);
	}
	
	public void clientSubmitROTran(ROTransactionMsg msg)
	{
		//GlobalPara.logger.warning("Get RO Transaction with ClientsideTID: " + msg.getClientsideTID());
		
		if(TPNode.inRecovery){
			ROGlobalTM.abortROTranForRecovery(msg.getClientNodeID(), msg.getClientsideTID());
			return;
		}
			
		QueryPlan qp;
		if(msg.hasPlan()){
			//GlobalPara.logger.warning("The RO Transaction "+msg.getClientsideTID()+" has a query plan with out query");
			qp=new  QueryPlan(msg.getPlan());
			qp.isValid=true;
			for(JoiningTable t:qp.tables.values()){
				t.result=new ResultOfTable();
			}
			qp.rootTable.result.candidateRows.addAll(qp.rootTable.allowedRows);
		}
		else{
			//GlobalPara.logger.warning("The RO Transaction "+msg.getClientsideTID()+" has a query: " + msg.getQuery());
			qp = ROGlobalTM.planROTransaction(msg.getQuery(), msg.getDbName());
		}

		//GlobalPara.logger.warning("Start executing one RO Transaction ClientsideTID: " + msg.getClientsideTID());
		ClientToROGTM_TranMsg tranmsg=new ClientToROGTM_TranMsg(qp, msg.getClientNodeID(), msg.getClientsideTID());
		GlobalPara.roComp.sendMsg(tranmsg);
	}
	
}
