package org.zhouw.stps.master;

import java.util.Hashtable;
import java.util.Random;
import java.util.TreeSet;
import java.util.Vector;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.zhouw.stps.*;
import org.zhouw.stps.conf.NodeConfig.role;
import org.zhouw.stps.tp.*;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Coordinate the TPNodes to finish the recovery process
 * @author zhouw
 *
 */
public class RecoveryManager {

	TreeSet<Integer> TPNodeSet;
	
	public enum STATUS{NOMRAL, TRANRECOVERY, TOTALRECOVERY}
	private ReentrantLock lock;;
	
	private RecoveryManager.STATUS state;
	public void startMonitoring()
	{
		ExecutorService se=Executors.newSingleThreadExecutor();
		long thinkTime=new Long(GlobalPara.sysConf.MonitorPeriod*1000);
		MonitoringTask task=new MonitoringTask(thinkTime, this.TPNodeSet);
    	se.submit(task);
	}
	
	public RecoveryManager(TreeSet<Integer> TPNodeSet)
	{
		tranRecoveryComplete=new Hashtable<Integer, Boolean>();
		totalRecoveryComplete=new Hashtable<Integer, Boolean>();
//		dataReplacementComplete=new Hashtable<Integer, Boolean>();
//		failedBackupRecoveryComplete=new Hashtable<Integer, Boolean>();
		for(Integer nodeID:TPNodeSet)
			{
				tranRecoveryComplete.put(nodeID, false);
				totalRecoveryComplete.put(nodeID, false);	
//				dataReplacementComplete.put(nodeID, false);
//				failedBackupRecoveryComplete.put(nodeID, false);
			}
		this.TPNodeSet=TPNodeSet;
		duration=new Vector<Long>();
		this.failedNodes=new TreeSet<Integer>();
		this.state=STATUS.NOMRAL;
		this.lock=new ReentrantLock();
		this.failedNodesDuringRecovery=new Vector<Integer>();
	}
	
	private Hashtable<Integer, Boolean> tranRecoveryComplete;
	private Hashtable<Integer, Boolean> totalRecoveryComplete;
//	private Hashtable<Integer, Boolean> dataReplacementComplete;
//	private Hashtable<Integer, Boolean> failedBackupRecoveryComplete;
	
	private long startTime;
	private long tranFinishTime;
	private long endTime;
	
	private TreeSet<Integer> failedNodes;
	private Vector<Integer> failedNodesDuringRecovery;//failed nodes during recovery process
	private Vector<Long> duration;

	/**
	 * update the TPNode list
	 * Send the message to all the still alive TPNodes: TPNode.reportFailedNode
	 * @param failedVirtualNodes
	 */
	public void detectNodeFailures(int[] newFailedNodes)
	{
		if(newFailedNodes==null||newFailedNodes.length==0)
		{
			GlobalPara.logger.warning("Master receive failed node report with no failed nodes");
			return;
		}
		this.lock.lock();
		try{
			for(int node:newFailedNodes)
				this.failedNodes.add(node);
			
			String msg="Node Failure report received, failed nodes are";
			for(int faileNode: newFailedNodes)
			{
				msg+=" "+faileNode;
			}
			msg+="\nCurrent TP node Set: ";
			for(int tpNode: this.TPNodeSet)
			{
				msg+=" "+tpNode;
			}
			GlobalPara.logger.warning(msg);
			
			
			for(int faileNode: newFailedNodes)
			{
				TPNodeSet.remove(faileNode);
				tranRecoveryComplete.remove(faileNode);
				totalRecoveryComplete.remove(faileNode);
	//			dataReplacementComplete.remove(faileNode);
	//			failedBackupRecoveryComplete.remove(faileNode);
			}
			
			msg="After deleting falied nodes from TP node set: ";
			for(int tpNode: this.TPNodeSet)
			{
				msg+=" "+tpNode;
			}
			GlobalPara.logger.warning(msg);
			
			if(this.state==STATUS.NOMRAL)
				{
					startTime=System.currentTimeMillis();
					this.state=STATUS.TRANRECOVERY;
					for(Integer nodeID:this.TPNodeSet)
					{
						ReportNodeFailureTask task=new ReportNodeFailureTask(nodeID,newFailedNodes);
						MsgBroker.executor.submit(task);
					}
				}
			else if(this.state==STATUS.TRANRECOVERY)
				{
					for(int faileNode: newFailedNodes)
						{
							this.failedNodesDuringRecovery.add(faileNode);
						}
					this.reportTranRecoveryCompleteInside(newFailedNodes[0]);//to prevent recovery blocked by concurrent failure event

				}
			else if(this.state==STATUS.TOTALRECOVERY)
				{
					for(int faileNode: newFailedNodes)
						{
							this.failedNodesDuringRecovery.add(faileNode);
						}
					this.reportTotalRecoveryCompleteInside(newFailedNodes[0]);
				}
			
			//////////////////
			// report to clients, so clients can stop sending messages to failed nodes immediately
			for(Integer nodeID:GlobalPara.sysConf.clientNodeSet)
			{
				ReportToClientNodeFailureTask task=new ReportToClientNodeFailureTask(nodeID,newFailedNodes);
				MsgBroker.executor.submit(task);
			}
		}
		finally{
			this.lock.unlock();
		}
	}
	
	private int startHandlingFailedNodesDuringRecovery()
	{
		int result=0;
		startTime=System.currentTimeMillis();
		if(this.failedNodesDuringRecovery.size()>0)
			{
				String msg="start Handling FailedNodes During Recovery: ";
				
				this.failedNodes.addAll(this.failedNodesDuringRecovery);
				this.failedNodesDuringRecovery.clear();
				
				int count=0;
				int[] newFailedNodes=new int[this.failedNodes.size()];
				for(Integer node:this.failedNodes)
					{
						newFailedNodes[count++]=node;
						msg+=" "+node;
					}
				GlobalPara.logger.warning(msg);
				
				for(Integer nodeID:this.TPNodeSet)//report the client only when in 
					{
						ReportNodeFailureTask task=new ReportNodeFailureTask(nodeID,newFailedNodes);
						MsgBroker.executor.submit(task);
					}
				this.state=STATUS.TRANRECOVERY;
			}
		return result;
	}
	
	private int sendAllTranRecoveryFinishMsg()
	{
		int result=0;
		for(int node:this.TPNodeSet)
			{
				ReportAllTranRecoveryFinishTask task=new ReportAllTranRecoveryFinishTask(node,this.failedNodes);
				MsgBroker.executor.submit(task);
			}
		return result;
	}
	
	/**
	 * Wait until all tasks returns
	 * means the total recovery process has actually finished
	 * @return
	 */
	private int sendBackToNormalMsg()
	{
		int result=0;
		Future<Integer>[] taskResults=new Future[this.TPNodeSet.size()]; 
		int count=0;
		for(int node:this.TPNodeSet)
			{
			ReportBackToNormalTask task=new ReportBackToNormalTask(node);
				taskResults[count++]=MsgBroker.executor.submit(task);
			}
		for(Future<Integer> t:taskResults)
			{
				try {
					t.get();
				} catch (InterruptedException e) {
					e.printStackTrace();
				} catch (ExecutionException e) {
					e.printStackTrace();
				}
			}
		return result;
	}
	
	
	
	/**
	 * 
	 * @param tpNodeID the reporter's TP node id
	 * @return 0 if successful, -1 if the reporter is not in the tp node list
	 */
	public int reportTranRecoveryComplete(int tpNodeID)
	{
		lock.lock();
		try{
			return reportTranRecoveryCompleteInside(tpNodeID);
		}
		finally{
			lock.unlock();
		}
		
	}
	
	/**
	 * 
	 * @param tpNodeID the reporter's TP node id
	 * @return 0 if successful, -1 if the reporter is not in the tp node list
	 */
	private int reportTranRecoveryCompleteInside(int tpNodeID)
	{
		int result=0;
		GlobalPara.logger.warning("Node  "+tpNodeID+" report that its ongoing transaction recovery has completed");
		if(this.TPNodeSet.contains(tpNodeID))
			{
				this.tranRecoveryComplete.put(tpNodeID, true);
			}
		else
			{
				result=-1;
			}
		
		if(this.isTranRecoveryFinish())
			{
				GlobalPara.logger.warning("All nodes have report that its ongoing transaction recovery has completed");
				sendAllTranRecoveryFinishMsg();
				this.state=STATUS.TOTALRECOVERY;
				result=0;
				tranFinishTime=System.currentTimeMillis();
			}
		return result;
	}
	
	public int reportTotalRecoveryComplete(int tpNodeID)
	{
		lock.lock();
		try{
			return reportTotalRecoveryCompleteInside(tpNodeID);
		}
		finally{
			lock.unlock();
		}

	}
	
	private int reportTotalRecoveryCompleteInside(int tpNodeID)
	{
		int result=0;
		GlobalPara.logger.warning("Node  "+tpNodeID+" report that its total recovery has successfully completed");
		if(this.TPNodeSet.contains(tpNodeID))
			{
				this.totalRecoveryComplete.put(tpNodeID, true);
			}
		else
			{
				result=-1;
			}
		
		if(this.isAllCompleted())//record the time spended
			{
				GlobalPara.logger.warning("All nodes have report that its total recovery has completed");
				sendBackToNormalMsg();
				this.failedNodes.clear();
				initDataStructure();
				endTime=System.currentTimeMillis();
				duration.add(this.tranFinishTime-startTime);
				duration.add(endTime-startTime);
				GlobalPara.logger.warning("The "+duration.size()+"th failure recovery succeeded:\n"+ this.printDuration());
				this.state=STATUS.NOMRAL;
				result=0;
				startHandlingFailedNodesDuringRecovery();
			}
		
		return result;
	}
	
	private void initDataStructure()
	{
		for(Integer nodeID:TPNodeSet)
		{
			tranRecoveryComplete.put(nodeID, false);
			totalRecoveryComplete.put(nodeID, false);	
		}
	}
	
	private String printDuration()
	{
		String result="";
		for(int i=0;i<duration.size();i++)
			{
				result+=" "+duration.get(i);
				if(i%2!=0)
					result+="\n";
			}
		
		return result;
	}
	
	/**
	 * all alive tp nodes finish all phases of recovery
	 * @return
	 */
	private boolean isTranRecoveryFinish()
	{
		boolean result=true;
		for(Integer nodeID:TPNodeSet)
			{
				Boolean tmp=tranRecoveryComplete.get(nodeID);
				if(tmp==null)
					{
						GlobalPara.logger.warning("The nodes in TPNodeSet is not found in tranRecoveryComplete table");
						continue;
					}
				if(!tmp)
					{
						result=false;
						break;
					}
			}
		return result;
	}
	
	/**
	 * all alive tp nodes finish all phases of recovery
	 * @return
	 */
	private boolean isAllCompleted()
	{
		boolean result=true;
		for(Integer nodeID:TPNodeSet)
			{
//				if(!tranRecoveryComplete.get(nodeID))
//					{
//						result=false;
//						break;
//					}
			Boolean tmp=totalRecoveryComplete.get(nodeID);
			if(tmp==null)
				{
					GlobalPara.logger.warning("The nodes in TPNodeSet is not found in totalRecoveryComplete table");
					continue;
				}
				if(!tmp)
					{
						result=false;
						break;
					}
			}
		return result;
	}

	/**
	 * @return the duration
	 */
	public Vector<Long> getDuration() {
		return duration;
	}
}
