package org.zhouw.stps.membership;

import java.util.concurrent.Semaphore; 

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.MsgBroker;
import org.zhouw.stps.SubTransaction;
import org.zhouw.stps.Transaction;
import java.util.concurrent.TimeUnit;

import java.io.IOException;
import java.io.InputStream;
import java.util.Hashtable;
import java.util.List;
import java.util.Random;
import java.util.TreeSet;
import java.util.Vector;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.locks.ReentrantLock;

import org.zhouw.stps.tp.*;
import org.zhouw.stps.*;
import org.zhouw.stps.client.GetLatestMembershipTask;
import org.zhouw.stps.client.TransExecutor;
import org.zhouw.stps.membership.msg.MembershipMsg.LastestMembershipMsg;
import org.zhouw.stps.membership.msg.MembershipMsg.SendAbortMTMsg;
import org.zhouw.stps.membership.msg.MembershipMsg.SendDataReorganizationFinishToCoordinatorMsg;
import org.zhouw.stps.membership.msg.MembershipMsg.SendOnGoingTranRecoveryFinishToCoordinatorMsg;
import org.zhouw.stps.membership.msg.MembershipMsg.SendStartAcceptingTransactionsMsg;
import org.zhouw.stps.membership.msg.MembershipMsg.SendStartDataReorganizationMsg;
import org.zhouw.stps.msg.SubTransactionGroupProtos.TransactionResultSetMsg;

public class MembershipManager {

	private static volatile int versionNumber=0;
	private static ReentrantLock versionLock=new ReentrantLock();
	public static volatile int failureTimes=0;
	
	/**
	 * The latest membership can only be obtained by clients only when the MT is completed
	 */
//	public static Semaphore semaphoreOnVersion=new Semaphore(1); 
	
	
	protected static boolean ro_complete=false;
	protected static boolean rw_complete=false;
	protected static boolean data_complete=false;
	
	protected static boolean isCoordinator=false;
	protected static RingMembership newRing;
//	protected static Vector<Integer> failednodesVector;
	protected static Vector<Integer> compeletedNodes;
	
	public static void reportROComplete()
	{
		ro_complete=true;
		GlobalPara.logger.warning("[MemberRecover]RO Tran Recovered Complete");
		reportTranRecoverComplete();
	}
	
	public static void reportRWComplete()
	{
		rw_complete=true;
		GlobalPara.logger.warning("[MemberRecover]RW Tran Recovered Complete");
		reportTranRecoverComplete();
	}
	
	public static void reportDataComplete()
	{
		data_complete=true;
		GlobalPara.logger.warning("[MemberRecover]Data Recovered Complete");
		reportTranRecoverComplete();
	}
	
	protected static void reportTranRecoverComplete()
	{
		if(ro_complete&&rw_complete&&data_complete){
			if(MembershipManager.isCoordinator){
				checkIfStartCommitUpdates(GlobalPara.sysConf.localNodeID);
			} else {
				sendOnGoingTranRecoveryFinishToCoordinator();
			}
		}
	}
	
	protected static void checkIfStartCommitUpdates(int completedNodeID)
	{
		
		compeletedNodes.add(completedNodeID);
		GlobalPara.logger.warning("Node "+completedNodeID+ "Finish the phase, there are "+compeletedNodes.size()+"nodes completed, total "+newRing.TPNodeSet.size()+"required.");
		if(compeletedNodes.size()==newRing.TPNodeSet.size()){
			GlobalPara.logger.warning("Finish the phase of Recovering transactions, start commit recovered updates");
			compeletedNodes.clear();
			MembershipManager.updateMembership(newRing);
			sendStartCommitRecoveredUpdates();
			//TPNode.updateSystemMembership(newRing);
			GlobalPara.ltmComp.startCommitRecoveredUpdates();
		}
		
	}
	
	public static int sendOnGoingTranRecoveryFinishToCoordinator()
	{
		if(currentCoordinatorID==-1)
		{
			GlobalPara.logger.warning("current coordintor of MT is -1");
			return 0;
		}
		
		SendOnGoingTranRecoveryFinishToCoordinatorMsg msg=
			SendOnGoingTranRecoveryFinishToCoordinatorMsg.newBuilder()
			.setVersion(MembershipManager.getVersionNumber())
			.setSrcID(GlobalPara.sysConf.localNodeID).build();
		
		PostMessageTask task=new PostMessageTask(currentCoordinatorID, msg.toByteArray(), "ReceiveOnGoingTranRecoveryFinishToCoordinatorService");
		executor.submit(task);
		return 0;
	}
	
	public static int receiveTranRecoveryFinishToCoordinator(int srcID)
	{
		GlobalPara.logger.warning("Receive OnGoingTranRecoveryFinishToCoordinator from " + srcID);
		checkIfStartCommitUpdates(srcID);
		return 0;
	}
	
	public static int sendStartCommitRecoveredUpdates()
	{
		for(int destId:partition){
			if(destId!=GlobalPara.sysConf.localNodeID){
				SendStartDataReorganizationMsg msg=SendStartDataReorganizationMsg.newBuilder().setVersion(MembershipManager.getVersionNumber()-1).build();
				PostMessageTask task=new PostMessageTask(destId, msg.toByteArray(), "ReceiveStartDataReorganizationService");
				MembershipManager.executor.submit(task);
			}
		}
		return 0;
	}
	
	/*
	 * Commit Recovered updates is a local operation, 
	 * increase the version number will not trigger accidental exits.
	 */
	public static int receiveStartCommitRecoveredUpdates()
	{
		GlobalPara.logger.warning("Receive Commit Recovered Updates");
		MembershipManager.updateMembership(newRing);
		GlobalPara.ltmComp.startCommitRecoveredUpdates();
		return 0;
	}
	
	public static void reportCommitUpdatesComplete()
	{
		if(MembershipManager.isCoordinator){
			checkIfStartAcceptingTrans(GlobalPara.sysConf.localNodeID);
		} else {
			sendCommitRecoveredUpdatesCompleteToCoordinator();
		}
	}
	
	/**
	 * Note that this function should use a version number = current -1
	 * @return
	 */
	public static int sendCommitRecoveredUpdatesCompleteToCoordinator()
	{
		if(currentCoordinatorID==-1)
		{
			GlobalPara.logger.warning("current coordintor of MT is -1");
			return  0;
		}
		
		SendDataReorganizationFinishToCoordinatorMsg msg=
			SendDataReorganizationFinishToCoordinatorMsg.newBuilder()
			.setVersion(MembershipManager.getVersionNumber())
			.setSrcID(GlobalPara.sysConf.localNodeID).build();
		
		PostMessageTask task=new PostMessageTask(currentCoordinatorID, msg.toByteArray(), "ReceiveCommitRecoveredUpdatesCompleteToCoordinatorService");
		executor.submit(task);

		return 0;
	}
	
	protected static void checkIfStartAcceptingTrans(int completedNodeID)
	{
		compeletedNodes.add(completedNodeID);
		if(compeletedNodes.size()==newRing.TPNodeSet.size()){
			GlobalPara.logger.warning("Finish the phase of Recovering transactions, start commit recovered updates");
			compeletedNodes.clear();
			sendStartAcceptingTransactions();
			endRecoveryProcess();
			GlobalPara.rwComp.startAcceptTransactions();
			Concurrency.releaseMTLock();
		}
		
		
	}
	
	public static int receiveCommitRecoveredUpdatesCompleteToCoordinator(int srcID)
	{
		checkIfStartAcceptingTrans(srcID);
		return 0;
	}

	public static int sendStartAcceptingTransactions()
	{
		for(int destId:partition){
			if(destId!=GlobalPara.sysConf.localNodeID){
				SendStartAcceptingTransactionsMsg msg=
					SendStartAcceptingTransactionsMsg.newBuilder()
					.setVersion(MembershipManager.getVersionNumber()).build();
				
				PostMessageTask task=new PostMessageTask(destId, msg.toByteArray(), "ReceiveStartAcceptingTransactionsService");
				executor.submit(task);
			}
		}
		return 0;
	}
	
	public static int receiveStartAcceptingTransactions()
	{
		GlobalPara.logger.warning("Receive Start Accepting Transactions");
		endRecoveryProcess();
		GlobalPara.rwComp.startAcceptTransactions();
		Concurrency.releaseMTLock();
		return 0;
	}
	
	
	/**
	 * For messages by   
	 * @return the versionNumber
	 */
	public static int getVersionNumberWithoutLock() {
		return versionNumber;
	}
	
	public static LastestMembershipMsg getLatestMembership()
	{
		LastestMembershipMsg msg=null;
		versionLock.lock();
		try {
			msg=LastestMembershipMsg.newBuilder().setVersion(versionNumber)
			.addAllTPNode(TPNode.ring.TPNodeSet).build();
		} finally{
			versionLock.unlock();
		}
		return msg;
	}
	
	public static void updateSystemMembershipForClient(LastestMembershipMsg msg)
	{
//		if(TransExecutor.tranTable.size()>0)
//		{
//			GlobalPara.logger.warning("[ClientMembershipUpdate]Abort current ongoing client-side transactions complete to update client's membership, still have  " +TransExecutor.tranTable.size());
//			TransactionResultSetMsg abortmsg=TransactionResultSetMsg.newBuilder().setTranState("ABORT").build();
//			for(LinkedBlockingQueue<TransactionResultSetMsg> queue: TransExecutor.tranTable.values())
//			{
//				try {
//					queue.offer(abortmsg, 100, TimeUnit.MILLISECONDS);
//				} catch (InterruptedException e) {
//					e.printStackTrace();
//				}
//			}
//		}
		
		versionLock.lock();
		try{
			GlobalPara.logger.warning("[ClientMembershipUpdate]Client now udpate the system membership and its version number into ("+ msg.getVersion()+")");
			
			GlobalPara.sysConf.TPNodeSet.clear();
			GlobalPara.sysConf.TPNodeSet.addAll(msg.getTPNodeList());
			RingMembership newRing=new RingMembership(GlobalPara.sysConf.TPNodeSet,
					GlobalPara.sysConf.localNodeID,
					GlobalPara.numOfVirtualNodes, 
					GlobalPara.numOfReplicas);
			TPNode.ring=null;
			TPNode.ring=newRing;
			versionNumber=msg.getVersion();
			TransExecutor.isMembershipUptodate=true;
		
		} finally{
			versionLock.unlock();
		}
		GlobalPara.logger.warning("[ClientMembershipUpdate]Clients Back to issue transactions");
	}
		
	public static void handleOutdatedMembershipForClient(int destID, int outVersion) 
	{
		versionLock.lock();
		try{
			if(outVersion>versionNumber&&TransExecutor.isMembershipUptodate)
			{
				GlobalPara.logger.warning("Clients Receiving membership version ("+ outVersion 
						+") larger than its own ("+ versionNumber +"), obtain the latest membership");
				TransExecutor.isMembershipUptodate=false;
				GetLatestMembershipTask task=new GetLatestMembershipTask(destID);
				executor.submit(task);
			}
		}
		finally
		{
			versionLock.unlock();
		}
	}
	
	public static void handleOutdatedMembership() {
		GlobalPara.logger.warning("Receiving message whose membership version is larger than local LTM, terminate itself");
		GlobalPara.monitor.closeAll();
		System.exit(0);
	}
	/**
	 * for membership transactions 
	 * @return 0: equal, 1: ours is larger, -1 ours is smaller
	 */
	public static int compareVersionNumber(int otherVersion) {
		versionLock.lock();
		try{
			if(versionNumber==otherVersion)
				return 0;
			else if(versionNumber>otherVersion)
				return 1;
			else 
				return -1;
		}
		finally{
			versionLock.unlock();
		}
	}
	
	/**
	 * for membership transactions 
	 * @return the versionNumber
	 */
	public static int getVersionNumber() {
		versionLock.lock();
		try{
			return versionNumber;
		}
		finally{
			versionLock.unlock();
		}
	}

	/**
	 * @param versionNumber the versionNumber to set
	 */
	public static void updateMembership(RingMembership newRing) {
		versionLock.lock();
		try{
			MembershipManager.versionNumber++;
			GlobalPara.sysConf.TPNodeSet.clear();
			GlobalPara.sysConf.TPNodeSet.addAll(newRing.TPNodeSet);
			TPNode.ring=null;
			TPNode.ring=newRing;
			TPNode.backupNodes=TPNode.getBackupTPNodes(GlobalPara.sysConf.localNodeID);
			TPNode.backupVirtualNodes=newRing.getBackupNodesOfDataItems();
		}
		finally{
			versionLock.unlock();
		}
	}	
	
	public static boolean isInFailureRecovery=false;
	//public static boolean isSendInvitationTaskStarted=false;
	//private static ReentrantLock lock=new ReentrantLock();
	
	private static ReentrantLock lockOfOtherPartitions=new ReentrantLock();
	
	//Partitions in this hashtable reject the invitation and return their current partition membership,
	//Partitions in this hashtable have empty join set
	private static Vector<Vector<Integer>> otherPartitions=new Vector<Vector<Integer>>();

	public static void addRejectedPartition(List<Integer> input)
	{ 
		if(input==null)
			return;
		Vector<Integer> in=new Vector<Integer>();
		in.addAll(input);
		addRejectedPartition(in);
	}
	public static void addRejectedPartition(Vector<Integer> input)
	{ 
		if(input==null)
			return;
		lockOfOtherPartitions.lock();
		try{
			Vector<Vector<Integer>> toBeRemoved=new  Vector<Vector<Integer>>();
			for(Vector<Integer> p:otherPartitions){ //does not allow duplicated partitions or intersected partitions
				for(int n:input)
					if(p.contains(n)){//if there are conflicts, always keep the latest one
						toBeRemoved.add(p);
						break;
					}
			}
			if(toBeRemoved.size()>0) { //multiple LTMs from another partition respond with reject.
//				String output="[Error]Different partitions share same members.\n";
//				output+="The input rejected partition:";
//				for(int n:input)
//					output+=n+",";
//				output+="\n";
//				for(Vector<Integer> to:toBeRemoved){
//					output+="Conflicted partition: ";
//					for(int n:to)
//						output+=n+",";
//					output+="\n";
//				}
				
				//GlobalPara.logger.warning(output);
				otherPartitions.removeAll(toBeRemoved);
			}
			otherPartitions.add(input);
		}
		finally{
			lockOfOtherPartitions.unlock();
		}
	}
	
	//Current partition which this LTM in
	private static Vector<Integer> partition=new Vector<Integer>();
	
	public static ExecutorService executor= Executors.newCachedThreadPool();
	
	
	public static int currentCoordinatorID;
	
	public static void cleanTemporary()
	{
		partition.clear();
		otherPartitions.clear();
	}
	
	/**
	 * 
	 * @return true: successfully started, false:already started
	 */
	private  static boolean isInFailureRecovery()
	{
		return isInFailureRecovery;
	}
	
	
	/**
	 * 
	 *
	 */
	private  static boolean startFailureRecovery()
	{
		if(isInFailureRecovery)
			return false;
		else {
			startFailureTime=System.currentTimeMillis();
			TPNode.inRecovery=true;
			isInFailureRecovery=true;
			partition.clear();
			return true;
		}
	}
	
	/**
	 * 
	 * @return true: successfully ended, false:already ended
	 */
	private static boolean endFailureRecovery()
	{
		if(isInFailureRecovery){
			failureTimes++;
			endFailureTime=System.currentTimeMillis();
			isInFailureRecovery=false;
			TPNode.inRecovery=false;
			GlobalPara.logger.warning("Printing Mointor Message");
			printMonitoredTime();
			clearMonitoredTime();
			return true;
		} else
			return false;
	}
	
	/**
	 * The function will not block for the other LTMs to execute abortMT.
	 * Just after insert into the work queue of membership component.
	 * 
	 * The assumption is that further failure or partition not happened in the recovery process
	 * @param agreed
	 * @return
	 */
	public static int sendAbortForMT(Vector<Integer> agreed)
	{
		if(agreed==null||agreed.size()==0)
			return 0;
		Future<Integer>[] tasks=new Future[agreed.size()];
		int count=0;
		for(int destId:agreed){
			if(destId!=GlobalPara.sysConf.localNodeID){
				SendAbortMTMsg msg=SendAbortMTMsg.newBuilder().setVersion(MembershipManager.getVersionNumber()).build();
				PostMessageTask task=new PostMessageTask(destId,msg.toByteArray(),"ReceiveAbortMTService");
				tasks[count++]=MembershipManager.executor.submit(task);
			}
		}
		
//		for(Future<Integer> t: tasks) {
//			try {
//				t.get();
//			} catch (InterruptedException e) {
//				e.printStackTrace();
//			} catch (ExecutionException e) {
//				e.printStackTrace();
//			}
//		}
		return 0;
	}
	
	
//	private static int SendCommitForInvitations(boolean isKeepLock, Vector<Integer> newPartition)
//	{
//		Future<Integer>[] tasks=new Future[newPartition.size()];
//		int count=0;
//		for(int destId:newPartition)
//		{
//			SendPartitionMembershipCommitToOneLTMTask task=new SendPartitionMembershipCommitToOneLTMTask(isKeepLock, destId, newPartition);
//			tasks[count++]=MembershipManager.executor.submit(task);
//		}
//		
//		for(Future<Integer> t: tasks)
//		{
//			try {
//				t.get();
//			} catch (InterruptedException e) {
//				e.printStackTrace();
//			} catch (ExecutionException e) {
//				e.printStackTrace();
//			}
//		}
//		return 0;
//	}
	
	/**
	 * 
	 * @return 1: agreed to join the partition, 0: abort as currently in another MT, -1: reject as already in a partition
	 * its current Partition returned in curPartition
	 */
	public static int receiveInvitation(int srcID, Vector<Integer> curPartition)
	{
		GlobalPara.logger.warning("receive Invitation from " + srcID + ", Try to obtain the lock");
		boolean isStartInvitationThead = startFailureRecovery();//receiving invitations is a type of failure reports
		
		if(isStartInvitationThead){
			reportedFailedNodeID=-1;
			StartInvitationTask task=new StartInvitationTask(failureTimes, false);
			executor.submit(task); 
		}
		
		boolean lockObtained=Concurrency.obtainMTLock(srcID);
		if(!lockObtained)
			return 0;
		
		GlobalPara.logger.warning("receive Invitation from " + srcID + ", lock obtained");
		
		 if(partition.contains(srcID) || partition.size()==0 )
			 return 1;
		 else //if(partition.size()>0)
			{
				curPartition.clear();
				curPartition.addAll(MembershipManager.partition);
				//GlobalPara.logger.warning("Releasing MT Lock ");
				Concurrency.releaseMTLock();
				return -1;
			}
	}
	
	public static int receivePartitionMembershipCommit(int srcID, Vector<Integer> curPartition, boolean isKeepLock)
	{
		GlobalPara.logger.warning("Receive Partition Membership Commit");
		partition.clear();
		partition.addAll(curPartition);
		
		if(isKeepLock) {// if there are following updates on partition membership, merge partitions for example, then hold the lock
			GlobalPara.logger.warning("The MT coordinator asks to keep the lock, so should start the recovering system");
			if(isAbleToRecoverFromNetPartitions()) {//enough number of LTMs joined to recover the system
				//should always moving in to this path
				MembershipManager.startRecoverSystem(false);
			} else 
				GlobalPara.logger.warning("[error]The MT coordinator asks to keep the lock, but the system is NOT ready to recover.");
		} else {
			GlobalPara.logger.warning("Releasing MT Lock");
			Concurrency.releaseMTLock();
		}
		return 0;
	}
	
	public static int receiveAbortMT(int srcID)
	{
		GlobalPara.logger.warning("Releasing MT Lock ");
		Concurrency.releaseMTLock();
		return 0;
	}
	
	/**
	 * 
	 * @param srcID 
	 * @param ourPartition
	 * @return 0: abort because of concurrency control, -1: partition does not match, the latest contained in ourPartition,
	 * 1: commit for merge 
	 */
	public static int receiveMergeRequest(int srcID, Vector<Integer> ourPartition)
	{
		GlobalPara.logger.warning("Receive Merge Request from " +srcID + ": Waiting to obtain lock");
		boolean locked=Concurrency.obtainMTLock(srcID);
		if(!locked)
			return 0; 

		GlobalPara.logger.warning("Receive Merge Request from " +srcID + ": Started");
		 if(partition.containsAll(ourPartition)&&partition.size()==ourPartition.size())
			  return 1;//still hold the lock, waiting to be merged
		 else
			 { // the partition does not match current partition membership
				ourPartition.clear();
				ourPartition.addAll(partition);
				Concurrency.releaseMTLock();
				return -1;
			 }
	}
	

//	public static int receiveMergeAbort(int srcID, Vector<Integer> ourPartition)
//	{
//		Concurrency.releaseMTLock();
//		return 1;
//	}
	
//	/**
//	 * Send the partition membership to all the members in the membership
//	 * @param partition
//	 * @return
//	 */
//	private static int sendMergeRequest(Vector<Integer> partition)
//	{
//		return 0;
//	}
	
//	/**
//	 * determine the final set of LTMs that can be merged, and the final partition membership 
//	 * try to merge with all the partitions in the list
//	 * If aborted, or dismatched, retry until successful.
//	 * Identify if multiple dismatched requests come from multiple nodes in the same partition
//	 * if finally not able to merge with one partition, abort merging with it and delete it from "otherPartitions"
//	 * The result mergable partitions are in the "otherPartitions"
//	 * 
//	 * MTLock should be firstly obtained by the invoker of this operation
//	 * @param list
//	 * @return 1
//	 */
//	private static int startMergePT()
//	{
//		//try to lock all members in the list
//		//Hold the MT in the nodes which return commit
//		//retry until the rest nodes also vote commit
//		//Commit the merge with all the members
//		
//		GlobalPara.logger.warning("Start  startMergePT");
//		
//		Future<Integer>[] tasks=new Future[otherPartitions.size()];
//		Vector<Integer>[] agreedMembers=new Vector[otherPartitions.size()];
//		SendMergeReqTask[] mergeTasks=new SendMergeReqTask[otherPartitions.size()];
//		Vector<Integer> partitionsToRetry=new Vector<Integer>();
//		Vector<Integer> partitionsCanBeMerged=new Vector<Integer>();
//		
//		int count=0;
//		for(Vector<Integer> partitionToMerge: otherPartitions){
//			partitionsToRetry.add(count);
//			agreedMembers[count]=new Vector<Integer>();
//			SendMergeReqTask task=new SendMergeReqTask(partitionToMerge, agreedMembers[count]); //updating this partition will also update the partition in otherPartitions
//			mergeTasks[count]=task;
//			count++;
//		}
//		GlobalPara.logger.warning("Send Merge requests and wait for responds...");
//		
//		boolean cancelRest=false;
//
//		while(!cancelRest&&partitionsToRetry.size()>0)
//		{
//			for(int i=0;i<tasks.length;i++){
//				if(partitionsToRetry.contains(i))
//					tasks[i]=executor.submit(mergeTasks[i]);
//			}
//			
//			for(int i=0;i<tasks.length;i++){
//				if(!partitionsToRetry.contains(i))
//					continue;
//				Future<Integer> t=tasks[i];
//				try {
//					int result=t.get();
//					Vector<Integer> partitionToMerge=mergeTasks[i].getPartition();
//					if(result==-2||result==-1) //failure
//						{
//							cancelRest=true;
//							otherPartitions.remove(partitionToMerge);
//							partitionsToRetry.remove(i);
//							if(result==-2)
//								MembershipManager.sendAbortForMT(agreedMembers[i]);
//							GlobalPara.logger.warning("Deadlock detected, this partition is to avoid it, the return code is " + result);
//						}
//					else if(result==1) //all vote yes
//						{
//							partitionsCanBeMerged.add(i);
//							partitionsToRetry.remove(i);
//						}
//					else if(result==0) //still some vote no
//						{
//						;//need to be retried 
//						}
//				} catch (InterruptedException e) {
//					e.printStackTrace();
//				} catch (ExecutionException e) {
//					e.printStackTrace();
//				}
//			}
//			
//			if(cancelRest)
//			{
//				for(Integer k:partitionsToRetry)
//				{
//					MembershipManager.sendAbortForMT(agreedMembers[k]);
//					Vector<Integer> partitionToMerge=mergeTasks[k].getPartition();
//					otherPartitions.remove(partitionToMerge);
//				}
//				partitionsToRetry.clear();
//			}
//		}
//		
//		GlobalPara.logger.warning("All Mergable partitions requests  are determined, go into the commit process");
//		return 1;
//	}
	

	private static Vector<Long> startPartitionVector=new Vector<Long>();
	private static Vector<Long> endPartitionVector=new Vector<Long>();
	
	private static Vector<Long> startSendInvitationVector=new Vector<Long>();
	private static Vector<Long> endSendInvitationVector=new Vector<Long>();
	
	private static Vector<Long> startMergeVector=new Vector<Long>();
	private static Vector<Long> endMergeVector=new Vector<Long>();
	
	public static Vector<Long> startInviteReportedFailedNodeVector=new Vector<Long>();
	public static Vector<Long> endInviteReportedFailedNodeVector=new Vector<Long>();
	
	public static long startRecoverTran=0;
	public static long endRecoverTran=0;
	public static long startDataOrganization=0;
	public static long endDataOrganization=0;
	
	private static long startFailureTime=0;
	private static long endFailureTime=0;
	private static long startSendInvitation=0;
	private static long startPartition=0;
	
	private static void startInvitationPT()
	{
		//GlobalPara.logger.warning("Start  startInvitationPT: Try to get Lock");
			
		Vector<Integer> agreed=new Vector<Integer>(); 
		Vector<Integer> rejected=new Vector<Integer>();
		Vector<Integer> aborted=new Vector<Integer>();
		
		startSendInvitation=System.currentTimeMillis();
		startSendInvitationVector.add(startSendInvitation);

		Vector<Integer>members=new Vector<Integer>();
		members.addAll(GlobalPara.sysConf.TPNodeSet); 
		if(countOfInvitation==1||isEnoughNodesResponded){
			GlobalPara.logger.warning("Reported failed node "+reportedFailedNodeID +" is ignored as countOfInvitation= " 
					+countOfInvitation+ ", isEnoughNodesResponded="+isEnoughNodesResponded);
			members.remove(new Integer(reportedFailedNodeID));
		}
		
		SendInvitationsTask task=new SendInvitationsTask(members, agreed, rejected, aborted); 
		executor.submit(task);
	}
	
	/**
	 * @param agreed do not contain itself
	 * @return 1: all nodes in current partition, -1: still need more LTMs to join the partition, 0: aborted because of the lock
	 */
	public static int handleInvitationResponses(Vector<Integer> agreed, Vector<Integer> rejected, Vector<Integer> aborted) {
		GlobalPara.logger.warning("After Invitation agreed: " + agreed.size()+"rejected:" + rejected.size() +"aborted:" + aborted.size());
		
		if(countOfInvitation==1) //The first time will ignore the reportedFailedTPNode 
		{
			int numOfResponded=agreed.size()+rejected.size()+aborted.size()+1;//including itself
			int miniMumNum=GlobalPara.sysConf.TPNodeSet.size()-GlobalPara.sysConf.numOfReplicas+1;
			String msg="numOfResponded="+numOfResponded+" ,  miniMumNum= "+miniMumNum;
			if(numOfResponded>=miniMumNum){
				isEnoughNodesResponded=true;
			}
			msg+=", so isEnoughNodesResponded="+isEnoughNodesResponded;
			GlobalPara.logger.warning(msg);
		}
			
		boolean isCommitForInvitation=true;
		
		//If the node currently is not in any partitions, then join as many available LTMs immediately
		if(partition.size()==0){
			if(agreed.size()>0){
				isCommitForInvitation=true;
			} else {
				isCommitForInvitation=false;
			}
		} else	{
			//if any LTMs in the same partition vote abort, abort the invitation
			if(aborted.size()>0) {
				for(int pID:partition){
					if(aborted.contains(pID)) {
						isCommitForInvitation=false;
						break;
					}
				}
			}
			//if there is no new nodes agree to attend
			if(isCommitForInvitation&&agreed.size()
					==(partition.size()-1)&&otherPartitions.size()==0)//there is no new LTMs to be added, so abort
				isCommitForInvitation=false;
		}
		
		
		long endSendInvitation=System.currentTimeMillis();
		endSendInvitationVector.add(endSendInvitation);

		long duration=endSendInvitation-startSendInvitation;
		
		GlobalPara.logger.warning("[Monitor]Duration of Send invitation process is:" +duration);

		long startMerge=System.currentTimeMillis();
		startMergeVector.add(startMerge);
		
		boolean isKeepLock=false;
		if(!isCommitForInvitation){
			if(agreed.size()>0)
				MembershipManager.sendAbortForMT(agreed);
		} else { 
//			delay the commit of invitation until the merge process finishes, when merge using the "agreed" rather than the "partition"
			if(otherPartitions.size()>0){
				//startMergePT();//determine the final set of LTMs that can be merged, so the final partition membership can be determined
				PerformMergeTask task=new PerformMergeTask(agreed, MembershipManager.otherPartitions);
				MembershipManager.executor.submit(task);
				isKeepLock=true;
			} else{
				isKeepLock=commitPartitionMembership(agreed);
			}
		}
		
		if(!isKeepLock)
			endInvitationPT();
		
		return 0;
	}

	private static void endInvitationPT()
	{
		GlobalPara.logger.warning("Releasing MT Lock ");
		Concurrency.releaseMTLock();
	}
	
	public static int commitPartitionMembershipAfterMerge(Vector<Integer> agreed)
	{
		if(!commitPartitionMembership(agreed))
			endInvitationPT();
		return 0;
	}
	
	public static boolean commitPartitionMembership(Vector<Integer> agreed)
	{
		boolean isKeepLock=false;
		Vector<Integer> finalPartition=new Vector<Integer>(); 
		finalPartition.addAll(agreed);
		finalPartition.add(GlobalPara.sysConf.localNodeID);
		
		if(otherPartitions.size()>0){
			for(Vector<Integer> p:otherPartitions){
				finalPartition.addAll(p);
			}
			
			long endMerge=System.currentTimeMillis();
			endMergeVector.add(endMerge);
			long startMerge=startMergeVector.lastElement();
			long duration=endMerge-startMerge;
			GlobalPara.logger.warning("[Monitor]Duration of Merge process is:" +duration);
		}
		
		partition.clear();
		partition.addAll(finalPartition);

		String msg="The updated partition membership is ";
		for(int nn:partition)
			msg+=" " + nn;
		GlobalPara.logger.warning(msg);
		
		isKeepLock=isAbleToRecoverFromNetPartitions();//enough number of LTMs joined to recover the system

		sendPartitionMembershipCommit(partition, isKeepLock);
				
		long endPartition=System.currentTimeMillis();
		endPartitionVector.add(endPartition);
		
		long  duration=endPartition-startPartition;
		GlobalPara.logger.warning("[Monitor]Duration of invitation process is:" +duration);
		
		
		GlobalPara.logger.warning("send Inivitation completed");

		if(isKeepLock){
			GlobalPara.logger.warning("Enough number of LTMs reached, start recovering system");
			startRecoverSystem(true);
		}
		return isKeepLock;
	}
	private static boolean isAbleToRecoverFromNetPartitions()
	{
		boolean result;
		int curSize=partition.size();
		int originSize=GlobalPara.sysConf.TPNodeSet.size();
		if((originSize-curSize)<curSize) //it is the majority partition
		{
			if(curSize>originSize-GlobalPara.sysConf.numOfReplicas)
				result=true;
			else
				result=false;
		}
		else
			result=false;
		return result;
	}
	/**
	 * Recovering ongoing transactions, move data to fit the new system membership
	 * in the end make this LTM available
	 * The MTLock is obtained before invocating this function
	 * @return
	 */
	public static int startRecoverSystem(boolean isCoordinator)
	{
		//first Recovering the ongoing transactions, even though no LTM failures,
		//still need to wait for the recovery to complete before accepting any other transactions
		//FIXME the result of the completed transaction may be lost in sending back to the client, 
		// but here we assume that there is no such problem 
		//All transactions need to be send failure recovery msg
		
		ro_complete=false;
		rw_complete=false;
		data_complete=false;
		compeletedNodes=new Vector<Integer>();
		MembershipManager.isCoordinator=isCoordinator;
		
		Vector<Integer> failedNodes=new Vector<Integer>();
		if(partition.size()<GlobalPara.sysConf.TPNodeSet.size()) {
			for(Integer s:GlobalPara.sysConf.TPNodeSet) {
				if(!partition.contains(s))
					failedNodes.add(s);
			}
		}
		
		TPNode.failedNodes=failedNodes;
		
		String msg="New Consistent membership established, Start recovering the system, there are "+failedNodes.size() +" failed nodes:";
		for(int n:failedNodes){
			msg+=" " + n;
		}
		GlobalPara.logger.warning(msg);
		
		Vector<Integer> aliveBackupNodes=new Vector<Integer>(); 
		for(int n:TPNode.backupNodes)
			if(!failedNodes.contains(n))
				aliveBackupNodes.add(n);
		
		if(aliveBackupNodes.size()<TPNode.backupNodes.length){ //update backup nodes for TC
			TPNode.backupNodes=new int[aliveBackupNodes.size()];
			int count=0;
			for(int n: aliveBackupNodes)
				TPNode.backupNodes[count++]=n;
		}
		
		//TPNode.removeFailedNodeFromBackupVnodes(failedNodes);//update backup nodes for data items
		//Now prepare work for recovering completed by deleting failed nodes from the backup lists
		//////////////////////////////////////////////////////////////////
		
		int[] failedTPNodes=new int[failedNodes.size()];
		for(int i=0;i<failedTPNodes.length;i++)
			failedTPNodes[i]=failedNodes.get(i);
		
		TreeSet<Integer> newTPNodes=new TreeSet<Integer>();
		newTPNodes.addAll(TPNode.ring.TPNodeSet);
		
		for(int node:failedTPNodes){
			newTPNodes.remove(node);
		}
			
		newRing=new RingMembership(newTPNodes,
				GlobalPara.sysConf.localNodeID,
				GlobalPara.numOfVirtualNodes, 
				GlobalPara.numOfReplicas);
		
		GlobalPara.logger.warning("Start recover transactions");
		
		GlobalPara.rwComp.startRecoverTransactions(newRing, failedNodes);
		GlobalPara.roComp.reportFailureRecovery();
		GlobalPara.ltmComp.startRecoverDataItems(newRing, failedNodes);
		
		return 0;
	}
	
	private static void endRecoveryProcess()
	{
		TPNode.failedNodes.clear();
		MembershipManager.cleanTemporary();
		
		GlobalPara.logger.warning("Going to end Recovery Phase");
		endFailureRecovery();
		GlobalPara.logger.warning("Recovery Phase Ended, back to normal");
		
		if(GlobalPara.sysConf.doMemoryManagement){
			TPNode.MM.reportBufferSizeChange();
		}
	}
	
	private static void clearMonitoredTime()
	{
		 startPartitionVector.clear();
		 endPartitionVector.clear();
		
		 startSendInvitationVector.clear();
		 endSendInvitationVector.clear();
		
		 startMergeVector.clear();
		 endMergeVector.clear();
		
		 startInviteReportedFailedNodeVector.clear();
		 endInviteReportedFailedNodeVector.clear();
		 
		startRecoverTran=0;
		endRecoverTran=0;
		startDataOrganization=0;
		endDataOrganization=0;
		
		startFailureTime=0;
		endFailureTime=0;
		
		reportedFailedNodeID=-1;
		countOfInvitation=0;
		isEnoughNodesResponded=false;
		triedInvitationTime.clear();

	}
	
	private static void printMonitoredTime()
	{
		String msg="";
		
		try{
		 msg+="[PrintMonitor]startFailureTime= "+startFailureTime+",endFailureTime="+endFailureTime+",duration="+ (endFailureTime-startFailureTime) + "\n";
		 msg+="[PrintMonitor] Total Number of tried Invitation Process: " + triedInvitationTime.size() + "\n";
		
		 //GlobalPara.logger.warning(msg);
		 
		 long first;
		 long second;
		//for(Long time:triedInvitationTime)
		 
		 if(triedInvitationTime.size()>0)
		 {
			 second=triedInvitationTime.get(0);
			for(int i=0;i<triedInvitationTime.size();i++)
			{
				long d=0;
				if(i>0)
					{
					first=second;
					second=triedInvitationTime.get(i);
					d=second-first;
					}
				msg+="[PrintMonitor]"+triedInvitationTime.get(i);
				if(i>0)
					msg+="   "+d+"\n";
				else
					msg+="\n";
			}
		 }
		
		// GlobalPara.logger.warning(msg);
		 
		 msg+="[PrintMonitor] Total Number of successful Invitation Process: " + countOfInvitation + "\n";
		 
		 //GlobalPara.logger.warning(msg);
		 
		for(int i=0;i<startPartitionVector.size();i++)
		{
			if(i==0)
			{
				msg+="[PrintMonitor] Duration from failure detected to the 1st successful start of Invitation process:" + (startPartitionVector.get(i)-startFailureTime)+"\n";
			}
			else if(i==startPartitionVector.size()-1)
			{
				msg+="[PrintMonitor] Duration from failure detected to the LAST successful start of Invitation process:" + (startPartitionVector.get(i)-startFailureTime)+"\n";
			}
			msg+=(i+1)+"th time\n";
			long startP=startPartitionVector.get(i);
			long endP;
			if(endPartitionVector.size()>i)
				endP=endPartitionVector.get(i);
			else
				endP=0;
			msg+="[PrintMonitor]startPartition= "+startP+",endPartition="+endP+",duration="+ (endP-startP) + "\n";
			
			long startS;
			long endS;
			if(startSendInvitationVector.size()>i)
				startS=startSendInvitationVector.get(i);
			else
				startS=0;
			
			if(endSendInvitationVector.size()>i)
				endS=endSendInvitationVector.get(i);
			else
				endS=0;
			
			msg+="[PrintMonitor]startSendInvitation= "+startS+",endSendInvitation="+endS+",duration="+ (endS-startS) + "\n";
			
			long startM;
			long endM;
			if(startMergeVector.size()>i)
				startM=startMergeVector.get(i);
			else
				startM=0;
			
			if(endMergeVector.size()>i)
				endM=endMergeVector.get(i);
			else
				endM=0;
			msg+="[PrintMonitor]startMerge= "+startM+",endMerge="+endM+",duration="+ (endM-startM) + "\n";
			if(startInviteReportedFailedNodeVector.size()>0&&i>0)
				msg+="[PrintMonitor]startInviteReportedFailedNodeVector= "+startInviteReportedFailedNodeVector.get(i-1)+",endInviteReportedFailedNodeVector="+endInviteReportedFailedNodeVector.get(i-1)+",duration="+ (endInviteReportedFailedNodeVector.get(i-1)-startInviteReportedFailedNodeVector.get(i-1)) + "\n";
		}
	
		//GlobalPara.logger.warning(msg);
		
		msg+="[PrintMonitor]startRecoverTran= "+startRecoverTran+",endRecoverTran="+endRecoverTran+",duration="+ (endRecoverTran-startRecoverTran) + "\n";
		msg+="[PrintMonitor]startDataOrganization= "+startDataOrganization+",endDataOrganization="+endDataOrganization+",duration="+ (endDataOrganization-startDataOrganization) + "\n";
		}
		catch(java.lang.Exception e)
		{
			e.printStackTrace();
		}
		GlobalPara.logger.warning(msg);
	}
	
	private static int sendPartitionMembershipCommit(Vector<Integer> newPartition, boolean isKeepLock)
	{
		if(newPartition==null||newPartition.size()==0){
			GlobalPara.logger.warning("empty input");
			return 0;
		}
		Future<Integer>[] tasks=new Future[newPartition.size()-1];
		int count=0;
		for(int destId:newPartition)
		{
			if(GlobalPara.sysConf.localNodeID==destId)
				continue;
			SendPartitionMembershipCommitToOneLTMTask task=new SendPartitionMembershipCommitToOneLTMTask(isKeepLock, destId, newPartition);
			tasks[count++]=MembershipManager.executor.submit(task);
		}
		
		for(Future<Integer> t: tasks)
		{
			try {
				t.get();
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
		}
		return 0;
	}
	
	private static Random ran=new Random(); 
	public static long getRetryInterval()
	{
		double retryInterval_var=GlobalPara.sysConf.retryInterval_var;
		double retryInterval_mini=GlobalPara.sysConf.retryInterval_mini;
		
		Double interval=(double)ran.nextInt(10000)/(double)10000*retryInterval_var+retryInterval_mini;
		return interval.longValue();
	}
	
	public static void recoverFailure(String url, String IP, int failureTimes)
	{
		if(url==null||IP==null){
			GlobalPara.logger.warning("Invalid inputs for Failure report");
			return;
		}

		if((!GlobalPara.isInServletRole))
			return;
		
		
		
		//GlobalPara.logger.warning("Inside recoverFailure process started: "+url);
		
		int nodeId=GlobalPara.sysConf.getID(IP);
		
		if(!GlobalPara.sysConf.isDoFaultTolerant){
			GlobalPara.logger.warning("Fault tolerant is not enabled, report node "+nodeId+" failed:"+url);
			return;
		}

		if(failureTimes!=MembershipManager.failureTimes){
			GlobalPara.logger.warning("[NPStart]Ignoring outed failure report for LTM "+ nodeId +" when getting "+url);
			return;
		} else {
			if(TPNode.ring.TPNodeSet.contains(nodeId))
				;//GlobalPara.logger.warning("[NPStart]failure report for LTM "+ nodeId +" when getting "+url);
			else {
				GlobalPara.logger.warning("[NPStart]Ignoring failure report for node "+ nodeId +" when getting "+url);
				return;
			}
		}
		
		
		
//		if(!GlobalPara.sysConf.TPNodeSet.contains(nodeId))
//		{
//			GlobalPara.logger.warning("[NPStart]Node "+nodeId+" Already removed from membership in the previous recovery process");
//			return;
//		}
		
		//FIXME if multiple nodes fail, only one of them will be recorded as the failed id
		boolean isContinueRecovery=startFailureRecovery();
		
		if(isContinueRecovery) {
			GlobalPara.logger.warning("[NPStart]failure report for LTM "+ nodeId +" when getting "+url);
			reportedFailedNodeID=nodeId;
			StartInvitationTask task=new StartInvitationTask(failureTimes, true);
			executor.submit(task); 
		} else
			return; //already in failure recovery status
	}

	/**
	 * If this task can be applied, then apply it.
	 * @param nodeId
	 */
	public static void beginInvitationTask()
	{
		if(!MembershipManager.isInFailureRecovery()){ //FIXME, the system may be in another failure recovery.
			GlobalPara.logger.warning("[Error]An invitation thread started when system is not in Failure Recovery");
			return;
		}
		
		boolean lockObtained=Concurrency.obtainMTLock(GlobalPara.sysConf.localNodeID);
		if(!lockObtained){
			return;
		}
	
		startPartition=System.currentTimeMillis();
		triedInvitationTime.add(startPartition);
		startPartitionVector.add(startPartition);
		countOfInvitation++;
		GlobalPara.logger.warning("Start  startInvitationPT: Lock obtained, start times: " +countOfInvitation);
		
		startInvitationPT();
	}
	

	
	//FIXME currently only support optimization for one LTM failure
	public static int reportedFailedNodeID=-1;
	private static int countOfInvitation=0;
	private static boolean isEnoughNodesResponded=false;
	private static Vector<Long> triedInvitationTime=new Vector<Long>();
	
//	public static synchronized void handleFailure(int failedNodeID)
//	{
//		reportedFailedNodeID=failedNodeID;
//		//countOfInvitation=0;
//		//isEnoughNodesResponded=false;
//		
//		do{
//			if(!MembershipManager.isInFailureRecovery())
//			{
//				return;//The system has been recovered when this process is sleeping
//			}
//			
//			triedInvitationTime.add(System.currentTimeMillis());
//			
//			startInvitationPT();
//			
//			try {
//				Thread.sleep(getRetryInterval()); //to avoid conflicts because of concurrency control 
//			} catch (InterruptedException e) {
//				e.printStackTrace();
//			}
//		}
//		while(true);
//	}

	/**
	 * @return the currentCoordinatorID
	 */
	public static int getCurrentCoordinatorID() {
		return currentCoordinatorID;
	}

//	private static boolean getMTLock()
//	{
//		boolean result=false;
//		boolean getMT=Concurrency.obtainMTLock();
//		int count=0;
//		Random ran=new Random();
//		while(!getMT)
//		{
//			count++;
//			if(count==5)
//				{
//					return false;
//				}
//			long interval=ran.nextInt(Concurrency.retryInterval)+Concurrency.retryInterval;
//			try {
//				Thread.sleep(interval);
//			} catch (InterruptedException e) {
//				e.printStackTrace();
//			}
//			getMT=Concurrency.obtainMTLock();
//		}
//		result=true;
//		return result;
//	}
	
	
//	private void addLTMsToMembeship(String[] IPs, int[] IDs)
//	{
//		//update sysConf.NodeConfig, IPMapping ...
//		//update TPNodeSet, ringMembership
//		//DataMovement module should backup the old ring membership to determine the data movement
//	}
//	
//	private void deleteLTMsFromMembeship(String[] IPs, int[] IDs)
//	{
//		
//	}
}
