package org.zhouw.stps.membership;

import java.util.Vector;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.zhouw.stps.GlobalPara;

///**
//* determine the final set of LTMs that can be merged, and the final partition membership 
//* try to merge with all the partitions in the list
//* If aborted, or dismatched, retry until successful.
//* Identify if multiple dismatched requests come from multiple nodes in the same partition
//* if finally not able to merge with one partition, abort merging with it and delete it from "otherPartitions"
//* The result mergable partitions are in the "otherPartitions"
//* 
//* MTLock should be firstly obtained by the invoker of this operation
//* @param list
//* @return 1
//*/
public class PerformMergeTask implements Runnable {

	
	private Vector<Vector<Integer>> otherPartitions;
	private Vector<Integer> agreed;
	
	public PerformMergeTask(Vector<Integer> agreed, Vector<Vector<Integer>> otherPartitions) {
		this.otherPartitions=otherPartitions;
		this.agreed=agreed;
	}

	@Override
	public void run() {
		GlobalPara.logger.warning("Start  startMergePT");
		
		Future<Integer>[] tasks=new Future[otherPartitions.size()];
		Vector<Integer>[] agreedMembers=new Vector[otherPartitions.size()];
		SendMergeReqTask[] mergeTasks=new SendMergeReqTask[otherPartitions.size()];
		//Vector<Integer> partitionsToRetry=new Vector<Integer>();
		Vector<Integer> partitionsCanBeMerged=new Vector<Integer>();
		
		int count=0;
		for(Vector<Integer> partitionToMerge: otherPartitions){
			//partitionsToRetry.add(count);
			agreedMembers[count]=new Vector<Integer>();
			SendMergeReqTask task=new SendMergeReqTask(partitionToMerge, agreedMembers[count]); //updating this partition will also update the partition in otherPartitions
			mergeTasks[count]=task;
			count++;
		}
		GlobalPara.logger.warning("Send Merge requests and wait for responds...");
		
		
		for(int i=0;i<tasks.length;i++){
			tasks[i]=MembershipManager.executor.submit(mergeTasks[i]);
		}
		
		for(int i=0;i<tasks.length;i++){
			Future<Integer> t=tasks[i];
			try {
				int result=t.get();
				Vector<Integer> partitionToMerge=mergeTasks[i].getPartition();
				if(result==-2||result==-1||result==0){ //Abort the merge
					otherPartitions.remove(partitionToMerge);
					MembershipManager.sendAbortForMT(agreedMembers[i]);
					GlobalPara.logger.warning("Deadlock detected, this partition is to avoid it, the return code is " + result);
				} else if(result==1) {//all vote yes
					partitionsCanBeMerged.add(i);
				}
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
		}
		
		GlobalPara.logger.warning("All Mergable partitions requests  are determined, go into the commit process");
		GlobalPara.memberComp.reportReceiveMergeResponse(agreed);
	}

}
