package org.zhouw.stps.membership;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.Callable;
import org.zhouw.stps.membership.msg.MembershipMsg.SendInvitationRequestMsg;
import org.zhouw.stps.membership.msg.MembershipMsg.ReceiveInvitationResponseMsg;
import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.MsgBroker;
import org.zhouw.stps.membership.MembershipManager;

public class SendInvitationToOneLTM implements Callable<Integer> {

	private int destID;
	private Vector<Integer> agreed;
	private Vector<Integer> rejected;
	private Vector<Integer> aborted;
	
	/**
	 * @return -2: failure, -3: unexpected exception
	 */
	public Integer call() throws Exception {
		int result=1;
		if(GlobalPara.sysConf.localNodeID==destID)
			return 1;
		long start=0;
		long end=0;
		if(destID==MembershipManager.reportedFailedNodeID){
			start=System.currentTimeMillis();
			MembershipManager.startInviteReportedFailedNodeVector.add(start);	
		}
		
		SendInvitationRequestMsg msg=SendInvitationRequestMsg.newBuilder().
			setVersion(MembershipManager.getVersionNumber()).setSrcID(GlobalPara.sysConf.localNodeID).build();
		byte[] input=msg.toByteArray();
		InputStream in=MsgBroker.sendMsgToOneLTM(destID, "ReceiveInvitationService", input);
		if(in!=null){
			ReceiveInvitationResponseMsg response=ReceiveInvitationResponseMsg.parseFrom(in);
			int version=response.getVersion();
			int compareResult=MembershipManager.compareVersionNumber(version);
			if(compareResult==-1){
				GlobalPara.logger.warning("Receiving membership version larger than local LTMs from LTM "+msg.getSrcID());
				MembershipManager.handleOutdatedMembership();
			} else if(compareResult==1) {
				GlobalPara.logger.warning("Receiving membership version smaller than local LTMs");
				result=-3;
			} else {
				int res=response.getResult();
				if(res==1)
					this.agreed.add(destID);
				else if(res==0)
					this.aborted.add(destID);
				else if(res==-1){
					this.rejected.add(destID);
					List<Integer> list=response.getCurPartitionList();
					MembershipManager.addRejectedPartition(list);
				}
				result=res;
			}
			MsgBroker.cleanupInputStream(in);
		} else
			result=-2;
		
		
		if(destID==MembershipManager.reportedFailedNodeID){
			end=System.currentTimeMillis();
			MembershipManager.endInviteReportedFailedNodeVector.add(end);
			GlobalPara.logger.warning("[Monitor]Invite reported failed node returns within " + (end-start));
		}
		
		return result;
	}

	public SendInvitationToOneLTM(int destID, Vector<Integer> agreed,
			Vector<Integer> rejected, Vector<Integer> aborted) {
		super();
		this.destID = destID;
		this.agreed = agreed;
		this.rejected = rejected;
		this.aborted = aborted;
	}

}
