package org.zhouw.stps.membership;

import java.util.Vector;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import org.zhouw.stps.GlobalPara;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Test if the partitionToMerge matches the current partitionToMerge membership of current nodes.
 * @author zhouw
 *
 */
public class SendMergeReqTask implements Callable<Integer> {

	//private volatile boolean isContinue; //controlled by the globalProcess
	private Vector<Integer> partitionToMerge;
	
	private ReentrantLock lockOfNewPartition;
	
	private Vector<Integer> currentPartition;
	private Vector<Integer> abortedMembers;
	private Vector<Integer> agreedMembers;
	
	public void setCurrentPartition(Vector<Integer> newPartition)
	{
		lockOfNewPartition.lock();
		try{
			currentPartition.clear();
			currentPartition.addAll(newPartition);
		}
		finally
		{
			lockOfNewPartition.unlock();
		}
	}
	
	public SendMergeReqTask(Vector<Integer> partition, Vector<Integer> agreedMembers) {
		this.partitionToMerge=partition;
		currentPartition=new Vector<Integer>();
		abortedMembers=new Vector<Integer>();
		this.agreedMembers=agreedMembers;
		this.lockOfNewPartition=new ReentrantLock();
		//isContinue=true;
	}
	
	/**
	 * if meeting membership dismatch problem, retry with the new membership 
	 * return 1: all LTMs in this partitionToMerge vote yes, 0: some abort and some yes and some reject, 
	 * -1: all votes aborts, -2: failure when try to reach some members
	 * -3: exception
	 * 
	 */
	public Integer call() {
		if(partitionToMerge==null||partitionToMerge.size()==0)
			return -1;
		Vector<Integer> listsToSend=new Vector<Integer>();
		listsToSend.addAll(partitionToMerge);
		listsToSend.removeAll(this.agreedMembers);
		
		int result=sendMergReq(listsToSend);
		if(result==-2)
			return -2;
		
		if(result==-1){
			if(this.agreedMembers.size()>0){
				GlobalPara.logger.warning("Merging partitions are not consistent in their membership");
				return -2;
			} else {//dismatch so restart the merge with the new membership
//					abortedMembers.clear();
//					this.partitionToMerge.clear();
//					this.partitionToMerge.addAll(this.currentPartition);
//					listsToSend.clear();
//					listsToSend.addAll(this.currentPartition);
//					currentPartition.clear();
				return 0;
			}
		} else if (result==1) {
			if(this.agreedMembers.size()==partitionToMerge.size())
				return 1;
				
			if(this.abortedMembers.size()==partitionToMerge.size())
				return -1;
			else //membership match, but some still aborted
				return 0;
				
//					listsToSend.clear();
//					listsToSend.addAll(this.abortedMembers);
//					this.abortedMembers.clear();
				//this.agreedMembers remains untouched.
		}
		
		GlobalPara.logger.warning("Exceptions: not expected behaviors");
		return -3;
	}

	/**
	 * 1: normal all respond agreed or aborted, -1: dismatched & rejected -2: failure
	 * @param destIDs the list of node ids to be merged
	 * @return
	 */
	private int sendMergReq(Vector<Integer> destIDs)
	{
		Future<Integer>[] tasks=new Future[destIDs.size()];
		int count=0;
		for(int destId:destIDs)
		{
			SendMergeReqToOneLTMTask task=new SendMergeReqToOneLTMTask(destId, partitionToMerge, abortedMembers, agreedMembers, this);
			tasks[count++]=MembershipManager.executor.submit(task);
		}
		boolean isDismatched=false;
		boolean isFailure=false;
		for(Future<Integer> t: tasks)
		{
			try {
				int result=t.get();
				if(result==-1)
					isDismatched=true;
				else if(result==-2)
					isFailure=true;
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
		}
		
		if(isFailure)
			return -2;
		else if(isDismatched)
			return -1;
		return 1;
	}
	
	

	/**
	 * @return the partitionToMerge
	 */
	public Vector<Integer> getPartition() {
		return partitionToMerge;
	}
}
