package org.zhouw.stps.membership;

import java.io.InputStream;
import java.util.Vector;
import java.util.concurrent.Callable;

import org.zhouw.stps.*;
import org.zhouw.stps.membership.msg.MembershipMsg.SendMergeRequestMsg;
import org.zhouw.stps.membership.msg.MembershipMsg.ReceiveMergeResponseMsg;

public class SendMergeReqToOneLTMTask implements Callable<Integer> {

	private Vector<Integer> PartitionToMerge;
	
	private Vector<Integer> currentPartition;
	private Vector<Integer> abortedMembers;
	private Vector<Integer> agreedMembers;
	private SendMergeReqTask parentTask;
	
	private int destID;
	
	
	/**
	 * return 1:vote match, so can commit, 0: aborted, so add itself to the abortedMembers, -1: dismatch, new partition returned in currentPartition
	 * -2: failure when try to contact
	 */
	public Integer call() throws Exception {
	
		if(GlobalPara.sysConf.localNodeID==destID)
			return 1;
		
		SendMergeRequestMsg.Builder msgBuilder=SendMergeRequestMsg.newBuilder().setVersion(MembershipManager.getVersionNumber())
		.setSrcID(GlobalPara.sysConf.localNodeID);
		msgBuilder.addAllPartitionToMerge(PartitionToMerge);
		
		SendMergeRequestMsg msg=msgBuilder.build();
		
		byte[] input=msg.toByteArray();
		InputStream in=MsgBroker.sendMsgToOneLTM(destID, "ReceiveMergeRequestService", input);
		
		if(in!=null)
			{
			ReceiveMergeResponseMsg response=ReceiveMergeResponseMsg.parseFrom(in);
			int result=response.getResult();
			if(result==1)
				agreedMembers.add(destID);
			else if(result==0)
				abortedMembers.add(destID);
			else if(result==-1)
				{
				currentPartition.addAll(response.getCurPartitionList());
				parentTask.setCurrentPartition(currentPartition);
				}
			MsgBroker.cleanupInputStream(in);
			return result;
			}
		else
			return -2;
	}



	public SendMergeReqToOneLTMTask(int destID, Vector<Integer> PartitionToMerge,
			Vector<Integer> abortedMembers, Vector<Integer> agreedMembers, SendMergeReqTask parentTask) {
		super();
		this.PartitionToMerge = PartitionToMerge;
		this.currentPartition = new Vector<Integer>();
		this.abortedMembers = abortedMembers;
		this.agreedMembers=agreedMembers;
		this.destID = destID;
		this.parentTask=parentTask;
	}

}
