package org.zhouw.stps.monitor;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Date;
import org.zhouw.stps.*;
import org.zhouw.stps.conf.SystemConfigure;

public class LogAnalyzer {

	public SystemConfigure sysConf;
	
	public LogAnalyzer(String inPath, String outPath, String confPath)
	{
		if(outPath!=null)
			{
			if(!outPath.endsWith("/"))
				outPath+="/";
			}
		this.outPath=outPath;
		if(inPath!=null)
			{
			if(!inPath.endsWith("/"))
				inPath+="/";
			}
		this.inPath=inPath;
		sysConf=new SystemConfigure(confPath);
		System.out.println("Ready to anaylze log in path:" +inPath);
		System.out.println("Configuration file path is:" +confPath);
		System.out.println("output path is:" +outPath);
	}
	
	private String outPath;
	private String inPath;
	
	private static final long responseTime_interval=10; //ms
	private static final int responseTime_slotAmount=100; //ms

	private static final long checkPointResponseTime_interval=100; //ms
	private static final int checkPointResponseTime_slotAmount=300; //ms
	
	private static final long concurrency_interval=1000; //ms
	private static final long throughput_interval=1000; //ms
	private static final long warmup_interval=300000; //ms
	private static final long endup_interval=120000; //ms
	
	public void analyzeDataServiceLog(String fileName)
	{
		analyzeClientTransactionLog(fileName);
		System.out.println("analyzeDataServiceLog has been successfully analyzed");
	}
	public void analyzeIDServiceLog(String fileName)
	{
		int total=0;
		int[] rstSlots=new int[responseTime_slotAmount];
		for(int c:rstSlots)
			c=0;
		
		long firstTimestamp=0;
		long lastTimestamp=0;
		long totalCommittedLatency=0;
		
		long curTime=0;
		int curNum=0;
		String fileNameOverview=fileName+"_Overview";
		String fileNameRs=fileName+"_Rs";
		String fileNameThroughputbyend=fileName+"_Throughputbyend";
		String fileNameRsInSlot=fileName+"_RsInSlot";
		try {
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			BufferedWriter writerRs=new BufferedWriter(new FileWriter(outPath+fileNameRs,false));
			BufferedWriter writerThroughputbyend=new BufferedWriter(new FileWriter(outPath+fileNameThroughputbyend,false));
			BufferedWriter writerRsInSlot=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot,false));
		 
			
			FileInputStream fstream = new FileInputStream(inPath+fileName);
			BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
			IDServiceLogItem log=new IDServiceLogItem();
			if(log.readFrom(in)!=-1)
				{
					firstTimestamp=log.timestamp;
					curTime=log.timestamp;
					do
					{
						
						//skip the load in warmup
						if(log.timestamp<(firstTimestamp+warmup_interval))
							continue;
						
						//skip the last time period which is doing ending
						if(log.timestamp>(firstTimestamp+ GlobalPara.testDuration -endup_interval))
							continue;
						
						lastTimestamp=log.timestamp;
						
						total++;
						totalCommittedLatency+=log.latency;
							
						writerRs.write((log.timestamp-firstTimestamp)
								+" "+ log.latency);
						writerRs.newLine();
						
						//
						if((log.timestamp-curTime)<throughput_interval)
							curNum++;
						else
							{
								do{
									writerThroughputbyend.write((log.timestamp-firstTimestamp)+" "+ curNum);
									writerThroughputbyend.newLine();
									curTime+=throughput_interval;
									curNum=0;
								}
								while((log.timestamp-curTime)>throughput_interval);
								curNum=1;
							}
						
						
						int index=(int)(log.latency/responseTime_interval);
						if(index>=responseTime_slotAmount)
							index=responseTime_slotAmount-1;
						rstSlots[index]++;
							
					}while(log.readFrom(in)!=-1);
				}
			double totalAmount=0;
			for(int i=rstSlots.length-1;i>=0;i--)
				{
					totalAmount+=rstSlots[i];
				}
			int amount=0;
			for(int i=rstSlots.length-1;i>=0;i--)
				{
					amount+=rstSlots[i];
					writerRsInSlot.write((i+1)*responseTime_interval+" "+((double)amount/totalAmount)*100);
					writerRsInSlot.newLine();
				}
			
			long interval=lastTimestamp-firstTimestamp-warmup_interval;
			double totalThroughput=(double)total/((double)interval/1000);
			double avgResponseTime=(double)totalCommittedLatency/(double)total;
			
			writerOverview.write("Experiment last for "+interval/1000 + " seconds");
			writerOverview.newLine();
			writerOverview.write("There are totally "+total +" transaction executed");
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput + " TPS");
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgResponseTime + " ms");
			writerOverview.newLine();
			
			in.close();
			
			writerOverview.flush();
			writerRs.flush();
			writerThroughputbyend.flush();
			//writerThroughputbystart.flush();
			writerRsInSlot.flush();
			
			writerOverview.close();
			writerRs.close();
			writerThroughputbyend.close();
			//writerThroughputbystart.close();
			writerRsInSlot.close();
			
			System.out.println("IDServiceLog has been successfully analyzed");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Generate percentage of requests response time. per 50ms
	 * Detailed points about response time by endtime timestamp  
	 * Detailed points about response time by startTime timestamp
	 * Detailed graph about throughput by timestamp per 30 seconds
	 * Percentage of committed 
	 * @param fileName
	 */
	public void analyzeClientTransactionLog(String fileName)
	{
		String fileNameOverview=fileName+"_Overview";
		String fileNameRs=fileName+"_Rs";
		//String fileNameThroughputbystart=fileName+"_Throughputbystart";
		String fileNameThroughputbyend=fileName+"_Throughputbyend";
		String fileNameRsInSlot=fileName+"_RsInSlot";
		int numCommitted=0;
		int total=0;
		int[] rstSlots=new int[responseTime_slotAmount];
		for(int c:rstSlots)
			c=0;
		
		long firstTimestamp=0;
		long lastTimestamp=0;
		long totalCommittedLatency=0;
		
		long curTime=0;
		int curNum=0;
		long curTimebyStart=0;
		int curNumbyStart=0;
		try {
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			BufferedWriter writerRs=new BufferedWriter(new FileWriter(outPath+fileNameRs,false));
			BufferedWriter writerThroughputbyend=new BufferedWriter(new FileWriter(outPath+fileNameThroughputbyend,false));
			//BufferedWriter writerThroughputbystart=new BufferedWriter(new FileWriter(outPath+fileNameThroughputbystart,false));
			BufferedWriter writerRsInSlot=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot,false));
			 
			
			FileInputStream fstream = new FileInputStream(inPath+fileName);
			BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
			ClientTransactionLogItem log=new ClientTransactionLogItem();
			if(log.readFrom(in)!=-1)
				{
					firstTimestamp=log.timestamp;
					curTime=log.timestamp;
					curTimebyStart=log.timestamp-log.latency;
					do
					{
						
						//skip the load in warmup
						if(log.timestamp<(firstTimestamp+warmup_interval))
							
							continue;
						//skip the last time period which is doing ending
						if(log.timestamp>(firstTimestamp+ GlobalPara.testDuration -endup_interval))
							continue;
						
						lastTimestamp=log.timestamp;
						
						total++;
						if(log.isCommitted)
							{
							numCommitted++;
							totalCommittedLatency+=log.latency;
							//
							//Date curDate=new Date(log.timestamp);
							writerRs.write((log.timestamp-firstTimestamp)
									+" "+ log.latency);
							writerRs.newLine();
							
							//
							if((log.timestamp-curTime)<throughput_interval)
								curNum++;
							else
								{
									do{
										writerThroughputbyend.write((log.timestamp-firstTimestamp)+" "+ curNum);
										writerThroughputbyend.newLine();
										curTime+=throughput_interval;
										curNum=0;
									}
									while((log.timestamp-curTime)>throughput_interval);
									curNum=1;
								}
							
							//
							if(((log.timestamp-log.latency)-curTimebyStart)<throughput_interval)
								curNumbyStart++;
							else
								{
//									do{
//										writerThroughputbystart.write((log.timestamp-firstTimestamp)+" "+ curNumbyStart);
//										writerThroughputbystart.newLine();
//										curTimebyStart+=throughput_interval;
//										curNumbyStart=0;
//									}
//									while(((log.timestamp-log.latency)-curTimebyStart)>throughput_interval);
//									curNumbyStart=1;
								}
							
							//
							int index=(int)(log.latency/responseTime_interval);
							if(index>=responseTime_slotAmount)
								index=responseTime_slotAmount-1;
							rstSlots[index]++;
							}
					}while(log.readFrom(in)!=-1);
				}
			double totalAmount=0;
			for(int i=rstSlots.length-1;i>=0;i--)
				{
					totalAmount+=rstSlots[i];
				}
			int amount=0;
			for(int i=rstSlots.length-1;i>=0;i--)
				{
					amount+=rstSlots[i];
					writerRsInSlot.write((i+1)*responseTime_interval+" "+((double)amount/totalAmount)*100);
					writerRsInSlot.newLine();
				}
			
			double percentageOfCommit=((double)numCommitted/(double)total)*100;
			long interval=lastTimestamp-firstTimestamp-warmup_interval;
			double commitThroughput=(double)numCommitted/((double)interval/1000);
			double totalThroughput=(double)total/((double)interval/1000);
			double avgResponseTime=(double)totalCommittedLatency/(double)numCommitted;
			
			writerOverview.write("Experiment last for "+interval/1000 + " seconds");
			writerOverview.newLine();
			writerOverview.write("There are totally "+total +" transaction executed, "+ numCommitted +" of them are successfully committed");
			writerOverview.newLine();
			writerOverview.write("percentageOfCommit="+percentageOfCommit +"%");
			writerOverview.newLine();
			writerOverview.write("commitThroughput="+commitThroughput + " TPS");
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput + " TPS");
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgResponseTime + " ms");
			writerOverview.newLine();
			
			in.close();
			
			writerOverview.flush();
			writerRs.flush();
			writerThroughputbyend.flush();
			//writerThroughputbystart.flush();
			writerRsInSlot.flush();
			
			writerOverview.close();
			writerRs.close();
			writerThroughputbyend.close();
			//writerThroughputbystart.close();
			writerRsInSlot.close();
			
			System.out.println("ClientTransactionLog has been successfully analyzed");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public void analyzeCheckPointLog(String fileName)
	{
		String fileNameOverview=fileName+"_Overview";
		String fileNameRs=fileName+"_Rs"; //include total, copy and flush latency 
		String fileNameRsInSlot=fileName+"_RsInSlot"; //include total, copy and flush latency inslot
		String fileNameRsInSlot_copy=fileName+"_RsInSlotCopyLatency"; //include total, copy and flush latency inslot
		String fileNameMem=fileName+"_memory"; //include total, copy and flush latency 
		int total=0;
		int[] totalLatencySlots=new int[checkPointResponseTime_slotAmount];
		int[] copyLatencySlots=new int[checkPointResponseTime_slotAmount];
		int[] flushLatencySlots=new int[checkPointResponseTime_slotAmount];
		int[] copyLatencySlotsInSingle=new int[responseTime_slotAmount];
		
		for(int c:totalLatencySlots)
			c=0;
		for(int c:copyLatencySlots)
			c=0;
		for(int c:flushLatencySlots)
			c=0;
		for(int c:copyLatencySlotsInSingle)
			c=0;
		
		long firstTimestamp=0;
		long lastTimestamp=0;
		long totalLatency=0;
		long totalCopyLatency=0;
		long totalFlushLatency=0;
		
		try {
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			BufferedWriter writerRs=new BufferedWriter(new FileWriter(outPath+fileNameRs,false));
			BufferedWriter writerMemory=new BufferedWriter(new FileWriter(outPath+fileNameMem,false));
			BufferedWriter writerRsInSlot=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot,false));
			BufferedWriter writerRsInSlot_copy=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot_copy,false));
			 
			
			FileInputStream fstream = new FileInputStream(inPath+fileName);
			BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
			CheckpointLogItem log=new CheckpointLogItem();
			if(log.readFrom(in)!=-1)
				{
					firstTimestamp=log.timestamp;
					do
					{
						
//						//skip the load in warmup
//						if(log.timestamp<(firstTimestamp+warmup_interval))
//							
//							continue;
//						//skip the last time period which is doing ending
//						if(log.timestamp>(firstTimestamp+ GlobalPara.testDuration -endup_interval))
//							continue;
						
						lastTimestamp=log.timestamp;
						
						total++;
						totalLatency+=log.totalLatency;
						totalCopyLatency+=log.copyLatency;
						totalFlushLatency+=log.totalLatency;
						
						//
						//Date curDate=new Date(log.timestamp);
						writerRs.write((log.timestamp-firstTimestamp)
								+" "+ log.totalLatency + " " + log.copyLatency + " " + log.flushLatency);
						writerRs.newLine();
						writerMemory.write((log.timestamp-firstTimestamp)
								+" "+ log.freeMemoryBefore + " " + log.totalMemoryBefore);
						writerRs.newLine();
						writerMemory.write((log.timestamp-firstTimestamp+log.totalLatency)
								+" "+ log.freeMemoryAfter + " " + log.totalMemoryAfter);
						writerRs.newLine();
						
						
						//
						int index=(int)(log.totalLatency/checkPointResponseTime_interval);
						if(index>=checkPointResponseTime_slotAmount)
							index=checkPointResponseTime_slotAmount-1;
						totalLatencySlots[index]++;
						
						index=(int)(log.copyLatency/checkPointResponseTime_interval);
						if(index>=checkPointResponseTime_slotAmount)
							index=checkPointResponseTime_slotAmount-1;
						copyLatencySlots[index]++;
						
						index=(int)(log.copyLatency/responseTime_interval);
						if(index>=responseTime_slotAmount)
							index=responseTime_slotAmount-1;
						copyLatencySlotsInSingle[index]++;
						
						index=(int)(log.flushLatency/checkPointResponseTime_interval);
						if(index>=checkPointResponseTime_slotAmount)
							index=checkPointResponseTime_slotAmount-1;
						flushLatencySlots[index]++;
							
					}while(log.readFrom(in)!=-1);
				}
			
			double totalAmount1=0;
			double totalAmount2=0;
			double totalAmount3=0;
			for(int i=totalLatencySlots.length-1;i>=0;i--)
				{
					totalAmount1+=totalLatencySlots[i];
					totalAmount2+=copyLatencySlots[i];
					totalAmount3+=flushLatencySlots[i];
				}
			int amount1=0;
			int amount2=0;
			int amount3=0;
			for(int i=totalLatencySlots.length-1;i>=0;i--)
				{
					amount1+=totalLatencySlots[i];
					amount2+=copyLatencySlots[i];
					amount3+=flushLatencySlots[i];
					writerRsInSlot.write((i+1)*checkPointResponseTime_interval+" "
							+((double)amount1/totalAmount1)*100 + " "
							+((double)amount2/totalAmount2)*100 + " "
							+((double)amount3/totalAmount3)*100 );
					writerRsInSlot.newLine();
				}
			
			double totalAmount=0;
			for(int i=copyLatencySlotsInSingle.length-1;i>=0;i--)
				{
					totalAmount+=copyLatencySlotsInSingle[i];
				}
			int amount=0;
			for(int i=copyLatencySlotsInSingle.length-1;i>=0;i--)
				{
					amount+=copyLatencySlotsInSingle[i];
					writerRsInSlot_copy.write((i+1)*responseTime_interval+" "+((double)amount/totalAmount)*100);
					writerRsInSlot.newLine();
				}
			
			long interval=lastTimestamp-firstTimestamp;
			double totalThroughput=(double)total/((double)interval/1000);
			double avgTotalResponseTime=(double)totalLatency/(double)total;
			double avgCopyLatency=(double)totalCopyLatency/(double)total;
			double avgFlushResponseTime=(double)totalFlushLatency/(double)total;
			
			writerOverview.write("Experiment last for "+interval/1000 + " seconds");
			writerOverview.newLine();
			writerOverview.write("There are totally "+total +" checkpointed executed, "+ totalThroughput +" times per second");
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgTotalResponseTime + " ms");
			writerOverview.newLine();
			writerOverview.write("avgCopyLatency="+avgCopyLatency + " ms");
			writerOverview.newLine();
			writerOverview.write("avgFlushResponseTime="+avgFlushResponseTime + " ms");
			writerOverview.newLine();
			
			in.close();
			
			writerOverview.flush();
			writerRs.flush();
			writerMemory.flush();
			writerRsInSlot.flush();
			writerRsInSlot_copy.flush();
			
			writerOverview.close();
			writerRs.close();
			writerMemory.close();
			writerRsInSlot.close();
			writerRsInSlot_copy.close();
			
			System.out.println("analyzeCheckPointLog has been successfully analyzed");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	public void analyzeTPNodeLog(String fileName)
	{
		String fileNameOverview=fileName+"_Overview";
		String fileNameRs=fileName+"_Rs";
		String fileNameThroughputbyend=fileName+"_Throughputbyend";
		String fileNameRsInSlot=fileName+"_RsInSlot";
		//String fileNameConcurrency=fileName+"_Concurrency";
		String fileNameConcurrencyInSlot=fileName+"__ConcurrencyInSlot";
		
		int numCommitted=0;
		int total=0;
		int[] totalRstSlots=new int[responseTime_slotAmount];
		int[] submitRstSlots=new int[responseTime_slotAmount];
		int[] waitRstSlots=new int[responseTime_slotAmount];
		int[] commitRstSlots=new int[responseTime_slotAmount];
		int[] abortRstSlots=new int[responseTime_slotAmount];
		for(int c:totalRstSlots)
			c=0;
		for(int c:submitRstSlots)
			c=0;
		for(int c:waitRstSlots)
			c=0;
		for(int c:commitRstSlots)
			c=0;
		for(int c:abortRstSlots)
			c=0;
		
		long firstTimestamp=0;
		long lastTimestamp=0;
		long totalCommittedLatency=0;
		long totalCommittedSubmitLatency=0;
		long totalCommittedWaitLatency=0;
		long totalCommittedCommitLatency=0;
		long totalCommittedAbortLatency=0;
		int totalRestartTimes=0;
		int[] totalTPNodeInvoked=new int[GlobalPara.sysConf.nodes.getTotalNumberOfNodes()];
		int[] totalTPNodeInvokedForCommitted=new int[GlobalPara.sysConf.nodes.getTotalNumberOfNodes()];
		int[] numOfTPNodeInslot=new int[GlobalPara.sysConf.nodes.getServletNodes(null).length];
		
		for(int c:totalTPNodeInvoked)
			c=0;
		for(int c:totalTPNodeInvokedForCommitted)
			c=0;
		for(int c:numOfTPNodeInslot)
			c=0;
		
		long curTime=0;
		int curNum=0;
		int concuNum=0;
		try {
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			BufferedWriter writerRs=new BufferedWriter(new FileWriter(outPath+fileNameRs,false));
			BufferedWriter writerThroughputbyend=new BufferedWriter(new FileWriter(outPath+fileNameThroughputbyend,false));
			BufferedWriter writerRsInSlot=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot,false));
			
			//BufferedWriter writerConcurrency=new BufferedWriter(new FileWriter(outPath+fileNameConcurrency,false));
			BufferedWriter writerConcurrencyInSlot=new BufferedWriter(new FileWriter(outPath+fileNameConcurrencyInSlot,false));
			
			FileInputStream fstream = new FileInputStream(inPath+fileName);
			BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
			ServerTransactionLogItem log=new ServerTransactionLogItem();
			
			writerRs.write("#timestamp totalLatency submitLatency waitLatency doCommitLatency abortLatency restartTimes transaction_name");
			writerRs.newLine();
			
			if(log.readFrom(in)!=-1)
				{
					firstTimestamp=log.timestamp;
					curTime=log.timestamp;
					do
					{
						
						//skip the load in warmup
						if(log.timestamp<(firstTimestamp+warmup_interval))
							
							continue;
						//skip the last time period which is doing ending
						if(log.timestamp>(firstTimestamp+ GlobalPara.testDuration -endup_interval))
							continue;
						
						lastTimestamp=log.timestamp;
						
						total++;
//						for(int i=0;i<log.STGDestTPNodeID.length;i++)
//							totalTPNodeInvokedForCommitted[log.STGDestTPNodeID[i]]++;
//						numOfTPNodeInslot[log.STGDestTPNodeID.length]++;
						
						if(log.isCommitted)
							{
							numCommitted++;
							totalCommittedLatency+=log.totalLatency;
//							 totalCommittedSubmitLatency+=log.submitLatency;
//							 totalCommittedWaitLatency+=log.waitLatency;
//							 totalCommittedCommitLatency+=log.doCommitLatency;
//							 totalCommittedAbortLatency+=log.abortLatency;
//							 totalRestartTimes+=log.restartTimes;
//							 for(int i=0;i<log.STGDestTPNodeID.length;i++)
//									totalTPNodeInvoked[log.STGDestTPNodeID[i]]++;
							
							Date curDate=new Date(log.timestamp);
							writerRs.write((log.timestamp-firstTimestamp)
									+" "+ log.totalLatency
//									+" "+ log.submitLatency
//									+" "+ log.waitLatency
//									+" "+ log.doCommitLatency
//									+" "+ log.abortLatency
//									+" "+ log.restartTimes
									+" "+ log.name
									);
							writerRs.newLine();
							
							
							//
							if((log.timestamp-curTime)<throughput_interval)
								{
								curNum++;
								//concuNum+=log.STGDestTPNodeID.length;
								}
							else
								{
									do{
										writerThroughputbyend.write((log.timestamp-firstTimestamp)+" "
												+ curNum + " "
												+ concuNum);
										writerThroughputbyend.newLine();
										curTime+=throughput_interval;
										curNum=0;
									}
									while((log.timestamp-curTime)>throughput_interval);
									curNum=1;
								}
							
						
							int index=(int)(log.totalLatency/responseTime_interval);
							if(index>=responseTime_slotAmount)
								index=responseTime_slotAmount-1;
							totalRstSlots[index]++;
							
//							index=(int)(log.submitLatency/responseTime_interval);
//							if(index>=responseTime_slotAmount)
//								index=responseTime_slotAmount-1;
//							submitRstSlots[index]++;
//							
//							index=(int)(log.waitLatency/responseTime_interval);
//							if(index>=responseTime_slotAmount)
//								index=responseTime_slotAmount-1;
//							waitRstSlots[index]++;
//							
//							index=(int)(log.doCommitLatency/responseTime_interval);
//							if(index>=responseTime_slotAmount)
//								index=responseTime_slotAmount-1;
//							commitRstSlots[index]++;
//							
//							index=(int)(log.abortLatency/responseTime_interval);
//							if(index>=responseTime_slotAmount)
//								index=responseTime_slotAmount-1;
//							abortRstSlots[index]++;
							}
					}while(log.readFrom(in)!=-1);
				}
			double totalAmount1=0;
			double totalAmount2=0;
			double totalAmount3=0;
			double totalAmount4=0;
			double totalAmount5=0;
			for(int i=totalRstSlots.length-1;i>=0;i--)
				{
					totalAmount1+=totalRstSlots[i];
					totalAmount2+=submitRstSlots[i];
					totalAmount3+=waitRstSlots[i];
					totalAmount4+=commitRstSlots[i];
					totalAmount5+=abortRstSlots[i];
				}
			int amount1=0;
			int amount2=0;
			int amount3=0;
			int amount4=0;
			int amount5=0;
			for(int i=totalRstSlots.length-1;i>=0;i--)
				{
					amount1+=totalRstSlots[i];
					amount2+=submitRstSlots[i];
					amount3+=waitRstSlots[i];
					amount4+=commitRstSlots[i];
					amount5+=abortRstSlots[i];
					writerRsInSlot.write((i+1)*checkPointResponseTime_interval+" "
							+((double)amount1/totalAmount1)*100 + " "
							+((double)amount2/totalAmount2)*100 + " "
							+((double)amount3/totalAmount3)*100 + " "
							+((double)amount4/totalAmount4)*100 + " "
							+((double)amount5/totalAmount5)*100);
					writerRsInSlot.newLine();
				}
			
			double percentageOfCommit=((double)numCommitted/(double)total)*100;
			long interval=lastTimestamp-firstTimestamp-warmup_interval;
			double commitThroughput=(double)numCommitted/((double)interval/1000);
			double totalThroughput=(double)total/((double)interval/1000);
			double avgResponseTime=(double)totalCommittedLatency/(double)numCommitted;
			double avgSubmitResponseTime=(double)totalCommittedSubmitLatency/(double)numCommitted;
			double avgWaitResponseTime=(double)totalCommittedWaitLatency/(double)numCommitted;
			double avgCommitResponseTime=(double)totalCommittedCommitLatency/(double)numCommitted;
			double avgAbortResponseTime=(double)totalCommittedAbortLatency/(double)numCommitted;
			double avgRestartTimes=(double)totalRestartTimes/(double)numCommitted;
			
			writerOverview.write("#Experiment last for "+interval/1000 + " seconds");
			writerOverview.newLine();
			writerOverview.write("totalTransactionExecuted="+ total);
			writerOverview.newLine();
			writerOverview.write("totalTransactionCommitted="+ numCommitted);
			writerOverview.newLine();
			writerOverview.write("percentageOfCommit="+percentageOfCommit);
			writerOverview.newLine();
			writerOverview.write("commitThroughput="+commitThroughput);
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput);
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgSubmitResponseTime="+avgSubmitResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgWaitResponseTime="+avgWaitResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgCommitResponseTime="+avgCommitResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgAbortResponseTime="+avgAbortResponseTime);
			writerOverview.newLine();
			writerOverview.write("avgRestartTimes="+avgRestartTimes);
			writerOverview.newLine();
			int[] TPNodes=GlobalPara.sysConf.nodes.getServletNodes(null);
			for(int i=0;i<TPNodes.length;i++)
				{
				writerOverview.write("node"+ TPNodes[i] +"InvokedForCommit="+ totalTPNodeInvokedForCommitted[TPNodes[i]]);
				writerOverview.newLine();
				}
			for(int i=0;i<TPNodes.length;i++)
				{
				writerOverview.write("node"+ TPNodes[i] +"Invoked="+ totalTPNodeInvoked[TPNodes[i]]);
				writerOverview.newLine();
				}
			
			for(int i=0;i<TPNodes.length;i++)
				{
				writerConcurrencyInSlot.write("NumOfTransactionInvoke"+i+"TPNode="+ numOfTPNodeInslot[i]);
				writerOverview.newLine();
				}
			
			in.close();
			
			writerOverview.flush();
			writerRs.flush();
			writerThroughputbyend.flush();
			writerConcurrencyInSlot.flush();
			writerRsInSlot.flush();
			
			writerOverview.close();
			writerRs.close();
			writerThroughputbyend.close();
			writerConcurrencyInSlot.close();
			writerRsInSlot.close();
			
			System.out.println("ClientTransactionLog has been successfully analyzed");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public void analyzeLoadDataLog(String fileName)
	{
		int total=0;
		int[] rstSlots=new int[responseTime_slotAmount];
		for(int c:rstSlots)
			c=0;
		
		long firstTimestamp=0;
		long lastTimestamp=0;
		long totalCommittedLatency=0;
		
		long curTime=0;
		int curNum=0;
		String fileNameOverview=fileName+"_Overview";
		String fileNameRs=fileName+"_Rs";
		String fileNameThroughputbyend=fileName+"_Throughputbyend";
		String fileNameRsInSlot=fileName+"_RsInSlot";
		try {
			BufferedWriter writerOverview=new BufferedWriter(new FileWriter(outPath+fileNameOverview,false));
			BufferedWriter writerRs=new BufferedWriter(new FileWriter(outPath+fileNameRs,false));
			BufferedWriter writerThroughputbyend=new BufferedWriter(new FileWriter(outPath+fileNameThroughputbyend,false));
			BufferedWriter writerRsInSlot=new BufferedWriter(new FileWriter(outPath+fileNameRsInSlot,false));
		 
			
			FileInputStream fstream = new FileInputStream(inPath+fileName);
			BufferedReader in = new BufferedReader(new InputStreamReader(fstream));
			LoadDataLogItem log=new LoadDataLogItem();
			if(log.readFrom(in)!=-1)
				{
					firstTimestamp=log.timestamp;
					curTime=log.timestamp;
					do
					{
						
						//skip the load in warmup
						if(log.timestamp<(firstTimestamp+warmup_interval))
							continue;
						
						//skip the last time period which is doing ending
						if(log.timestamp>(firstTimestamp+ GlobalPara.testDuration -endup_interval))
							continue;
						
						lastTimestamp=log.timestamp;
						
						total++;
						totalCommittedLatency+=log.latency;
							
						writerRs.write((log.timestamp-firstTimestamp)
								+" "+ log.latency);
						writerRs.newLine();
						
						//
						if((log.timestamp-curTime)<throughput_interval)
							curNum++;
						else
							{
								do{
									writerThroughputbyend.write((log.timestamp-firstTimestamp)+" "+ curNum);
									writerThroughputbyend.newLine();
									curTime+=throughput_interval;
									curNum=0;
								}
								while((log.timestamp-curTime)>throughput_interval);
								curNum=1;
							}
						
						
						int index=(int)(log.latency/responseTime_interval);
						if(index>=responseTime_slotAmount)
							index=responseTime_slotAmount-1;
						rstSlots[index]++;
							
					}while(log.readFrom(in)!=-1);
				}
			double totalAmount=0;
			for(int i=rstSlots.length-1;i>=0;i--)
				{
					totalAmount+=rstSlots[i];
				}
			int amount=0;
			for(int i=rstSlots.length-1;i>=0;i--)
				{
					amount+=rstSlots[i];
					writerRsInSlot.write((i+1)*responseTime_interval+" "+((double)amount/totalAmount)*100);
					writerRsInSlot.newLine();
				}
			
			long interval=lastTimestamp-firstTimestamp-warmup_interval;
			double totalThroughput=(double)total/((double)interval/1000);
			double avgResponseTime=(double)totalCommittedLatency/(double)total;
			
			writerOverview.write("Experiment last for "+interval/1000 + " seconds");
			writerOverview.newLine();
			writerOverview.write("There are totally "+total +" transaction executed");
			writerOverview.newLine();
			writerOverview.write("totalThroughput="+totalThroughput + " TPS");
			writerOverview.newLine();
			writerOverview.write("averageResponseTime="+avgResponseTime + " ms");
			writerOverview.newLine();
			
			in.close();
			
			writerOverview.flush();
			writerRs.flush();
			writerThroughputbyend.flush();
			//writerThroughputbystart.flush();
			writerRsInSlot.flush();
			
			writerOverview.close();
			writerRs.close();
			writerThroughputbyend.close();
			//writerThroughputbystart.close();
			writerRsInSlot.close();
			
			System.out.println("LoadDataLog has been successfully analyzed");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
