package org.zhouw.stps.monitor;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.concurrent.locks.ReentrantLock;

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.conf.NodeConfig.role;

public class PerformanceMonitor {
	
	
	public PerformanceMonitor(String logPath)
	{
		if(logPath==null)
			{
			GlobalPara.logger.severe("Input logpath is null");
			return;
			}
		if(!logPath.endsWith("/"))
			logPath+="/";
		this.logPath=logPath;
		this.tpNodeLogFile=logPath+"tpnode"+GlobalPara.sysConf.getLocalNodeID()+".log";
		this.DSLogFile=logPath+"ds"+GlobalPara.sysConf.getLocalNodeID()+".log";
		this.checkPointLogFile=logPath+"checkpoint"+GlobalPara.sysConf.getLocalNodeID()+".log";
		this.loadDataLogFile=logPath+"loadData"+GlobalPara.sysConf.getLocalNodeID()+".log";
		this.idServiceLogFile=logPath+"idservice"+GlobalPara.sysConf.getLocalNodeID()+".log";
		this.clientTranLogFile=logPath+"clientTran"+GlobalPara.sysConf.getLocalNodeID()+".log";
		
		this.doDSLog=GlobalPara.sysConf.doDSLog;
		this.dotpNodeLog=GlobalPara.sysConf.dotpNodeLog;
		this.docheckPointLog=GlobalPara.sysConf.docheckPointLog;
		this.doloadDataLog=GlobalPara.sysConf.doloadDataLog;
		this.doIdServiceLog=GlobalPara.sysConf.doIdServiceLog;
		this.doClientTransactionLog=GlobalPara.sysConf.doClientTransactionLog;
		
		role[] a= GlobalPara.sysConf.nodes.getNodeRole(GlobalPara.sysConf.getLocalNodeID());
		if(a[0]==role.CLIENT)
			{
			//this.docheckPointLog=false;
			this.doIdServiceLog=false;
			//this.dotpNodeLog=false;
			//this.doloadDataLog=false;
			}
		else if(a[0]==role.IDService)
			{
			this.doDSLog=false;
			this.dotpNodeLog=false;
			this.docheckPointLog=false;
			this.doloadDataLog=false;
			this.doClientTransactionLog=false;
			this.doIdServiceLog=false;
			}
		else if(a[0]==role.SERVLET)
			{
			this.doDSLog=false;
			this.doClientTransactionLog=false;
			}
		
		try {		
		if(this.doDSLog)
			{
				this.dslog=new BufferedWriter(new FileWriter(this.DSLogFile,false), 8192);
			}
		if(this.dotpNodeLog)
			{
				this.tplog=new BufferedWriter(new FileWriter(this.tpNodeLogFile,false), 8192);
			}
		if(this.docheckPointLog)
			{
				this.cplog=new BufferedWriter(new FileWriter(this.checkPointLogFile,false), 8192);
			}
		if(this.doloadDataLog)
			{
				this.ldlog=new BufferedWriter(new FileWriter(this.loadDataLogFile,false), 8192);
			}
		if(this.doIdServiceLog)
			{
				this.idlog=new BufferedWriter(new FileWriter(this.idServiceLogFile,false), 8192);
			}
		if(this.doClientTransactionLog)
			{
				this.dctlog=new BufferedWriter(new FileWriter(this.clientTranLogFile,false), 8192);
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	
	public void closeAll()
	{
		try{
			if(this.doDSLog&&this.dslog!=null)
				{
					this.dslog.flush();
					this.dslog.close();
				}
			if(this.dotpNodeLog&&this.tplog!=null)
				{
					this.tplog.flush();
					this.tplog.close();
				}
			if(this.docheckPointLog&&this.cplog!=null)
				{
					this.cplog.flush();
					this.cplog.close();
				}
			if(this.doloadDataLog&&this.ldlog!=null)
				{
					this.ldlog.flush();
					this.ldlog.close();
				}
			if(this.doIdServiceLog&&this.idlog!=null)
				{
					this.idlog.flush();
					this.idlog.close();
				}
			if(this.doClientTransactionLog&&this.dctlog!=null)
				{
					this.dctlog.flush();
					this.dctlog.close();
				}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private String logPath;
	
	private String clientTranLogFile;
	private boolean doClientTransactionLog;
	private BufferedWriter dctlog;
	private ReentrantLock dctock=new ReentrantLock();
	
	private String DSLogFile;
	private boolean doDSLog;
	private BufferedWriter dslog;
	private ReentrantLock dslock=new ReentrantLock();
	
	private String tpNodeLogFile;
	private boolean dotpNodeLog;
	private BufferedWriter tplog;
	private ReentrantLock tplock=new ReentrantLock();
	
	private String checkPointLogFile;
	private boolean docheckPointLog;
	private BufferedWriter cplog;
	private ReentrantLock cplock=new ReentrantLock();
	
	private String loadDataLogFile;
	private boolean doloadDataLog;
	private BufferedWriter ldlog;
	private ReentrantLock ldlock=new ReentrantLock();
	
	private String idServiceLogFile;
	private boolean doIdServiceLog;
	private BufferedWriter idlog;
	private ReentrantLock idlock=new ReentrantLock();
	
	/**
	 * Client Transaction logging
	 * @param t
	 */
	public void writelnCTLog(String t)
	{
		if(!this.doClientTransactionLog)
			return;
		dctock.lock();
		try {
			this.dctlog.write(t);
			this.dctlog.newLine();
		} catch (IOException e) {
			e.printStackTrace();
		} finally
		{
			dctock.unlock();
		}
	}
	
	/**
	 * Data service invocation logging
	 * @param t
	 */
	public void writelnDSLog(String t)
	{
		if(!this.doDSLog)
			return;
		dslock.lock();
		try {
			this.dslog.write(t);
			this.dslog.newLine();
		} catch (IOException e) {
			e.printStackTrace();
		} finally
		{
			dslock.unlock();
		}
	}
	
	/**
	 * Transaction execution log at TPnode 
	 * @param t
	 */
	public void writelnTPLog(String t)
	{
		if(!this.dotpNodeLog)
			return;
		tplock.lock();
		try {
			this.tplog.write(t);
			this.tplog.newLine();
		} catch (IOException e) {
			e.printStackTrace();
		} finally
		{
			tplock.unlock();
		}
	}
	
	/**
	 * checkpoint log
	 * @param t
	 */
	public void writelnCPLog(String t)
	{
		if(!this.docheckPointLog)
			return;
		cplock.lock();
		try {
			this.cplog.write(t);
			this.cplog.newLine();
		} catch (IOException e) {
			e.printStackTrace();
		} finally
		{
			cplock.unlock();
		}
	}
	
	/**
	 * Load data logging
	 * @param t
	 */
	public void writelnLDLog(String t)
	{
		if(!this.doloadDataLog)
			return;
		ldlock.lock();
		try {
			this.ldlog.write(t);
			this.ldlog.newLine();
		} catch (IOException e) {
			e.printStackTrace();
		} finally
		{
			ldlock.unlock();
		}
	}
	
	/**
	 * ID service or timestamp service logging
	 * @param t
	 */
	public void writelnIDLog(String t)
	{
		if(!this.doIdServiceLog)
			return;
		idlock.lock();
		try {
			this.idlog.write(t);
			this.idlog.newLine();
		} catch (IOException e) {
			e.printStackTrace();
		} finally
		{
			idlock.unlock();
		}
	}
}
