package org.zhouw.stps.recovery;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Vector;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.zhouw.stps.tp.LocalTransManager;
import org.zhouw.stps.tp.UpdatesOfST;
import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.MsgBroker;
import org.zhouw.stps.PostMessageTask;
import org.zhouw.stps.msg.SubTransactionGroupProtos.*;

public class BatchBackupDataItemsTask implements Callable<Integer> {

	HashMap<Integer, Vector<UpdatesOfST>> dataBuffer;
	private int type;
	
	public Integer call() {
		int result=0;
		
		//GlobalPara.logger.warning("[DataRecover]Pushing data to to new backup LTMs with type"+result);
		
		Future<Integer>[] tasks=new Future[dataBuffer.size()];
		int count=0;
		for(int destID:dataBuffer.keySet()){
			GlobalPara.logger.warning("[DataRecover]Pushing data to to new backup LTM "+ destID);
			
			Vector<UpdatesOfST> data=dataBuffer.get(destID);
			
			if(data!=null)
				GlobalPara.logger.warning("[DataRecover]Pushing "+data.size()+" of data items to to new backup LTM "+ destID);
			else
				GlobalPara.logger.warning("[DataRecover]Pushing null to to new backup LTM "+ destID);
			
			NewBackupData.Builder msgBuilder=NewBackupData.newBuilder();
			for(int i=0;i<data.size();i++){
				UpdatesOfST item=data.get(i);
				UpdateOnDataRowMsg tempmsg=item.getMsg();
				msgBuilder.addUpdates(tempmsg);
			}
			NewBackupData msg=msgBuilder.build();
			
			//GlobalPara.logger.warning("[DataRecover]Start the task of pushing data to to new backup LTM "+ destID);
			PostMessageTask task=new PostMessageTask(destID, msg.toByteArray(), "BatchBackUpDataItemsService");
			try{
				tasks[count++]=MsgBroker.executor.submit(task);
			}catch(java.util.concurrent.RejectedExecutionException e){
				e.printStackTrace();
			}
		}
		
		for(Future<Integer> t:tasks){
			try {
				int r=t.get();
				if(r==-1){
					GlobalPara.logger.warning("Failures happen when pushing data to the new backups");
					result=-1;
				}
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
			GlobalPara.logger.warning("[info]Task of pushing batch data item backups returned.");
		}
		
		dataBuffer.clear();
		dataBuffer=null;
		
		if(type==1){//1: working DataPool, 2: backup data
			GlobalPara.ltmComp.reportWorkingDataComplete();
		}
		else{
			GlobalPara.ltmComp.reportDataBackupComplete();
		}
			
		return result;
	}

	public BatchBackupDataItemsTask(int type,
			HashMap<Integer, Vector<UpdatesOfST>> dataBuffer) {
		super();
		this.dataBuffer=dataBuffer;
		this.type=type;
	}

}
