package org.zhouw.stps.rotran;

import java.util.Hashtable;
import java.util.Vector;

import org.zhouw.stps.app.*;
import org.zhouw.stps.rotran.ROTranMsg.JoinEdgeMsg;
import org.zhouw.stps.rotran.ROTranMsg.JoiningTableMsg;
import org.zhouw.stps.rotran.ROTranMsg.QueryPlanMsg;
import org.zhouw.stps.*;

/**
 * No ring or circle of execution path is allowed.
 * Create two joiningTable with the same table name if necessary
 * @author zhouw
 *
 */
public class QueryPlan {

	public JoiningTable rootTable; //static
	public Hashtable<Integer,  JoiningTable> tables; //static <ID, table>
	public Vector<JoinEdge> edges; //static
	public boolean isValid;
	public final String dbName;
	
	public void inputEdge(JoinEdge e)
	{
		if(e!=null){
			this.edges.add(e);
			JoiningTable table=this.tables.get(e.tableIDA);
			if(table!=null)
				table.edges.add(e);
			else
				GlobalPara.logger.warning("Table ID not exist");
			JoiningTable table2=this.tables.get(e.tableIDB);
			if(table2!=null)
				table2.edges.add(e);
			else
				GlobalPara.logger.warning("Table ID not exist");
		}
	}
	
	public void transformIntoPhysicalTableName()
	{
		for(JoiningTable t: tables.values()){
			t.tableName=DataModel.getPhysicalTableName(dbName, t.tableName);
//			List<String> temp=new LinkedList<String>();
//			for(String s:t.projections){
//				temp.add(s.toUpperCase());
//			}
//			t.projections.clear();
//			t.projections.addAll(temp);
		}
	}
	
	public  QueryPlan(String dbName)
	{
		this.edges=new Vector<JoinEdge>();
		this.tables=new Hashtable<Integer,  JoiningTable>();
		this.rootTable=null;
		this.isValid=false;
		this.dbName=dbName;
	}
	
	public void clear()
	{
		this.rootTable=null;
		this.edges.clear();
		this.tables.clear();
		this.isValid=false;
	}
	
	private boolean analyzeProjections(String[] paras, DataModel dm, Vector<String> allProjections)
	{
		boolean result=true;
		int count=this.curOffset;
		if(allProjections==null)
			return false;
		
		while(count<paras.length)
		{
			if(paras[count].compareTo("")!=0)
			{
				if(paras[count].compareToIgnoreCase("select")==0)
				{;} 
				else if(paras[count].compareToIgnoreCase("from")==0)
				{
					count++;
					break;
				}
				else if(paras[count].compareToIgnoreCase(",")==0)
				{;}
				else
				{
					paras[count]=paras[count].replaceAll(",", "");
					allProjections.add(paras[count]);
				}
			}
			count++;
		}
		
		if(allProjections.size()==0)
		{
			this.errorMsg="None of projection columns are provided";
			return false;
		}
		
		this.curOffset=count;
		return result;
	}
	
	private boolean analyzeFromTables(String[] paras, DataModel dm)
	{
		boolean result=true;
		int count=this.curOffset;
		int tableID=0;
		//Between from and where
		while(count<paras.length)
		{
			if(paras[count].compareTo("")!=0)
			{
				if(paras[count].compareToIgnoreCase("where")==0)
				{
					count++;
					break;
				}
				else if(paras[count].compareToIgnoreCase(",")==0)
				{;}
				else
				{
					paras[count]=paras[count].replaceAll(",", "");
					if(!dm.hasTable(paras[count])) //invalid table name
					{
						this.errorMsg="Table \""+paras[count]+"\" in the query not defined in the \""+dm.getDBName()+"\" database";
						return false;
					}
					JoiningTable table=new JoiningTable(tableID, paras[count]);
					this.tables.put(tableID, table);
					tableID++;
				}
			}
			count++;
		}
		this.curOffset=count;
		if(this.tables.size()==0)
			return false;
		return result;
	}
	
	/**
	 * Allows "*" to indicate select all columns, table.* means all columns of this table
	 * 
	 * @param dm
	 * @param allProjections
	 * @return
	 */
	private boolean applyProjectionsToTables(DataModel dm, Vector<String> allProjections)
	{
		boolean result=true;
		for(String column:allProjections)
		{
			if(column.compareTo("*")==0)
			{
				if(allProjections.size()!=1)
				{
					this.errorMsg="Other projection columns are provided along with \"*\" ";
					return false;
				}
				for(JoiningTable table:this.tables.values())
				{
					table.projections.add("*");
				}
			}
			else
			{
				if(column.contains("."))
				{
					String ops[]=column.split("\\.");
					if(ops.length!=2)
					{
						this.errorMsg="Error column name in the projection: " + column;
						return false;
					}
					JoiningTable table=this.getTable(ops[0]);
					if(table==null)
					{
						this.errorMsg="Table name in the projection  not given in the query: " + column;
						return false;
					}
//					DataTable t=dm.getTable(ops[0]);
//					if(t==null)
//					{
//						this.errorMsg="The table name of the projection not defined in the selected database "+dm.name+" : " + column;
//						return false;
//					}
					if(ops[1].compareToIgnoreCase("*")==0)
					{
						table.projections.add(ops[1]);
					}
					else
					{
						DataTable dt=dm.getTable(table.tableName);
						if(dt.hasColumn(ops[1]))
							table.projections.add(ops[1]); //the parameter can be "*"
						else
						{
							this.errorMsg="The column "+ ops[1] +" is not contained in the table "+ dt.tableName;
							return false;
						}
					}
				}
				else
				{
					 Vector<DataTable> tables=dm.getTableByColumn(column);
					 if(tables.size()>0)
					 {
						 Vector<JoiningTable> jTables=new Vector<JoiningTable>();
						 for(DataTable dt: tables)
						 {
							 JoiningTable table=this.getTable(dt.tableName);
							 if(table!=null)
								 jTables.add(table);
						 }
							 
						 if(jTables.size()>1) {
							 this.errorMsg="Ambiguious column names defined in multiple tables of the query: " + column;
							 return false;
						 } else if(jTables.size()==0) {
								this.errorMsg="Column names not defined in any tables accessed by the query : " + column;
								return false;
						 } else { //size=1
							 jTables.get(0).projections.add(column);
						 }
					 }
					 else 
					 {
						this.errorMsg="Column names not defined in any tables in the database  "+dm.getDBName()+" : " + column;
						return false;
					 }
				}
			}
		}
		return result;
	}
	
	/**
	 * For each predicate, there must be blanks before and after the operator
	 * @param paras
	 * @param dm
	 * @return
	 */
	private boolean analyzePredicates(String[] paras, DataModel dm)
	{
		boolean result=true;
		int count=this.curOffset;
		
		while(count<paras.length)
		{
			if(paras[count].compareTo("")!=0)
			{
				if(paras[count].compareToIgnoreCase("and")==0)
				{;}
				else
				{
					String operand1=paras[count];
					
					count++;
					while(count<paras.length)
					{
						if(paras[count].compareTo("")!=0)
						{
							break;
						}
						count++;
					}
					
					if(count>=paras.length)
						break;
					
					if(paras[count].compareToIgnoreCase("=")!=0) //Currently only support "="
					{
						this.errorMsg="Incorrect predicate for \""+operand1+" "+paras[count]+"\"";
						return false;
					}
					
					count++;
					while(count<paras.length)
					{
						if(paras[count].compareTo("")!=0)
						{
							break;
						}
						count++;
					}
					
					if(count>=paras.length)
						break;
					String operand2=paras[count];
					
					if(!analyzeOneEquiCondition(operand1, operand2, dm))
						return false;
				}
			}
			count++;
		}
		this.curOffset=count;
		return result;
	}
	
	//used only for make plan
	private int curOffset;
	
	/**
	 * Incomplete query parser, only allow "Select, from, where, and, ="
	 * Currently not supporting self join and alias name
	 * In the section of predicates, there must be blanks before and after the operator "="
	 * @param paras
	 * @param dm
	 * @return false if the input query is not valid
	 */
	public boolean makePlan(String[] paras, DataModel dm)
	{
		this.clear();
		this.curOffset=0;
		
		
		//////////////////////////////////////////
		//Between select and from, record the projections
		Vector<String> allProjections=new Vector<String>();
		if(!this.analyzeProjections(paras, dm, allProjections))
			return false;
		
		if(!this.analyzeFromTables(paras, dm))
			return false;
		
		/////////////////////////////////////
		// assign projections to the tables
		if(!this.applyProjectionsToTables(dm, allProjections))
			return false;
		
		this.table=new DataTable[2];
		this.colName=new String[2];
		this.joinTable=new JoiningTable[2];
		this.isValue=new boolean[2];
		this.values=new String[2];
		
		if(!this.analyzePredicates(paras, dm))
			return false;
		
		this.transformIntoPhysicalTableName();
		
		if(this.rootTable==null)
			this.errorMsg="Unable to find a root table for the query";
		else
			this.isValid=true;
		return this.isValid;
	}
		
	private DataTable[] table=new DataTable[2];
	private String[] colName=new String[2];
	private JoiningTable[] joinTable=new JoiningTable[2];
	private boolean[] isValue=new boolean[2];
	private String[] values=new String[2];
	
	private boolean analyzeOperand(String operand, DataModel dm, int num)
	{
		boolean result=true;
		if(num<0||num>1)
		{
			this.errorMsg="Num given is out of range:"+num;
			return false;
		}
	
		if(!operand.startsWith("\""))
		{//it is not a value
			this.isValue[num]=false;
			if(operand.indexOf(".")!=-1)
			{
				String[] ops= operand.split("\\.");
				if(ops.length!=2) //NO "." in the table name or column name
				{
					this.errorMsg="Error format for operand of predicate:"+operand;
					return false;
				}
				table[num]=dm.getTable(ops[0]);
				if(table[num]==null||(!table[num].hasColumn(ops[1])))
				{
					this.errorMsg="Table name and column name defined in the operand not found:"+operand;
					return false;
				}
				joinTable[num]=this.getTable(table[num].tableName);
				if(joinTable[num]==null)
				{
					this.errorMsg="Table name and column name defined in the operand not in the accessed table of query:"+operand;
					return false;
				}
				colName[num]=ops[1];
			}
			else
			{
				Vector<DataTable> ts=dm.getTableByColumn(operand);
				if(ts.size()==0)
				{
					this.errorMsg="Column name defined in the operand not defined in the database:"+operand;
					return false;
				}
				
				colName[num]=operand;
				Vector<JoiningTable> jts=new Vector<JoiningTable>();
				for(DataTable dt:ts)
				{
					JoiningTable jt=this.getTable(dt.tableName);
					if(jt!=null)
						jts.add(jt);
				}
				
				if(jts.size()==1)
				{
					this.joinTable[num]=jts.get(0);
					table[num]=dm.getTable(this.joinTable[num].tableName);
				}
				else if(jts.size()==0)
				{
					this.errorMsg="Column name defined in the operand not defined in the database:"+operand;
					return false;
				} else {
					this.errorMsg="Ambiguious column names defined in the operand:"+operand;
					return false;
				}
			}
		}
		else
		{
			values[num]=operand.substring(1, operand.length()-1);
			this.isValue[num]=true;
		}
		return result;
	}
	
	private boolean handleJoinOperator(String operand1, String operand2)
	{
		boolean result=true;
		int col1Type;
		int col2Type;
		if(table[0].isPK(colName[0]))
		{
			col1Type=JoinEdge.PK;
		}
		else
		{
			ForeighKey fk=table[0].getSelfFK(colName[0]);
			if(fk!=null)
			{
				if(fk.BKTable==table[1]&&fk.BKName.compareTo(colName[1])==0)
				{
					col1Type=JoinEdge.FK;
				}
				else
				{
					errorMsg="Reference of Join attributes are not defined in the data model " +operand1+ " and " + operand1;
					return false;
				}
			}
			else
				col1Type=JoinEdge.SK;
		}
		
		if(table[1].isPK(colName[1]))
			col2Type=JoinEdge.PK;
		else
		{
			ForeighKey fk=table[1].getSelfFK(colName[1]);
			if(fk!=null)
			{
				if(fk.BKTable==table[0]&&fk.BKName.compareTo(colName[0])==0)
				{
					col2Type=JoinEdge.FK;
				}
				else
				{
					errorMsg="Reference of Join attributes are not defined in the data model " +operand1+ " and " + operand1;
					return false;
				}
			}
			else
				col2Type=JoinEdge.SK;
		}
		
		if(col1Type==JoinEdge.SK&&col2Type==JoinEdge.SK)
		{
			errorMsg="Reference of Join attributes are not defined in the data model " +operand1+ " and " + operand1;
			return false;
		}
			
		JoinEdge edge=new JoinEdge(joinTable[0].ID, colName[0], col1Type, joinTable[1].ID,colName[1], col2Type);
		this.edges.add(edge);
		joinTable[0].edges.add(edge);
		joinTable[1].edges.add(edge);
		
		return result;
	}
	
	/**
	 * 
	 * @param index The operand which is not a value.
	 * @return
	 */
	private boolean handlePredicate(int index)
	{
		boolean result=true;
		if(index!=1&&index!=0)
		{
			this.errorMsg="Index out of range for handling Predicate: "+index;
			return false;
		}
		int valueIndex=1-index;
		String value=this.values[valueIndex];
		if(table[index].isPK(colName[index]))
		{
			joinTable[index].allowedRows.add(value);
			if(this.rootTable==null)
			{
				this.rootTable=joinTable[index]; //decide the root table
				this.rootTable.result.candidateRows.add(value);
			}
		}
		else
		{
			String p=colName[index]+" = "+value;
			joinTable[index].predicates.add(p);	
		}

		return result;
	}
	
	private boolean analyzeOperator(String operand1, String operand2) //DataModel dm
	{
		boolean result=true;
		if(!this.isValue[0]&&!this.isValue[1]){ //this is a join operator
			this.handleJoinOperator(operand1, operand2);
		} else if(!this.isValue[0]&&this.isValue[1]) { //this is a predicate
			if(!handlePredicate(0))
				return false;
		} else if(operand1.startsWith("\"")&&!operand2.startsWith("\"")) {//this is a predicate
			if(!handlePredicate(1))
				return false;
		} else{
			this.errorMsg="\"Value = Value\" is not a valid predicate: " +operand1+ " and " + operand2;
			return false;
		}
		return result;
	}
	
	private boolean analyzeOneEquiCondition(String operand1, String operand2, DataModel dm)
	{
		boolean result=true;
		if(!analyzeOperand(operand1, dm, 0))
			return false;
		
		if(!analyzeOperand(operand2, dm, 1))
			return false;
		
		if(!this.analyzeOperator(operand1, operand2))
			return false;
		
		return result;
	}
	
	public String errorMsg="'";
	
	public void restart()
	{
		for(JoiningTable table:tables.values())
		{
			if(table==rootTable)
				table.result.clearAsRootTable();
			else
				table.result.clearAll();
		}
	}
	
	public void getParent(int tableID)
	{}
	
	public void getChilds(int tableID)
	{}
	
	public JoiningTable getTable(int ID)
	{
		return tables.get(ID);
	}
	
	public JoiningTable getTable(String tblName)
	{
		for(JoiningTable t: this.tables.values())
		{
			if(t.tableName.compareTo(tblName)==0)
				return t;
		}
		return null;
	}
	
	public JoinEdge getEdge(int tableID1, int tableID2)
	{
		for(JoinEdge e:edges)
		{
			if((tableID1==e.tableIDA&&tableID2==e.tableIDB)||(tableID2==e.tableIDA&&tableID1==e.tableIDB))
			{
				return e;
			}
		}
		return null;
	}
	
	public  QueryPlan(QueryPlanMsg msg)
	{
		this.tables=new Hashtable<Integer, JoiningTable>();
		this.edges=new Vector<JoinEdge>();
		this.dbName=msg.getDbName();
		for(JoiningTableMsg jMsg: msg.getTablesList())
		{
			this.tables.put(jMsg.getID(), new JoiningTable(jMsg));
		}
		
		for(JoinEdgeMsg eMsg: msg.getEdgesList())
		{
			JoinEdge e=new JoinEdge(eMsg);
			this.edges.add(e);
			JoiningTable table=this.tables.get(eMsg.getTableIDA());
			if(table!=null)
				table.edges.add(e);
			else
				GlobalPara.logger.warning("Inconsistent query plan");
			table=this.tables.get(eMsg.getTableIDB());
			if(table!=null)
				table.edges.add(e);
			else
				GlobalPara.logger.warning("Inconsistent query plan");
		}
		
		this.rootTable=this.tables.get(msg.getRootTableID());
		if(this.rootTable==null)
			GlobalPara.logger.warning("Invalid Root table ID in the query plan");
	}
	

	/**
	 * 
	 * @param plan
	 * @param curPositionTableID This will overwrite the value in param plan.
	 * @return
	 */
	public QueryPlanMsg constructQueryPlanMsg()
	{
		ROTranMsg.QueryPlanMsg.Builder qpBuilder=ROTranMsg.QueryPlanMsg.newBuilder();
		qpBuilder.setRootTableID(this.rootTable.ID).setDbName(this.dbName);
		for(Integer tableID: this.tables.keySet())
		{
			JoiningTable table=this.tables.get(tableID);
			qpBuilder.addTables(table.constructMsg());
		}
		for(JoinEdge e: this.edges)
		{
			qpBuilder.addEdges(e.constructMsg());
		}
		return qpBuilder.build();
	}
	
	
	

}
