package org.zhouw.stps.rotran;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.MsgBroker;
import org.zhouw.stps.TransCoodinator;
import org.zhouw.stps.Transaction;
import org.zhouw.stps.components.ClientToROGTM_TranMsg;
import org.zhouw.stps.rotran.ROTranMsg.*;

public class ROGlobalTM {
	
	private static Hashtable<Long, ROTranCoordinator> roTrans= new Hashtable<Long, ROTranCoordinator>();
	private static Hashtable<Long, ROTranCoordinator> tmpRoTrans= new Hashtable<Long, ROTranCoordinator>();
	
	public static void printDebugMessage()
	{
		GlobalPara.logger.warning("[Debug]Start print debug message for RO Transaction");
		String msg="[Debug]ROTransactions \n-----------------------------------\n\n";
		msg+="RO Transactions waiting for timestamps\n---------------------------\n";
		for(ROTranCoordinator tran: tmpRoTrans.values()){
			msg+="RO Tran from client " + tran.clientNodeID + " ClientsideTID " +tran.clientsideTID + "\n";
		}
		msg+="\n\n";
		msg+="RO Transactions waiting for votes\n---------------------------\n";
		for(ROTranCoordinator tran: roTrans.values()){
			msg+="RO Tran from client " + tran.clientNodeID + " ClientsideTID " +tran.clientsideTID + "\n";
			for(JoiningTable table:tran.plan.tables.values()){
				if(!table.result.checkedRows.equals(table.result.candidateRows)){
					msg+="Following rows of table " + table.tableName + " are not responded:\n";
					for(String rowkey: table.result.candidateRows){
						msg+=rowkey+", ";
					}
					msg+="\n";
				}
			}
			msg+="\n";
		}
		
		System.out.print(msg);
		System.out.flush();
		
	}
	
	//public static LinkedBlockingQueue<VoteROTranMsg> queue=new LinkedBlockingQueue<VoteROTranMsg>();
	
//	public static void submitVoteMsg(VoteROTranMsg msg)
//	{
//		boolean success=ROGlobalTM.queue.offer(msg);
//		while(!success)
//		{
//			GlobalPara.logger.warning("Unable to insert Votes into the queue for ROTransactions");
//			try {
//				Thread.sleep(100);
//			} catch (InterruptedException e) {
//				e.printStackTrace();
//			}
//			success=ROGlobalTM.queue.offer(msg);
//		}
//	}
	
//	public static void beginHandlingVoteMsg()
//	{
//		while(true)
//		{
//			try {
//				VoteROTranMsg msg=queue.take();
//				long tranID=msg.getTranID();
//				ROTranCoordinator tran=roTrans.get(tranID);;
//				if(tran!=null)
//				{
//					int result=tran.handlingVoteMsg(msg);
//					if(result==1)
//					{
//						roTrans.remove(tranID);
//					}
//				}
//			} catch (InterruptedException e) {
//				e.printStackTrace();
//			}
//		}
//	}
	
	public static int processROTran(long compID,QueryPlan qp, int clientNodeID, long clientsideTID)
	{
		MsgBroker.getNewTimestamp(GlobalPara.roComp,compID, false);
		ROTranCoordinator rotran=new ROTranCoordinator(clientNodeID, clientsideTID, compID, compID, qp);
		tmpRoTrans.put(compID, rotran);
		return 0;
	}
	
	public static void beginExecutingROTran(long compID,long timestamp)
	{
		ROTranCoordinator rotran = tmpRoTrans.get(compID);
		tmpRoTrans.remove(compID);
		if(rotran!=null){
			rotran.timestamp=timestamp;
			rotran.ID=timestamp;
			roTrans.put(timestamp, rotran);
			rotran.start();	
		} else {
			GlobalPara.logger.warning("ROTransaction with compID "+ compID +" not found in the waiting buffer");
		}
	}
	
	public static void handleRestart(long GID,long newTimestamp)
	{
		ROTranCoordinator rotran = roTrans.get(GID);
		if(rotran!=null){
			rotran.restart(newTimestamp);
		} else {
			GlobalPara.logger.warning("ROTransaction with GID "+ GID +" not found in the waiting buffer");
		}
	}
	
	public static void handleVoteMsg(VoteROTranMsg msg){
		long tranID=msg.getTranID();
		ROTranCoordinator tran=roTrans.get(tranID);
		if(tran!=null){
			int result=tran.handlingVoteMsg(msg);
			if(result==1){
				roTrans.remove(tranID);
			}
		} else {
			;//GlobalPara.logger.info("[info]The ROCoordinator "+tranID+" is not found in the buffer, can be caused by delayed outdated vote message due to restart");
		}
	}
	
	public static QueryPlan planROTransaction(String query, String databaseName)
	{
		QueryPlan qp=QueryParser.parse(query, databaseName);
		return qp;
	}
	
	public static void abortROTranForRecovery(int clientNodeID, long clientsideTID)
	{
		ROTranResultMsg.Builder returnMsg=ROTranResultMsg.newBuilder()
		.setClientsideID(clientsideTID);
		returnMsg.setTranState(3);
		returnMsg.setErrMsg("Abort the transaction because system is in the recovery process ");
		MsgBroker.returnROTran(clientNodeID, returnMsg.build());
	}
	
	public static void abortAllOnGoingROTrans()
	{
		for(ROTranCoordinator rotran: ROGlobalTM.roTrans.values()){
			abortROTranForRecovery(rotran.clientNodeID,rotran.clientsideTID);
		}
		ROGlobalTM.roTrans.clear();
		for(ROTranCoordinator rotran: ROGlobalTM.tmpRoTrans.values()){
			abortROTranForRecovery(rotran.clientNodeID,rotran.clientsideTID);
		}
		ROGlobalTM.tmpRoTrans.clear();
	}
}
