package org.zhouw.stps.rotran;

import java.util.HashMap;
import java.util.Hashtable;

import org.zhouw.stps.SubTransaction;
import org.zhouw.stps.TranState;
import org.zhouw.stps.app.DataModel;
import org.zhouw.stps.app.DataTable;
import org.zhouw.stps.rotran.ROTranMsg.*;
import org.zhouw.stps.*;

public class ROSubTran extends SubTransaction {

	/////////////////////////////////
	// Input
	public int curTableID;
	public ROSubTranGroup parentROTranGroup;
		
	/////////////////////////////////
	// Output
	public boolean isThisRowValid;
	public DataItemMsg dataOutputForROTran;
	public volatile boolean isExecuted;
	
	public ROSubTran(String strDBName, String strTblName, String strRowKey, int curTableID, 
			ROSubTranGroup parentGroup)
	{
		super("ROTransaction", strDBName, strTblName, strRowKey, null, false,
				false, true);
		this.parentROTranGroup=parentGroup;
		this.curTableID=curTableID;
		this.isThisRowValid=false;
		this.isExecuted=false;
	}
	
	public TranState run()
	{
		this.isThisRowValid=false;
		TranState result=TranState.COMMIT; //always commit
		try{
			if(super.run()==TranState.ABORT) //The row is not founded
			{
				this.isExecuted=true;
				return result;
			}
			HashMap<String, String> row= this.rootEntry.getCommittedData();
			
			if(checkWithPredicates(row))
				{
				this.isThisRowValid=true;
				this.dataOutputForROTran=generateOutput(row);
				}

		}catch(java.lang.Exception e)
		{
			e.printStackTrace();
		}
		
		this.isExecuted=true;
		return result;
	}

	/**
	 * Column1=? AND Column2=? AND ...
	 * @param data
	 * @return
	 */
	private boolean checkWithPredicates(HashMap<String, String> data)
	{
		boolean result=true;
		JoiningTable table=parentROTranGroup.plan.getTable(this.curTableID);
		for(String condition: table.predicates)
		{
			String[] operands=condition.split("=");
			if(operands.length!=2) //invalid predicates
				continue;
			operands[0]=operands[0].trim();
			operands[1]=operands[1].trim();
			String value=data.get(operands[0]);
			if(value==null||value.compareTo(operands[1])!=0)
			{
				result=false;
				break;
			}
		}
		return result;
	}
	
	/**
	 * Contains only the attributes in the projection
	 * @return
	 */
	private DataItemMsg generateOutput(HashMap<String, String> data)
	{
		DataItemMsg.Builder msgBuilder=DataItemMsg.newBuilder();
		JoiningTable table=parentROTranGroup.plan.getTable(this.curTableID);
		msgBuilder.setPK(this.strRowKey);
		
		DataModel dm=DataModel.getDataModel(this.parentROTranGroup.plan.dbName);
		DataTable tableSchema=dm.getTableByPhysicalName(this.strTblName);
		String pkName=tableSchema.primaryKeyName;
		if(table.projections.size()==1&&table.projections.get(0).compareTo("*")==0)
		{
			AttrValPairMsg.Builder pkAttr=AttrValPairMsg.newBuilder()
			.setKey(pkName).setValue(this.strRowKey);
			msgBuilder.addData(pkAttr);
			for(String column:data.keySet())
			{
				String value=data.get(column);
				if(value!=null)
				{
					AttrValPairMsg.Builder dataItem=AttrValPairMsg.newBuilder()
					.setKey(column).setValue(value);
					msgBuilder.addData(dataItem);
				}
			}
		}
		else
		{			
			for(String column:table.projections)
			{
				if(column.compareTo(pkName)==0){
					AttrValPairMsg.Builder dataItem=AttrValPairMsg.newBuilder()
					.setKey(column).setValue(this.strRowKey);
					msgBuilder.addData(dataItem);
				} else {
					String value=data.get(column);
					if(value!=null)
					{
						AttrValPairMsg.Builder dataItem=AttrValPairMsg.newBuilder()
						.setKey(column).setValue(value);
						msgBuilder.addData(dataItem);
					}
				}
			}
		}
		
		//////////////////////////////////////////////////////////////////
		// Obtaining expected rows to join further
		for(JoinEdge e:table.edges)
		{
			int otherTableID=e.getJoiningTable(this.curTableID);
			JoiningTable otherTable=this.parentROTranGroup.plan.getTable(otherTableID);
			
			if(otherTableID==-1)
			{
				GlobalPara.logger.warning("Incorrect edges of a joining table ");
				continue;
			}
			if(this.parentROTranGroup.inputMsg.hasParentTableID())
				if(otherTableID==this.parentROTranGroup.inputMsg.getParentTableID())
					continue;
						
			ExpectedRowsMsg.Builder expTablemsg=ExpectedRowsMsg.newBuilder();
			expTablemsg.setCurPositionTableID(otherTableID);
			
			String joinColumnCur=e.getJoinAttribute(this.curTableID);
			String joinColumnOther=e.getJoinAttribute(otherTableID);
			int joinColTypeCur=e.getJoinAttributeType(this.curTableID);
			int joinColTypeOther=e.getJoinAttributeType(otherTableID);
			
			if(joinColTypeCur==JoinEdge.PK&&joinColTypeOther==JoinEdge.PK) //PK-PK relationship
			{
				if(otherTable.allowedRows.size()>0&&(!otherTable.allowedRows.contains(this.strRowKey)))
					continue;
				if(!expTablemsg.getCurPositionRowPKList().contains(this.strRowKey))
					expTablemsg.addCurPositionRowPK(this.strRowKey);
			}
			else if(joinColTypeCur==JoinEdge.FK&&joinColTypeOther==JoinEdge.PK)
			{
				String pk=data.get(joinColumnCur);
				if(otherTable.allowedRows.size()>0&&(!otherTable.allowedRows.contains(pk)))
					continue;
				if(pk!=null)
				{
					if(!expTablemsg.getCurPositionRowPKList().contains(pk))
						expTablemsg.addCurPositionRowPK(pk);
				}
				else
				{
					GlobalPara.logger.warning("Foreign Key value of "+table.tableName+"."+ joinColumnCur +" is null, violated data integrity");
				}
			} else if(joinColTypeCur==JoinEdge.PK&&joinColTypeOther==JoinEdge.FK)
			{
				for(String column:data.keySet())
				{
					if(column.startsWith(DataModel.getReferenceRowsCacheColumnPrefix(
							DataModel.getLogicalTableName(this.parentROTranGroup.plan.dbName, 
									this.parentROTranGroup.plan.getTable(otherTableID).tableName), joinColumnOther)))
					{
						String value=data.get(column);
						if(otherTable.allowedRows.size()>0&&(!otherTable.allowedRows.contains(value)))
							continue;
						if(!expTablemsg.getCurPositionRowPKList().contains(value))
							expTablemsg.addCurPositionRowPK(value);
					}
				}
			}
			else
			{
				;//SK-FK, SK-PK is possible in the query, but the generated query plan only contains the PK-PK, PK-FK 
			}
			
			if(expTablemsg.getCurPositionRowPKCount()>0)
			{
				msgBuilder.addExpectedChildRows(expTablemsg);
			}
		}
		
		return msgBuilder.build();
	}
}
