package org.zhouw.stps.rotran;

import java.util.Hashtable;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.concurrent.locks.ReentrantLock;

import org.zhouw.stps.rotran.ROTranMsg.*;
import org.zhouw.stps.tp.LTMActions;
import org.zhouw.stps.tp.LocalTransManager;
import org.zhouw.stps.tp.ProcessGroupOfRow;
import org.zhouw.stps.tp.TPNode;
import org.zhouw.stps.tp.TableDataBuffer;
import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.MsgBroker;
import org.zhouw.stps.STGState;
import org.zhouw.stps.SubTransaction;
import org.zhouw.stps.SubTransactionGroup;
import org.zhouw.stps.TranState;

public class ROSubTranGroup {

	// message VoteROTranMsg
	// {
	// optional int32 parentTableID=1;
	// optional string parentRowPK=2;
	// repeated OneTableResultRowsMsg result=3;
	// required int32 vote=6; //1: Commit, 2: Restart, 3 abort if met faiure
	// required int64 timestamp=7;
	// required int64 tranID=8;
	// optional int32 version=9;
	// }

	// message SubROTranGroupMsg
	// {
	// required int32 coordinatorID=1;
	// optional int32 version=6;
	// required int64 timestamp=7;
	// required int64 tranID=8;
	// required QueryPlanMsg plan=2;
	// repeated ExpectedRowsMsg rows=3;
	// optional int32 parentTableID=4; //only required for non-root positions
	// optional string parentRowPK=5; //only required for non-root positions
	// }

	public QueryPlan plan;
	public QueryPlanMsg planMsg;
	public long timestamp;
	public SubROTranGroupMsg inputMsg;
	public LinkedList<ROSubTran> tranList;
	public Hashtable<Integer, OneTableResultRowsMsg.Builder> result; 

	public STGState state;
	private String[] accessedDataItems;
	
	public ReentrantLock lock=new ReentrantLock();
	
	public ROSubTranGroup() {
	}

	public ROSubTranGroup(SubROTranGroupMsg msg) {
		this.result=new Hashtable<Integer, OneTableResultRowsMsg.Builder>();
		this.planMsg = msg.getPlan();
		this.plan = new QueryPlan(this.planMsg);
		this.inputMsg = msg;
		this.timestamp = msg.getTimestamp();
		this.tranList = new LinkedList<ROSubTran>();
		for (ExpectedRowsMsg tranMsg : msg.getRowsList()) {
			int tid = tranMsg.getCurPositionTableID();
			String tname = this.plan.getTable(tid).tableName;
			for (String rowPK : tranMsg.getCurPositionRowPKList()) {
				ROSubTran subTran = new ROSubTran(this.plan.dbName, tname, rowPK, tid, this);
				this.tranList.add(subTran);
			}
		}
		this.state=STGState.START;
	}

//	/**
//	 * Abort of a RO transaction occurs when a given row is not existed, so the referential integrity is violated.
//	 * A second case is meeting machine failures. 
//	 * @param timestamp
//	 */
//	public void reportAbort(long timestamp, ROSubTran t) {
//
//		if(this.state==STGState.ABORT)
//			return;
//		this.state=STGState.ABORT;
//		this.tranList.remove(t);
//
//		
///**		VoteROTranMsg.Builder builder = VoteROTranMsg.newBuilder();
//		builder.setTimestamp(this.timestamp).setTranID(
//				this.inputMsg.getTranID()).setVote(Vote.Abort);
//		if (this.inputMsg.hasParentRowPK())
//			builder.setParentRowPK(this.inputMsg.getParentRowPK());
//		if (this.inputMsg.hasParentTableID())
//			builder.setParentTableID(this.inputMsg.getParentTableID());
//		MsgBroker.voteForROTrans(this.inputMsg.getCoordinatorID(), builder
//				.build());
//				**/
//	}
	
	public void reportCommit(long timestamp, ROSubTran t)
	{
		if(this.state==STGState.ABORT)//the STG restarted
			return;
	
		int tid=t.curTableID;
		OneTableResultRowsMsg.Builder tableResult=this.result.get(tid);
		if(tableResult==null) {
			tableResult=OneTableResultRowsMsg.newBuilder().setCurPositionTableID(tid);
			this.result.put(tid, tableResult);
		}
		
		if(t.isThisRowValid) {
			tableResult.addMatchedRows(t.dataOutputForROTran);
			joinChildTables(t);
		}
		else
			tableResult.addRejectedRows(t.strRowKey);
		
		this.tranList.remove(t);
		
		if(this.tranList.size()==0) {
			this.returnFinalVote();
			doTerminate();
		}
	}
	
	protected void joinChildTables(ROSubTran t)
	{
		LinkedHashMap<Integer, ROTranMsg.SubROTranGroupMsg.Builder> jobs;
		jobs=ROTranCoordinator.prepareROSubTranGroups(t.dataOutputForROTran.getExpectedChildRowsList()
					, planMsg, plan, 
					this.inputMsg.getTranID(), 
					this.inputMsg.getCoordinatorID()
					, timestamp, t.curTableID, t.strRowKey);
		for(int destID: jobs.keySet())
		{
			MsgBroker.submitROTran(destID, jobs.get(destID).build());
		}
	}
	
	protected void returnFinalVote()
	{
		this.state=STGState.COMMITTED;
		VoteROTranMsg.Builder builder = VoteROTranMsg.newBuilder();
		builder.setTimestamp(this.timestamp).setTranID(
				this.inputMsg.getTranID()).setVote(Vote.Commit);
		if (this.inputMsg.hasParentRowPK())
			builder.setParentRowPK(this.inputMsg.getParentRowPK());
		if (this.inputMsg.hasParentTableID())
			builder.setParentTableID(this.inputMsg.getParentTableID());
		for(OneTableResultRowsMsg.Builder oneTableBuilder:this.result.values())
		{
			builder.addResult(oneTableBuilder);
		}
		
		MsgBroker.voteForROTrans(this.inputMsg.getCoordinatorID(), builder
				.build());
	}
	

	public void reportRestart(long timestamp) {
		if(this.state==STGState.WAIT){
			this.state=STGState.ABORT;
			VoteROTranMsg.Builder builder = VoteROTranMsg.newBuilder();
			builder.setTimestamp(this.timestamp).setTranID(
					this.inputMsg.getTranID()).setVote(Vote.Restart);
			if (this.inputMsg.hasParentRowPK())
				builder.setParentRowPK(this.inputMsg.getParentRowPK());
			if (this.inputMsg.hasParentTableID())
				builder.setParentTableID(this.inputMsg.getParentTableID());
			MsgBroker.voteForROTrans(this.inputMsg.getCoordinatorID(), builder
					.build());
			doTerminate();
		} else {
			GlobalPara.logger.warning("[error]Unable to perform the reportRestart in the state:"+this.state);
		}
	}
	
	public void doStart()
	{
		this.state=STGState.WAIT;
		
//		if(GlobalPara.sysConf.doMemoryManagement)
//		{
//			accessedDataItems=new String[this.tranList.size()];
//			int count=0;
//			for(SubTransaction st:this.tranList)
//			{
//				accessedDataItems[count++]=TPNode.LTM.getDataItemID(st.strTblName, st.strRowKey);
//			}
//			//TPNode.MM.accessDataItems(this.accessedDataItems);
//		}
		SubTransaction[] trans=this.tranList.toArray(new SubTransaction[this.tranList.size()]);
		int addresult=LTMActions.AddGroupSTs(trans, false);
		if(addresult==1){
			LTMActions.ProcessGroupOfSTs(trans);
		} else if(addresult==0) {
			this.reportRestart(timestamp);
		} else {
			GlobalPara.logger.warning("[error]Exception happens during adding ROSubTran.");
		}
	}
	
	
	private void doTerminate()
	{
//		if(GlobalPara.sysConf.doMemoryManagement)
//		{
//			//TPNode.MM.unaccessDataItems(accessedDataItems);
//		}
	}
}
