package org.zhouw.stps.rotran;

import org.zhouw.stps.*;
import org.zhouw.stps.tp.*;
import org.zhouw.stps.membership.MembershipManager;
import org.zhouw.stps.monitor.ServerTransactionLogItem;
import org.zhouw.stps.msg.SubTransactionGroupProtos.TransactionResultSetMsg;
import org.zhouw.stps.rotran.ROTranMsg.*;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Set;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.LinkedHashMap;

public class ROTranCoordinator {
	public final int clientNodeID; 
	public final long clientsideTID; 
	public long ID;
	public long timestamp;
	public QueryPlan plan;
	
	public ServerTransactionLogItem logItem;
	private long startTime;
	private long endTime;
	
	//public static enum VOTE {COMMIT, RESTART };
	public static final int COMMIT=1;
	public static final int ABORT=2;
	
	public static LinkedBlockingQueue<ROVoteMsg> roTranQueue=new LinkedBlockingQueue<ROVoteMsg>();

	public ROTranCoordinator(int clientNodeID, long clientsideTID, long id,
			long timestamp, QueryPlan plan) {
		super();
		if(GlobalPara.sysConf.dotpNodeLog){
			startTime=System.currentTimeMillis();
			logItem=new ServerTransactionLogItem();
			logItem.isReadOnly=true;
			logItem.name="ROTran";
		}
		this.clientNodeID = clientNodeID;
		this.clientsideTID = clientsideTID;
		ID = id;
		this.timestamp = timestamp;
		this.plan = plan;
	}
	
	
	/**
	 * Given a start row, and a number of candidate joinable rows of other tables
	 * Construct the messages for the LTMs which contain these candidate rows.
	 * These candidate rows are also called expected rows. 
	 * At the very beginning, there is no such a row. So the expected rows are given by the query plan directly.
	 */
	public static LinkedHashMap<Integer, ROTranMsg.SubROTranGroupMsg.Builder> prepareROSubTranGroups(
			Iterable <? extends ExpectedRowsMsg> expectedRows, //can be multiple rows of multiple tables
			QueryPlanMsg planMsg, QueryPlan plan, //these two must reflect the same plan
			long tranID,
			int coordinatorID,
			long timestamp,
			Integer parentTableID, //null if at the very beginning
			String parentRowPK //null if at the very beginning
			)
	{
		LinkedHashMap<Integer, ROTranMsg.SubROTranGroupMsg.Builder> subROTranGroups=new LinkedHashMap<Integer, ROTranMsg.SubROTranGroupMsg.Builder>();
		if(expectedRows==null)
			return subROTranGroups;
		
		for(ExpectedRowsMsg tableMsg: expectedRows)
		{
			int tableID=tableMsg.getCurPositionTableID();
			JoiningTable table=plan.getTable(tableID);
			List<String> startRows=tableMsg.getCurPositionRowPKList();
			LinkedHashMap<Integer, Vector<String>> rows=new LinkedHashMap<Integer, Vector<String>>();
			
			for(String rowKey: startRows)
			{
				VirtualTPNode node=TPNode.getVirtualNode(table.tableName, rowKey);
				if(node!=null)
				{
					int destID=node.physicalID;
					SubROTranGroupMsg.Builder groupBuilder;
					if(!subROTranGroups.containsKey(destID))
					{
						groupBuilder= SubROTranGroupMsg.newBuilder();
						groupBuilder.setCoordinatorID(coordinatorID).setPlan(planMsg).
						setTimestamp(timestamp).setTranID(tranID).
						setVersion(MembershipManager.getVersionNumberWithoutLock());
						if(parentTableID!=null&&parentRowPK!=null)
							groupBuilder.setParentTableID(parentTableID).setParentRowPK(parentRowPK);
						subROTranGroups.put(destID, groupBuilder);
					}
					
					if(rows.containsKey(destID))
					{
						Vector<String> r=rows.get(destID);
						r.add(rowKey);
					}
					else
					{
						Vector<String> r=new Vector<String>();
						r.add(rowKey);
						rows.put(destID, r);
					}
				}
				else
				{
					GlobalPara.logger.warning("return virtual node null for " + table.tableName +"- "+ rowKey);
				}
			}
			
			for(Integer destID: rows.keySet())
			{
				SubROTranGroupMsg.Builder groupBuilder;
				groupBuilder=subROTranGroups.get(destID);
				Vector<String> r=rows.get(destID);
//				if(r==null)
//					r=new Vector<String>();
				ExpectedRowsMsg.Builder expRows=ExpectedRowsMsg.newBuilder();
				expRows.setCurPositionTableID(tableID).addAllCurPositionRowPK(r);
				groupBuilder.addRows(expRows);
			}
			rows=null;
		}
		
		return subROTranGroups;
	}
	
	//private Set<String> startRows;
	//private LinkedHashMap<Integer, ROTranMsg.SubROTranGroupMsg.Builder> jobs;
	private QueryPlanMsg planMsg;
	
	public int restart(long newTimestamp)
	{
		this.timestamp=newTimestamp;
		JoiningTable root = plan.rootTable;
		Set<String> startRows=root.result.candidateRows;
		
		if(GlobalPara.sysConf.dotpNodeLog){
			for(String r:startRows)
				this.logItem.accessedRecords.add(root.tableName+":"+r);
		}

		ExpectedRowsMsg.Builder rows=ExpectedRowsMsg.newBuilder();
		rows.setCurPositionTableID(root.ID).addAllCurPositionRowPK(startRows);
		Vector<ExpectedRowsMsg> expectedRows=new Vector<ExpectedRowsMsg>();
		expectedRows.add(rows.build());
		
		LinkedHashMap<Integer, ROTranMsg.SubROTranGroupMsg.Builder> jobs=
			ROTranCoordinator.prepareROSubTranGroups(expectedRows, planMsg, plan, this.ID, GlobalPara.sysConf.localNodeID, timestamp, null, null);
		for(int destID: jobs.keySet())
		{
			MsgBroker.submitROTran(destID, jobs.get(destID).build());
		}
		return 0;
	}
	
	public int start()
	{
		JoiningTable root = plan.rootTable;
		Set<String> startRows=root.result.candidateRows;
		this.planMsg=plan.constructQueryPlanMsg();
		
		this.count=startRows.size();
		
		if(GlobalPara.sysConf.dotpNodeLog){
			for(String r:startRows)
				this.logItem.accessedRecords.add(root.tableName+":"+r);
		}
		
		ExpectedRowsMsg.Builder rows=ExpectedRowsMsg.newBuilder();
		rows.setCurPositionTableID(root.ID).addAllCurPositionRowPK(startRows);
		Vector<ExpectedRowsMsg> expectedRows=new Vector<ExpectedRowsMsg>();
		expectedRows.add(rows.build());
		
		LinkedHashMap<Integer, ROTranMsg.SubROTranGroupMsg.Builder> jobs=
			ROTranCoordinator.prepareROSubTranGroups(expectedRows, planMsg, plan, this.ID, GlobalPara.sysConf.localNodeID, timestamp, null, null);
		
		for(int destID: jobs.keySet()){
			MsgBroker.submitROTran(destID, jobs.get(destID).build());
		}
		
		return 0;
	}
	
	/**
	 * 
	 * @return 1: transaction finished, 0: transaction still continue wait for more votes
	 */
	public int handlingVoteMsg(VoteROTranMsg msg)
	{
		int result=0;
		try{
			long tt=msg.getTimestamp();
			if(tt!=this.timestamp)
				return 0;
			int vote=msg.getVote();
			if(vote==Vote.Commit) //COMMIT
			{
				this.updateQueryResult(msg.getResultList());
				if(isCompleted())
				{
					ROTranResultMsg returnMsg=constructReturnMsg();
					MsgBroker.returnROTran(this.clientNodeID, returnMsg);
					
					if(GlobalPara.sysConf.dotpNodeLog){
						endTime=System.currentTimeMillis();
						logItem.totalLatency=endTime-startTime;
						logItem.isCommitted=true;
						logItem.timestamp=endTime;
						GlobalPara.monitor.writelnTPLog(this.logItem.writeToString());
					}
//					GlobalPara.logger.warning("RO Transaction ClientsideTID: " + returnMsg.getClientsideID()+ " finish execution return result to the client");
					
					result=1;
				}
			}
			else if(vote==Vote.Restart) //RESTART
			{
				this.plan.restart(); //discard all previous results
				this.timestamp++; //so the rest vote messages with the old timestamp will be rejected
				MsgBroker.getNewTimestamp(GlobalPara.roComp,this.ID,true);
				if(GlobalPara.sysConf.dotpNodeLog)
					logItem.accessedRecords.clear();
			}
			else if(vote==Vote.Abort) //ABORT
			{
				ROTranResultMsg returnMsg=constructAbortMsg();
				MsgBroker.returnROTran(this.clientNodeID, returnMsg);
				result=1;
				if(GlobalPara.sysConf.dotpNodeLog){
					endTime=System.currentTimeMillis();
					logItem.totalLatency=endTime-startTime;
					logItem.isCommitted=false;
					logItem.timestamp=endTime;
					GlobalPara.monitor.writelnTPLog(this.logItem.writeToString());
				}
			}
			else
			{
				GlobalPara.logger.warning("Bad vote result");
			}
		}catch(java.lang.NullPointerException e){
			e.printStackTrace();
			result=0;
		}
		return result;
	}
	
	private int count=0;
	
	private void updateQueryResult(List<OneTableResultRowsMsg> results)
	{
		for(OneTableResultRowsMsg tableResultMsg:results)
		{
			int curTableID=tableResultMsg.getCurPositionTableID();
			JoiningTable curTable=plan.getTable(curTableID);
			ResultOfTable data=curTable.result;
			data.checkedRows.addAll(tableResultMsg.getRejectedRowsList());
			data.rejectedRows.addAll(tableResultMsg.getRejectedRowsList());
			for(DataItemMsg itemMsg: tableResultMsg.getMatchedRowsList())
			{
				String pk=itemMsg.getPK();
				//multiple rows may map to the same row, so a row may be loaded more than once
				//so if a row already exists, then skip the update process
				if(!data.returnedRows.containsKey(pk))
				{
					data.checkedRows.add(pk);
					//the returned data row can be 
					ResultOfRow newRow=new ResultOfRow(itemMsg.getDataList(), curTable, pk, 
							itemMsg.getExpectedChildRowsList());
					data.returnedRows.put(pk, newRow);
					
					for(ExpectedRowsMsg childRowMsg:itemMsg.getExpectedChildRowsList())
					{
						int childTableID=childRowMsg.getCurPositionTableID();
						JoiningTable childTable=this.plan.getTable(childTableID);
						List<String> newCandidates=childRowMsg.getCurPositionRowPKList();
						childTable.result.candidateRows.addAll(newCandidates);
						count+=newCandidates.size();
						if(GlobalPara.sysConf.dotpNodeLog){
							for(String r:newCandidates)
								this.logItem.accessedRecords.add(childTable.tableName+":"+r);
						}
					}
				}
			}
		}
	}
	
	private boolean isCompleted()
	{
		boolean result=true;
		for(JoiningTable table:this.plan.tables.values())
		{
			if(!table.result.checkedRows.equals(table.result.candidateRows))
			{
				result=false;
				break;
			}
		}
		return result;
	}
	
	private ROTranResultMsg constructReturnMsg()
	{
		int order=0;
		ROTranResultMsg.Builder msgBuilder=ROTranResultMsg.newBuilder();
		msgBuilder.setTranState(ROTranCoordinator.COMMIT).setClientsideID(this.clientsideTID);
		Vector<DataRowMsg> result=new Vector<DataRowMsg>();
		for(String pk:this.plan.rootTable.result.returnedRows.keySet())
		{
			ResultOfRow row=this.plan.rootTable.result.returnedRows.get(pk);
			Vector<DataRowMsg.Builder> joinedRows= joinReturnedRows(row, true);
			for(DataRowMsg.Builder r :joinedRows)
			{
				result.add(r.setOrder(order++).build());
			}
		}
		msgBuilder.addAllResult(result);
		
		msgBuilder.setNumberOfAccessedItems(count);
		return msgBuilder.build();
	}
	
	//only inner join supported
	//now we assume no same column names across all tables
	//all columns remain their column name in the returned result
	private Vector<DataRowMsg.Builder> joinReturnedRows(ResultOfRow parentRow, boolean isRoot)
	{
		Vector<DataRowMsg.Builder> result=new Vector<DataRowMsg.Builder>();
		DataRowMsg.Builder baseRow=DataRowMsg.newBuilder();
		baseRow.addAllColumns(parentRow.data);
		
		LinkedHashMap<Integer, List<String>> expectedChildRows=parentRow.expectedChildRows;
		if(expectedChildRows==null) //it is the leaf node, add itself in
		{
			result.add(baseRow);
		}
		else
		{
			int childTableNum;
			if(isRoot)
				childTableNum = parentRow.table.edges.size();
			else
				childTableNum = parentRow.table.edges.size()-1;
			if(expectedChildRows.size()==childTableNum) //otherwise, return empty vector as one child table has no matched rows
			{
				boolean isAnyTableHasNoMatchRows=false;
				HashMap<Integer, Vector<DataRowMsg.Builder>> allChildRows=new HashMap<Integer, Vector<DataRowMsg.Builder>>(); 
				for(Integer childTableID:expectedChildRows.keySet())
				{
					Vector<DataRowMsg.Builder> rows=new Vector<DataRowMsg.Builder>();
					allChildRows.put(childTableID, rows);
					List<String> pks=expectedChildRows.get(childTableID);
					JoiningTable childTable=plan.getTable(childTableID);

					for(String pk:pks)
					{
						if(childTable.result.returnedRows.containsKey(pk))
						{
							ResultOfRow childRow=childTable.result.returnedRows.get(pk);
							Vector<DataRowMsg.Builder> childChildRows=joinReturnedRows(childRow,false);
							rows.addAll(childChildRows);
						}
					}
					if(rows.size()==0)
					{
						isAnyTableHasNoMatchRows=true;
						break;
					}
				}
				
				if(!isAnyTableHasNoMatchRows)
				{
					result.add(baseRow);
					Vector<DataRowMsg.Builder> temp=new Vector<DataRowMsg.Builder>();
					for(Vector<DataRowMsg.Builder> childRows:allChildRows.values())
					{
						for(DataRowMsg.Builder resultRow:result)
						{
							for(DataRowMsg.Builder crow:childRows)
							{
								DataRowMsg.Builder mergedRow=DataRowMsg.newBuilder();
								mergedRow.addAllColumns(resultRow.getColumnsList())
								.addAllColumns(crow.getColumnsList());
								temp.add(mergedRow);
							}
						}
						result.clear();
						result.addAll(temp);
						temp.clear();
					}
					temp=null;
				}
				allChildRows=null;
			}
		}
		return result;
	}
	
	private ROTranResultMsg constructAbortMsg()
	{
		ROTranResultMsg.Builder msgBuilder=ROTranResultMsg.newBuilder();
		msgBuilder.setTranState(ROTranCoordinator.ABORT).setClientsideID(this.clientsideTID);
		return msgBuilder.build();
	}
}
