package org.zhouw.stps.tp;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.*;

import org.zhouw.stps.SubTransaction;
import org.zhouw.stps.MsgBroker;
import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.SubTransactionGroup;
import org.zhouw.stps.app.DataTable;

public class BufferForCheckpoint {
	
	public final String[] tables;
	private boolean doImmediateCP; 
	
	public BufferForCheckpoint(String tables[], int initSnapShot)
	{
		this.lock=new ReentrantLock();
		this.buffer=new Hashtable<Long, Hashtable<String, ConcurrentLinkedQueue<UpdatedDataRow>>>();
		this.MaxTimestampSubmitted=0;
		this.minAcceptableTimestamp=initSnapShot*GlobalPara.intervalOfSnapshot;
		this.currentSnapShot=initSnapShot;
		counters=new Hashtable<Long, Long[]>();
		this.tables=tables;
		
		this.doImmediateCP=true;
		
//		if(GlobalPara.sysConf.DB.compareTo("SimpleDB")==0)
//			this.doImmediateCP=true;
//		else
//			this.doImmediateCP=false;
	}
	
	private Long[] initSnapshotCounter(long snapshotID)
	{
		Long[] c=new Long[2];
		c[0]=0L;
		c[1]=0L;
		counters.put(snapshotID, c);
		return c;
	}
	
	
//	private void setInsideBufferValue(long snapshotID, String tableName, String rowKey, String[][] dataToPut, String[] dataTodelete,DataRowContainer row )
//	{
//		Hashtable<String, ConcurrentLinkedQueue<UpdatedDataRow>> middle=buffer.get(snapshotID);
//		if(middle==null)
//			{
//			middle=new Hashtable<String,ConcurrentLinkedQueue<UpdatedDataRow>>();
//			for(String name:tables)
//				{
//				middle.put(name, new ConcurrentLinkedQueue<UpdatedDataRow>());
//				}
//			buffer.put(snapshotID, middle);
//			}
//		ConcurrentLinkedQueue<UpdatedDataRow> inside=middle.get(tableName);
//		if(inside==null)
//			{
//			middle.put(tableName, new ConcurrentLinkedQueue<UpdatedDataRow>());
//			GlobalPara.logger.warning("Table "+tableName+"not found in the given list "+ this.tables[0]+ " ..."); 
//			}
//		UpdatedDataRow value=new UpdatedDataRow(rowKey);
//		value.rowData=(Hashtable<String,String>)row.rowData.clone();//to support delete of one column family
//		value.update(dataToPut, dataTodelete);
//		inside.add(value);
//	}
	
	/**
	 * Three Tiered Hashtable,
	 * Outside Hashtable: key: snapshot id, value: Middle hashtable
	 * Middle hashtable: key: TableName, value: Inside hashtable
	 * Inside hashtable: key: rowKey, value: UpdatedDataRow
	 */
	//private Hashtable<Long, Hashtable<String, Hashtable<String, UpdatedDataRow>>> buffer;
	private Hashtable<Long, Hashtable<String, ConcurrentLinkedQueue<UpdatedDataRow>>> buffer;
		
	/**
	 * key: snapshot id
	 * value: long[0]: count of active transactions
	 * long[1]: latest submit time in this snapshot
	 */
	private Hashtable<Long, Long[]> counters;
	
	private long MaxTimestampSubmitted;
	private long minAcceptableTimestamp;
	private long currentSnapShot;
	
	private ReentrantLock lock;
	
	/**
	 * Change the counter and check if it needs to be restarted
	 * even if it needs to be restarted, it would still add it to the list.
	 * @param timestamp
	 * @return 1, needs to restart, 0: no need to restart
	 */
	public int doSubmit(long timestamp)
	{
		if(true)//FIXME currently always no need to restart
			return 0;
		
		//TPNode.LTM.lock.readLock().lock();
		this.lock.lock();
		try{
			long snapshotID=GlobalPara.timestampToSnapshot(timestamp);
			if(timestamp>MaxTimestampSubmitted)
				MaxTimestampSubmitted=timestamp;
			if(snapshotID<this.currentSnapShot)
				return 1;
			Long[] values = counters.get(snapshotID);
			if(values==null)
				values=initSnapshotCounter(snapshotID);
			values[0]++;
			values[1]=System.currentTimeMillis();
			//counters.put(snapshotID, values);
			if(timestamp<minAcceptableTimestamp)
				return 1;
			else
				return 0;
		}
		finally{
			this.lock.unlock();
			//TPNode.LTM.lock.readLock().unlock();
		}
	}
	
	/**
	 * Change the counter of two snapshot involved
	 * @param timestamp
	 * @return 1, still needs to restart, 0: no need to restart
	 */
	public int doRestart(long oldTimestamp, long newTimestamp)
	{
		if(true)
			return 0; //FIXME always no need to restart now
		//TPNode.LTM.lock.readLock().lock();
		this.lock.lock();
		try{
			long oldSnapshotID=GlobalPara.timestampToSnapshot(oldTimestamp);
			long newSnapshotID=GlobalPara.timestampToSnapshot(newTimestamp);
			if(newSnapshotID<this.currentSnapShot)
				return 1;
			if(oldSnapshotID!=newSnapshotID)
				{
				if(oldSnapshotID>=this.currentSnapShot)
					{
					Long[] values = counters.get(oldSnapshotID);
					if(values==null)//although it is not possible
						{
						values=initSnapshotCounter(oldSnapshotID);
						GlobalPara.logger.warning("oldTimestamp "+oldTimestamp + " With oldSnapshotID "+oldSnapshotID+ " not in the range of counter" );
						}
					values[0]--;
					//counters.put(oldSnapshotID, values);
					}
				if(newSnapshotID>=this.currentSnapShot)
					{
					Long[] values = counters.get(newSnapshotID);
					if(values==null)
						{
						values=initSnapshotCounter(newSnapshotID);
						}
					values[0]++;
					values[1]=System.currentTimeMillis();
					//counters.put(newSnapshotID, values);
					}
				}
			if(newTimestamp<minAcceptableTimestamp)
				return 1;
			else
				return 0;
		}
		finally{
			this.lock.unlock();
			//TPNode.LTM.lock.readLock().unlock();
		}
	}
	
	public void doAbort(long timestamp)
	{
		if(true)//FIXME currently always no need to restart
			return;
		//TPNode.LTM.lock.readLock().lock();
		this.lock.lock();
		try{
			long snapshotID=GlobalPara.timestampToSnapshot(timestamp);
//			if(snapshotID<this.currentSnapShot)
//				{
//				return;
//				}
			Long[] values = counters.get(snapshotID);
			if(values!=null)
				{
				values[0]--;
				//counters.put(snapshotID, values);
				}
			else
				{
				GlobalPara.logger.warning("Abort of Transaction with timestamp "+timestamp + " in snapshotID "+snapshotID+ " not in the range of counter" );
				}
		}
		finally{
			this.lock.unlock();
			//TPNode.LTM.lock.readLock().unlock();
		}
	}
	
	/**
	 * for read-only transactions, no data is required to insert
	 */
	public void doCommitted(long timestamp)
	{
		if(true)//FIXME currently always no need to restart
			return;
		//TPNode.LTM.lock.readLock().lock();
		this.lock.lock();
		try{
			long snapshotID=GlobalPara.timestampToSnapshot(timestamp);
			Long[] values = counters.get(snapshotID);
			if(values!=null)
				{
				values[0]--;
				//counters.put(snapshotID, values);
				}
			else
				{
				GlobalPara.logger.warning("Commit of read-only Transaction with timestamp "+timestamp + " in snapshotID "+snapshotID+ " not in the range of counter" );
				}
		}
		finally{
			this.lock.unlock();
			//TPNode.LTM.lock.readLock().unlock();
		}
	}
	

//	/**
//	 * Scan counter and decide if the snapshot has been constructed
//	 * @return >=0 the id of nextSI, -1  if current snapshot is not completed
//	 */
//	public long refreshSnapshot()
//	{
//		//TPNode.LTM.lock.readLock().lock();
//		//GlobalPara.logger.warning("Before obtaining lock of checkpoing buffer");
//		this.lock.lock();
//		//GlobalPara.logger.warning("After obtaining lock of checkpoing buffer");
//		long nextSI=-1;
//		try{
//			long threshold=GlobalPara.beginTimestampOfSnapshot(this.currentSnapShot+2);
//			if(this.MaxTimestampSubmitted>=threshold)
//			{
//				long n=this.currentSnapShot;
//				long m=GlobalPara.timestampToSnapshot(this.MaxTimestampSubmitted);
//				//GlobalPara.logger.warning("Start computing next Snapshot");
//				while(n<m)
//					{
//						Long[] values=this.counters.get(n);
//						if(values!=null&&values[0]!=0)//if counter for snapshot dose not exist, then pass through
//							break;
//						n++;
//					}
//				nextSI=n;
//				//GlobalPara.logger.warning("After computing next Snapshot");
//			}
//			if(nextSI>this.currentSnapShot)
//				{
//				this.currentSnapShot=nextSI;
//				this.minAcceptableTimestamp=GlobalPara.beginTimestampOfSnapshot(nextSI);
//				this.garabageCollectInCounters();
//				}
//			else
//				nextSI=-1;
//		}
//		finally{
//			//GlobalPara.logger.warning("Before releasing lock of checkpoing buffer");
//			this.lock.unlock();
//			//GlobalPara.logger.warning("After releasing lock of checkpoing buffer");
//			//TPNode.LTM.lock.readLock().unlock();
//		}
//		return nextSI;
//	}
	
//	/**
//	 * Delete counters of which snapshot id are less than current Snapshot ID 
//	 * @return
//	 */
//	private int garabageCollectInCounters()
//	{
//		int result=0;
////		TPNode.LTM.lock.readLock().lock();
////		this.lock.lock();
////		try{
//			Enumeration<Long> e=this.counters.keys();
//			while(e.hasMoreElements())
//			{
//				Long id=e.nextElement();
//				if(id<this.currentSnapShot)
//					this.counters.remove(id);
//			}
//			
////			Enumeration<Long> be=this.buffer.keys();
////			while(be.hasMoreElements())
////			{
////				Long id=be.nextElement();
////				if(id<this.currentSnapShot)
////					this.buffer.remove(id);
////			}
////		}
////		finally{
////			this.lock.unlock();
////			TPNode.LTM.lock.readLock().unlock();
////		}
//		return result;
//	}
	
//	/**
//	 * Delete data buffers of which snapshot id are less than current Snapshot ID
//	 * It is insynchronized, as hashtable is thread-safe and no transactions would access the data items
//	 * that belongs to the past snapshots 
//	 * @return
//	 */
//	public int garabageCollectInBuffers()
//	{
//		int result=0;
//
//		Enumeration<Long> be=this.buffer.keys();
//		while(be.hasMoreElements())
//		{
//			Long id=be.nextElement();
//			if(id<this.currentSnapShot)
//				this.buffer.remove(id);
//		}
//
//		return result;
//	}
	
//	/**
//	 * This method is invoked in the assumption that TPNode.LTM.lock.readLock() has been invoked
//	 * 
//	 * @return
//	 */
//	public Hashtable<Long, Hashtable<String, ConcurrentLinkedQueue<UpdatedDataRow>>> checkpointDataWithoutLock()
//	{
//		return buffer;
//	}
	
}
