package org.zhouw.stps.tp;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import org.zhouw.stps.GlobalPara;

public class DataRowContainer {

	public long getSize()
	{
		this.size=8+3*4+2*8+DataRowEntry.getSizeOfAHashMap(rowData)+DataRowEntry.getSizeOfAHashMap(backupRowData);
		return this.size;
	}
	
	public String[][] getRowData()
	{
		if(rowData==null)
			return null;
		
		LinkedHashMap<String, String> outputData;
		
		if(isDataBackuped){
			outputData=this.backupRowData;
		}else{
			outputData=this.rowData;
		}
		//GlobalPara.logger.warning("The size of rowData "+rowData.size());
		String[][] result=new String[outputData.size()][2];
		int count=0;
		//GlobalPara.logger.warning("Before counting");
		for(String key:outputData.keySet()){
			result[count][0]=key;
			result[count][1]=outputData.get(key);
			count++;
		}
		return result;
	}
	
	public LinkedHashMap<String, String> backupRowData;
	public LinkedHashMap<String, String> rowData;
	public long lastWriteTimestamp; //for checkpoint
	public long size;
	
	public boolean isDataBackuped;
	
	public DataRowContainer()
	{
		rowData=new LinkedHashMap<String, String>();
		backupRowData=new LinkedHashMap<String, String>();
		this.lastWriteTimestamp=-1;
		this.isDataBackuped=false;
		this.size=1;
	}
	
//	public DataRowContainer(int countColumn)
//	{
//		rowData=new Hashtable<String, String>(countColumn*2);
//		backupRowData=new Hashtable<String, String>(countColumn*2);
//	}
//	
	public String printInfo()
	{
		String result="RowData:{\n";
		String[] keys=rowData.keySet().toArray(new String[rowData.keySet().size()]);
		for(int i=0;i<keys.length;i++)
			result+=keys[i]+":" + this.rowData.get(keys[i])+"\n";
		result+="}\n";
		return result;
	}

	/**
	 * Restore the backup data item if there is any backup,
	 * if there is no backups, doing nothing
	 * 
	 * If multiple subtransactions update the same data item,
	 * this function may be invoked multiple times.
	 */
	public void restoreToBackup()
	{
		if(this.isDataBackuped){
			this.rowData.clear();
			this.rowData.putAll(this.backupRowData);
			this.backupRowData.clear();
			this.isDataBackuped=false;
		}
	}
	
	public void makeBackup()
	{
		if(!this.isDataBackuped){
			backupRowData.clear();
			backupRowData.putAll(rowData);
			this.isDataBackuped=true;
		}
	}
	
	public void clearBackup()
	{
		if(this.isDataBackuped){
			backupRowData.clear();
			this.isDataBackuped=false;
		}
	}
	/**
	 * It is not a final commit. Will be reversed, 
	 * so not updating the timestamp and make a backup.
	 * @param dataToPut
	 * @param dataTodelete
	 * @param timestamp
	 */
	public void updateForTemp(Iterable<Map.Entry<String,String>> dataToPut, Iterable<String> dataTodelete)
	{
		if(dataToPut!=null)
			for(Map.Entry<String,String> s: dataToPut)
				rowData.put(s.getKey(),s.getValue());
		if(dataTodelete!=null)
			for(String s: dataTodelete) {
//				if((s.length()>0)&&(s.indexOf(":")==(s.length()-1))) {//if s is a column or a column-family
//				Enumeration<String> e= rowData.keys();
//				while(e.hasMoreElements()) {
//					String col=e.nextElement();
//					if(col.startsWith(s))//check if the string belongs to the column family
//						rowData.remove(col);
//					}
//				}
//			else
				rowData.remove(s);
			}
	}
	
	public void updateForTemp(String[][] dataToPut, String[] dataTodelete)
	{
		if(dataToPut!=null)
			for(String[] s: dataToPut)
				rowData.put(s[0],s[1]);

		if(dataTodelete!=null)
			for(String s: dataTodelete) {
//				if((s.length()>0)&&(s.indexOf(":")==(s.length()-1))) {//if s is a column or a column-family
//						for(String col: rowData.keySet())
//						if(col.startsWith(s))//check if the string belongs to the column family
//							rowData.remove(col);
//					}
//				else
				rowData.remove(s);
			}
	}
	
	
	public void commit(String[][] dataToPut, String[] dataTodelete, long timestamp)
	{
		this.lastWriteTimestamp=timestamp;
		updateForTemp(dataToPut,dataTodelete);
	}
	
//	public void commit(Iterable<Map.Entry<String,String>> dataToPut, Iterable<String> dataTodelete, long timestamp)
//	{
//		this.lastWriteTimestamp=timestamp;
//		updateForTemp(dataToPut,dataTodelete);
//	}
}
