package org.zhouw.stps.tp;


import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.locks.*;
import org.zhouw.stps.*;
import org.zhouw.stps.app.DataTable;
import org.zhouw.stps.rotran.ROSubTran;

import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.Map;

public class DataRowEntry {
	
	protected long updateSize()
	{
		this.state.sizeOfRootDataRowEntry=8 + 7*4+8*2 + (8+2*4+(8+4*4)) + getSizeOfAString(rowKey)+ DataItemState.memorySize+committedData.getSize();
		this.state.isDataSizeChanged=false;
		return this.state.sizeOfRootDataRowEntry;
	}
	
	protected long updateSizeWithoutRecomputeDataItemSize(){
		this.state.sizeOfRootDataRowEntry=8 + 7*4+8*2 + (8+2*4+(8+4*4)) + getSizeOfAString(rowKey)+ DataItemState.memorySize+committedData.size;
		this.state.isDataSizeChanged=false;
		return this.state.sizeOfRootDataRowEntry;
	}
	
	public static long getSizeOfAString(String input)
	{
		return 8+16+2*input.length()+4+4+4+4;
	}
	
	public static long getSizeOfAHashMap(LinkedHashMap<String,String> input)
	{
		int len=input.size();
		long result=8+7*4+8+16+4*len;
		for(Map.Entry<String,String> e: input.entrySet()){ //the size of entries
			result+=8+4*6+getSizeOfAString(e.getKey())+getSizeOfAString(e.getValue());
		}
		
		return result;
	}
	
	private final LinkedList<SubTransaction> list=new LinkedList<SubTransaction>();
	public String rowKey;
	public TableDataBuffer dataTable;
	public DataItemState state; 
	
//	private LinkedHashMap<String,String> dataToPut;
//	private List<String> dataToDelete;
	
	/**
	 * When this data object first created, it is not in the queue of buffer manager until 
	 * the sub-transaction ends
	 */
	private boolean isInitiated;
	
	public long latestWriteTransaction;
	public long latestTransaction; //both read and write
	
	public DataRowContainer committedData;
	
	//1:loaded, 0: loading, -1: not loaded, -2: just created during commit recovered updates and not loaded
	public int isDataLoaded;
	
	protected static LinkedBlockingQueue<UpdatesOfST> updateQueueForReplicas=new LinkedBlockingQueue<UpdatesOfST>();
	protected static HashSet<String> inProgressReplications=new HashSet<String>();
	protected static HashMap<String,LinkedList<UpdatesOfST>> delayedReplications=new HashMap<String,LinkedList<UpdatesOfST>>();
	
	protected static void sendUpdateToReplicas(UpdatesOfST backupData)
	{
		VirtualTPNode vnode=TPNode.getVirtualNode(backupData.tableName, backupData.rowKey);
		Vector<VirtualTPNode> backupNodeList=TPNode.backupVirtualNodes.get(vnode.virtualID);
//		GlobalPara.logger.warning("Send backup to "+backupNodeList.size()+" backup LTMs");
		
		MsgBroker.DoBackUpCommitService(backupNodeList, backupData);
	}
	
	public static void removeReplicaTask(UpdatesOfST backupData)
	{
		backupData.isSucceeded=false;
		backupData.isToRemove=true;
		updateQueueForReplicas.offer(backupData);
	}
	
	public static void reportRemoveReplicaSucceed(UpdatesOfST backupData)
	{
		backupData.isSucceeded=true;
		updateQueueForReplicas.offer(backupData);
	}
	
	public static void newUpdateToReplicaTask(UpdatesOfST backupData)
	{
		backupData.isSucceeded=false;
		updateQueueForReplicas.offer(backupData);
	}
	
	public static void reportSendUpdateSucceed(UpdatesOfST backupData)
	{
		backupData.isSucceeded=true;
		updateQueueForReplicas.offer(backupData);
	}
	
	
//	public static void clearBackupQueue()
//	{
//		updateQueueForReplicas.clear();
//		inProgressReplications.clear();
//		delayedReplications.clear();
//	}
	
	
	private static void printDebugForPushUpdateQueue()
	{
		String result="\n\nThe queue of pushing data to backups\n-----------\n ";
		
		for(String key:inProgressReplications){
			result+="In Progresss: " +key + "\n";
			LinkedList<UpdatesOfST> list=delayedReplications.get(key);
			if(list!=null&&list.size()>0){
				result+="Delayed updates for key "+key+" from STGs: ";
				for(UpdatesOfST u:list){
					if(u.tran!=null)
						result+=u.tran.parentGroup.key+",";
					else
						result+=null+",";
				}
			} 
			result+="\n";
			delayedReplications.remove(key);
		}
		
		result+="\nDelayed backup updates without in progress key\n";
		if(delayedReplications.size()>0){
			for(String key: delayedReplications.keySet()){
				result+="For key "+key +":";
				LinkedList<UpdatesOfST> list=delayedReplications.get(key);
				if(list!=null||list.size()>0){
					for(UpdatesOfST u:list){
						result+=u.tran.parentGroup.key+",";
					}
				} else {
					result+="null";
				}
				result+="\n";
			}
		} else {
			result+="null\n";
		}
		
		System.out.println(result);
		System.out.flush();
		
	}
	
	/**
	 * Send updates  to the replicas, guarantee that updates to the same data items 
	 * are sent in FIFO.
	 * @return
	 */
	public static int managePushUpdatesToReplicas()
	{
		while(true) {
			try{
				UpdatesOfST backupData=null;
				try {
					backupData = updateQueueForReplicas.take();
				} catch (InterruptedException e1) {
					e1.printStackTrace();
				}
				
				if(TPNode.inRecovery){ //in case of recovery, stop the work immediately, because the system will perform data reorganization
					updateQueueForReplicas.clear();
					inProgressReplications.clear();
					delayedReplications.clear();
					continue;
				}
				
//				GlobalPara.logger.warning("Receiving new tasks for backup data items");
				
				if(backupData==null)
					continue;
				else {
					String key=backupData.getKey();
					if(key==null){
						printDebugForPushUpdateQueue();
						continue;
					}
					
					LinkedList<UpdatesOfST> delayed=delayedReplications.get(key);
					
					if(backupData.isSucceeded){
						if(delayed!=null){
							UpdatesOfST delayedUpdate=delayed.poll();
							if(delayedUpdate!=null){
								sendUpdateToReplicas(delayedUpdate);
							} else {
								inProgressReplications.remove(key);
								delayedReplications.remove(key);
							}
						} else {
							inProgressReplications.remove(key);
						}
					} else {
						if(inProgressReplications.contains(key)){
							if(delayed==null){
								delayed=new LinkedList<UpdatesOfST>();
								delayedReplications.put(key, delayed);
							}
							delayed.offer(backupData);
						}else{
							inProgressReplications.add(key);
							sendUpdateToReplicas(backupData);
						}
					}
				}
			}catch(java.lang.NullPointerException e){
				e.printStackTrace();
			}
		}
	}
	
//	public static int managePushUpdatesToReplicas()
//	{
//		while(true) {
//			try{
//				UpdatesOfST backupData=null;
//				try {
//					backupData = updateQueueForReplicas.take();
//				} catch (InterruptedException e1) {
//					e1.printStackTrace();
//				}
//				if(backupData==null)
//					continue;
//				else {
//					if(backupData.isSucceeded){
////						GlobalPara.ltmComp.
//						
//						backupData.tran.parentGroup.reportCommitDone(backupData.tran);
//					}
//					else {
//						sendUpdateToReplicas(backupData);
//					}
//				}
//			}catch(java.lang.NullPointerException e){
//				e.printStackTrace();
//			}
//		}
//	}

	public DataRowEntry(TableDataBuffer table, String rowKey)
	{
		committedData=new DataRowContainer();
		//updated=false;
		isDataLoaded=-1;
		latestWriteTransaction=-1;
		this.latestTransaction=-1;
		
		this.dataTable=table;
		this.rowKey=rowKey;
		this.state=new DataItemState(this, latestWriteTransaction);
		this.isInitiated=true;
	}
	
	public DataRowEntry(TableDataBuffer table, String rowKey, DataRowContainer dataItem)
	{
		this.committedData=dataItem;
		this.isDataLoaded=1;
		this.latestWriteTransaction=dataItem.lastWriteTimestamp;
		this.latestTransaction=dataItem.lastWriteTimestamp;
		this.dataTable=table;
		this.rowKey=rowKey;
		this.state=new DataItemState(this, latestWriteTransaction);
		if(GlobalPara.sysConf.doMemoryManagement){
			this.state.sizeOfRootDataRowEntry=this.updateSizeWithoutRecomputeDataItemSize();
			this.state.isEvictable=true;
			this.state.isHValueOutdate=true;
			this.state.latestInflation=TPNode.MM.inflation;
		}
		this.isInitiated=true;
	}
	
	public LinkedList<SubTransaction> getSTList()
	{
		return list;
	}
	
	
	
	public boolean isSTInOfList(SubTransaction st)
	{
		if(list!=null){
			return list.contains(st);
		}
		return false;
	}
	
	public boolean isSTHeadOfList(SubTransaction st)
	{
		if(list!=null&&list.size()>0){
			if(list.get(0)==st)
				return true;
		}
		return false;
	}
	
	public SubTransaction getHeadOfST()
	{
		if(list!=null&&list.size()>0){
			return list.get(0);
		}
		return null;
	}
	
	public void resetState()
	{
		this.list.clear();
		this.committedData.restoreToBackup();
	}
	
	/**
	 * Just commit the collected updates to the data item.
	 *
	 * @param timestamp The timestamp of the RW tran which do commit.
	 * @param 0: no updates need to be performed, 1: need to push updates to the replicas, -1 error that t cannot be committed 
	 */
	public int doCommit(SubTransaction t){
		
		if(t==null){
			GlobalPara.logger.warning("Input subtransaction to commit is null");
			return -1;
		}
		
		LinkedHashMap<String,String> dataToPut=new LinkedHashMap<String,String>();
		List<String> dataToDelete=new LinkedList<String>();
		
		int result =1;
		Vector<SubTransaction> transToBeComitted=new Vector<SubTransaction>();
		try{
			long timestamp=t.parentGroup.timestamp;
			
			if(!this.list.contains(t)){
				if(this.latestWriteTransaction!=t.parentGroup.timestamp){
					GlobalPara.logger.warning("[Error]The subtransaction is missing when do commit");
				} else {
					;//it is possible when multiple subtrans with the same timestamp on one data item
				}
				t.parentGroup.reportCommitDone(t);
				t.parentGroup.reportCheckpointDone(t);
				return 0; 
			} else {
				
				for(int i=0;i<this.list.size();i++){
					SubTransaction tInList=list.get(i);
					if(tInList.isReadOnlyGlobalTran)
						continue;
					if(tInList.parentGroup.timestamp==timestamp){ //multiple SubTrans, only RW tran can be in ASKCOMMIT
						if(tInList.DataToPut!=null){
							for(String[] s:tInList.DataToPut){
								dataToPut.put(s[0],s[1]); //FIXME now it is correct because the derived transactions only updates index entries, so will not update the same data columns
							}
						}
						if(tInList.DataToDelete!=null){
							for(String s:tInList.DataToDelete) {
								dataToDelete.add(s);
							}
						}
						transToBeComitted.add(tInList);
					} else {
						if(tInList.parentGroup.timestamp>timestamp)
							break;
					}
				}

				this.committedData.clearBackup();
				this.committedData.lastWriteTimestamp=timestamp;
				
				if(dataToPut.size()==0&&dataToDelete.size()==0){
					//the transactions have not change this data item, therefore the checkpoint, replicaion should be avoided.
					//however, they should still be deleted from the queue.
					//GlobalPara.logger.warning("[Info] not collecting any updates for the data item.");
					t.parentGroup.reportCommitDone(t);
					t.parentGroup.reportCheckpointDone(t);
					
				} else {
					
					String[][] dataPutArray;
					String[] dataDeleteArray;
					
					if(this.latestWriteTransaction==-1) { //if it is the new loaded data, the whole row should be replicated
						dataPutArray=this.committedData.getRowData();
						dataDeleteArray=null;
					} else {
						dataPutArray=new String[dataToPut.size()][2];
						dataDeleteArray=dataToDelete.toArray(new String[dataToDelete.size()]);
						int count=0;
						for(String key:dataToPut.keySet()) {
							dataPutArray[count][0]=key;
							dataPutArray[count][1]=dataToPut.get(key);
							count++;
						}	
					}
					
					this.latestTransaction=timestamp;
					this.latestWriteTransaction=timestamp;
					if(GlobalPara.sysConf.doMemoryManagement)
						this.state.updateBufferSizeChange();
					
					UpdatesOfST backupData=new UpdatesOfST(this.dataTable.tblName,this.rowKey,
							dataPutArray, dataDeleteArray, GlobalPara.sysConf.localNodeID,
							timestamp,t, this.committedData.size);
					
//					GlobalPara.logger.warning("Docommit push backup to replicas");
					newUpdateToReplicaTask(backupData);
					
					//long snapshotID=GlobalPara.timestampToSnapshot(timestamp);
					UpdatedDataRow value=new UpdatedDataRow(rowKey, dataPutArray, dataDeleteArray);
					GlobalPara.adapterComp.checkpointOneRow(timestamp, this.dataTable.tableSchema, rowKey, value, t);
					
//					TPNode.LTM.checkpointBuffer.doCommitDoneWithoutLock(timestamp, 
//							this.dataTable.tableSchema, this.rowKey,
//							dataPutArray,
//							dataDeleteArray);
				}
				this.list.removeAll(transToBeComitted);
				
//				if(GlobalPara.sysConf.doMemoryManagement){
//					this.state.transactionLeaves(transToBeComitted.size());
//				}
			}
		}catch(java.lang.NullPointerException e){
			e.printStackTrace();
		}
		return result;
	}
	
	public String printListSTS(Iterable <? extends SubTransaction> list)
	{
		String result="";
		int count=0;
		for(SubTransaction st:list) {
			if(!st.isReadOnlyGlobalTran){
				result+=printListSTS(st);
				count++;
			}
		}
		if(count>0){
			String str="=================\n";
			str+=result;
			str+="======================\n";
			return str;
		} else {
			return null;
		}
		
	}
	
	public String printListSTS(SubTransaction st)
	{
		if(st.isReadOnlyGlobalTran)
			return null;
		else
			return "Remove one ST in status "+ st.parentGroup.state.toString() +"\n" + st.printInfo() + "\n\n";
		
	}
//	private boolean isDataBackuped;
//	public boolean getIsDataBackuped()
//	{
//		return isDataBackuped;
//	}
	
	/**
	 * Try to process as many sub-transaction as possible for this row
	 * The process will generate a number of messages to the LTM components
	 * to vote commit or abort for transaction
	 * 
	 * @param timestamp
	 */
	public void execRest()
	{
		if(this.isDataLoaded!=1) //wait for the data loaded before executing the transactions
			return;
		try{
			if(list.size()==0)
				return;
			
			int count=0;
			SubTransaction t=null;
			Vector<SubTransaction> transToBeRemoved=new Vector<SubTransaction>();
			long executingRWTranTimestamp=-1;
			if(list.size()>0) {
				while(count<list.size()) {
					t=list.get(count++);
					long timestamp;
					STGState state;
					ROSubTran rot=null;
					if(!t.isReadOnlyGlobalTran) {
						state=t.parentGroup.state;
						timestamp=t.parentGroup.timestamp;
						//GlobalPara.logger.info("[EXEC] About to execute one ST of the STG "+t.parentGroup.key);
					} else {
						rot=(ROSubTran)t;
						state=rot.parentROTranGroup.state; //always execute RO transaction as soon as possible
						timestamp=rot.parentROTranGroup.timestamp;
					}
					
					if(state== STGState.ABORT||state== STGState.ASKRESTART||state== STGState.COMMITTED) {
//						transToBeRemoved.add(t);
//						if(state== STGState.ABORT)
//							numOfAborted++;
						GlobalPara.logger.warning("[ERROR]SubT with state"+state+" is present in the transaction queue of data item");
					} else if(state== STGState.ASKCOMMIT){ //multiple SubTrans, only RW tran can be in ASKCOMMIT 
						executingRWTranTimestamp=timestamp;
					} else if(state== STGState.WAIT) {
						TranState result=null;
						if(executingRWTranTimestamp!=-1&&executingRWTranTimestamp!=timestamp) //Allows RW SubTran with the same timestamp to execute together
							break;
						if(!t.isReadOnlyGlobalTran) {
							executingRWTranTimestamp=timestamp;
							if(!t.parentGroup.isTExeced(t)){
								result=t.run();
								if(result==TranState.ABORT) {
									t.parentGroup.reportAbort(t);
									TPNode.LTM.checkpointBuffer.doAbort(timestamp);
									transToBeRemoved.add(t);
									break;
								} else {
									t.postRun(); //postRun should share the same state with the run();
									if(!this.committedData.isDataBackuped){
										this.committedData.makeBackup();
									}
									this.committedData.updateForTemp(t.DataToPut,t.DataToDelete);
									t.parentGroup.reportCommit(timestamp,t);
								}
							}
						} else {
							result=t.run(); //even t.run returns ABORT, the globalROTran will not abort, it leads to a invalid row.
							rot.parentROTranGroup.reportCommit(timestamp, rot);
							if(timestamp>this.latestTransaction)
								this.latestTransaction=timestamp;
							transToBeRemoved.add(t);
							if(GlobalPara.sysConf.doMemoryManagement)
								this.state.transactionLeaves();
						}
					}
				}
				list.removeAll(transToBeRemoved);
				//if(GlobalPara.sysConf.doMemoryManagement)
				//	this.state.transactionLeaves(transToBeRemoved.size()); //for aborted rw transaction and completed ro transactions
			}
		}catch(java.lang.NullPointerException e){
			e.printStackTrace();
		}
	}

	
//	/**
//	 * Only can be invoked by RW transactions. Not RO transaction.
//	 * However, subTransactions of a RW transaction can be read-only
//	 * It is possible that a RW transaction are not updating any data items.  
//	 * @param t
//	 * @return
//	 */
//	public boolean commitDoneST(SubTransaction t)
//	{
//		this.lock.lock();
//		try{
//			if(!(t.DataToPut==null&&t.DataToDelete==null)){
//				this.latestWriteTransaction=t.parentGroup.timestamp;	
//			}
//			this.latestTransaction=t.parentGroup.timestamp;
//			if(list.remove(t)) {
//				this.transactionComplete();
//				return true;
//			}
//			return false;
//		}
//		finally{
//			this.lock.unlock();
//		}
//	}
	
	public boolean removeSubTransaction(SubTransaction t, boolean isRestart)
	{
		if(!this.list.contains(t)){
			return false;
		} else {

			if(!t.isReadOnlyGlobalTran){
				if(t.parentGroup.isTExeced(t)){
					this.committedData.restoreToBackup();
					t.parentGroup.cancelCommit(t);
				}
			} else {
				if(GlobalPara.sysConf.doMemoryManagement&&isRestart)
					this.state.transactionLeaves();
			}
			list.remove(t);
		}
		
		return true;
	}

	
//	public void restart(SubTransaction t)
//	{
//		list.remove(t);
//		String str=printListSTS(t);
//		add(t);
//	}

//	/**
//	 * Obtain all the locks of parentSTG which will be accessed in the process
//	 * @param lockGroup
//	 * @param t The new Sub-transaction to be added , null if no new ones, invoked by "exec".
//	 */
//	private void obtainAllNeededLocks(Vector<ReentrantLock> lockGroup, SubTransaction t)
//	{
//		int retryLockCount=0;
//		long executingSTTimestamp=-1;
//		if(lockGroup==null)
//			return;
//		long timestampOfT=-1;
//		while(true) {
//			boolean isSucceed=true; 
//			for(int i=-1;i<list.size();i++) {
//				SubTransaction s;
//				ReentrantLock lock;
//				if(i==-1) { //also lock SubTransaction t
//					if(t==null)
//						continue;
//					else
//						s=t;
//				} else
//					s=list.get(i);
//				
//				long curTimestamp;
//				STGState state;
//				if(!s.isReadOnlyGlobalTran) {
//					lock=s.parentGroup.lock;
//					curTimestamp=s.parentGroup.timestamp;
//					state=s.parentGroup.state;
//				} else {
//					ROSubTran ros=(ROSubTran) s;
//					lock=ros.parentROTranGroup.lock;
//					curTimestamp=ros.parentROTranGroup.timestamp;
//					state=ros.parentROTranGroup.state;
//				}
//				if(s==t)
//					timestampOfT=curTimestamp;
//				
//				if(lock.tryLock()) {
//					lockGroup.add(lock);
//					if(t!=null){
//						if(timestampOfT<curTimestamp)
//							break;//further STs in the queue will not be accessed in the "ADD a ST"
//					} else {
//						if(state==STGState.ASKCOMMIT||state==STGState.WAIT){
//							if(executingSTTimestamp==-1)
//								executingSTTimestamp=curTimestamp;
//							else {
//								if(curTimestamp>executingSTTimestamp)
//									break;//further STs in the queue will not be accessed in the "ADD a ST"
//							}
//						}
//					}
//				} else {
//					isSucceed=false;
//					for(ReentrantLock l:lockGroup) {
//						l.unlock();
//					}
//					lockGroup.clear();
//					break;
//				}
//			}
//			
//			//if(lockGroup.size()>0||(lockGroup.size()==0&&t==null)) {
//			if(isSucceed) {
//				if(retryLockCount>0)
//					GlobalPara.logger.warning("[Add SubTransaction] Succeed in obtaining all locks in "+retryLockCount +" times.");		
//				break;
//			} else {
//				retryLockCount++;
//				GlobalPara.logger.warning("[Add SubTransaction] Retry to to obtain all locks in "+retryLockCount +" times.");
//				try {
//					Thread.sleep(20);
//				} catch (InterruptedException e) {
//					e.printStackTrace();
//				}
//			}
//		}
//	}

	/**
	 * Two situations that a transaction may restart:
	 * Timestamp of AskCommit/committed transactions or latest accessed transaction timestamp 
	 * The input transaction timestamp is smaller
	 * For RO transaction, it is latest write transaction timestamp
	 *  
	 * @param 
	 * @return 1: success, 0: restart, -1: error
	 */
	public int add(SubTransaction t, boolean isRestart)
	{
		if(t==null){
			GlobalPara.logger.warning("[Error]Receiving null subtransaction to add");
			return -1;
		}
		
		if(this.list.contains(t)){
			GlobalPara.logger.warning("[Error]The to-be-added subtransaction is already in the queue");
			return -1;
		}
		
		Vector<SubTransaction> transToBeRemoved=new Vector<SubTransaction>();
			
		long timestampT;
		ROSubTran rot=null;
		STGState stateT;
		if(!t.isReadOnlyGlobalTran) {
			stateT=t.parentGroup.state;
			if(stateT!=STGState.WAIT){
				GlobalPara.logger.warning("[Error]The state of the to-be-added subtransaction is not right, it is: " +stateT );
				return -1;
			}
			timestampT=t.parentGroup.timestamp;
			if(timestampT<this.latestTransaction) {
				return 0;
			}
		} else {
			rot=(ROSubTran)t;
			stateT=rot.parentROTranGroup.state;
			if(stateT!=STGState.WAIT) {
				GlobalPara.logger.warning("[Error]The state of the to-be-added subtransaction is not right, it is: " +stateT );
				return -1;
			}
			timestampT=rot.parentROTranGroup.timestamp;
			if(timestampT<=this.latestWriteTransaction) { //Here compare to the latest write transaction
				if(timestampT==this.latestWriteTransaction){
					GlobalPara.logger.warning("[Error]A RO transaction has the same timestamp as a read-write transaction");
					return -1;
				} else
					return 0;
			}
		}

		int result=1;
		if(list.size()==0) {
			list.add(t);
		} else {
//			long executingRWTranTimestamp=-1;
			for(int i=0;i<list.size();i++) {
				SubTransaction s=list.get(i);
				long timestampS;
				ROSubTran ros=null;
				STGState stateS;
				if(!s.isReadOnlyGlobalTran) {
					stateS=s.parentGroup.state;
					timestampS=s.parentGroup.timestamp;
				} else {
					ros=(ROSubTran)s;
					timestampS=ros.parentROTranGroup.timestamp;
					stateS=ros.parentROTranGroup.state; //A RO transaction in the queue always with State W  				if(timestampT<timestampS) { //in case of equality, applying FIFO
				}
				
				if(timestampT==timestampS){
					t.isUpdateSameDataItem=true;
				} else if(timestampT<timestampS){
					if(stateS== STGState.ABORT || stateS== STGState.ASKRESTART)	{ //not remove them now, remove them when execRest
						transToBeRemoved.add(s);
						GlobalPara.logger.warning("[ERROR]Unexpected state "+stateS + " present in the queue");
						continue;
					} else if(stateS== STGState.ASKCOMMIT || stateS== STGState.COMMITTED) { 
						result=0;//an ROTran in the queue can only be in the status of WAIT
						break;
					} else if(stateS== STGState.WAIT) {
						if(s.isReadOnlyGlobalTran){
							list.add(i,t);
						} else {
							if(s.parentGroup.isTExeced(s))
								result=0;
							else
								list.add(i,t);
						}
						break;
					}
				}
				
				if(i==(list.size()-1)) {
					list.addLast(t);
					break;
				}
			}
		}
			
		list.removeAll(transToBeRemoved);
		return result;
	}
	
	public void reportCheckpointSuccess(long timestamp){
		this.state.checkpointedTimestamp=timestamp;
		checkIfEvictable();
	}
	
	public void reportReplicaSynchronized(long timestamp){
		this.state.replicaSynchronizedTimestamp=timestamp;
		checkIfEvictable();
	}
	
	public boolean checkIfEvictable(){
		boolean result=true;
		checkIfEvictable(true);
		return result;
	}
	
	public boolean checkIfEvictable(boolean isReportBufferSizeChange){
		boolean result=true;
		if(this.state.checkpointedTimestamp!=this.latestWriteTransaction)
			result=false;
		else if(this.state.replicaSynchronizedTimestamp!=this.latestWriteTransaction)
			result=false;
		else if(this.isDataLoaded!=1)
			result=false;
		else if(this.state.referencesOfSTs!=0)
			result=false;
		this.state.isEvictable=result;

		if(result){
			TPNode.MM.reportRowBecomeEvictable(this.state, isReportBufferSizeChange);
		}
		
		return result;
	}

//	/**
//	 * @return the updated
//	 */
//	public boolean isUpdated() {
//		return updated;
//	}

	/**
	 * @return the committedData
	 */
	public HashMap<String, String> getCommittedData() {
		return committedData.rowData;
	}

	/**
	 * @return the committedData
	 */
	public HashMap<String, String> getBackupedData() {
		return committedData.backupRowData;
	}
	
	/**
	 * Check if the committed data
	 * @return the isDataEmpty
	 */
	public boolean isDataEmpty() {
		if(this.committedData.rowData.size()==0)
			return true;
		else
			return false;
	}

}
