package org.zhouw.stps.tp;

import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Vector;
import java.util.Map.Entry;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import org.zhouw.stps.*;
import org.zhouw.stps.TransCoodinator.STATUS;
import org.zhouw.stps.msg.SubTransactionGroupProtos.*;
import org.zhouw.stps.recovery.AbortAllQueueingTransTask;
import org.zhouw.stps.rotran.*;

public class GlobalTransManager {

//public volatile int currentTransaction;
//	private static long counter=0;
//	private static long readCounter=0;
//	private static final java.util.concurrent.locks.ReentrantLock lock=new ReentrantLock();
//	
//	public static long getTimestamp(){
//		lock.lock();
//		try{
//			return counter++;	
//		}
//		finally{lock.unlock();}
//		}
	

	//public static CountDownLatch signal;
	
	public static void printDebugMsg()
	{
		String msg="[DEBUG]RW transactions\n-------------------------------\n\n";
		msg+="RW transactions waiting for timestamps\n-------------------------------\n";
		msg+="There are total "+ tranBufferWaitingForTimestamps.size()+ " transactions blocked waiting for timestamps\n";
		
		for(Transaction t:tranBufferWaitingForTimestamps.values()){
			msg+="RW Transaction from Client "+ t.clientNodeID + " ClientSideTID "+ t.clientsideTID+"\n";
		}
		//System.out.println(msg);
		
		msg+="RW transactions waiting for responses\n-------------------------------\n";
		msg+="There are total "+ tranTable.size()+ " transactions blocked waiting for responses\n";
		
		for(TransCoodinator tc:tranTable.values()){
			msg+=tc.getDebugMsg();
			msg+="\n\n";
		}
		System.out.println(msg);
		System.out.flush();
	}
	
	public static void endRecovery()
	{
		isInRecovery=false;
	}
	
	/**
	 * Recovering ongoing transactions
	 * @param failedNodes
	 * @param nextNodes
	 * @return
	 */
	public static int recoverOnGoingTransactions(Vector<Integer> failedTPNodes,
			RingMembership newMembership)
			//Vector<VirtualTPNode> failedNodes, Vector<VirtualTPNode> nextNodes)
	{
		isInRecovery=true;
		AbortAllQueueingTransTask task=new AbortAllQueueingTransTask(GlobalTransManager.tranBufferWaitingForTimestamps);
		MsgBroker.executor.submit(task);
		
		tranBufferWaitingForTimestamps=new HashMap<Long, Transaction>(100);
		List<Long> toDeleteList=new LinkedList<Long>();
		for(TransCoodinator tc:GlobalTransManager.tranTable.values()){
			int result=tc.recoverFromFailure(newMembership);
			if(result==-1) {
//				if(tc.isResultReturnedToClient){
					toDeleteList.add(tc.ID);
					tc.end();
//				}
			} 
//			else if(result==1||result==0){
//				tc.reportSuccessReturnResultToClient(); //make sure this transaction will end.
//			}  
		}
		
		for(Long id:toDeleteList){
			GlobalTransManager.tranTable.remove(id);
		}
		
		GlobalPara.logger.warning("[Info]There are "+GlobalTransManager.tranTable.size()+ "ongoing transactions to be recovered by commit");
		
		checkIfCommitDone();
		
		toDeleteList.clear();
		for(TransCoodinator tc:GlobalTransManager.replicaTranTable.values()){
			boolean isCommit=false;
			if(failedTPNodes.contains(tc.primaryID)){
				int[] backupNodes=TPNode.getBackupTPNodes(tc.primaryID);
				int newPrimaryNode=-1;
				for(int i=0;i<backupNodes.length;i++){
					if(!failedTPNodes.contains(backupNodes[i])){
						newPrimaryNode=backupNodes[i];
						break;
					}
				}
				if(newPrimaryNode==-1){
					GlobalPara.logger.warning("[error]Unable to find the new not failed primary node");
				}
				if(newPrimaryNode==GlobalPara.sysConf.localNodeID){
					int result=tc.recoverBackupedTC(newMembership);
					if(result==1) //if it is commit
						isCommit=true;
				}
			}
			if(!isCommit){
				toDeleteList.add(tc.ID);
			}
		}
		
		for(Long id:toDeleteList){
			GlobalTransManager.replicaTranTable.remove(id);
		}
		GlobalPara.logger.warning("[Info]There are "+GlobalTransManager.replicaTranTable.size()+ "backuped transactions to be recovered by commit");
		
		checkIfBackupedTCRecoverDone();
		return 0;
	}
	
	public static int updateBackupTC(BackupTransactionMsg msg)
	{
		TransCoodinator tc=new TransCoodinator(msg);
		replicaTranTable.put(tc.ID, tc);
		
		tc.setState(TransCoodinator.STATUS.COMMITTING);

		ResultSetMsg data=msg.getData();
		List<Parameter> list=data.getResultList();
		for(int i=0;i<list.size();i++){
			Parameter p=list.get(i);
			tc.output.add(new String[]{p.getKey(), p.getValue()});
		}

		List<UpdateOnDataRowMsg> list2=data.getUpdatesList();
		for(int i=0;i<list2.size();i++){
			tc.updates.add(new UpdatesOfST(list2.get(i)));
		}
		return 0;
	}
	
	
	public static int removeBackupTC(long id) {
		replicaTranTable.remove(id);
		return 0;
	}

	
	/**
	 * Get the timestamp for this transaction
	 * return 0 means successful
	 */
	public static void runTransaction(long curCompID, Transaction t)
	{
		if(TPNode.inRecovery) { //!TPNode.auditTranInRecovery(this.stgs))
			TransactionResultSetMsg.Builder msgReturnBuilder=TransactionResultSetMsg.newBuilder();
			msgReturnBuilder.setTranState("ABORT");
			msgReturnBuilder.setClientsideID(t.clientsideTID);
			TransactionResultSetMsg msgReturn=msgReturnBuilder.build();
			MsgBroker.DoReturnTransactionResult(t.clientNodeID, msgReturn, -1);
		}
		tranBufferWaitingForTimestamps.put(curCompID, t);
		MsgBroker.getNewTimestamp(GlobalPara.rwComp,curCompID,false);
	}
	
	/**
	 * Assign the timestmp to the transaction, start-->  waitvote, aborting
	 * Replicate the transaction coordinator
	 * @param compID
	 * @param timestamp
	 */
	public static void begin2PC(long compID, long timestamp)
	{
//		if(timestamp%10000==0){
//			printDebugMsg();
//		}
		
		Transaction t = tranBufferWaitingForTimestamps.get(compID);
		tranBufferWaitingForTimestamps.remove(compID);
		if(t!=null){
			TransCoodinator tc=new TransCoodinator(timestamp,t);
			tranTable.put(tc.ID, tc);
			tc.prepareStart();
			tc.submitTransaction();
		} else {
			GlobalPara.logger.warning("RWTransaction with compID "+ compID +" not found in the waiting buffer");
		}
	}
	
//	/**
//	 * Submit the subtransactions, CREATINGBACKUP -->waitvote
//	 * @param compID
//	 * @param timestamp
//	 */
//	public static void begin2PC(long GID, int[] succeededIDs)
//	{
//		TransCoodinator tc=tranTable.get(GID);
//		if(tc==null){
//			GlobalPara.logger.warning("The RW transactino coordinator "+GID+" is not found in the GT buffer");
//			return;
//		} else {
//			tc.submitTransaction();
//		}
//	}
		
	public static void restartGlobalTransaction(long GID, long newTimestamp)
	{
		TransCoodinator tc=tranTable.get(GID);
		if(tc==null){
			GlobalPara.logger.info("The RW transactino coordinator "+GID+" is not found in the GT buffer when doing restarting");
			return;
		} else {
			tc.doRestart(newTimestamp);
		}  
	}
	
	/**
	 * Submit the subtransactions, waitvote--> UPDATINGBACKUP, aborting, RESTARTING
	 * @param compID
	 * @param timestamp
	 */
	public static void handleVote(long GID, DTMsg msg)
	{
		TransCoodinator tc=tranTable.get(GID);
		if(tc==null){
			GlobalPara.logger.info("The RW transactino coordinator "+GID+" is not found in the GT buffer when getting a vote");
			return;
		} else {
			tc.handleMsg(msg);
		}                   
	}
	
//	/**
//	 * BAKCUPING_DERIVED --> Waitvote
//	 * @param GID
//	 */
//	public static void submitDerivedTransaction(long GID)
//	{
//		TransCoodinator tc=tranTable.get(GID);
//		if(tc==null){
//			GlobalPara.logger.warning("The RW transactino coordinator "+GID+" is not found in the GT buffer");
//			return;
//		} else {
//			tc.submitDerivedTransaction();
//		}
//	}
	
	
	/**
	 * UPDATINGBACKUP-->COMMITTING
	 * @param GID
	 */
	public static void startDoCommitPhase(long GID)
	{
		TransCoodinator tc=tranTable.get(GID);
		if(tc==null){
			GlobalPara.logger.warning("The RW transactino coordinator "+GID+" is not found in the GT buffer for do commit");
			return;
		} else {
			tc.doCommit();
		}   
	}
	
	
	
	public static void handleReportAbortDone(long GID, int stgIndex, long timestamp)
	{
		TransCoodinator tc=tranTable.get(GID);
		if(tc==null){
			//this is possible for a aborted derived subtransaction because of restarting and the global transaction has committed/aborted 
			GlobalPara.logger.info("The RW transactino coordinator "+GID+" is not found in the GT buffer");
			return;
		} else {
			tc.handleReportAbortDone(stgIndex,timestamp);
			if(tc.isResultReturnedToClient&&(tc.state==STATUS.ABORTED)) {
				//GlobalPara.logger.info("TC "+this.timestamp+"complete execution, exiting");
				//finally remove the TC backups after the transaction complete
				tc.end();
				tranTable.remove(GID);
			}
		}
	}
	
	public static void handleReportCommitDone(long GID, int stgIndex)
	{
		TransCoodinator tc=tranTable.get(GID);
		if(tc==null){
			GlobalPara.logger.warning("The RW transactino coordinator "+GID+" is not found in the GT buffer");
			return;
		} else {
			tc.handleReportCommitDone(stgIndex);
			if(tc.isResultReturnedToClient&&(tc.state==STATUS.COMMITTED)) {
				//GlobalPara.logger.info("TC "+this.timestamp+"complete execution, exiting");
				//finally remove the TC backups after the transaction complete
				tc.end();
				tranTable.remove(GID);
			}
		}
	}
	
	public static void handleReportSuccessReturnResultToClient(long GID)
	{
		TransCoodinator tc=tranTable.get(GID);
		if(tc==null){
			if(!isInRecovery){
				GlobalPara.logger.warning("The RW transactino coordinator "+GID+" is not found in the GT buffer");
			} else {
//				tc=replicaTranTable.get(GID);
//				if(tc!=null){
//					tc.reportSuccessReturnResultToClient();
//					if(tc.state==STATUS.ABORTED||tc.state==STATUS.COMMITTED) {
//						replicaTranTable.remove(GID);
//						checkIfBackupedTCRecoverDone();
//					}
//				} else {
//					GlobalPara.logger.warning("The RW TC "+GID+" is still not found in Backuped TC buffer");
//				}
			}
				
			return;
		} else {
			tc.reportSuccessReturnResultToClient();
			if(tc.state==STATUS.ABORTED||tc.state==STATUS.COMMITTED) {
				//GlobalPara.logger.info("TC "+this.timestamp+"complete execution, exiting");
				//finally remove the TC backups after the transaction complete
				tc.end();
				tranTable.remove(GID);
//				checkIfCommitDone();
			}
		}   
	}
	
//	public static void handleReportRecoverCommitDone(long GID)
//	{
//		TransCoodinator tc=tranTable.get(GID);
//		if(tc==null){
//			GlobalPara.logger.warning("The RW transactino coordinator "+GID+" is not found in the GT buffer");
//			return;
//		} else {
//			tc.end();
//			tranTable.remove(GID);
//		}
//		
//		if(isInRecovery){
//			checkIfCommitDone();
//		}
//	}
	
	public static void handleReportTCRecoveredCommitDone(long GID)
	{
		TransCoodinator tc=replicaTranTable.get(GID);
		if(tc==null){
			tc=tranTable.get(GID);
			if(tc==null)
				GlobalPara.logger.warning("The RW transactino coordinator "+GID+" is not found in the GT buffer");
			else {
				tc.state=STATUS.COMMITTED;
				tranTable.remove(GID);
				checkIfCommitDone();
			}
		} else {
			tc.state=STATUS.COMMITTED;
			replicaTranTable.remove(GID);
			checkIfBackupedTCRecoverDone();
		}
	}
	
	protected static void checkIfCommitDone()
	{
		if(tranTable.size()==0){
			GlobalPara.rwComp.reportOnGoingRWTranRecoverComplete();
		}
	}
	
	protected static void checkIfBackupedTCRecoverDone()
	{
		if(replicaTranTable.size()==0){
			GlobalPara.rwComp.reportBackupedRWTranRecoverComplete();
		}
	}

	public static HashMap<Long, Transaction> tranBufferWaitingForTimestamps=new HashMap<Long, Transaction>(100);
	public static HashMap<Long, TransCoodinator> tranTable=new HashMap<Long, TransCoodinator>(300);
	public static HashMap<Long, TransCoodinator> replicaTranTable=new HashMap<Long, TransCoodinator>(300);

	protected static boolean isInRecovery=false;
}
