package org.zhouw.stps.tp;

import org.zhouw.stps.*;
import org.zhouw.stps.membership.MembershipManager;
import org.zhouw.stps.msg.SubTransactionGroupProtos.NewBackupData;
import org.zhouw.stps.msg.SubTransactionGroupProtos.UpdateOnDataRowMsg;
import org.zhouw.stps.recovery.BatchBackupDataItemsTask;
import org.zhouw.stps.app.join.JoinAppDataModel;
import  org.zhouw.stps.app.*;
import  org.zhouw.stps.app.otpcw.*;
import org.zhouw.stps.components.RWGTMtoLTM_NewSTG;

import java.util.*;
import java.util.concurrent.*; 
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.TreeMap;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.locks.ReentrantLock;


public class LocalTransManager {
	
	//public enum BACKUPSTATUS{NORMAL, REPLICATEDATA, REPLICATEUPDATES}
	//public volatile BACKUPSTATUS isInReplicateNewNode;
	//public volatile VirtualTPNode[] from;
	
	public List<String> activeDatabases;
	
	public BufferForCheckpoint checkpointBuffer;
	
//	public String[] inputTables;
	
	public final ReentrantLock lockForRecoveredCommit=new ReentrantLock(true);
	//public static ExecutorService bgService= Executors.newFixedThreadPool(GlobalPara.TPNodeBGThreadPoolSize);
	//public static ExecutorService bgService= Executors.newCachedThreadPool();
	
	//public ReentrantLock lockOfHit=new ReentrantLock();
	public int Miss=0;
	public int hit=0;
	public int countN=0;
	
	//public ReentrantLock lockOfTotalBufferSize=new ReentrantLock();
	public int totalNumInDataPool=0;
	public int totalNumInBackupDataPool=0;
	
	public HashMap<String, TableDataBuffer> dataPool;
	public HashMap<String, SubTransactionGroup> STGPool;
	
	//<Key, Vector<Timestamp>>
	public HashMap<String, Vector<Long>> doAbortMsgBuffer;
	
	//<Key, newTimestamp>
	public HashMap<String,Long> doRestartMsgBuffer;
	
	public HashMap<String, List<SubTransactionGroup>> outdateSTGs;
	
	public String[] tables;
	
	/**
	 * <Primary Node ID, data pool> 
	 */
	public HashMap<Integer, BackUpDataPool> replicaDataPool;
	
	//sorted by the timestamp of UpdatesOfST
	public TreeMap<Long, Vector<UpdatesOfST>> arrivedUpdatesInRecoveryAsPrimary=new TreeMap<Long,Vector<UpdatesOfST>>();
	public TreeMap<Long,Vector<UpdatesOfST>> arrivedUpdatesInRecoveryAsBackup=new TreeMap<Long,Vector<UpdatesOfST>>();
	
	//public UpdateBufferOnSnapshot updateBuffer; 
	
	///////////////////////////////////////////////////////////
	//accumulate updates for sending to new node later
	//public ConcurrentHashMap<Integer,  Vector<UpdatesOfST>> recoveredCommit;
	
	public LocalTransManager()
	{
		activeDatabases=new LinkedList<String>();//TODO Be defined configuration file in the future
		activeDatabases.addAll(GlobalPara.sysConf.database);
		
//		for(String dbname:activeDatabases){
//			GlobalPara.logger.warning("Databasde " +dbname+ " is loaded into the system");
//		}
		
//		activeDatabases.add(JoinAppDataModel.name);	
//		activeDatabases.add(TPCWAPPDataModel.name);

		
		doAbortMsgBuffer=new HashMap<String, Vector<Long>>();
		doRestartMsgBuffer=new HashMap<String, Long>();
		outdateSTGs = new HashMap<String, List<SubTransactionGroup>>();
		
		dataPool=new HashMap<String, TableDataBuffer>();
		for(String dbName:activeDatabases){
			DataModel db=DataModel.getDataModel(dbName);
			if(db!=null){
				GlobalPara.logger.info("Databasde " +db.dbname+ " is loaded into the system");
				for(DataTable table: db.getAllTables()) {
					TableDataBuffer h=new TableDataBuffer(table);
					dataPool.put(h.tblName, h); //the table name is the phsical table name
					//GlobalPara.logger.warning("Table "+ h.tblName+ " is loaded");
				}
			} else
				GlobalPara.logger.warning("Database "+ dbName+ " not exist");
		}
		
		tables=new String[dataPool.size()];
		int count=0;
		for(TableDataBuffer table:dataPool.values()) {
			tables[count++]=table.tblName;
		}
		
		this.STGPool=new HashMap<String, SubTransactionGroup>(500);
		this.checkpointBuffer=new BufferForCheckpoint(tables,1);
		replicaDataPool=new HashMap<Integer, BackUpDataPool>();

		for(int nodeID:GlobalPara.sysConf.TPNodeSet)
		{
			BackUpDataPool pool=new BackUpDataPool(tables, nodeID);
			replicaDataPool.put(nodeID, pool);
		}
	}

//	private int moveBackupDataToNewNodeID(int oldTPNodeID, int newTPNodeID)
//	{
//		BackUpDataPool pool=this.replicaDataPool.get(newTPNodeID);
//		BackUpDataPool oldpool=this.replicaDataPool.get(oldTPNodeID);
//		if(pool==null)
//			{
//				oldpool.primaryNodeID=newTPNodeID;
//				this.replicaDataPool.remove(oldpool);
//				this.replicaDataPool.put(newTPNodeID, oldpool);
//			}
//		else
//			{
//				pool.merge(oldpool);
//			}
//		this.replicaDataPool.remove(oldTPNodeID);
//		return 0;
//	}
	
	public TableDataBuffer getTable(String physicalTableName)
	{
		return this.dataPool.get(physicalTableName);
	}
	
	/**
	 * Buffered the recovered updates
	 * 
	 * @param isPrimary Is this LTM host it in the working pool or backup pool
	 * @param data
	 * @return
	 */
	public int doRecoverDataToCommit(boolean isPrimary, UpdatesOfST[] datas)
	{
		int result=0;
		lockForRecoveredCommit.lock();
		try{
			if(datas==null||datas.length==0)
				return 0;
			TreeMap<Long,Vector<UpdatesOfST>> buffer;
			if(isPrimary){
				buffer=this.arrivedUpdatesInRecoveryAsPrimary;
			} else {
				buffer=this.arrivedUpdatesInRecoveryAsBackup;
			}
			for(UpdatesOfST data:datas){
				Vector<UpdatesOfST> updates=buffer.get(data.timestamp);
				if(updates==null){
					updates=new Vector<UpdatesOfST>();
					buffer.put(data.timestamp, updates);
				}
				if(data.dataRow!=null)
					updates.add(0, data);//the update in the head will be commit firstly
				else
					updates.add(data);
			}
		} finally{
			lockForRecoveredCommit.unlock();
		}
		
//		Vector<UpdatesOfST> input=new Vector<UpdatesOfST>();
//
//		/////////////////////////
//		///Update local backup data
//		
//		for(UpdatesOfST data:datas){
//			input.add(data);
//			BackUpDataPool pool=this.replicaDataPool.get(data.nodeID);
//			pool.update(data);
//		}
		
//		/////////////////////////
//		///Update the rest of still alive backup replicas right now
//		
//		VirtualTPNode primaryVNodeID=TPNode.getVirtualNodeWithoutLock(datas[0].tableName, datas[0].rowKey);
//		Vector<VirtualTPNode> replicas=TPNode.getRestOfAliveReplicasWithoutLock(primaryVNodeID);
//		
//		Future<Integer>[] returnTasks=new Future[replicas.size()];
//		int count=0;
//		for(VirtualTPNode node:replicas){
//			BatchBackupDataItemsTask task=new BatchBackupDataItemsTask(node.physicalID, input);
//			returnTasks[count++]=LocalTransManager.bgService.submit(task);
//		}
//		
//		for(Future<Integer> r: returnTasks)
//			{
//				try {
//					if(r.get()==-1)
//						result=-1;
//				} catch (InterruptedException e) {
//					e.printStackTrace();
//					result=-1;
//				} catch (ExecutionException e) {
//					e.printStackTrace();
//					result=-1;
//				}
//			}
		return result;
	}
	
	/**
	 * when transaction happens in primary Node commit, invoke following function to backup the updates
	 * @param primaryNodeID
	 * @param data
	 * @return
	 */
	public int doBackUp(UpdatesOfST data)
	{
		//GlobalPara.logger.warning("Receive one data backup for Primary LTM "+data.nodeID);
		BackUpDataPool pool=this.replicaDataPool.get(data.nodeID);
		if(pool!=null)
			pool.update(data, true);
		else
			GlobalPara.logger.warning("Backup data pool for LTM "+data.nodeID+ " is not found");
		return 0;
	}
	
	public int removeBackUp(UpdatesOfST data)
	{
		BackUpDataPool pool=this.replicaDataPool.get(data.nodeID);
		if(pool!=null){
			pool.remove(data.tableName, data.rowKey, data.timestamp, data.nodeID);
		} else
			GlobalPara.logger.warning("Backup Data pool not found for  TP node "+ data.nodeID);
		return 0;
	}
	
	/**
	 * multiple updates on multiple rows, do them in batch
	 * @param data
	 * @return
	 */
	public int bactchBackUp(Vector<UpdatesOfST> data)
	{
		GlobalPara.logger.warning("[DataRecover]Start adding "+data.size()+" data items as new backups");
		for(UpdatesOfST item:data){
			BackUpDataPool pool=this.replicaDataPool.get(item.nodeID);
			if(pool==null){
				pool=new BackUpDataPool(tables, item.nodeID);
				this.replicaDataPool.put(item.nodeID, pool);
				//System.out.println("[DEBUG]New pool for node "+item.nodeID+" Created");
			}
			//System.out.println("[DEBUG]New backup data item for node "+item.nodeID+" added: "+item.tableName+":"+item.rowKey + " Pool PrimaryNodeID="+pool.primaryNodeID);
			if(pool.update(item, false)==1)
				GlobalPara.logger.warning("[Error]The backup data item should not exist when pushing data to new backups  ");
		}
		GlobalPara.logger.warning("[DataRecover]Adding new backups done");
		return 0;
	}
	
//	public String[][] commitReadOnlySTG(SubTransactionGroup stg)
//	{
//		//this.lock.writeLock().lock();
//		//this.lock.readLock().lock();
//		try{
//			log.info("Readonly STG received:" + stg.printInfo());
//			int count=0;
//			for(int i=0;i<stg.trans.length;i++)
//				{
//					//GlobalPara.logger.info("Return the latest committed data");
//					TableDataBuffer buffer=TPNode.LTM.dataPool.get(stg.trans[i].strTblName);
//					buffer.executeReadOnlySTWithoutLock(stg.trans[i]);
//					if(stg.trans[i].DataToReturn!=null)
//						count+=stg.trans[i].DataToReturn.length;
//				}
//			String[][] result=null;
//			if(count!=0)
//				{
//				result=new String[count][2];
//				count=0;
//				for(int i=0;i<stg.trans.length;i++)
//					{
//						if(stg.trans[i].DataToReturn!=null)
//							for(int j=0;j<stg.trans[i].DataToReturn.length;j++)
//								result[count++]=stg.trans[i].DataToReturn[j];
//					}
//				}
//			
//			//GlobalPara.logger.info("ReadOnly Transaction Successfully committed with "+count +" items");
//			return result;
//		}
//		finally
//		{
//			//this.lock.writeLock().unlock();
//			//this.lock.readLock().unlock();
//		}
//	}
	
	public String getDataItemID(String strTblName, String strRowKey)
	{
		return strTblName+ "__++__"+strRowKey;
	}
	
	//////////////////////
	// Handling Incoming  Messages
	
	public void addOneSTG(SubTransactionGroup stg)
	{
		//GlobalPara.logger.info("Start "+stg.getSTGkey());
		//log.warning("LTM start add new STG:" + stg.printInfo());
		
		String key=stg.getSTGkey();
		
		Long newTimestamp=this.doRestartMsgBuffer.get(key);
		
		if(newTimestamp!=null){
			if(stg.timestamp>= newTimestamp){
				GlobalPara.logger.warning("[Error]STG key ("+stg.key+"), with timestamp "+ stg.timestamp+ " meets a prearrived dorestart with smaller timestamp ("+ newTimestamp +")");
			} else { 
				stg.timestamp=newTimestamp;
			}
			this.doRestartMsgBuffer.remove(key);
		}
		
		long timestamp=stg.timestamp;
		
		boolean isExistEqual=false;
		boolean isExistHigher=false;
		if(this.doAbortMsgBuffer.keySet().contains(key)){
			Vector<Long> abortTimestamps=this.doAbortMsgBuffer.get(key);
			for(long at:abortTimestamps){
				if(at==timestamp)
					isExistEqual=true;
				else if(at>timestamp)
					isExistHigher=true;
			}
			
			if(isExistEqual){
				MsgBroker.reportAbortDone(stg);
				abortTimestamps.remove(timestamp);
				if(abortTimestamps.size()==0){
					this.doAbortMsgBuffer.remove(key);
				}
				return;
			}
			
			if(isExistHigher){
				List<SubTransactionGroup> list=this.outdateSTGs.get(key);
				if(list==null){
					list=new LinkedList<SubTransactionGroup>();
					this.outdateSTGs.put(key, list);
				}
				list.add(stg);
			}
		}

		boolean needProcessRestForAbort=false;
		boolean isContinueAdd=true;
		
		if(isExistHigher)
			isContinueAdd=false;
		
		
		SubTransactionGroup oldSTG=STGPool.get(key);
		if(oldSTG!=null) {
			List<SubTransactionGroup> list=this.outdateSTGs.get(key);
			if(list==null){
				list=new LinkedList<SubTransactionGroup>();
				this.outdateSTGs.put(key, list);
			}
			
			//GlobalPara.logger.warning("[info] STG with the key "+ key + " already existed");
			if(oldSTG.timestamp>stg.timestamp){
				isContinueAdd=false;
				if(!isExistHigher){
					list.add(stg);
				}
			} else if(oldSTG.timestamp<stg.timestamp){
				oldSTG.doAbort();
				if(GlobalPara.sysConf.doMemoryManagement)
					oldSTG.terminateForMM();
				
				list.add(oldSTG);
				needProcessRestForAbort=true;
				if(!isExistHigher){
					STGPool.put(key, stg);
				}
			} else {
				GlobalPara.logger.warning("[Error] Two STGs with the same key "+key+ "and same timestamp "+timestamp);
			}
		} else{
			if(!isExistHigher)
				STGPool.put(key, stg);	
		}
		
		int result=checkpointBuffer.doSubmit(timestamp);
		
//		if(GlobalPara.sysConf.doMemoryManagement) {
//			String[] accessedDataItems=new String[stg.trans.length];
//			int count=0;
//			for(SubTransaction st:stg.trans)
//				accessedDataItems[count++]=getDataItemID(st.strTblName, st.strRowKey);
//			TPNode.MM.accessDataItems(accessedDataItems);
//		}
		
		if(isContinueAdd){
			int addresult=stg.doStart();
			if(addresult!=1){
				stg.reportRestart();
			} else {
				LTMActions.ProcessGroupOfSTs(stg.trans);
			}
		} else {
			if(needProcessRestForAbort)
				LTMActions.ProcessGroupOfSTs(stg.trans);
		}
	}
	
	public void printDebugMsg()
	{
		String result="[DEBUG]Pring STGs in the data pool:\nThere are still "+this.STGPool.size()+" STGs in the pool\n";
		
		Vector<SubTransactionGroup> toPrintSTGs=new Vector<SubTransactionGroup>();
		//scan STG pool
		for(SubTransactionGroup stg:this.STGPool.values()){
			boolean isToPrint=false;
			boolean isAllInListHead=true;
			for(SubTransaction st:stg.trans){
				if(st.rootEntry!=null){
					if(!st.rootEntry.isSTInOfList(st)){
						isToPrint=true;
					} else {
						if(!st.rootEntry.isSTHeadOfList(st)){
							isAllInListHead=false;
						}
					}
				} else {
					isToPrint=true;
				}
			}
			if(isAllInListHead)
				isToPrint=true;
			if(isToPrint)
				toPrintSTGs.add(stg);
		}
		
		boolean isReadOnlyTranHeadOfList=false;

		//scan data pool
		
		for(TableDataBuffer table:this.dataPool.values()){
			for(DataRowEntry row: table.data.values()){
				SubTransaction st=row.getHeadOfST();
				if(st!=null){
					if(st.isReadOnlyGlobalTran){
						isReadOnlyTranHeadOfList=true;
					} else {
						SubTransactionGroup stg=st.parentGroup;
						if(!this.STGPool.containsValue(stg))
							toPrintSTGs.add(stg);
						else {
							; //if a STG contains in the pool, and all its STs are in the list head, it already in the toPrintSTGs
						}
					}
				}
			}
		}
		
		UpdatesOfST debugMsg=new UpdatesOfST(null, null, 0, null, 0,1);
		
		//print queue for pushing data backups
		DataRowEntry.updateQueueForReplicas.offer(debugMsg);
		
		//print buffered abort and restarts
		result+="\nThe STGs which are the root of blocking all other STGs:\n";
		
		for(SubTransactionGroup stg:toPrintSTGs){
			result+="\n<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>\n"+stg.printDebugInfo()+"\n<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>\n";
		}
		 
		
		//Find the root STGs which cause the system to halt.
		//And find the STGs which are not able to commit
		//Also display the abort and restart buffer
		System.out.println(result);
		System.out.flush();
		
	}
	
	public String doDebug(String strGID)
	{
		SubTransactionGroup a=TPNode.LTM.STGPool.get(strGID);
		if(a!=null){
	    String result="<<<<<<<<<<<<<<<<<<<<<\n"+a.printInfo()+"\n";
	    for(int i=0;i<a.trans.length;i++)
	    	result+="\n {{{{{{{{{{{{{{{{{{{{{{{{{{{{\n The "+i+"th ST is "+ a.trans[i].className  
			+ "\n this ST of STG is:" +  a.trans[i].printInfo()
			+ "\n the DataRowEntity of this ST is:" + a.trans[i].printTheDataListInfo()
			+ "\n }}}}}}}}}}}}}}}}}}}}}}}}}}}}}";
	    //GlobalPara.logger.warning(result);
	    result+="\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n";
	    return result;
		}
		else
			return "<<<<<<<<<<<<<<<\nnull for STG :" + strGID+"\n >>>>>>>>>>>>>>>\n";
	}
	
	//the parameter might not be the same stg in data pool.
	public void doAbortSTG(String key, long timestamp)
	{
		//GlobalPara.logger.info("Start "+key);
		//GlobalPara.logger.warning("Receive doAbortSTG request for SubTransactionGroup with key "+key+" to abort");
		
		SubTransactionGroup s=this.STGPool.get(key);
		if(s!=null&&s.timestamp==timestamp) {
			s.doAbort();
			MsgBroker.reportAbortDone(s);
			if(GlobalPara.sysConf.doMemoryManagement)
				s.terminateForMM();
			this.STGPool.remove(key);
			this.checkpointBuffer.doAbort(s.timestamp);
			LTMActions.ProcessGroupOfSTs(s.trans);
			return;
		} else {
			List<SubTransactionGroup> outSTGs=this.outdateSTGs.get(key);
			if(outSTGs!=null){
				for(SubTransactionGroup stg:outSTGs){
					if(stg.timestamp==timestamp){
						MsgBroker.reportAbortDone(stg);
						outSTGs.remove(stg);
						if(outSTGs.size()==0){
							this.outdateSTGs.remove(key);
						}
						return;
					}
				}
			}
			
			Vector<Long> abortTimestamps=this.doAbortMsgBuffer.get(key);
			if(abortTimestamps==null){
				abortTimestamps=new Vector<Long>();
				this.doAbortMsgBuffer.put(key, abortTimestamps);
			}
			abortTimestamps.add(timestamp);
		}
	}
	
	public void doRestartSTG(long timestamp,String key,long newTimestamp)
	{
		Long oldRestartTimestamp=this.doRestartMsgBuffer.get(key);
		if(oldRestartTimestamp==null||oldRestartTimestamp<newTimestamp){
			if(oldRestartTimestamp!=null)
				this.doRestartMsgBuffer.remove(key);
			SubTransactionGroup s=this.STGPool.get(key);
			if(s==null){
				this.doRestartMsgBuffer.put(key, newTimestamp);
			} else {
				if(s.timestamp>newTimestamp){
					return;
				} else if(s.timestamp<newTimestamp){
					Vector<Long> abortTimestamps=this.doAbortMsgBuffer.get(key);
					if(abortTimestamps!=null){
						for(Long at:abortTimestamps){
							if(at==newTimestamp){
								s.doAbort();
								s.timestamp=newTimestamp;
								this.STGPool.remove(key);
								if(GlobalPara.sysConf.doMemoryManagement)
									s.terminateForMM();
								MsgBroker.reportAbortDone(s);
								LTMActions.ProcessGroupOfSTs(s.trans);
								abortTimestamps.remove(at);
								if(abortTimestamps.size()==0)
									this.doAbortMsgBuffer.remove(key);
								return;
							}
						}
					}
					
					int result=checkpointBuffer.doRestart(timestamp, newTimestamp);
					int restartResult=s.doRestart(newTimestamp);
					
					if(restartResult==0){
						MsgBroker.reportRestart(s);
					} else if(restartResult==-1){
						GlobalPara.logger.warning("[errors]STG blocked when restarting because error happens when adding transactions.");
					}
					LTMActions.ProcessGroupOfSTs(s.trans);
				} else {
					GlobalPara.logger.warning("[error]DoRestart message for STG "+ key +" has the equal timestamp "+ newTimestamp +" with current stg");	
				}
			}
		} else if(oldRestartTimestamp>newTimestamp){
			return;
		} else if(oldRestartTimestamp==newTimestamp){
			GlobalPara.logger.warning("[error]two doRestart message for STG "+ key +" with the same timestamp "+ newTimestamp +" received");
			return;
		}
	}
	
	public void doCommitOneSTG(String key)
	{
		SubTransactionGroup s=this.STGPool.get(key);
		if(s==null){
			GlobalPara.logger.warning("[error]Unable to find the STG "+key+" in the buffer");
			return;
		}
		s.doCommit();
		//this.STGPool.remove(key);
		this.checkpointBuffer.doCommitted(s.timestamp);
		LTMActions.ProcessGroupOfSTs(s.trans);
	}
	
//	public ReentrantLock doCommitLock=new ReentrantLock();
//	public int numOfInDoCommit=0; 
//	public void increaseOneDoCommit()
//	{
//		doCommitLock.lock();
//		try{
//			numOfInDoCommit++;
//		}
//		finally{
//			doCommitLock.unlock();
//		}
//	}
//	public void releaseOneDoCommit()
//	{
//		doCommitLock.lock();
//		try{
//			numOfInDoCommit--;
//		}
//		finally{
//			doCommitLock.unlock();
//		}
//	}
//	
//	public void doCommitSTGDirectly(SubTransactionGroup stg)
//	{
//		if(stg==null)
//			return;
////		GlobalPara.logger.warning("Receive doCommitSTG request for SubTransactionGroup with key "+key+" to commit");
//		//stg.doCommit(stg.timestamp);
//
//		Vector<Future> commitTasks=new Vector<Future>();
//		for(int i=0;i<stg.trans.length;i++) {
//			DoLTMCommitTask task=new DoLTMCommitTask(stg.trans[i]);
//			commitTasks.add(bgService.submit(task));
//		}
//			
//		for(Future task:commitTasks){
//			try {
//				task.get();
//			} catch (InterruptedException e) {
//				e.printStackTrace();
//			} catch (ExecutionException e) {
//				e.printStackTrace();
//			}
//		}
//
//		TPNode.LTM.checkpointBuffer.doCommittedWithoutLock(stg.timestamp);
//		
//		
////		for(int i=0;i<stg.trans.length;i++)
////		{
////			TableDataBuffer buffer=TPNode.LTM.dataPool.get(stg.trans[i].strTblName);
////			DataRowEntry row=buffer.getRowWithoutLockForDoCommit(stg.trans[i].strRowKey);//it is safe as no other transaction can read or modify the data
////			if(row==null) {
////				GlobalPara.logger.severe("The data row is not found in the databuffer when committing updates");
////				continue;
////			}
////			row.latestTransaction=stg.timestamp;
////			buffer.commitDoneST(stg.trans[i]);
////		}
//		//stg.doneCommit(timestamp);
//		//GlobalPara.logger.info("Returned information successfully collected with "+count +" items");
//		
//		if(GlobalPara.sysConf.doMemoryManagement) {
//			String[] unaccessedDataItems=new String[stg.trans.length];
//			int count=0;
//			for(SubTransaction st:stg.trans)
//			{
//				unaccessedDataItems[count++]=getDataItemID(st.strTblName, st.strRowKey);
//			}
//			TPNode.MM.unaccessDataItems(unaccessedDataItems);
//		}
//	}
//	/**
//	 * The commit process is synchronized such that only committed data is visible
//	 * @return Combine all the dataToReturn in each ST
//	 */
//	public String doCommitSTG(String key)
//	{
//		SubTransactionGroup stg=this.STGPool.get(key);
//		if(stg==null) {
//			GlobalPara.logger.warning("The SubTransactionGroup with key "+key+" is not found in this TPNode");
//			return null;
//		}
//		increaseOneDoCommit();//this must be in the first
//		if(MembershipManager.isInFailureRecovery){
//			UpdatesOfST data=new UpdatesOfST(stg.timestamp, stg.id,key);
//			this.arrivedUpdatesInRecovery.add(data);
//			releaseOneDoCommit();
//			return null;
//		}
//		
//		doCommitSTGDirectly(stg);
//		
//		releaseOneDoCommit();
//		//GlobalPara.logger.info("End "+key);
//		return null;
//	}
	
	/**
	 * Use to sort the sequence of process a groupe of subtransactions
	 * Smallest first
	 * @param t
	 * @return
	 */
	public static String getSTkey(SubTransaction t)
	{
		return t.strTblName+"-" +t.strRowKey;
	}
	
	
	public void reportCommitDone(String tableName, String rowKey,SubTransaction t, long timestamp)
	{
		if(t!=null)
			t.parentGroup.reportCommitDone(t);
		
		if(GlobalPara.sysConf.doMemoryManagement){
			TableDataBuffer data=this.dataPool.get(tableName);
			DataRowEntry row=data.getOneRowEntry(rowKey);
			row.reportReplicaSynchronized(timestamp);
		}
	}
	
	public void reportCheckpointDone(String tableName, String rowKey,SubTransaction t, long timestamp)
	{
		if(t!=null)
			t.parentGroup.reportCheckpointDone(t);
		
		if(GlobalPara.sysConf.doMemoryManagement){
			TableDataBuffer data=this.dataPool.get(tableName);
			DataRowEntry row=data.getOneRowEntry(rowKey);
			if(row!=null)
				row.reportCheckpointSuccess(timestamp);
			else
				GlobalPara.logger.warning("[Error]Null data row when report checkpoint success");
		}
		
	}
	
	public void doDataReorganization(RingMembership newMembership, Vector<Integer> failedNodes)
	{
		this.STGPool.clear();
		this.doAbortMsgBuffer.clear();
		this.doRestartMsgBuffer.clear();
		this.outdateSTGs.clear();
		workingDataPoolComplete=false;
		backupDataPoolComplete=false;
		
		Vector<Integer> currentPrimaryVNodes=getVNodeBackupedByFailedNodes(failedNodes);
		
		Hashtable<Integer, Vector<VirtualTPNode>> backupnodes=newMembership.getBackupNodesOfDataItems();
		
//		HashMap<VirtualTPNode,VirtualTPNode> primaryFailedNodes=getVNodePrimaryFailed(failedNodes);
		
		//<old Virtual ID of local primary, List of Physical IDs>
		HashMap<Integer, Vector<Integer>>  destIDsForPrimary=new HashMap<Integer, Vector<Integer>>();
		//<old Virtual ID of failed primary, the new primary is local and the list of new backup LTMs>
		HashMap<Integer, Vector<Integer>>  destIDsForBecomePrimaryBackup=new HashMap<Integer, Vector<Integer>>();
		//<old Virtual ID of failed primary, the physical ID of the new primary>
		HashMap<Integer, Integer>  destIDsForNormalBackup=new HashMap<Integer, Integer>();
		
		
//		Hashtable<Integer, Vector<VirtualTPNode>> newVirtualBackups=newMembership.getBackupNodesOfDataItems();
		
		for(int vid:currentPrimaryVNodes){
			
			
//			String debugString="Backup TPNodes in new membership for VID"+vid+": ";
//			for(int nid:newids){
//				debugString+=nid+",";
//			}
//			debugString+="\n";
//			
//			debugString+="The old backup TPNodes for VID"+vid+":";
//			for(int nid:oldZone.getTPNodes()){
//				debugString+=nid+",";
//			}
//			debugString+="\n";
//			GlobalPara.logger.warning(debugString);
			
			Vector<Integer> newBackupIDs;
			
			if(GlobalPara.sysConf.doMemoryManagement){ //to all replicas
				Vector<VirtualTPNode> newBackupVirtualIDs=backupnodes.get(vid);
				newBackupIDs=new Vector<Integer>();
				for(VirtualTPNode vnode:newBackupVirtualIDs){
					newBackupIDs.add(vnode.physicalID);
				}
			} else { //to only the new backup LTMs
				RecoveryZone oldZone=TPNode.ring.getZoneForPrimaryNode(vid);
				RecoveryZone newZone=newMembership.getZoneForPrimaryNode(vid);
				int[] newids=newZone.getTPNodes();
				newBackupIDs=oldZone.excludes(newids);
			}
			
			if(newBackupIDs.size()>0){
				destIDsForPrimary.put(vid, newBackupIDs);
			}
		}
		
//		String output="Pushing working data pool to these new backups:\n";
//		for(int vid: destIDsForPrimary.keySet()){
//			output+=vid+" to new backup LTMs: ";
//			for(int pid:destIDsForPrimary.get(vid)){
//				output+=pid+", ";
//			}
//			output+="\n";
//		}
//		GlobalPara.logger.warning(output);
		
		
		Vector<RecoveryZone> zones=TPNode.ring.getAllInvolvedRecoveryZones();
		for(RecoveryZone zone: zones){
			VirtualTPNode nextVnode=zone.getSecondaryPrimayWhenFailure(failedNodes);
			if(nextVnode==null){
				continue;
			} else if(nextVnode.physicalID==GlobalPara.sysConf.localNodeID){
				Vector<Integer> newBackupIDs;
				if(GlobalPara.sysConf.doMemoryManagement){ //push all replicas
					Vector<VirtualTPNode> newBackupVirtualIDs=backupnodes.get(nextVnode.virtualID);
					newBackupIDs=new Vector<Integer>();
					for(VirtualTPNode vnode:newBackupVirtualIDs){
						newBackupIDs.add(vnode.physicalID);
					}
				} else { //push only to new replicas
					RecoveryZone newZone=newMembership.getZoneForPrimaryNode(nextVnode.virtualID);
					int[] newids=newZone.getTPNodes();
					newBackupIDs=zone.excludes(newids);	
				}
				if(newBackupIDs.size()>0){
					destIDsForBecomePrimaryBackup.put(zone.getPrimaryVNode().virtualID, newBackupIDs);
				}
			} else {
				destIDsForNormalBackup.put(zone.getPrimaryVNode().virtualID, nextVnode.physicalID);
			}
		}
			
//		output="Pushing BACKUP data pool to these new backups:\n";
//		for(int vid: destIDsForBecomePrimaryBackup.keySet()){
//			output+=vid+" to new backup LTMs: ";
//			for(int pid:destIDsForBecomePrimaryBackup.get(vid)){
//				output+=pid+", ";
//			}
//			output+="\n";
//		}
//		GlobalPara.logger.warning(output+"\n");
		
		
//		 output="Moving BACKUP data pool under new primary LTM:\n";
//		for(int vid: destIDsForNormalBackup.keySet()){
//			output+=vid+" to new primary LTM: ";
//			output+=destIDsForNormalBackup.get(vid)+", ";
//			output+="\n";
//		}
//		GlobalPara.logger.warning(output);
		
		//<Dest TPNodeID, The data item value >
		HashMap<Integer, Vector<UpdatesOfST>> dataBuffer =new HashMap<Integer, Vector<UpdatesOfST>>(); 
		cleanWorkDataPool(newMembership, destIDsForPrimary, dataBuffer);
		pushDataToNewBackups(1, dataBuffer);
		HashMap<Integer, Vector<UpdatesOfST>> dataBuffer2 =new HashMap<Integer, Vector<UpdatesOfST>>(); 
		scanBackupDataPool(newMembership, failedNodes, destIDsForBecomePrimaryBackup, destIDsForNormalBackup, dataBuffer2);
		pushDataToNewBackups(2, dataBuffer2);
	}
	
	protected boolean workingDataPoolComplete=false;
	protected boolean backupDataPoolComplete=false;
	
	public void reportWorkingDataPoolReorganizationComplete(){
		workingDataPoolComplete=true;
		GlobalPara.logger.warning("[LTMRecover] report Working Data Pool Reorganization Complete");
		if(workingDataPoolComplete&&backupDataPoolComplete){
			GlobalPara.memberComp.reportDataReorganizationComplete();
		}
	}
	
	public void reportBackupDataPoolReorganizationComplete(){
		backupDataPoolComplete=true;
		GlobalPara.logger.warning("[LTMRecover] report Backup Data Pool Reorganization Complete");
		if(workingDataPoolComplete&&backupDataPoolComplete){
			GlobalPara.memberComp.reportDataReorganizationComplete();
		}
	}
	
	/**
	 * 
	 * @param type 1: working DataPool, 2: backup data
	 * @param dataBuffer
	 */
	protected void pushDataToNewBackups(int type, HashMap<Integer, Vector<UpdatesOfST>> dataBuffer)
	{
		BatchBackupDataItemsTask task=new BatchBackupDataItemsTask(type, dataBuffer);
		MsgBroker.executor.submit(task);
	}
	
	protected void cleanWorkDataPool(RingMembership newMembership, 
			HashMap<Integer, Vector<Integer>>  destIDsForPrimary, 
			HashMap<Integer, Vector<UpdatesOfST>> dataBuffer)
	{
		int count=0;
		int totalCount=0;
		for(TableDataBuffer table:this.dataPool.values()){
			for(DataRowEntry dataItem:table.data.values()){
				totalCount++;
				dataItem.resetState();
				if(GlobalPara.sysConf.doMemoryManagement&&!dataItem.state.isEvictable){
					TPNode.MM.toBeResetDataItems.add(dataItem.state);
				}
				
				if(dataItem.isDataLoaded==0){
					dataItem.isDataLoaded=-1;
				} else if(dataItem.isDataLoaded==-1){
					;
				} else { //==1
					VirtualTPNode oldVNode=TPNode.getVirtualNode(dataItem.dataTable.tblName, dataItem.rowKey);
					if(destIDsForPrimary.keySet().contains(oldVNode.virtualID)){
						count++;
						if(GlobalPara.sysConf.doMemoryManagement){ 
							//dataItem.state.transactionAccess(); //not evictable as it will be pushed to new backups
							//TPNode.MM.toBeResetDataItems.add(dataItem.state);
							if(dataItem.state.isDataSizeChanged){//possible for loaded but not modified data items
								if(dataItem.state.sizeOfRootDataRowEntry>1)
									TPNode.MM.bufferSize-=dataItem.state.sizeOfRootDataRowEntry;
								dataItem.updateSize(); //push the up-to-date size value to the replica
								TPNode.MM.bufferSize+=dataItem.state.sizeOfRootDataRowEntry;
								//don't report buffer size change now, because not all affected data items are marked unevictable.  
							}
						}
						
						UpdatesOfST updateData=new UpdatesOfST(dataItem.rowKey, 
								dataItem.dataTable.tblName, dataItem.committedData.lastWriteTimestamp, 
								dataItem.committedData.getRowData(), GlobalPara.sysConf.localNodeID, dataItem.committedData.size);
						
						Vector<Integer> tpnodeIDs=destIDsForPrimary.get(oldVNode.virtualID);

						for(int dest:tpnodeIDs){
							Vector<UpdatesOfST> backupDataCollection=dataBuffer.get(dest);
							if(backupDataCollection==null){
								backupDataCollection=new Vector<UpdatesOfST>();
								dataBuffer.put(dest, backupDataCollection);
							}
							backupDataCollection.add(updateData);
						}
					}
				}
			}
		}
		
//		if(GlobalPara.sysConf.doMemoryManagement)
//			TPNode.MM.reportBufferSizeChange();
		
		GlobalPara.logger.warning("Result: "+count+" of total "+totalCount+" data items in the working data pool will be pushed to the new backups");
	}
	
	
	/**
	 * Move the backups of the failed TPnodes into the right container:
	 * 1. to the working data POOL if local is the new primary: move to work data pool and push to new backup LTMs
	 * 2. still in the backup data pool but under the new primary TP Node
	 * 
	 * @param destIDsForBecomePrimaryBackup
	 * @param destIDsForNormalBackup
	 * @param dataBuffer
	 */
	protected void scanBackupDataPool(RingMembership newMembership, 
			Vector<Integer> failedNodes,
			HashMap<Integer, Vector<Integer>>  destIDsForBecomePrimaryBackup, 
			HashMap<Integer, Integer>  destIDsForNormalBackup,
			HashMap<Integer, Vector<UpdatesOfST>> dataBuffer)
	{
//		Set<Integer> failedNodes=new TreeSet<Integer>();
//		for(Integer oldVnode:destIDsForBecomePrimaryBackup.keySet()){
//			if(!failedNodes.contains(oldVnode.physicalID))
//				failedNodes.add(oldVnode.physicalID);
//		}
//		for(Integer oldVnode:destIDsForNormalBackup.keySet()){
//			if(!failedNodes.contains(oldVnode.physicalID))
//				failedNodes.add(oldVnode.physicalID);
//		}
		
		int totalCount=0;
		int moveToWorkCount=0;
		int moveToNewBackupCount=0;
		
		for(Integer failedTPNode: failedNodes){
			BackUpDataPool data=this.replicaDataPool.get(failedTPNode);
			GlobalPara.logger.warning("[info]Scan backup data pool for failed LTM " + failedTPNode
					+" \nThere are  " +data.dataPool.keySet().size()+ " Tables in his backup data pool");
			for(String tableName:data.dataPool.keySet()){
				HashMap<String, DataRowContainer> backupTable=data.dataPool.get(tableName);
				TableDataBuffer workTable=this.dataPool.get(tableName);
				if(workTable==null){
					GlobalPara.logger.warning("[Error] WorkingDataPool doesn't contain the table "+tableName+" in the backup data pool");
					return;
				}
				//GlobalPara.logger.warning("[info]There are "+backupTable.size()+"data items in backup table "+tableName +" For LTM "+failedTPNode);
				
				for(String rowKey:backupTable.keySet()){
					totalCount++;
					DataRowContainer dataitem=backupTable.get(rowKey);
					UpdatesOfST updateData=new UpdatesOfST(rowKey, tableName, dataitem.lastWriteTimestamp, 
							dataitem.getRowData(), GlobalPara.sysConf.localNodeID, dataitem.size);
					VirtualTPNode oldVnode=TPNode.getVirtualNode(tableName, rowKey);
					
//					//////For debug
//					VirtualTPNode newVnode=newMembership.getVirtualNode(tableName, rowKey);
//					if(newVnode.physicalID==GlobalPara.sysConf.localNodeID&&!destIDsForBecomePrimaryBackup.containsKey(oldVnode.virtualID)){
//						GlobalPara.logger.warning("[Error] The data entry ("+tableName+","+ rowKey+") should be written to working data pool but not been done.");
//					}
//					////////end for debug
					
					if(destIDsForBecomePrimaryBackup.containsKey(oldVnode.virtualID)){
						moveToWorkCount++;
						Vector<Integer> destIDs=destIDsForBecomePrimaryBackup.get(oldVnode.virtualID);
						for(int destid:destIDs){
							Vector<UpdatesOfST> backupDataCollection=dataBuffer.get(destid);
							if(backupDataCollection==null){
								backupDataCollection=new Vector<UpdatesOfST>();
								dataBuffer.put(destid, backupDataCollection);
							}
							backupDataCollection.add(updateData);
						}
						workTable.addNewRow(rowKey, dataitem);
					} else if(destIDsForNormalBackup.containsKey(oldVnode.virtualID)){
						moveToNewBackupCount++;
						Integer newTPNodeID=destIDsForNormalBackup.get(oldVnode.virtualID);
						BackUpDataPool newBackupDataPool=this.replicaDataPool.get(newTPNodeID);
						if(newBackupDataPool==null){
							newBackupDataPool=new BackUpDataPool(tables,newTPNodeID);
							this.replicaDataPool.put(newTPNodeID, newBackupDataPool);
						}
						newBackupDataPool.insert(tableName, rowKey, dataitem);
					} else {
						GlobalPara.logger.warning("[error]One backup data item of the failed LTM is not replaced.");
					}
				}
			}
			this.replicaDataPool.remove(failedTPNode);
		}
		
		if(GlobalPara.sysConf.doMemoryManagement){
			TPNode.LTM.totalNumInDataPool+=moveToWorkCount;
			TPNode.LTM.totalNumInBackupDataPool-=moveToWorkCount;
		}
		
		GlobalPara.logger.warning("Total "+totalCount+" are scanned, and "
				+moveToWorkCount+"data items are moved to working data pool, "
				+ moveToNewBackupCount+" are replaced as backup for a new LTM");
		
	}
	
	/**
	 * 
	 * @param failedNodes
	 * @return collection of Virtual IDs
	 */
	protected Vector<Integer> getVNodeBackupedByFailedNodes(Vector<Integer> failedNodes){
		Vector<Integer> result=new Vector<Integer>();
		for(VirtualTPNode vnode:TPNode.ring.localVNodes){
			int vid=vnode.virtualID;
			Vector<VirtualTPNode> list=TPNode.backupVirtualNodes.get(vid);
			if(list==null){
				GlobalPara.logger.warning("[Error] There are no backup LTMs for Vnode " +vid);
			} else {
				for(VirtualTPNode backupNode:list){
					if(failedNodes.contains(backupNode.physicalID)){
						result.add(vid);
						break;
					}
				}	
			}
		}
		
		if(TPNode.backupVirtualNodes.size()!=TPNode.ring.localVNodes.length){
			GlobalPara.logger.warning("[Error] The length of backupVirtualNodes ("+TPNode.backupVirtualNodes.size()+") is not equal to localVNodes" +
					"("+TPNode.ring.localVNodes.length+")");
		}
		return result;
	}
	
//	/**
//	 * 
//	 * @param failedNodes
//	 * @return <old failed primary virtual node, new primary virtual node>
//	 */
//	protected HashMap<VirtualTPNode,VirtualTPNode> getVNodePrimaryFailed(Vector<Integer> failedNodes){
//		HashMap<VirtualTPNode,VirtualTPNode> result=new HashMap<VirtualTPNode,VirtualTPNode>();
//		Vector<RecoveryZone> zones=TPNode.ring.getAllInvolvedRecoveryZones();
//		for(RecoveryZone zone: zones){
//			VirtualTPNode nextVnode=zone.getSecondaryPrimayWhenFailure(failedNodes);
//			if(nextVnode!=null) //&&nextVnode.physicalID==GlobalPara.sysConf.localNodeID
//				result.put(zone.getPrimaryVNode(), nextVnode);
//		}
//		return result;
//	}
	
	public void commitRecoveredUpdates()
	{
		for(Long timestamp:this.arrivedUpdatesInRecoveryAsPrimary.keySet()){
			Vector<UpdatesOfST> data=this.arrivedUpdatesInRecoveryAsPrimary.get(timestamp);
			for(int i=0;i<data.size();i++){
				UpdatesOfST update=data.get(i);
				TableDataBuffer table=this.dataPool.get(update.tableName);
				DataRowEntry row=table.data.get(update.rowKey);
				
				UpdatedDataRow value=null;
				if(row==null){ //must firstly load the row, multiple updates may access the same row
					if(!update.isContainsDataRow){
						if(GlobalPara.sysConf.doMemoryManagement){
							GlobalPara.logger.warning("[MMinfo]Data item evicted while TC not terminated");
						} else {
							//////////for debug
							VirtualTPNode vnode=TPNode.getVirtualNode(update.tableName, update.rowKey);
							if(vnode.physicalID!=GlobalPara.sysConf.localNodeID){
								GlobalPara.logger.warning("[Error]The datarow entry ("+update.tableName+", "+ update.rowKey 
										+") should be primary at LTM "+ vnode.physicalID +" not this node "+ GlobalPara.sysConf.localNodeID 
										+". So it is missing in the working data pool and update backups of TC");
							} else
								GlobalPara.logger.warning("[Error]The datarow entry ("+update.tableName+", "+ update.rowKey 
										+")is missing in working data pool and update backups of TC");
							/////end for debug
						}
						continue; //these updates will not be checkpointed
					} else {
						row=new DataRowEntry(table, update.rowKey);
						row.isDataLoaded=1;
						table.data.put(update.rowKey,row);
						if(update.dataRow!=null)
							row.committedData.updateForTemp(update.dataRow, null);
						row.committedData.commit(update.dataToPut, update.dataTodelete, timestamp);
						row.latestWriteTransaction=row.latestTransaction=timestamp;
						
						if(GlobalPara.sysConf.doMemoryManagement){
							TPNode.LTM.totalNumInDataPool++;
							row.state.isDataSizeChanged=true;
							row.state.transactionAccess(); //not evictable until the complete process completes, because the other replicas may not be ready
							TPNode.MM.toBeResetDataItems.add(row.state);
							TPNode.MM.accessedDataItems.add(row.state);
						}
						value=new UpdatedDataRow(row.rowKey, row.committedData.getRowData(), null);
					}
				} else {
					if(row.isDataLoaded==1){
						if(update.timestamp>row.latestWriteTransaction){ //it is possible when backup is updated, but not able to report commit done
							row.committedData.commit(update.dataToPut, update.dataTodelete, timestamp);
							row.latestWriteTransaction=row.latestTransaction=timestamp;
							
							value=new UpdatedDataRow(row.rowKey, update.dataToPut, update.dataTodelete);
							
							if(GlobalPara.sysConf.doMemoryManagement){
								row.state.isDataSizeChanged=true;
								row.state.transactionAccess();
								TPNode.MM.toBeResetDataItems.add(row.state);
							}
						} else
							continue;
					} else {
						GlobalPara.logger.warning("[ERROR] The data item is not loaded when commit recovered updates.");
						continue;
					}
				}
				
				GlobalPara.adapterComp.checkpointOneRow(timestamp, row.dataTable.tableSchema, row.rowKey, value, null);
			}
		}
//		if(GlobalPara.sysConf.doMemoryManagement)
//			TPNode.MM.reportBufferSizeChange();
		
		for(Long timestamp:this.arrivedUpdatesInRecoveryAsBackup.keySet()){
			Vector<UpdatesOfST> data=this.arrivedUpdatesInRecoveryAsBackup.get(timestamp);
			for(int i=0;i<data.size();i++){
				UpdatesOfST update=data.get(i);//the one with dataRow is always in the head of Vector
				VirtualTPNode vnode=TPNode.getVirtualNode(update.tableName, update.rowKey);
				BackUpDataPool dataPool=this.replicaDataPool.get(vnode.physicalID);
				if(dataPool==null){
					GlobalPara.logger.warning("[Error] This LTM doesn't have backup data pool for LTM " + vnode.physicalID);
					continue;
				}
				HashMap<String, DataRowContainer> table=dataPool.dataPool.get(update.tableName);
				if(table==null){
					GlobalPara.logger.warning("[Error] This backup data pool does not have table " + update.tableName);
					continue;
				}
				DataRowContainer row=table.get(update.rowKey);
				if(row==null){
					if(!update.isContainsDataRow){
						if(GlobalPara.sysConf.doMemoryManagement){
							GlobalPara.logger.warning("[info]Data item evicted while TC not terminated");
						} else {
							/////////////////for debug
							RecoveryZone zone=TPNode.ring.getZoneForPrimaryNode(vnode.virtualID);
							if(zone.getTPNodes()[1]!=GlobalPara.sysConf.localNodeID){
								GlobalPara.logger.warning("[Error]The datarow entry ("+update.tableName+", "+ update.rowKey 
										+") from STG "+update.stgKey+" should be backup at LTM "+ zone.getTPNodes()[1] +" not this node "+ GlobalPara.sysConf.localNodeID 
										+". So it is missing in the backup data pool and update backups of TC");
							} else
								GlobalPara.logger.warning("[Error]The datarow entry ("+update.tableName+", "+ update.rowKey 
										+") from STG "+update.stgKey+" is missing in bakcup data pool and update backups of TC!!!!!!!!!!!");
							/////////////////for debug
						}
						continue;
					} else {
						row=new DataRowContainer();
						table.put(update.rowKey, row);
						if(row.rowData!=null)
							row.updateForTemp(update.dataRow, null);
						row.commit(update.dataToPut, update.dataTodelete, timestamp);
						if(GlobalPara.sysConf.doMemoryManagement){
							TPNode.MM.backupDataSize+=row.getSize();
							TPNode.LTM.totalNumInBackupDataPool++;
						}
					}
				} else {
					if(update.timestamp>row.lastWriteTimestamp){
						if(GlobalPara.sysConf.doMemoryManagement){
							TPNode.MM.backupDataSize-=row.size;
							TPNode.MM.backupDataSize+=row.getSize();
						}
						row.commit(update.dataToPut, update.dataTodelete, timestamp);
					}
				}
			}
		}
		
		if(GlobalPara.sysConf.doMemoryManagement){
			GlobalPara.logger.warning("[info]There are "+ TPNode.MM.toBeResetDataItems.size() +" data items to be reset after recoverysuccess");
			for(DataItemState row:TPNode.MM.toBeResetDataItems){
				row.referencesOfSTs=0;
				row.replicaSynchronizedTimestamp=row.data.latestWriteTransaction;
				row.data.checkIfEvictable(false);
			}
			TPNode.MM.toBeResetDataItems.clear();
		}
		
		this.arrivedUpdatesInRecoveryAsBackup.clear();
		this.arrivedUpdatesInRecoveryAsPrimary.clear();
		
		GlobalPara.memberComp.reportRecoveredUpdatesCommitComplete();
		
//		if(delayedUpdatesForPrimary.size()==0&&delayedUpdatesForBackup.size()==0)
//			GlobalPara.memberComp.reportRecoveredUpdatesCommitComplete();
//		else {
//			for(DataRowEntry row:delayedUpdatesForPrimary.keySet()){
//				GlobalPara.adapterComp.loadDataForPrimaryWhenRecovery(row);
//			}
//			for(DataRowContainer row:delayedUpdatesForBackup.keySet()){
//				UpdatesOfST update=delayedUpdatesForBackup.get(row).firstEntry().getValue();
//				DataTable table=this.dataPool.get(update.tableName).tableSchema;
//				GlobalPara.adapterComp.loadDataForBackupWhenRecovery(table, update.rowKey, row);
//			}
//		}
	}
	
//	public void delayedLoadDataForCommitRecoveredComplete(DataRowEntry row)
//	{
//		TreeMap<Long, UpdatesOfST> updates=delayedUpdatesForPrimary.get(row);
//		row.isDataLoaded=1;
//		for(UpdatesOfST update:updates.values()){
//			row.committedData.commit(update.dataToPut, update.dataTodelete, update.timestamp);
//		}
//		delayedUpdatesForPrimary.remove(row);
//		checkIfDelayedCommitForRecoveryAllComplete();
//	}
//	
//	public void delayedLoadDataForCommitRecoveredComplete(DataRowContainer row)
//	{
//		TreeMap<Long, UpdatesOfST> updates=delayedUpdatesForBackup.get(row);
//		for(UpdatesOfST update:updates.values()){
//			row.commit(update.dataToPut, update.dataTodelete, update.timestamp);
//		}
//		row.isCreatedInRecoverCommit=false;
//		delayedUpdatesForBackup.remove(row);
//		checkIfDelayedCommitForRecoveryAllComplete();
//	}
	
//	protected void checkIfDelayedCommitForRecoveryAllComplete(){
//		if(delayedUpdatesForPrimary.size()==0&&delayedUpdatesForBackup.size()==0)
//			GlobalPara.memberComp.reportRecoveredUpdatesCommitComplete();
//	}
	
	
}
