package org.zhouw.stps.tp;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Comparator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.concurrent.locks.ReentrantLock;
import java.util.Hashtable;
import java.util.HashSet;

import org.zhouw.stps.GlobalPara;
import org.zhouw.stps.MsgBroker;
import org.zhouw.stps.conf.NodeConfig.role;

public class MemoryManager {

	public double inflation;
	public PriorityBlockingQueue<DataItemState> queue;
	public HashSet<DataItemState> accessedDataItems;
	
	public LinkedHashSet<DataItemState> toBeResetDataItems;
	
	public LinkedBlockingQueue<MMMsg.MSGTYPE> workQueue;

	
	public void updateInflation(double newValue)
	{
		if(inflation<newValue)
			inflation=newValue;
	}
	
	private long startTime=0;
	private int monitorCount=0;
	private BufferedWriter writer=null;
	private BufferedWriter writer_hitmiss=null;
	private int lastHit=0;
	private int lastMiss=0;
	
	public void printMonitor(){
		
		long curTime=(System.currentTimeMillis()-startTime)/1000;//seconds
		
		monitorCount++;
		
		try {
			long totalMemory=Runtime.getRuntime().totalMemory()/1024;
			long freeMemory=Runtime.getRuntime().freeMemory()/1024;
			writer.write(curTime+" "+freeMemory+ " "+totalMemory);
			writer.newLine();
				
			int curHit;
			int curMiss;
			int totalHit;
			int totalMiss;
	
			totalHit=TPNode.LTM.hit;
			totalMiss=TPNode.LTM.Miss;
			
			curHit=totalHit-lastHit;
			curMiss=totalMiss-lastMiss;
			double curRate=0;
			double totalRate=0;
			
			if((curHit+curMiss)!=0)
				curRate=((double)curHit/(double)(curHit+curMiss))*100;
			
			if((totalHit+totalMiss)!=0)
				totalRate=((double)totalHit/(double)(totalHit+totalMiss))*100;
	
			int numofinqueue=TPNode.MM.getNumOfInqueue();
	
					
			String output=String.format("%1$5d %2$10.1f %3$10.1f %4$10d %5$10d %6$10d %7$10d  %8$10d %9$10d %10$10d %11$10d %12$10d", 
					curTime, curRate, totalRate, TPNode.MM.bufferSize,TPNode.MM.backupDataSize, TPNode.LTM.totalNumInDataPool, TPNode.LTM.totalNumInBackupDataPool,
					numofinqueue, curHit, totalHit, curMiss, totalMiss);
			writer_hitmiss.write(output);
			writer_hitmiss.newLine();
			lastHit=totalHit;
			lastMiss=totalMiss;
					
			if(monitorCount%20==0){
				writer_hitmiss.flush();
				writer.flush();
			}
		} catch (IOException e1) {
			e1.printStackTrace();
		}
	}
	
	public void startMonitor(){
		startTime=System.currentTimeMillis();
		monitorCount=0;
		
		String filename=GlobalPara.sysConf.outPerfLogPath+GlobalPara.sysConf.localNodeID+".memoryLog";
		String filename_HitMiss=GlobalPara.sysConf.outPerfLogPath+GlobalPara.sysConf.localNodeID+".hitmiss";

		try {
			writer=new BufferedWriter(new FileWriter(filename,false));
			writer_hitmiss=null;
			
			writer_hitmiss=new BufferedWriter(new FileWriter(filename_HitMiss,false));
			writer_hitmiss.write("#time hit_rate hit_rate_total BufferSize BackupSize NumInBuffer NumInBackup NumOfInQueue Current_hit total_hit Current_miss total_miss");
			writer_hitmiss.newLine();
			
			writer.write("#time freeMemory totalMemory");
			writer.newLine();
		} catch (IOException e1) {
			e1.printStackTrace();
		}
	}
	
//	public void start()
//	{
//		ExecutorService se=Executors.newSingleThreadExecutor();
//		HandleMemoryMissTask task=new HandleMemoryMissTask();
//		se.submit(task);
//		se.shutdown();
//	}
	
	public int getNumOfInqueue()
	{
		return numofInQueue;
	}
	
	private int numofInQueue=0;
	
	public long bufferSize;
	public long backupDataSize;
	
	public void reportBufferSizeChange(){
		while(this.bufferSize>GlobalPara.sysConf.bufferLimit){
			DataItemState row=this.queue.poll();
			if(row==null){
				GlobalPara.logger.warning("[Error]There no rows in the queue to be evicted");
				break;
			} else {
				if(row.isEvictable){
					if(row.isHValueOutdate){
						row.updateHValue();
						this.queue.offer(row);
					} else {
						this.removeOneRowFromBuffer(row);
						TPNode.LTM.totalNumInDataPool--;
						this.bufferSize-=row.sizeOfRootDataRowEntry;
						numofInQueue--;
					}
				} else {
					this.accessedDataItems.add(row);
					numofInQueue--;
				}
			}
		}
	}
	
	public void removeOneRowFromBuffer(DataItemState row){
		row.data.dataTable.data.remove(row.data.rowKey);
		TPNode.MM.updateInflation(row.H);
		//GlobalPara.logger.warning("Row "+row.data.dataTable.tblName+ "-"+row.data.rowKey+" is evicted with H="+row.H);
		if(row.data.latestWriteTransaction>-1){ //if==-1, then it don't have backup
			UpdatesOfST backupData=new UpdatesOfST(row.data.rowKey, row.data.dataTable.tblName, row.data.latestWriteTransaction, null, GlobalPara.sysConf.localNodeID, 1);
			DataRowEntry.removeReplicaTask(backupData);
		}
	}
	
	public void reportRowBecomeEvictable(DataItemState row, boolean isReportBufferSizeChange)
	{
		if(this.accessedDataItems.remove(row)){
			if(GlobalPara.sysConf.minimumMemory){
				removeOneRowFromBuffer(row);
			} else {
				numofInQueue++;
				boolean isDataChanged=row.updateHValue();
				this.queue.offer(row);
				if(isReportBufferSizeChange&&isDataChanged)
					TPNode.MM.reportBufferSizeChange();
			}
		}
	}
	
	public void addNewRowToQueueForRecovery(DataItemState row)
	{
		numofInQueue++;
		row.updateHValue();
		this.queue.offer(row);
	}
	
//	public void putRowInQueue(DataItemState row)
//	{
//		numofInQueue++;
//		this.queue.offer(row);
//	}
	
	public MemoryManager()
	{
//		this.accessedDataItems=new Hashtable<String, Integer>(200);
//		this.totalAccessedDataItems=new HashSet<String>(300);
//		this.periodAccessedDataItems=new HashSet<String>(300);
//		this.lockForaccessedDataItems=new ReentrantLock();
//		this.lock=new ReentrantLock();
		//this.lockOfQueue=new ReentrantLock();
		
		this.accessedDataItems=new HashSet<DataItemState>();
		this.toBeResetDataItems=new LinkedHashSet<DataItemState>();
		this.bufferSize=0;
		this.backupDataSize=0;
		this.workQueue=new LinkedBlockingQueue<MMMsg.MSGTYPE>();
		this.inflation=0.0;
		this.queue=new PriorityBlockingQueue<DataItemState>(10000, new Comparator<DataItemState>()
				{
		     public int compare(DataItemState a, DataItemState b)
		     {
		       double A = a.H;
		       double B = b.H;
		       if (B<A)
		            return 1;
		       else if (B>A)
		            return -1;
		       else 
		            return 0; 
		     }
		}
		);
		
//		if(GlobalPara.sysConf.doMemoryManagement)
//		{
//			this.start();
//		}
	}
}
