package org.zhouw.stps.tp;

import java.util.Vector;

import org.zhouw.stps.*;

public class RecoveryZone {

	//the nodes in the ring is sorted, with indexOfPrimary always at the position 0
	public VirtualTPNode[] nodes;
	
	public int[] getTPNodes()
	{
		int[] result=null;
		if(nodes!=null){
			result=new int[nodes.length];
			result[0]=nodes[0].physicalID;
			for(int i=1;i<nodes.length;i++){
				result[i]=nodes[i].physicalID;
			}
		}
		return result;
	}
	
	
	public String printZone()
	{
		String msg="";
		for(int i=0;i<nodes.length;i++)
			{
				VirtualTPNode node=nodes[i];
				msg+=nodes[i].virtualID+" "+nodes[i].physicalID + "_"+nodes[i].index+"\n";
			}
		return msg;
	}
	
	/**
	 * the primary node index
	 */
	public int indexOfPrimary;
	
	public VirtualTPNode getPrimaryVNode()
	{
		if(nodes==null||nodes.length==0||this.indexOfPrimary<0||this.indexOfPrimary>nodes.length-1)
			return null;
		return nodes[this.indexOfPrimary];
	}
	
	/**
	 * check if current primary node fails and the local node will become the new primary node
	 * @param localNodeID
	 * @param failedTPNodes
	 * @return
	 */
	public boolean isLocalNextPrimary(int localNodeID, Vector<Integer> failedTPNodes)
	{
		if(nodes==null||nodes.length==0||nodes.length==1)
			return false;
		
		if(nodes[0].physicalID==localNodeID)
			return false; //it is the current primary
		else if(!failedTPNodes.contains(nodes[0].physicalID))
			return false; //the primary node is not failed
		
		for(int i=1;i<nodes.length;i++){//start from the 2nd in the list
			if(!failedTPNodes.contains(nodes[i].physicalID)){
				if(localNodeID==nodes[i].physicalID)
					return true;
				else
					return false;
			}
		}
		
		GlobalPara.logger.warning("There are no LTMs to become primary in this zone");
		return false;
	}
	
	public VirtualTPNode getSecondaryPrimayWhenFailure(Vector<Integer> failedTPNodes)
	{
		if(nodes==null||nodes.length==0||nodes.length==1)
			return null;
		
		if(!failedTPNodes.contains(nodes[0].physicalID))
			return null; //the primary node is not failed
		
		for(int i=1;i<nodes.length;i++){//start from the 2nd in the list
			if(!failedTPNodes.contains(nodes[i].physicalID)){
				return nodes[i];
			}
		}
		
		GlobalPara.logger.warning("There are no LTMs to become primary in this zone");
		return null;
	}
//	
//	public VirtualTPNode getSecondaryPrimay()
//	{
//		if(nodes==null||nodes.length==0||nodes.length==1)
//			return null;
//		return nodes[1];
//	}
//	
//	public boolean isPrimary(int tpNode)
//	{
//		if(nodes==null)
//			return false;
//		if(nodes[indexOfPrimary].physicalID==tpNode)
//			return true;
//		else
//			return false;
//	}
	
	/**
	 * Check if this zone contains any of the failed nodes
	 * @param inputNodes 
	 * @return the nodes that not contained in the zone
	 */
	public Vector<Integer> excludes(int[] inputNodes)
	{
		Vector<Integer> included=new Vector<Integer>();
		Vector<Integer> result=new Vector<Integer>();
		if(nodes==null||inputNodes==null)
			return result;
		for(int i=0;i<nodes.length;i++)
			for(int j=0;j<inputNodes.length;j++) {
				if(nodes[i].physicalID==inputNodes[j]){
					included.add(inputNodes[j]);
					break;
				}
			}
		for(int node:inputNodes){
			if(!included.contains(node))
				result.add(node);
		}
		return result;
	}
//	
//	/**
//	 * Check if this zone contains any of the failed nodes
//	 * @param failedTPNodes failed physical node ids
//	 * @return
//	 */
//	public int countOfContained(int[] failedTPNodes)
//	{
//		if(nodes==null)
//			return 0;
//		int count=0;
//		for(int i=0;i<nodes.length;i++)
//			for(int j=0;j<failedTPNodes.length;j++)
//				{
//					if(nodes[i].physicalID==failedTPNodes[j])
//						{
//						count++;
//						break;
//						}
//				}
//		return count;
//	}
	
	public RecoveryZone(VirtualTPNode[] nodes, int indexOfPrimary) {
		super();
		this.nodes = nodes;
		this.indexOfPrimary = indexOfPrimary;
	}
	
}
