package org.zhouw.stps.tp;

import org.zhouw.stps.*;
import org.zhouw.stps.app.DataModel;
import org.zhouw.stps.app.DataTable;

import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Future;
import java.util.concurrent.locks.*;

public class TableDataBuffer {
	
	/**
	 * the initial cache size for getting data from database
	 * currently assign a global percentage for all tables
	 * what percentage of rows can be sustained in memory of TP-Node
	 */
	public static final double percentage=0.1;
	public String tblName;
	public HashMap<String, DataRowEntry> data;
	public DataTable tableSchema;
	
	
//	/**
//	 * locked when adding a new row to it
//	 */
//	private final ReentrantReadWriteLock lock;
//	private final Lock r;
//    private final Lock w;
	
//    private final ReentrantReadWriteLock localLock=new ReentrantReadWriteLock();
//	private final Lock lr=localLock.readLock();
//    private final Lock lw=localLock.writeLock();

    /**
     * 
     * @param tblName
     * @param count as the a factor to initial size of inside hashtable
     * @param lock the global lock from LTM, for checkpoint or committing data to multiple TableDataBuffers . 
     */
    TableDataBuffer(DataTable schema)
	{
    		int count =1000;
		this.tblName=schema.getPhysicalTableName();
		data=new HashMap<String, DataRowEntry>((int)(count*percentage/0.75));
		this.tableSchema=schema;
//		this.lock=lock;
//		this.r=this.lock.readLock();
//		this.w=this.lock.writeLock();
	}
    
    
    public void commitST(SubTransaction t) {
    	
    	if(t==null){
    		GlobalPara.logger.warning("The input Subtransaction is null");
    		return;
    	}
    	DataRowEntry row=this.data.get(t.strRowKey);
    	if(row==null){
			GlobalPara.logger.severe("The row does not exist when trying to remove a SubTransaction for docommit");
		} else {
			row.doCommit(t); //commitDoneST(t);
		}
    }
    
    
    public static String printHashTable(Hashtable<String, String> table)
	{
		String result="";
		Enumeration<String> e= table.keys();
		while(e.hasMoreElements())
		{
			String key=e.nextElement();
			String value=table.get(key);
			result+=key+"  "+value + "\n";
		}
		return result;
	}
    
    
    public void handleOneRowLoaded(String rowkey,HashMap<String, String> rowData)
    {
    	DataRowEntry row=this.data.get(rowkey);
    	if(row==null){
    		GlobalPara.logger.warning("There is no data row entry for the loaded row content");
    		return;
    	}
    	if(row.isDataLoaded==1){
    		GlobalPara.logger.warning("[Error]The row of " + rowkey+ " in table "+ this.tblName +" has already been loaded");
    		return;
    	}
    	row.committedData.rowData.clear();
    	row.committedData.rowData.putAll(rowData);
    	row.isDataLoaded=1;
    	if(GlobalPara.sysConf.doMemoryManagement)
    		row.state.reportDataLoaded();
    	row.execRest();
    }
    
    public void precommitRest(String rowkey)
    {
		DataRowEntry row=this.data.get(rowkey);
		if(row!=null)
			row.execRest();
		else {
			GlobalPara.logger.warning("[error]Table " + this.tblName+ " does not have the entry of " + rowkey + " present in memory.");
			java.lang.NullPointerException e=new java.lang.NullPointerException();
			e.printStackTrace();
		}
    }
//    
//    public void commitST(SubTransaction t)
//    {
//    	r.lock();
//    	try{
//    		DataRowEntry row=getRow(t.strRowKey);
//    		DoBgPreCommit task=new DoBgPreCommit(t.strTblName,t.strRowKey);
//    		LocalTransManager.bgService.submit(task);
//    	}
//    	finally{
//    		r.unlock();
//    		}
//    }
//    
//    public void abortST(SubTransaction t)
//    {
//    	r.lock();
//    	try{
//    		DataRowEntry row=getRow(t.strRowKey);
//    		int result=row.abort(t);
//    		if(result==2)
//    			{
//    			DoBgPreCommit task=new DoBgPreCommit(t.strTblName,t.strRowKey);
//    			LocalTransManager.bgService.submit(task);
//    			}
//    	}
//    	finally{
//    		r.unlock();
//    		}
//    }
//    
    /**
     * 
     * @param t
     * @return 1 committed, 0 abort, -1 still wait
     */
    public int addSubTransaction(SubTransaction t, boolean isRestart)
    {
		DataRowEntry row=getRow(t.strRowKey, isRestart);
		t.setRootEntry(row);
		return row.add(t, isRestart);
    }
    
    
//    /**
//     * This function will not trigger the reload of the data item
//     * @param t
//     * @return 1 committed, 0 abort, -1 still wait
//     */
//    public int readdASubTransaction(SubTransaction t)
//    {
//    	DataRowEntry row=this.data.get(t.strRowKey);
//    	if(row==null){
//    		GlobalPara.logger.warning("The row is not existed when restarting a subtransaction of data row: "+ t.strTblName+"--"+t.strRowKey);
//    		return -1;
//    	}
//    	else
//    		return row.add(t);
//    }
    
    /**
     * 
     * @param t
     * @return 1 success, 0 the subtran not in the queue, -1 not data item is not existed
     */
    public int removeSubTransaction(SubTransaction t, boolean isRestart)
    {
    	DataRowEntry row=this.data.get(t.strRowKey);
		if(row!=null){
			if(row.removeSubTransaction(t, isRestart))
				return 1;
			else
				return 0;
		}
		else
			return -1;
    }
    
//	/**
//	 * Copy a snapshot of the table about the data involved by the virtual nodes into result
//	 * note that this "snapshot" may not be consistent
//	 * because following updates will all be recorded and redo
//	 * so no synchronization is needed here 
//	 * @param from
//	 * @param to
//	 * @return
//	 */
//    public Hashtable<Integer, Vector<UpdatesOfST>> copySnapshotForVNodes
//    	(VirtualTPNode from[],Vector<VirtualTPNode[]> to, Hashtable<Integer, Vector<UpdatesOfST>> result)
//    {
//    	//lr.lock();
//    	try{
//    		 Set<Map.Entry<String, DataRowEntry>> entries=this.data.entrySet();
//    		 Iterator<Entry<String, DataRowEntry>> its=entries.iterator();
//    		 while(its.hasNext())
//    		 	{
//    			  Entry<String, DataRowEntry> en=its.next();
//    			  String rowKey=en.getKey();
//    			  VirtualTPNode vnode=TPNode.getVirtualNodeWithoutLock(this.tblName, rowKey);
//    			  for(int i=0;i<from.length;i++)
//    				  if(from[i].virtualID==vnode.virtualID) //if it is affected
//    				  	{
//    					  	VirtualTPNode[] dests=to.get(i);
//    					  	for(VirtualTPNode vDestNode:dests)
//    					  		{
//	    					  		 Vector<UpdatesOfST> buffer=result.get(vDestNode.physicalID);// the result is structured by the destination
//	    	    					  if(buffer==null)
//	    	    						  {
//	    	    						  	buffer=new Vector<UpdatesOfST>();
//	    	    						  	result.put(vDestNode.physicalID, buffer);
//	    	    						  }
//	    	    					  DataRowEntry row=en.getValue();
//	    	    					  UpdatesOfST updates=null;
//	    	    					  if(row.latestWriteTransaction!=-1)
//	    	    					  	{
//	    	    						  updates=new  UpdatesOfST(rowKey, this.tblName, 
//	    	    								  row.latestWriteTransaction, row.committedData.getRowData(),GlobalPara.sysConf.localNodeID);
//	    	    						  buffer.add(updates);
//	    	    					  	}
//    					  		}
//    					  break;//next data item
//    				  	}
//    		 	}
//    	}
//    	finally{
//    		//r.unlock();
//    		//lr.unlock();
//    		}
//		return result;
//    }
    
    /**
     * In this case, the row is always obtained
     * @param rowkey
     * @return
     */
    public DataRowEntry getOneRowEntry(String rowkey)
  {
    	return this.data.get(rowkey);
  }
	/**
	 * Get the row object, if not present, then retrieve the data from HBase
	 * @param rowkey
	 * @return
	 */
	private DataRowEntry getRow(String rowkey, boolean isRestart)
	{
		DataRowEntry row=this.data.get(rowkey);
		if(row==null) {//if this row is not in-memory, add a new row and load data then submit
			row=new DataRowEntry(this, rowkey);
			if(GlobalPara.sysConf.doMemoryManagement){
				TPNode.MM.accessedDataItems.add(row.state);
				TPNode.LTM.totalNumInDataPool++;
			}
			this.data.put(rowkey,row);
			TPNode.LTM.Miss++;
		} else {
			TPNode.LTM.hit++;
		}
		
		if(GlobalPara.sysConf.doMemoryManagement&&!isRestart)
			row.state.transactionAccess();
		
		if(row.isDataLoaded==-1){
			GlobalPara.adapterComp.loadData(this.tableSchema, rowkey);
			row.isDataLoaded=0;
		}

		return row;
	}
	
	/**
	 * Add a new row with data already loaded into the system.
	 * @param rowkey
	 * @param dataItem
	 * @return
	 */
	public DataRowEntry addNewRow(String rowkey, DataRowContainer dataItem)
	{
		DataRowEntry row=new DataRowEntry(this, rowkey, dataItem);
		if(GlobalPara.sysConf.doMemoryManagement){
			//row.state.transactionAccess();
			//TPNode.MM.toBeResetDataItems.add(row.state);
			TPNode.MM.addNewRowToQueueForRecovery(row.state);
			//TPNode.MM.accessedDataItems.add(row.state);
			TPNode.MM.backupDataSize-=dataItem.size;
			TPNode.MM.bufferSize+=row.state.sizeOfRootDataRowEntry;
//			TPNode.MM.reportBufferSizeChange();
		}
		this.data.put(rowkey,row);
		return row;
	}
	
}
