package org.zhouw.stps.tp;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.locks.ReentrantLock;

import org.zhouw.stps.GlobalPara;

/**
 * buffer the updates of transaction during backup data items to a new node
 * @author zhouw
 *
 */
public class UpdateBufferOnSnapshot {
	
	//used to stop commit updates when the system is replicating the updates to the new nodes
	//after replication complete, the backup list will be updated and new committed data to also 
	//pushed to the new nodes
	public CountDownLatch signal;
	public ReentrantLock lock;
	
	public void clear()
	{
		this.signal=null;
		this.pool.clear();
	}
	
	/**
	 * <DestID, Hashtable<tablename--rowkey, Integer>>
	 * records the data items id that has been updated to the snapshot
	 * the actual values are obtained from working set at runtime
	 */
	Hashtable<Integer, Hashtable<String, Integer>> pool;
	
//	public Hashtable<Integer, Vector<UpdatesOfST>> getUptodateDataUpdates(ConcurrentHashMap<String, TableDataBuffer> dataPool)
//	{
//		Hashtable<Integer, Vector<UpdatesOfST>> result=new  Hashtable<Integer, Vector<UpdatesOfST>>();
//		Enumeration<Integer> e= this.pool.keys();
//		while(e.hasMoreElements())
//			{
//				int destID=e.nextElement();
//				Vector<UpdatesOfST> item=new Vector<UpdatesOfST>();
//				Hashtable<String, Integer> table=pool.get(destID);
//				Enumeration<String> es =table.keys();
//				while(es.hasMoreElements())
//					{
//						String key=es.nextElement();
//						String[] temp=key.split("--");
//						String tableName=temp[0];
//						String rowKey=temp[1];
//						TableDataBuffer tdb=dataPool.get(tableName);
//						if(tdb==null)
//							{
//							GlobalPara.logger.warning("Unable to find table" + tableName);
//							continue;
//							}
//						DataRowEntry row=tdb.getRowWithoutLock(rowKey);
//						if(row==null)
//							{
//							GlobalPara.logger.warning("Unable to find row " + rowKey + " in table " +tableName);
//							continue;
//							}
//						UpdatesOfST updates=null;
//
//						//always replicate the entire row, because it is possible that some rows are inserted.
//  						  updates=new  UpdatesOfST(rowKey, tableName, 
//  								  row.latestWriteTransaction, row.committedData.getRowData(),GlobalPara.sysConf.localNodeID);
//  						  item.add(updates);
//					}
//				result.put(destID, item);
//			}
//		return result;
//	}
	
	public UpdateBufferOnSnapshot()
	{
		pool=new Hashtable<Integer, Hashtable<String, Integer>>();
		signal=new CountDownLatch(1);
		this.lock=new ReentrantLock();
	}
	
	private String getKey(String tableName, String rowKey)
	{
		return tableName+"--"+rowKey;
	}
	
	private String getTablename(String key)
	{
		String[] data=key.split("--");
		return data[0];
	}
	
	private String getRowkey(String key)
	{
		String[] data=key.split("--");
		return data[1];
	}
	
	/**
	 * 一个数据行在复制snapshot的过程有没有被更新过
	 * @param vnode
	 * @param tablename
	 * @param rowKey
	 * @return
	 */
	public boolean checkUpdateExisted(VirtualTPNode vnode, String tablename, String rowKey)
	{
		if(vnode==null)
			return false;
		Hashtable<String, Integer> item=pool.get(vnode.physicalID);
		
		if(item==null)
			{
				return false;
			}
		
		String key=this.getKey(tablename, rowKey);
		
		if(item.get(key)==null)
			return false;
		else
			return true;
	}
	
	
	public void recordUpdate(VirtualTPNode vnode,String tablename, String rowKey)
	{
		//VirtualTPNode vnode=TPNode.getVirtualNode(tablename, rowKey);
		if(vnode==null)
			return;
		Hashtable<String, Integer> item=pool.get(vnode.physicalID);
		
		if(item==null)
			{
				item=new Hashtable<String, Integer>();
				pool.put(vnode.physicalID, item);
			}
		
		String key=this.getKey(tablename, rowKey);
		
		if(item.get(key)==null)
			item.put(key, 1);
	}
}
